/* Generated by Cython 0.23.3 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_23_3"
#include <stddef.h>
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
#define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
#define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) (void_result, Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__bzrlib___btree_serializer_pyx
#define __PYX_HAVE_API__bzrlib___btree_serializer_pyx
#include "python-compat.h"
#include "stdlib.h"
#include "string.h"
#include "_static_tuple_c.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "bzrlib/_btree_serializer_pyx.pyx",
  "bzrlib/_static_tuple_c.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser;
struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
struct __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record;
typedef struct __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record;

/* "bzrlib/_btree_serializer_pyx.pyx":337
 * #       One slightly ugly option would be to cache block offsets in a global.
 * #       However, that leads to thread-safety issues, etc.
 * ctypedef struct gc_chk_sha1_record:             # <<<<<<<<<<<<<<
 *     long long block_offset
 *     unsigned int block_length
 */
struct __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record {
  PY_LONG_LONG block_offset;
  unsigned int block_length;
  unsigned int record_start;
  unsigned int record_end;
  char sha1[20];
};

/* "bzrlib/_btree_serializer_pyx.pyx":123
 * 
 * 
 * cdef class BTreeLeafParser:             # <<<<<<<<<<<<<<
 *     """Parse the leaf nodes of a BTree index.
 * 
 */
struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser {
  PyObject_HEAD
  struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_vtab;
  PyObject *bytes;
  int key_length;
  int ref_list_length;
  PyObject *keys;
  char *_cur_str;
  char *_end_str;
  char *_start;
  int _header_found;
};


/* "bzrlib/_btree_serializer_pyx.pyx":518
 * 
 * 
 * cdef class GCCHKSHA1LeafNode:             # <<<<<<<<<<<<<<
 *     """Track all the entries for a given leaf node."""
 * 
 */
struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode {
  PyObject_HEAD
  struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_vtab;
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *records;
  PyObject *last_key;
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *last_record;
  int num_records;
  unsigned char common_shift;
  unsigned char offsets[0x101];
};



/* "bzrlib/_btree_serializer_pyx.pyx":123
 * 
 * 
 * cdef class BTreeLeafParser:             # <<<<<<<<<<<<<<
 *     """Parse the leaf nodes of a BTree index.
 * 
 */

struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser {
  PyObject *(*extract_key)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *, char *);
  int (*process_line)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *);
};
static struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser;


/* "bzrlib/_btree_serializer_pyx.pyx":518
 * 
 * 
 * cdef class GCCHKSHA1LeafNode:             # <<<<<<<<<<<<<<
 *     """Track all the entries for a given leaf node."""
 * 
 */

struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode {
  StaticTuple *(*_record_to_value_and_refs)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *);
  StaticTuple *(*_record_to_item)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *);
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *(*_lookup_record)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *);
  int (*_count_records)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *);
  PyObject *(*_parse_bytes)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, PyObject *);
  char *(*_parse_one_entry)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *);
  int (*_offset_for_sha1)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *);
  PyObject *(*_compute_common)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *);
};
static struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode;

/* --- Runtime support code (head) --- */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb);

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

static CYTHON_INLINE long __Pyx_div_long(long, long);

static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_EqObjC(op1, op2, intval, inplace)\
    PyObject_RichCompare(op1, op2, Py_EQ)
    #endif

#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

static int __Pyx_SetVtable(PyObject *dict, void *vtable);

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value);

static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name);

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_extract_key(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, char *__pyx_v_last); /* proto*/
static int __pyx_f_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_process_line(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto*/
static StaticTuple *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs(CYTHON_UNUSED struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record); /* proto*/
static StaticTuple *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record); /* proto*/
static __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1); /* proto*/
static int __pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records(CYTHON_UNUSED struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_content, char *__pyx_v_c_end); /* proto*/
static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes); /* proto*/
static char *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry(CYTHON_UNUSED struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_cur, char *__pyx_v_c_end, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record); /* proto*/
static int __pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1); /* proto*/
static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto*/

/* Module declarations from 'bzrlib._static_tuple_c' */

/* Module declarations from 'bzrlib._static_tuple_c' */
static PyTypeObject *__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple = 0;

/* Module declarations from 'bzrlib._btree_serializer_pyx' */
static PyTypeObject *__pyx_ptype_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser = 0;
static PyTypeObject *__pyx_ptype_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode = 0;
static int __pyx_v_6bzrlib_21_btree_serializer_pyx__unhexbuf[0x100];
static char *__pyx_v_6bzrlib_21_btree_serializer_pyx__hexbuf;
static void *__pyx_f_6bzrlib_21_btree_serializer_pyx__my_memrchr(void *, int, size_t); /*proto*/
static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_safe_string_from_size(char *, Py_ssize_t); /*proto*/
static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_safe_interned_string_from_size(char *, Py_ssize_t); /*proto*/
static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx__populate_unhexbuf(void); /*proto*/
static int __pyx_f_6bzrlib_21_btree_serializer_pyx__unhexlify_sha1(char *, char *); /*proto*/
static void __pyx_f_6bzrlib_21_btree_serializer_pyx__hexlify_sha1(char *, char *); /*proto*/
static int __pyx_f_6bzrlib_21_btree_serializer_pyx__key_to_sha1(PyObject *, char *); /*proto*/
static StaticTuple *__pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_key(char *); /*proto*/
static unsigned int __pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_uint(char *); /*proto*/
static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx__format_record(__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *); /*proto*/
#define __Pyx_MODULE_NAME "bzrlib._btree_serializer_pyx"
int __pyx_module_is_main_bzrlib___btree_serializer_pyx = 0;

/* Implementation of 'bzrlib._btree_serializer_pyx' */
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_TypeError;
static char __pyx_k_i[] = "i";
static char __pyx_k_0_0[] = " 0 0";
static char __pyx_k__20[] = "\000";
static char __pyx_k_key[] = "key";
static char __pyx_k_out[] = "out";
static char __pyx_k_sys[] = "sys";
static char __pyx_k_val[] = "val";
static char __pyx_k_join[] = "join";
static char __pyx_k_line[] = "line";
static char __pyx_k_main[] = "__main__";
static char __pyx_k_node[] = "node";
static char __pyx_k_sha1[] = "sha1:";
static char __pyx_k_test[] = "__test__";
static char __pyx_k_bytes[] = "bytes";
static char __pyx_k_parse[] = "parse";
static char __pyx_k_value[] = "value";
static char __pyx_k_as_bin[] = "as_bin";
static char __pyx_k_as_hex[] = "as_hex";
static char __pyx_k_import[] = "__import__";
static char __pyx_k_parser[] = "parser";
static char __pyx_k_key_len[] = "key_len";
static char __pyx_k_ref_bit[] = "ref_bit";
static char __pyx_k_s_u_u_u[] = "%s %u %u %u";
static char __pyx_k_KeyError[] = "KeyError";
static char __pyx_k_flat_len[] = "flat_len";
static char __pyx_k_lu_u_u_u[] = "%lu %u %u %u";
static char __pyx_k_next_len[] = "next_len";
static char __pyx_k_node_len[] = "node_len";
static char __pyx_k_ref_list[] = "ref_list";
static char __pyx_k_refs_len[] = "refs_len";
static char __pyx_k_sha1_bin[] = "sha1_bin";
static char __pyx_k_TypeError[] = "TypeError";
static char __pyx_k_ref_lists[] = "ref_lists";
static char __pyx_k_reference[] = "reference";
static char __pyx_k_type_leaf[] = "type=leaf";
static char __pyx_k_value_len[] = "value_len";
static char __pyx_k_ValueError[] = "ValueError";
static char __pyx_k_as_bin_sha[] = "as_bin_sha";
static char __pyx_k_key_length[] = "key_length";
static char __pyx_k_py_hexlify[] = "_py_hexlify";
static char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static char __pyx_k_string_key[] = "string_key";
static char __pyx_k_ref_bit_len[] = "ref_bit_len";
static char __pyx_k_type_leaf_2[] = "type=leaf\n";
static char __pyx_k_flatten_node[] = "_flatten_node";
static char __pyx_k_py_unhexlify[] = "_py_unhexlify";
static char __pyx_k_AssertionError[] = "AssertionError";
static char __pyx_k_first_ref_list[] = "first_ref_list";
static char __pyx_k_parse_into_chk[] = "_parse_into_chk";
static char __pyx_k_py_key_to_sha1[] = "_py_key_to_sha1";
static char __pyx_k_py_sha1_to_key[] = "_py_sha1_to_key";
static char __pyx_k_first_reference[] = "first_reference";
static char __pyx_k_ref_list_length[] = "ref_list_length";
static char __pyx_k_reference_lists[] = "reference_lists";
static char __pyx_k_0123456789abcdef[] = "0123456789abcdef";
static char __pyx_k_last_self__start[] = "last < self._start";
static char __pyx_k_parse_leaf_lines[] = "_parse_leaf_lines";
static char __pyx_k_have_reference_lists[] = "have_reference_lists";
static char __pyx_k_key_r_is_not_present[] = "key %r is not present";
static char __pyx_k_We_failed_to_unhexlify[] = "We failed to unhexlify";
static char __pyx_k_not_a_40_byte_hex_digest[] = "not a 40-byte hex digest";
static char __pyx_k_Failed_to_parse_record_end[] = "Failed to parse record end";
static char __pyx_k_self_bytes_is_not_a_string[] = "self.bytes is not a string.";
static char __pyx_k_not_a_20_byte_binary_digest[] = "not a 20-byte binary digest";
static char __pyx_k_Failed_to_parse_block_length[] = "Failed to parse block length";
static char __pyx_k_Failed_to_parse_block_offset[] = "Failed to parse block offset";
static char __pyx_k_bzrlib__btree_serializer_pyx[] = "bzrlib._btree_serializer_pyx";
static char __pyx_k_Failed_to_find_the_value_area[] = "Failed to find the value area";
static char __pyx_k_only_1_null_not_2_as_expected[] = "only 1 null, not 2 as expected";
static char __pyx_k_line_did_not_start_with_sha1_r[] = "line did not start with sha1: %r";
static char __pyx_k_We_only_support_parsing_plain_8[] = "We only support parsing plain 8-bit strings.";
static char __pyx_k_home_vila_src_bzr_integration_2[] = "/home/vila/src/bzr/integration/2.7.0-dev/bzrlib/_btree_serializer_pyx.pyx";
static char __pyx_k_invalid_key_wanted_segment_from[] = "invalid key, wanted segment from ";
static char __pyx_k_Expected_a_plain_str_for_value_n[] = "Expected a plain str for value not: %s";
static char __pyx_k_Line_did_not_contain_40_hex_byte[] = "Line did not contain 40 hex bytes";
static char __pyx_k_Node_did_not_start_with_type_lea[] = "Node did not start with \"type=leaf\": %r";
static char __pyx_k_Pyrex_extensions_to_btree_node_p[] = "Pyrex extensions to btree node parsing.";
static char __pyx_k_Something_went_wrong_while_parsi[] = "Something went wrong while parsing.";
static char __pyx_k_We_expect_reference_bits_to_be_s[] = "We expect reference bits to be strings not: %s";
static char __pyx_k_We_expect_references_to_be_tuple[] = "We expect references to be tuples not: %s";
static char __pyx_k_We_expected_a_tuple_or_StaticTup[] = "We expected a tuple() or StaticTuple() for node not: %s";
static char __pyx_k_With_ref_lists_we_expected_4_ent[] = "With ref_lists, we expected 4 entries not: %s";
static char __pyx_k_Without_ref_lists_we_need_at_lea[] = "Without ref_lists, we need at least 3 entries not: %s";
static char __pyx_k_bytes_did_not_start_with_type_le[] = "bytes did not start with 'type=leaf\\n': %r";
static char __pyx_k_invalid_key_loop_counter_self_re[] = "invalid key, loop_counter != self.ref_list_length";
static char __pyx_k_sha1_bin_must_be_a_str_of_exactl[] = "sha1_bin must be a str of exactly 20 bytes";
static char __pyx_k_tried_to_create_a_string_with_an[] = "tried to create a string with an invalid size: %d @0x%x";
static char __pyx_k_unexpected_reference_data_presen[] = "unexpected reference data present";
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_kp_s_Expected_a_plain_str_for_value_n;
static PyObject *__pyx_kp_s_Failed_to_find_the_value_area;
static PyObject *__pyx_kp_s_Failed_to_parse_block_length;
static PyObject *__pyx_kp_s_Failed_to_parse_block_offset;
static PyObject *__pyx_kp_s_Failed_to_parse_record_end;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_s_Line_did_not_contain_40_hex_byte;
static PyObject *__pyx_kp_s_Node_did_not_start_with_type_lea;
static PyObject *__pyx_kp_s_Something_went_wrong_while_parsi;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s_We_expect_reference_bits_to_be_s;
static PyObject *__pyx_kp_s_We_expect_references_to_be_tuple;
static PyObject *__pyx_kp_s_We_expected_a_tuple_or_StaticTup;
static PyObject *__pyx_kp_s_We_failed_to_unhexlify;
static PyObject *__pyx_kp_s_We_only_support_parsing_plain_8;
static PyObject *__pyx_kp_s_With_ref_lists_we_expected_4_ent;
static PyObject *__pyx_kp_s_Without_ref_lists_we_need_at_lea;
static PyObject *__pyx_kp_s__20;
static PyObject *__pyx_n_s_as_bin;
static PyObject *__pyx_n_s_as_bin_sha;
static PyObject *__pyx_n_s_as_hex;
static PyObject *__pyx_n_s_bytes;
static PyObject *__pyx_kp_s_bytes_did_not_start_with_type_le;
static PyObject *__pyx_n_s_bzrlib__btree_serializer_pyx;
static PyObject *__pyx_n_s_first_ref_list;
static PyObject *__pyx_n_s_first_reference;
static PyObject *__pyx_n_s_flat_len;
static PyObject *__pyx_n_s_flatten_node;
static PyObject *__pyx_n_s_have_reference_lists;
static PyObject *__pyx_kp_s_home_vila_src_bzr_integration_2;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_kp_s_invalid_key_loop_counter_self_re;
static PyObject *__pyx_kp_s_invalid_key_wanted_segment_from;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_key_len;
static PyObject *__pyx_n_s_key_length;
static PyObject *__pyx_kp_s_key_r_is_not_present;
static PyObject *__pyx_kp_s_last_self__start;
static PyObject *__pyx_n_s_line;
static PyObject *__pyx_kp_s_line_did_not_start_with_sha1_r;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_next_len;
static PyObject *__pyx_n_s_node;
static PyObject *__pyx_n_s_node_len;
static PyObject *__pyx_kp_s_not_a_20_byte_binary_digest;
static PyObject *__pyx_kp_s_not_a_40_byte_hex_digest;
static PyObject *__pyx_kp_s_only_1_null_not_2_as_expected;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_parse;
static PyObject *__pyx_n_s_parse_into_chk;
static PyObject *__pyx_n_s_parse_leaf_lines;
static PyObject *__pyx_n_s_parser;
static PyObject *__pyx_n_s_py_hexlify;
static PyObject *__pyx_n_s_py_key_to_sha1;
static PyObject *__pyx_n_s_py_sha1_to_key;
static PyObject *__pyx_n_s_py_unhexlify;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_ref_bit;
static PyObject *__pyx_n_s_ref_bit_len;
static PyObject *__pyx_n_s_ref_list;
static PyObject *__pyx_n_s_ref_list_length;
static PyObject *__pyx_n_s_ref_lists;
static PyObject *__pyx_n_s_reference;
static PyObject *__pyx_n_s_reference_lists;
static PyObject *__pyx_n_s_refs_len;
static PyObject *__pyx_kp_s_self_bytes_is_not_a_string;
static PyObject *__pyx_n_s_sha1_bin;
static PyObject *__pyx_kp_s_sha1_bin_must_be_a_str_of_exactl;
static PyObject *__pyx_n_s_string_key;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_tried_to_create_a_string_with_an;
static PyObject *__pyx_kp_s_unexpected_reference_data_presen;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_value_len;
static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser___init__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_2parse(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx__parse_leaf_lines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_2_py_unhexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_hex); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_4_py_hexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_bin); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_6_py_key_to_sha1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_8_py_sha1_to_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static void __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static Py_ssize_t __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_sha1); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self); /* proto */
static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_10_parse_into_chk(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length); /* proto */
static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_12_flatten_node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node, PyObject *__pyx_v_reference_lists); /* proto */
static PyObject *__pyx_tp_new_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_2147483648;
static PyObject *__pyx_int_4294967295;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;

/* "bzrlib/_btree_serializer_pyx.pyx":81
 * 
 * # TODO: Find some way to import this from _dirstate_helpers
 * cdef void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     # It is not present in any win32 standard library
 */

static void *__pyx_f_6bzrlib_21_btree_serializer_pyx__my_memrchr(void *__pyx_v_s, int __pyx_v_c, size_t __pyx_v_n) {
  char *__pyx_v_pos;
  char *__pyx_v_start;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_my_memrchr", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":87
 *     cdef char *start
 * 
 *     start = <char*>s             # <<<<<<<<<<<<<<
 *     pos = start + n - 1
 *     while pos >= start:
 */
  __pyx_v_start = ((char *)__pyx_v_s);

  /* "bzrlib/_btree_serializer_pyx.pyx":88
 * 
 *     start = <char*>s
 *     pos = start + n - 1             # <<<<<<<<<<<<<<
 *     while pos >= start:
 *         if pos[0] == c:
 */
  __pyx_v_pos = ((__pyx_v_start + __pyx_v_n) - 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":89
 *     start = <char*>s
 *     pos = start + n - 1
 *     while pos >= start:             # <<<<<<<<<<<<<<
 *         if pos[0] == c:
 *             return <void*>pos
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_pos >= __pyx_v_start) != 0);
    if (!__pyx_t_1) break;

    /* "bzrlib/_btree_serializer_pyx.pyx":90
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
 */
    __pyx_t_1 = (((__pyx_v_pos[0]) == __pyx_v_c) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_btree_serializer_pyx.pyx":91
 *     while pos >= start:
 *         if pos[0] == c:
 *             return <void*>pos             # <<<<<<<<<<<<<<
 *         pos = pos - 1
 *     return NULL
 */
      __pyx_r = ((void *)__pyx_v_pos);
      goto __pyx_L0;

      /* "bzrlib/_btree_serializer_pyx.pyx":90
 *     pos = start + n - 1
 *     while pos >= start:
 *         if pos[0] == c:             # <<<<<<<<<<<<<<
 *             return <void*>pos
 *         pos = pos - 1
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":92
 *         if pos[0] == c:
 *             return <void*>pos
 *         pos = pos - 1             # <<<<<<<<<<<<<<
 *     return NULL
 * 
 */
    __pyx_v_pos = (__pyx_v_pos - 1);
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":93
 *             return <void*>pos
 *         pos = pos - 1
 *     return NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":81
 * 
 * # TODO: Find some way to import this from _dirstate_helpers
 * cdef void* _my_memrchr(void *s, int c, size_t n): # cannot_raise             # <<<<<<<<<<<<<<
 *     # memrchr seems to be a GNU extension, so we have to implement it ourselves
 *     # It is not present in any win32 standard library
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":97
 * 
 * # TODO: Import this from _dirstate_helpers when it is merged
 * cdef object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_safe_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("safe_string_from_size", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":98
 * # TODO: Import this from _dirstate_helpers when it is merged
 * cdef object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d @0x%x'
 */
  __pyx_t_1 = ((__pyx_v_size < 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":101
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d @0x%x'
 *             % (size, <int>s))             # <<<<<<<<<<<<<<
 *     return PyString_FromStringAndSize(s, size)
 * 
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 101; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(((int)__pyx_v_s)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 101; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 101; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_t_4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 101; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":99
 * cdef object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d @0x%x'
 *             % (size, <int>s))
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 99; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 99; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 99; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":98
 * # TODO: Import this from _dirstate_helpers when it is merged
 * cdef object safe_string_from_size(char *s, Py_ssize_t size):
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d @0x%x'
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":102
 *             'tried to create a string with an invalid size: %d @0x%x'
 *             % (size, <int>s))
 *     return PyString_FromStringAndSize(s, size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyString_FromStringAndSize(__pyx_v_s, __pyx_v_size); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":97
 * 
 * # TODO: Import this from _dirstate_helpers when it is merged
 * cdef object safe_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     if size < 0:
 *         raise AssertionError(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.safe_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":105
 * 
 * 
 * cdef object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     cdef PyObject *py_str
 *     if size < 0:
 */

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_safe_interned_string_from_size(char *__pyx_v_s, Py_ssize_t __pyx_v_size) {
  PyObject *__pyx_v_py_str;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("safe_interned_string_from_size", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":107
 * cdef object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     cdef PyObject *py_str
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d @0x%x'
 */
  __pyx_t_1 = ((__pyx_v_size < 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":110
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d @0x%x'
 *             % (size, <int>s))             # <<<<<<<<<<<<<<
 *     py_str = PyString_FromStringAndSize_ptr(s, size)
 *     PyString_InternInPlace(&py_str)
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_size); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 110; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int(((int)__pyx_v_s)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 110; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 110; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_t_4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 110; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":108
 *     cdef PyObject *py_str
 *     if size < 0:
 *         raise AssertionError(             # <<<<<<<<<<<<<<
 *             'tried to create a string with an invalid size: %d @0x%x'
 *             % (size, <int>s))
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 108; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 108; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 108; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":107
 * cdef object safe_interned_string_from_size(char *s, Py_ssize_t size):
 *     cdef PyObject *py_str
 *     if size < 0:             # <<<<<<<<<<<<<<
 *         raise AssertionError(
 *             'tried to create a string with an invalid size: %d @0x%x'
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":111
 *             'tried to create a string with an invalid size: %d @0x%x'
 *             % (size, <int>s))
 *     py_str = PyString_FromStringAndSize_ptr(s, size)             # <<<<<<<<<<<<<<
 *     PyString_InternInPlace(&py_str)
 *     result = <object>py_str
 */
  __pyx_v_py_str = PyString_FromStringAndSize(__pyx_v_s, __pyx_v_size);

  /* "bzrlib/_btree_serializer_pyx.pyx":112
 *             % (size, <int>s))
 *     py_str = PyString_FromStringAndSize_ptr(s, size)
 *     PyString_InternInPlace(&py_str)             # <<<<<<<<<<<<<<
 *     result = <object>py_str
 *     # Casting a PyObject* to an <object> triggers an INCREF from Pyrex, so we
 */
  PyString_InternInPlace((&__pyx_v_py_str));

  /* "bzrlib/_btree_serializer_pyx.pyx":113
 *     py_str = PyString_FromStringAndSize_ptr(s, size)
 *     PyString_InternInPlace(&py_str)
 *     result = <object>py_str             # <<<<<<<<<<<<<<
 *     # Casting a PyObject* to an <object> triggers an INCREF from Pyrex, so we
 *     # DECREF it to avoid geting immortal strings
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_py_str);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":116
 *     # Casting a PyObject* to an <object> triggers an INCREF from Pyrex, so we
 *     # DECREF it to avoid geting immortal strings
 *     Py_DECREF_ptr(py_str)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  Py_DECREF(__pyx_v_py_str);

  /* "bzrlib/_btree_serializer_pyx.pyx":117
 *     # DECREF it to avoid geting immortal strings
 *     Py_DECREF_ptr(py_str)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * # This sets up the StaticTuple C_API functionality
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":105
 * 
 * 
 * cdef object safe_interned_string_from_size(char *s, Py_ssize_t size):             # <<<<<<<<<<<<<<
 *     cdef PyObject *py_str
 *     if size < 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.safe_interned_string_from_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":153
 *     cdef int _header_found
 * 
 *     def __init__(self, bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *         self.bytes = bytes
 *         self.key_length = key_length
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytes = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bytes,&__pyx_n_s_key_length,&__pyx_n_s_ref_list_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bytes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_key_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 153; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ref_list_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 153; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 153; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_bytes = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 153; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.BTreeLeafParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser___init__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self), __pyx_v_bytes, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser___init__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":154
 * 
 *     def __init__(self, bytes, key_length, ref_list_length):
 *         self.bytes = bytes             # <<<<<<<<<<<<<<
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length
 */
  __Pyx_INCREF(__pyx_v_bytes);
  __Pyx_GIVEREF(__pyx_v_bytes);
  __Pyx_GOTREF(__pyx_v_self->bytes);
  __Pyx_DECREF(__pyx_v_self->bytes);
  __pyx_v_self->bytes = __pyx_v_bytes;

  /* "bzrlib/_btree_serializer_pyx.pyx":155
 *     def __init__(self, bytes, key_length, ref_list_length):
 *         self.bytes = bytes
 *         self.key_length = key_length             # <<<<<<<<<<<<<<
 *         self.ref_list_length = ref_list_length
 *         self.keys = []
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_key_length); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 155; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->key_length = __pyx_t_1;

  /* "bzrlib/_btree_serializer_pyx.pyx":156
 *         self.bytes = bytes
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length             # <<<<<<<<<<<<<<
 *         self.keys = []
 *         self._cur_str = NULL
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_ref_list_length); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 156; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->ref_list_length = __pyx_t_1;

  /* "bzrlib/_btree_serializer_pyx.pyx":157
 *         self.key_length = key_length
 *         self.ref_list_length = ref_list_length
 *         self.keys = []             # <<<<<<<<<<<<<<
 *         self._cur_str = NULL
 *         self._end_str = NULL
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 157; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->keys);
  __Pyx_DECREF(__pyx_v_self->keys);
  __pyx_v_self->keys = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":158
 *         self.ref_list_length = ref_list_length
 *         self.keys = []
 *         self._cur_str = NULL             # <<<<<<<<<<<<<<
 *         self._end_str = NULL
 *         self._header_found = 0
 */
  __pyx_v_self->_cur_str = NULL;

  /* "bzrlib/_btree_serializer_pyx.pyx":159
 *         self.keys = []
 *         self._cur_str = NULL
 *         self._end_str = NULL             # <<<<<<<<<<<<<<
 *         self._header_found = 0
 *         # keys are tuples
 */
  __pyx_v_self->_end_str = NULL;

  /* "bzrlib/_btree_serializer_pyx.pyx":160
 *         self._cur_str = NULL
 *         self._end_str = NULL
 *         self._header_found = 0             # <<<<<<<<<<<<<<
 *         # keys are tuples
 * 
 */
  __pyx_v_self->_header_found = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":153
 *     cdef int _header_found
 * 
 *     def __init__(self, bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *         self.bytes = bytes
 *         self.key_length = key_length
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.BTreeLeafParser.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":163
 *         # keys are tuples
 * 
 *     cdef extract_key(self, char * last):             # <<<<<<<<<<<<<<
 *         """Extract a key.
 * 
 */

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_extract_key(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self, char *__pyx_v_last) {
  char *__pyx_v_temp_ptr;
  int __pyx_v_loop_counter;
  StaticTuple *__pyx_v_key = 0;
  PyObject *__pyx_v_failure_string = NULL;
  PyObject *__pyx_v_key_element = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("extract_key", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":173
 *         cdef StaticTuple key
 * 
 *         key = StaticTuple_New(self.key_length)             # <<<<<<<<<<<<<<
 *         for loop_counter from 0 <= loop_counter < self.key_length:
 *             # grab a key segment
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(__pyx_v_self->key_length)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":174
 * 
 *         key = StaticTuple_New(self.key_length)
 *         for loop_counter from 0 <= loop_counter < self.key_length:             # <<<<<<<<<<<<<<
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 */
  __pyx_t_2 = __pyx_v_self->key_length;
  for (__pyx_v_loop_counter = 0; __pyx_v_loop_counter < __pyx_t_2; __pyx_v_loop_counter++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":176
 *         for loop_counter from 0 <= loop_counter < self.key_length:
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)             # <<<<<<<<<<<<<<
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:
 */
    __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\x00', (__pyx_v_last - __pyx_v_self->_start)));

    /* "bzrlib/_btree_serializer_pyx.pyx":177
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
 */
    __pyx_t_3 = ((__pyx_v_temp_ptr == NULL) != 0);
    if (__pyx_t_3) {

      /* "bzrlib/_btree_serializer_pyx.pyx":178
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:             # <<<<<<<<<<<<<<
 *                     # capture to last
 *                     temp_ptr = last
 */
      __pyx_t_3 = (((__pyx_v_loop_counter + 1) == __pyx_v_self->key_length) != 0);
      if (__pyx_t_3) {

        /* "bzrlib/_btree_serializer_pyx.pyx":180
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
 *                     temp_ptr = last             # <<<<<<<<<<<<<<
 *                 else:
 *                     # Invalid line
 */
        __pyx_v_temp_ptr = __pyx_v_last;

        /* "bzrlib/_btree_serializer_pyx.pyx":178
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:
 *                 if loop_counter + 1 == self.key_length:             # <<<<<<<<<<<<<<
 *                     # capture to last
 *                     temp_ptr = last
 */
        goto __pyx_L6;
      }

      /* "bzrlib/_btree_serializer_pyx.pyx":183
 *                 else:
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +             # <<<<<<<<<<<<<<
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
 */
      /*else*/ {

        /* "bzrlib/_btree_serializer_pyx.pyx":184
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +
 *                         repr(safe_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                    last - self._start)))
 *                     raise AssertionError(failure_string)
 */
        __pyx_t_1 = __pyx_f_6bzrlib_21_btree_serializer_pyx_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 184; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "bzrlib/_btree_serializer_pyx.pyx":183
 *                 else:
 *                     # Invalid line
 *                     failure_string = ("invalid key, wanted segment from " +             # <<<<<<<<<<<<<<
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
 */
        __pyx_t_1 = PyNumber_Add(__pyx_kp_s_invalid_key_wanted_segment_from, __pyx_t_4); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 183; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_failure_string = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "bzrlib/_btree_serializer_pyx.pyx":186
 *                         repr(safe_string_from_size(self._start,
 *                                                    last - self._start)))
 *                     raise AssertionError(failure_string)             # <<<<<<<<<<<<<<
 *             # capture the key string
 *             if (self.key_length == 1
 */
        __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 186; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_v_failure_string);
        __Pyx_GIVEREF(__pyx_v_failure_string);
        PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_failure_string);
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 186; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 186; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      __pyx_L6:;

      /* "bzrlib/_btree_serializer_pyx.pyx":177
 *             # grab a key segment
 *             temp_ptr = <char*>memchr(self._start, c'\0', last - self._start)
 *             if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                 if loop_counter + 1 == self.key_length:
 *                     # capture to last
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":188
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, 'sha1:', 5) == 0):
 */
    __pyx_t_5 = ((__pyx_v_self->key_length == 1) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":189
 *             # capture the key string
 *             if (self.key_length == 1
 *                 and (temp_ptr - self._start) == 45             # <<<<<<<<<<<<<<
 *                 and strncmp(self._start, 'sha1:', 5) == 0):
 *                 key_element = safe_string_from_size(self._start,
 */
    __pyx_t_5 = (((__pyx_v_temp_ptr - __pyx_v_self->_start) == 45) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":190
 *             if (self.key_length == 1
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, 'sha1:', 5) == 0):             # <<<<<<<<<<<<<<
 *                 key_element = safe_string_from_size(self._start,
 *                                                     temp_ptr - self._start)
 */
    __pyx_t_5 = ((strncmp(__pyx_v_self->_start, __pyx_k_sha1, 5) == 0) != 0);
    __pyx_t_3 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;

    /* "bzrlib/_btree_serializer_pyx.pyx":188
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, 'sha1:', 5) == 0):
 */
    if (__pyx_t_3) {

      /* "bzrlib/_btree_serializer_pyx.pyx":191
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, 'sha1:', 5) == 0):
 *                 key_element = safe_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                     temp_ptr - self._start)
 *             else:
 */
      __pyx_t_4 = __pyx_f_6bzrlib_21_btree_serializer_pyx_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_temp_ptr - __pyx_v_self->_start)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 191; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_key_element, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":188
 *                     raise AssertionError(failure_string)
 *             # capture the key string
 *             if (self.key_length == 1             # <<<<<<<<<<<<<<
 *                 and (temp_ptr - self._start) == 45
 *                 and strncmp(self._start, 'sha1:', 5) == 0):
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":194
 *                                                     temp_ptr - self._start)
 *             else:
 *                 key_element = safe_interned_string_from_size(self._start,             # <<<<<<<<<<<<<<
 *                                                          temp_ptr - self._start)
 *             # advance our pointer
 */
    /*else*/ {

      /* "bzrlib/_btree_serializer_pyx.pyx":195
 *             else:
 *                 key_element = safe_interned_string_from_size(self._start,
 *                                                          temp_ptr - self._start)             # <<<<<<<<<<<<<<
 *             # advance our pointer
 *             self._start = temp_ptr + 1
 */
      __pyx_t_4 = __pyx_f_6bzrlib_21_btree_serializer_pyx_safe_interned_string_from_size(__pyx_v_self->_start, (__pyx_v_temp_ptr - __pyx_v_self->_start)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 194; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_key_element, __pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L7:;

    /* "bzrlib/_btree_serializer_pyx.pyx":197
 *                                                          temp_ptr - self._start)
 *             # advance our pointer
 *             self._start = temp_ptr + 1             # <<<<<<<<<<<<<<
 *             Py_INCREF(key_element)
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 */
    __pyx_v_self->_start = (__pyx_v_temp_ptr + 1);

    /* "bzrlib/_btree_serializer_pyx.pyx":198
 *             # advance our pointer
 *             self._start = temp_ptr + 1
 *             Py_INCREF(key_element)             # <<<<<<<<<<<<<<
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 *         key = StaticTuple_Intern(key)
 */
    Py_INCREF(__pyx_v_key_element);

    /* "bzrlib/_btree_serializer_pyx.pyx":199
 *             self._start = temp_ptr + 1
 *             Py_INCREF(key_element)
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)             # <<<<<<<<<<<<<<
 *         key = StaticTuple_Intern(key)
 *         return key
 */
    StaticTuple_SET_ITEM(__pyx_v_key, __pyx_v_loop_counter, __pyx_v_key_element);
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":200
 *             Py_INCREF(key_element)
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 *         key = StaticTuple_Intern(key)             # <<<<<<<<<<<<<<
 *         return key
 * 
 */
  __pyx_t_4 = ((PyObject *)StaticTuple_Intern(__pyx_v_key)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF_SET(__pyx_v_key, ((StaticTuple *)__pyx_t_4));
  __pyx_t_4 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":201
 *             StaticTuple_SET_ITEM(key, loop_counter, key_element)
 *         key = StaticTuple_Intern(key)
 *         return key             # <<<<<<<<<<<<<<
 * 
 *     cdef int process_line(self) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_key));
  __pyx_r = ((PyObject *)__pyx_v_key);
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":163
 *         # keys are tuples
 * 
 *     cdef extract_key(self, char * last):             # <<<<<<<<<<<<<<
 *         """Extract a key.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.BTreeLeafParser.extract_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_failure_string);
  __Pyx_XDECREF(__pyx_v_key_element);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":203
 *         return key
 * 
 *     cdef int process_line(self) except -1:             # <<<<<<<<<<<<<<
 *         """Process a line in the bytes."""
 *         cdef char *last
 */

static int __pyx_f_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_process_line(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  char *__pyx_v_last;
  char *__pyx_v_temp_ptr;
  char *__pyx_v_ref_ptr;
  char *__pyx_v_next_start;
  int __pyx_v_loop_counter;
  Py_ssize_t __pyx_v_str_len;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  StaticTuple *__pyx_v_ref_lists = NULL;
  PyObject *__pyx_v_ref_list = NULL;
  StaticTuple *__pyx_v_node_value = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("process_line", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":212
 *         cdef Py_ssize_t str_len
 * 
 *         self._start = self._cur_str             # <<<<<<<<<<<<<<
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 */
  __pyx_t_1 = __pyx_v_self->_cur_str;
  __pyx_v_self->_start = __pyx_t_1;

  /* "bzrlib/_btree_serializer_pyx.pyx":214
 *         self._start = self._cur_str
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)             # <<<<<<<<<<<<<<
 *         if last == NULL:
 *             # Process until the end of the file
 */
  __pyx_v_last = ((char *)memchr(__pyx_v_self->_start, '\n', (__pyx_v_self->_end_str - __pyx_v_self->_start)));

  /* "bzrlib/_btree_serializer_pyx.pyx":215
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 *         if last == NULL:             # <<<<<<<<<<<<<<
 *             # Process until the end of the file
 *             last = self._end_str
 */
  __pyx_t_2 = ((__pyx_v_last == NULL) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":217
 *         if last == NULL:
 *             # Process until the end of the file
 *             last = self._end_str             # <<<<<<<<<<<<<<
 *             self._cur_str = self._end_str
 *         else:
 */
    __pyx_t_1 = __pyx_v_self->_end_str;
    __pyx_v_last = __pyx_t_1;

    /* "bzrlib/_btree_serializer_pyx.pyx":218
 *             # Process until the end of the file
 *             last = self._end_str
 *             self._cur_str = self._end_str             # <<<<<<<<<<<<<<
 *         else:
 *             # And the next string is right after it
 */
    __pyx_t_1 = __pyx_v_self->_end_str;
    __pyx_v_self->_cur_str = __pyx_t_1;

    /* "bzrlib/_btree_serializer_pyx.pyx":215
 *         # Find the next newline
 *         last = <char*>memchr(self._start, c'\n', self._end_str - self._start)
 *         if last == NULL:             # <<<<<<<<<<<<<<
 *             # Process until the end of the file
 *             last = self._end_str
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":221
 *         else:
 *             # And the next string is right after it
 *             self._cur_str = last + 1             # <<<<<<<<<<<<<<
 *             # The last character is right before the '\n'
 * 
 */
  /*else*/ {
    __pyx_v_self->_cur_str = (__pyx_v_last + 1);
  }
  __pyx_L3:;

  /* "bzrlib/_btree_serializer_pyx.pyx":224
 *             # The last character is right before the '\n'
 * 
 *         if last == self._start:             # <<<<<<<<<<<<<<
 *             # parsed it all.
 *             return 0
 */
  __pyx_t_2 = ((__pyx_v_last == __pyx_v_self->_start) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":226
 *         if last == self._start:
 *             # parsed it all.
 *             return 0             # <<<<<<<<<<<<<<
 *         if last < self._start:
 *             # Unexpected error condition - fail
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_btree_serializer_pyx.pyx":224
 *             # The last character is right before the '\n'
 * 
 *         if last == self._start:             # <<<<<<<<<<<<<<
 *             # parsed it all.
 *             return 0
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":227
 *             # parsed it all.
 *             return 0
 *         if last < self._start:             # <<<<<<<<<<<<<<
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 */
  __pyx_t_2 = ((__pyx_v_last < __pyx_v_self->_start) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":229
 *         if last < self._start:
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":227
 *             # parsed it all.
 *             return 0
 *         if last < self._start:             # <<<<<<<<<<<<<<
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":230
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 *         if 0 == self._header_found:             # <<<<<<<<<<<<<<
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp("type=leaf", self._start, last - self._start) == 0:
 */
  __pyx_t_2 = ((0 == __pyx_v_self->_header_found) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":232
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp("type=leaf", self._start, last - self._start) == 0:             # <<<<<<<<<<<<<<
 *                 self._header_found = 1
 *                 return 0
 */
    __pyx_t_2 = ((strncmp(__pyx_k_type_leaf, __pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)) == 0) != 0);
    if (__pyx_t_2) {

      /* "bzrlib/_btree_serializer_pyx.pyx":233
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp("type=leaf", self._start, last - self._start) == 0:
 *                 self._header_found = 1             # <<<<<<<<<<<<<<
 *                 return 0
 *             else:
 */
      __pyx_v_self->_header_found = 1;

      /* "bzrlib/_btree_serializer_pyx.pyx":234
 *             if strncmp("type=leaf", self._start, last - self._start) == 0:
 *                 self._header_found = 1
 *                 return 0             # <<<<<<<<<<<<<<
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "bzrlib/_btree_serializer_pyx.pyx":232
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp("type=leaf", self._start, last - self._start) == 0:             # <<<<<<<<<<<<<<
 *                 self._header_found = 1
 *                 return 0
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":236
 *                 return 0
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'             # <<<<<<<<<<<<<<
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
 */
    /*else*/ {

      /* "bzrlib/_btree_serializer_pyx.pyx":237
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'
 *                     % (safe_string_from_size(self._start, last - self._start)))             # <<<<<<<<<<<<<<
 * 
 *         key = self.extract_key(last)
 */
      __pyx_t_3 = __pyx_f_6bzrlib_21_btree_serializer_pyx_safe_string_from_size(__pyx_v_self->_start, (__pyx_v_last - __pyx_v_self->_start)); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 237; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Node_did_not_start_with_type_lea, __pyx_t_3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 237; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":236
 *                 return 0
 *             else:
 *                 raise AssertionError('Node did not start with "type=leaf": %r'             # <<<<<<<<<<<<<<
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
 */
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":230
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")
 *         if 0 == self._header_found:             # <<<<<<<<<<<<<<
 *             # The first line in a leaf node is the header "type=leaf\n"
 *             if strncmp("type=leaf", self._start, last - self._start) == 0:
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":239
 *                     % (safe_string_from_size(self._start, last - self._start)))
 * 
 *         key = self.extract_key(last)             # <<<<<<<<<<<<<<
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->extract_key(__pyx_v_self, __pyx_v_last); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 239; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_key = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":241
 *         key = self.extract_key(last)
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)             # <<<<<<<<<<<<<<
 *         if temp_ptr == NULL:
 *             # Invalid line
 */
  __pyx_v_temp_ptr = ((char *)__pyx_f_6bzrlib_21_btree_serializer_pyx__my_memrchr(__pyx_v_self->_start, '\x00', (__pyx_v_last - __pyx_v_self->_start)));

  /* "bzrlib/_btree_serializer_pyx.pyx":242
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 *         if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")
 */
  __pyx_t_2 = ((__pyx_v_temp_ptr == NULL) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":244
 *         if temp_ptr == NULL:
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")             # <<<<<<<<<<<<<<
 *         else:
 *             # Because of how conversions were done, we ended up with *lots* of
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":242
 *         # find the value area
 *         temp_ptr = <char*>_my_memrchr(self._start, c'\0', last - self._start)
 *         if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":252
 *             # something like '12607215 328306 0 0', which ends up consuming 1MB
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1             # <<<<<<<<<<<<<<
 *             if (str_len > 4
 *                 and strncmp(" 0 0", last - 4, 4) == 0):
 */
  /*else*/ {
    __pyx_v_str_len = ((__pyx_v_last - __pyx_v_temp_ptr) - 1);

    /* "bzrlib/_btree_serializer_pyx.pyx":253
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 */
    __pyx_t_5 = ((__pyx_v_str_len > 4) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L10_bool_binop_done;
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":254
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4
 *                 and strncmp(" 0 0", last - 4, 4) == 0):             # <<<<<<<<<<<<<<
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 *                 # For Launchpad 236MB => 232MB
 */
    __pyx_t_5 = ((strncmp(__pyx_k_0_0, (__pyx_v_last - 4), 4) == 0) != 0);
    __pyx_t_2 = __pyx_t_5;
    __pyx_L10_bool_binop_done:;

    /* "bzrlib/_btree_serializer_pyx.pyx":253
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 */
    if (__pyx_t_2) {

      /* "bzrlib/_btree_serializer_pyx.pyx":257
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 *                 # For Launchpad 236MB => 232MB
 *                 value = safe_interned_string_from_size(temp_ptr + 1, str_len)             # <<<<<<<<<<<<<<
 *             else:
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)
 */
      __pyx_t_4 = __pyx_f_6bzrlib_21_btree_serializer_pyx_safe_interned_string_from_size((__pyx_v_temp_ptr + 1), __pyx_v_str_len); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 257; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_value = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":253
 *             # of memory, just for those strings.
 *             str_len = last - temp_ptr - 1
 *             if (str_len > 4             # <<<<<<<<<<<<<<
 *                 and strncmp(" 0 0", last - 4, 4) == 0):
 *                 # This drops peak mem for bzr.dev from 87.4MB => 86.2MB
 */
      goto __pyx_L9;
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":259
 *                 value = safe_interned_string_from_size(temp_ptr + 1, str_len)
 *             else:
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)             # <<<<<<<<<<<<<<
 *             # shrink the references end point
 *             last = temp_ptr
 */
    /*else*/ {
      __pyx_t_4 = __pyx_f_6bzrlib_21_btree_serializer_pyx_safe_string_from_size((__pyx_v_temp_ptr + 1), __pyx_v_str_len); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 259; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_value = __pyx_t_4;
      __pyx_t_4 = 0;
    }
    __pyx_L9:;

    /* "bzrlib/_btree_serializer_pyx.pyx":261
 *                 value = safe_string_from_size(temp_ptr + 1, str_len)
 *             # shrink the references end point
 *             last = temp_ptr             # <<<<<<<<<<<<<<
 * 
 *         if self.ref_list_length:
 */
    __pyx_v_last = __pyx_v_temp_ptr;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":263
 *             last = temp_ptr
 * 
 *         if self.ref_list_length:             # <<<<<<<<<<<<<<
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0
 */
  __pyx_t_2 = (__pyx_v_self->ref_list_length != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":264
 * 
 *         if self.ref_list_length:
 *             ref_lists = StaticTuple_New(self.ref_list_length)             # <<<<<<<<<<<<<<
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:
 */
    __pyx_t_4 = ((PyObject *)StaticTuple_New(__pyx_v_self->ref_list_length)); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 264; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_ref_lists = ((StaticTuple *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":265
 *         if self.ref_list_length:
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0             # <<<<<<<<<<<<<<
 *             while loop_counter < self.ref_list_length:
 *                 ref_list = []
 */
    __pyx_v_loop_counter = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":266
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:             # <<<<<<<<<<<<<<
 *                 ref_list = []
 *                 # extract a reference list
 */
    while (1) {
      __pyx_t_2 = ((__pyx_v_loop_counter < __pyx_v_self->ref_list_length) != 0);
      if (!__pyx_t_2) break;

      /* "bzrlib/_btree_serializer_pyx.pyx":267
 *             loop_counter = 0
 *             while loop_counter < self.ref_list_length:
 *                 ref_list = []             # <<<<<<<<<<<<<<
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 */
      __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 267; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":269
 *                 ref_list = []
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1             # <<<<<<<<<<<<<<
 *                 if last < self._start:
 *                     raise AssertionError("last < self._start")
 */
      __pyx_v_loop_counter = (__pyx_v_loop_counter + 1);

      /* "bzrlib/_btree_serializer_pyx.pyx":270
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:             # <<<<<<<<<<<<<<
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
 */
      __pyx_t_2 = ((__pyx_v_last < __pyx_v_self->_start) != 0);
      if (__pyx_t_2) {

        /* "bzrlib/_btree_serializer_pyx.pyx":271
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:
 *                     raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

        /* "bzrlib/_btree_serializer_pyx.pyx":270
 *                 # extract a reference list
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:             # <<<<<<<<<<<<<<
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
 */
      }

      /* "bzrlib/_btree_serializer_pyx.pyx":273
 *                     raise AssertionError("last < self._start")
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)             # <<<<<<<<<<<<<<
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 */
      __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\t', (__pyx_v_last - __pyx_v_self->_start)));

      /* "bzrlib/_btree_serializer_pyx.pyx":274
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 *                 if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:
 */
      __pyx_t_2 = ((__pyx_v_temp_ptr == NULL) != 0);
      if (__pyx_t_2) {

        /* "bzrlib/_btree_serializer_pyx.pyx":276
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:             # <<<<<<<<<<<<<<
 *                         # Invalid line
 *                         raise AssertionError(
 */
        __pyx_t_2 = ((__pyx_v_loop_counter != __pyx_v_self->ref_list_length) != 0);
        if (__pyx_t_2) {

          /* "bzrlib/_btree_serializer_pyx.pyx":278
 *                     if loop_counter != self.ref_list_length:
 *                         # Invalid line
 *                         raise AssertionError(             # <<<<<<<<<<<<<<
 *                             "invalid key, loop_counter != self.ref_list_length")
 *                     else:
 */
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 278; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          {__pyx_filename = __pyx_f[0]; __pyx_lineno = 278; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

          /* "bzrlib/_btree_serializer_pyx.pyx":276
 *                 if temp_ptr == NULL:
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:             # <<<<<<<<<<<<<<
 *                         # Invalid line
 *                         raise AssertionError(
 */
        }

        /* "bzrlib/_btree_serializer_pyx.pyx":282
 *                     else:
 *                         # scan to the end of the ref list area
 *                         ref_ptr = last             # <<<<<<<<<<<<<<
 *                         next_start = last
 *                 else:
 */
        /*else*/ {
          __pyx_v_ref_ptr = __pyx_v_last;

          /* "bzrlib/_btree_serializer_pyx.pyx":283
 *                         # scan to the end of the ref list area
 *                         ref_ptr = last
 *                         next_start = last             # <<<<<<<<<<<<<<
 *                 else:
 *                     # scan to the end of this ref list
 */
          __pyx_v_next_start = __pyx_v_last;
        }

        /* "bzrlib/_btree_serializer_pyx.pyx":274
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 *                 if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                     # Only valid for the last list
 *                     if loop_counter != self.ref_list_length:
 */
        goto __pyx_L16;
      }

      /* "bzrlib/_btree_serializer_pyx.pyx":286
 *                 else:
 *                     # scan to the end of this ref list
 *                     ref_ptr = temp_ptr             # <<<<<<<<<<<<<<
 *                     next_start = temp_ptr + 1
 *                 # Now, there may be multiple keys in the ref list.
 */
      /*else*/ {
        __pyx_v_ref_ptr = __pyx_v_temp_ptr;

        /* "bzrlib/_btree_serializer_pyx.pyx":287
 *                     # scan to the end of this ref list
 *                     ref_ptr = temp_ptr
 *                     next_start = temp_ptr + 1             # <<<<<<<<<<<<<<
 *                 # Now, there may be multiple keys in the ref list.
 *                 while self._start < ref_ptr:
 */
        __pyx_v_next_start = (__pyx_v_temp_ptr + 1);
      }
      __pyx_L16:;

      /* "bzrlib/_btree_serializer_pyx.pyx":289
 *                     next_start = temp_ptr + 1
 *                 # Now, there may be multiple keys in the ref list.
 *                 while self._start < ref_ptr:             # <<<<<<<<<<<<<<
 *                     # loop finding keys and extracting them
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 */
      while (1) {
        __pyx_t_2 = ((__pyx_v_self->_start < __pyx_v_ref_ptr) != 0);
        if (!__pyx_t_2) break;

        /* "bzrlib/_btree_serializer_pyx.pyx":291
 *                 while self._start < ref_ptr:
 *                     # loop finding keys and extracting them
 *                     temp_ptr = <char*>memchr(self._start, c'\r',             # <<<<<<<<<<<<<<
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:
 */
        __pyx_v_temp_ptr = ((char *)memchr(__pyx_v_self->_start, '\r', (__pyx_v_ref_ptr - __pyx_v_self->_start)));

        /* "bzrlib/_btree_serializer_pyx.pyx":293
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr
 */
        __pyx_t_2 = ((__pyx_v_temp_ptr == NULL) != 0);
        if (__pyx_t_2) {

          /* "bzrlib/_btree_serializer_pyx.pyx":295
 *                     if temp_ptr == NULL:
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr             # <<<<<<<<<<<<<<
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 */
          __pyx_v_temp_ptr = __pyx_v_ref_ptr;

          /* "bzrlib/_btree_serializer_pyx.pyx":293
 *                     temp_ptr = <char*>memchr(self._start, c'\r',
 *                                              ref_ptr - self._start)
 *                     if temp_ptr == NULL:             # <<<<<<<<<<<<<<
 *                         # key runs to the end
 *                         temp_ptr = ref_ptr
 */
        }

        /* "bzrlib/_btree_serializer_pyx.pyx":297
 *                         temp_ptr = ref_ptr
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))             # <<<<<<<<<<<<<<
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))
 *                 Py_INCREF(ref_list)
 */
        __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->extract_key(__pyx_v_self, __pyx_v_temp_ptr); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = PyList_Append(__pyx_v_ref_list, __pyx_t_4); if (unlikely(__pyx_t_6 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 297; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }

      /* "bzrlib/_btree_serializer_pyx.pyx":298
 * 
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))             # <<<<<<<<<<<<<<
 *                 Py_INCREF(ref_list)
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 */
      __pyx_t_4 = PySequence_Tuple(__pyx_v_ref_list); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = ((PyObject *)StaticTuple_Intern(((StaticTuple *)__pyx_t_3))); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_ref_list, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":299
 *                     PyList_Append(ref_list, self.extract_key(temp_ptr))
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))
 *                 Py_INCREF(ref_list)             # <<<<<<<<<<<<<<
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 *                 # prepare for the next reference list
 */
      Py_INCREF(__pyx_v_ref_list);

      /* "bzrlib/_btree_serializer_pyx.pyx":300
 *                 ref_list = StaticTuple_Intern(StaticTuple(*ref_list))
 *                 Py_INCREF(ref_list)
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)             # <<<<<<<<<<<<<<
 *                 # prepare for the next reference list
 *                 self._start = next_start
 */
      StaticTuple_SET_ITEM(__pyx_v_ref_lists, (__pyx_v_loop_counter - 1), __pyx_v_ref_list);

      /* "bzrlib/_btree_serializer_pyx.pyx":302
 *                 StaticTuple_SET_ITEM(ref_lists, loop_counter - 1, ref_list)
 *                 # prepare for the next reference list
 *                 self._start = next_start             # <<<<<<<<<<<<<<
 *             node_value = StaticTuple(value, ref_lists)
 *         else:
 */
      __pyx_v_self->_start = __pyx_v_next_start;
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":303
 *                 # prepare for the next reference list
 *                 self._start = next_start
 *             node_value = StaticTuple(value, ref_lists)             # <<<<<<<<<<<<<<
 *         else:
 *             if last != self._start:
 */
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 303; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_value);
    __Pyx_INCREF(((PyObject *)__pyx_v_ref_lists));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_ref_lists));
    PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_ref_lists));
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 303; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_node_value = ((StaticTuple *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":263
 *             last = temp_ptr
 * 
 *         if self.ref_list_length:             # <<<<<<<<<<<<<<
 *             ref_lists = StaticTuple_New(self.ref_list_length)
 *             loop_counter = 0
 */
    goto __pyx_L12;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":305
 *             node_value = StaticTuple(value, ref_lists)
 *         else:
 *             if last != self._start:             # <<<<<<<<<<<<<<
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
 */
  /*else*/ {
    __pyx_t_2 = ((__pyx_v_last != __pyx_v_self->_start) != 0);
    if (__pyx_t_2) {

      /* "bzrlib/_btree_serializer_pyx.pyx":307
 *             if last != self._start:
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")             # <<<<<<<<<<<<<<
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 307; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 307; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

      /* "bzrlib/_btree_serializer_pyx.pyx":305
 *             node_value = StaticTuple(value, ref_lists)
 *         else:
 *             if last != self._start:             # <<<<<<<<<<<<<<
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":308
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")
 *             node_value = StaticTuple(value, StaticTuple())             # <<<<<<<<<<<<<<
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 *         return 0
 */
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 308; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 308; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_value);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 308; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_node_value = ((StaticTuple *)__pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L12:;

  /* "bzrlib/_btree_serializer_pyx.pyx":309
 *                 raise AssertionError("unexpected reference data present")
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = __pyx_v_self->keys;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 309; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_key);
  __Pyx_GIVEREF(__pyx_v_key);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_key);
  __Pyx_INCREF(((PyObject *)__pyx_v_node_value));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_node_value));
  PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_node_value));
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple), __pyx_t_4, NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 309; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = PyList_Append(__pyx_t_3, __pyx_t_7); if (unlikely(__pyx_t_6 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 309; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":310
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def parse(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":203
 *         return key
 * 
 *     cdef int process_line(self) except -1:             # <<<<<<<<<<<<<<
 *         """Process a line in the bytes."""
 *         cdef char *last
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.BTreeLeafParser.process_line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_ref_lists);
  __Pyx_XDECREF(__pyx_v_ref_list);
  __Pyx_XDECREF((PyObject *)__pyx_v_node_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":312
 *         return 0
 * 
 *     def parse(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t byte_count
 *         if not PyString_CheckExact(self.bytes):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_3parse(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_3parse(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parse (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_2parse(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_2parse(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_self) {
  Py_ssize_t __pyx_v_byte_count;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  char *__pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parse", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":314
 *     def parse(self):
 *         cdef Py_ssize_t byte_count
 *         if not PyString_CheckExact(self.bytes):             # <<<<<<<<<<<<<<
 *             raise AssertionError('self.bytes is not a string.')
 *         byte_count = PyString_Size(self.bytes)
 */
  __pyx_t_1 = __pyx_v_self->bytes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((!(PyString_CheckExact(__pyx_t_1) != 0)) != 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":315
 *         cdef Py_ssize_t byte_count
 *         if not PyString_CheckExact(self.bytes):
 *             raise AssertionError('self.bytes is not a string.')             # <<<<<<<<<<<<<<
 *         byte_count = PyString_Size(self.bytes)
 *         self._cur_str = PyString_AsString(self.bytes)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 315; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 315; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":314
 *     def parse(self):
 *         cdef Py_ssize_t byte_count
 *         if not PyString_CheckExact(self.bytes):             # <<<<<<<<<<<<<<
 *             raise AssertionError('self.bytes is not a string.')
 *         byte_count = PyString_Size(self.bytes)
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":316
 *         if not PyString_CheckExact(self.bytes):
 *             raise AssertionError('self.bytes is not a string.')
 *         byte_count = PyString_Size(self.bytes)             # <<<<<<<<<<<<<<
 *         self._cur_str = PyString_AsString(self.bytes)
 *         # This points to the last character in the string
 */
  __pyx_t_1 = __pyx_v_self->bytes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_byte_count = PyString_Size(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":317
 *             raise AssertionError('self.bytes is not a string.')
 *         byte_count = PyString_Size(self.bytes)
 *         self._cur_str = PyString_AsString(self.bytes)             # <<<<<<<<<<<<<<
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count
 */
  __pyx_t_1 = __pyx_v_self->bytes;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = PyString_AsString(__pyx_t_1); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 317; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->_cur_str = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":319
 *         self._cur_str = PyString_AsString(self.bytes)
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count             # <<<<<<<<<<<<<<
 *         while self._cur_str < self._end_str:
 *             self.process_line()
 */
  __pyx_v_self->_end_str = (__pyx_v_self->_cur_str + __pyx_v_byte_count);

  /* "bzrlib/_btree_serializer_pyx.pyx":320
 *         # This points to the last character in the string
 *         self._end_str = self._cur_str + byte_count
 *         while self._cur_str < self._end_str:             # <<<<<<<<<<<<<<
 *             self.process_line()
 *         return self.keys
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_self->_cur_str < __pyx_v_self->_end_str) != 0);
    if (!__pyx_t_2) break;

    /* "bzrlib/_btree_serializer_pyx.pyx":321
 *         self._end_str = self._cur_str + byte_count
 *         while self._cur_str < self._end_str:
 *             self.process_line()             # <<<<<<<<<<<<<<
 *         return self.keys
 * 
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_v_self->__pyx_vtab)->process_line(__pyx_v_self); if (unlikely(__pyx_t_4 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 321; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":322
 *         while self._cur_str < self._end_str:
 *             self.process_line()
 *         return self.keys             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->keys);
  __pyx_r = __pyx_v_self->keys;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":312
 *         return 0
 * 
 *     def parse(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t byte_count
 *         if not PyString_CheckExact(self.bytes):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.BTreeLeafParser.parse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":325
 * 
 * 
 * def _parse_leaf_lines(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(bytes, key_length, ref_list_length)
 *     return parser.parse()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_1_parse_leaf_lines(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6bzrlib_21_btree_serializer_pyx_1_parse_leaf_lines = {"_parse_leaf_lines", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_1_parse_leaf_lines, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_1_parse_leaf_lines(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytes = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_leaf_lines (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bytes,&__pyx_n_s_key_length,&__pyx_n_s_ref_list_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bytes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_key_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ref_list_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_parse_leaf_lines") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_bytes = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_leaf_lines", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._parse_leaf_lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx__parse_leaf_lines(__pyx_self, __pyx_v_bytes, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx__parse_leaf_lines(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *__pyx_v_parser = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_leaf_lines", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":326
 * 
 * def _parse_leaf_lines(bytes, key_length, ref_list_length):
 *     parser = BTreeLeafParser(bytes, key_length, ref_list_length)             # <<<<<<<<<<<<<<
 *     return parser.parse()
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 326; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_bytes);
  __Pyx_GIVEREF(__pyx_v_bytes);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_bytes);
  __Pyx_INCREF(__pyx_v_key_length);
  __Pyx_GIVEREF(__pyx_v_key_length);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_key_length);
  __Pyx_INCREF(__pyx_v_ref_list_length);
  __Pyx_GIVEREF(__pyx_v_ref_list_length);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_ref_list_length);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 326; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_parser = ((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":327
 * def _parse_leaf_lines(bytes, key_length, ref_list_length):
 *     parser = BTreeLeafParser(bytes, key_length, ref_list_length)
 *     return parser.parse()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_parser), __pyx_n_s_parse); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 327; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 327; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 327; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":325
 * 
 * 
 * def _parse_leaf_lines(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(bytes, key_length, ref_list_length)
 *     return parser.parse()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._parse_leaf_lines", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_parser);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":349
 * _hexbuf = '0123456789abcdef'
 * 
 * cdef _populate_unhexbuf():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i from 0 <= i < 256:
 */

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx__populate_unhexbuf(void) {
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_populate_unhexbuf", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":351
 * cdef _populate_unhexbuf():
 *     cdef int i
 *     for i from 0 <= i < 256:             # <<<<<<<<<<<<<<
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 */
  for (__pyx_v_i = 0; __pyx_v_i < 0x100; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":352
 *     cdef int i
 *     for i from 0 <= i < 256:
 *         _unhexbuf[i] = -1             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i
 */
    (__pyx_v_6bzrlib_21_btree_serializer_pyx__unhexbuf[__pyx_v_i]) = -1;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":353
 *     for i from 0 <= i < 256:
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 */
  for (__pyx_v_i = 0; __pyx_v_i < 10; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":354
 *         _unhexbuf[i] = -1
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i             # <<<<<<<<<<<<<<
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i
 */
    (__pyx_v_6bzrlib_21_btree_serializer_pyx__unhexbuf[(__pyx_v_i + '0')]) = __pyx_v_i;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":355
 *     for i from 0 <= i < 10: # 0123456789 => map to the raw number
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 */
  for (__pyx_v_i = 10; __pyx_v_i < 16; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":356
 *         _unhexbuf[(i + c'0')] = i
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i             # <<<<<<<<<<<<<<
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i
 */
    (__pyx_v_6bzrlib_21_btree_serializer_pyx__unhexbuf[((__pyx_v_i - 10) + 'a')]) = __pyx_v_i;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":357
 *     for i from 10 <= i < 16: # abcdef => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16             # <<<<<<<<<<<<<<
 *         _unhexbuf[(i - 10 + c'A')] = i
 * _populate_unhexbuf()
 */
  for (__pyx_v_i = 10; __pyx_v_i < 16; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":358
 *         _unhexbuf[(i - 10 + c'a')] = i
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i             # <<<<<<<<<<<<<<
 * _populate_unhexbuf()
 * 
 */
    (__pyx_v_6bzrlib_21_btree_serializer_pyx__unhexbuf[((__pyx_v_i - 10) + 'A')]) = __pyx_v_i;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":349
 * _hexbuf = '0123456789abcdef'
 * 
 * cdef _populate_unhexbuf():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     for i from 0 <= i < 256:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":362
 * 
 * 
 * cdef int _unhexlify_sha1(char *as_hex, char *as_bin): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Take the hex sha1 in as_hex and make it binary in as_bin
 * 
 */

static int __pyx_f_6bzrlib_21_btree_serializer_pyx__unhexlify_sha1(char *__pyx_v_as_hex, char *__pyx_v_as_bin) {
  int __pyx_v_top;
  int __pyx_v_bot;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_unhexlify_sha1", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":374
 *     # binascii does this using isupper() and tolower() and ?: syntax. I'm
 *     # guessing a simple lookup array should be faster.
 *     j = 0             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 */
  __pyx_v_j = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":375
 *     # guessing a simple lookup array should be faster.
 *     j = 0
 *     for i from 0 <= i < 20:             # <<<<<<<<<<<<<<
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 */
  for (__pyx_v_i = 0; __pyx_v_i < 20; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":376
 *     j = 0
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 */
    __pyx_v_top = (__pyx_v_6bzrlib_21_btree_serializer_pyx__unhexbuf[((unsigned char)(__pyx_v_as_hex[__pyx_v_j]))]);

    /* "bzrlib/_btree_serializer_pyx.pyx":377
 *     for i from 0 <= i < 20:
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 */
    __pyx_v_j = (__pyx_v_j + 1);

    /* "bzrlib/_btree_serializer_pyx.pyx":378
 *         top = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         if top == -1 or bot == -1:
 */
    __pyx_v_bot = (__pyx_v_6bzrlib_21_btree_serializer_pyx__unhexbuf[((unsigned char)(__pyx_v_as_hex[__pyx_v_j]))]);

    /* "bzrlib/_btree_serializer_pyx.pyx":379
 *         j = j + 1
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         if top == -1 or bot == -1:
 *             return 0
 */
    __pyx_v_j = (__pyx_v_j + 1);

    /* "bzrlib/_btree_serializer_pyx.pyx":380
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         if top == -1 or bot == -1:             # <<<<<<<<<<<<<<
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 */
    __pyx_t_2 = ((__pyx_v_top == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_bot == -1L) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "bzrlib/_btree_serializer_pyx.pyx":381
 *         j = j + 1
 *         if top == -1 or bot == -1:
 *             return 0             # <<<<<<<<<<<<<<
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 *     return 1
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "bzrlib/_btree_serializer_pyx.pyx":380
 *         bot = _unhexbuf[<unsigned char>(as_hex[j])]
 *         j = j + 1
 *         if top == -1 or bot == -1:             # <<<<<<<<<<<<<<
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":382
 *         if top == -1 or bot == -1:
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    (__pyx_v_as_bin[__pyx_v_i]) = ((unsigned char)((__pyx_v_top << 4) + __pyx_v_bot));
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":383
 *             return 0
 *         as_bin[i] = <unsigned char>((top << 4) + bot);
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":362
 * 
 * 
 * cdef int _unhexlify_sha1(char *as_hex, char *as_bin): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Take the hex sha1 in as_hex and make it binary in as_bin
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":386
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_3_py_unhexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_hex); /*proto*/
static char __pyx_doc_6bzrlib_21_btree_serializer_pyx_2_py_unhexlify[] = "For the test infrastructure, just thunks to _unhexlify_sha1";
static PyMethodDef __pyx_mdef_6bzrlib_21_btree_serializer_pyx_3_py_unhexlify = {"_py_unhexlify", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_3_py_unhexlify, METH_O, __pyx_doc_6bzrlib_21_btree_serializer_pyx_2_py_unhexlify};
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_3_py_unhexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_hex) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_unhexlify (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_2_py_unhexlify(__pyx_self, ((PyObject *)__pyx_v_as_hex));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_2_py_unhexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_hex) {
  PyObject *__pyx_v_as_bin = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_unhexlify", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":388
 * def _py_unhexlify(as_hex):
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyString_FromStringAndSize(NULL, 20)
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_as_hex); if (unlikely(__pyx_t_2 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = ((__pyx_t_2 != 40) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((!(PyString_CheckExact(__pyx_v_as_hex) != 0)) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":389
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):
 *         raise ValueError('not a 40-byte hex digest')             # <<<<<<<<<<<<<<
 *     as_bin = PyString_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyString_AS_STRING(as_hex), PyString_AS_STRING(as_bin)):
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":388
 * def _py_unhexlify(as_hex):
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyString_FromStringAndSize(NULL, 20)
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":390
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyString_FromStringAndSize(NULL, 20)             # <<<<<<<<<<<<<<
 *     if _unhexlify_sha1(PyString_AS_STRING(as_hex), PyString_AS_STRING(as_bin)):
 *         return as_bin
 */
  __pyx_t_4 = PyString_FromStringAndSize(NULL, 20); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 390; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_as_bin = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":391
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyString_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyString_AS_STRING(as_hex), PyString_AS_STRING(as_bin)):             # <<<<<<<<<<<<<<
 *         return as_bin
 *     return None
 */
  __pyx_t_1 = (__pyx_f_6bzrlib_21_btree_serializer_pyx__unhexlify_sha1(PyString_AS_STRING(__pyx_v_as_hex), PyString_AS_STRING(__pyx_v_as_bin)) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":392
 *     as_bin = PyString_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyString_AS_STRING(as_hex), PyString_AS_STRING(as_bin)):
 *         return as_bin             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_as_bin);
    __pyx_r = __pyx_v_as_bin;
    goto __pyx_L0;

    /* "bzrlib/_btree_serializer_pyx.pyx":391
 *         raise ValueError('not a 40-byte hex digest')
 *     as_bin = PyString_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyString_AS_STRING(as_hex), PyString_AS_STRING(as_bin)):             # <<<<<<<<<<<<<<
 *         return as_bin
 *     return None
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":393
 *     if _unhexlify_sha1(PyString_AS_STRING(as_hex), PyString_AS_STRING(as_bin)):
 *         return as_bin
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_None);
  __pyx_r = Py_None;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":386
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._py_unhexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_bin);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":396
 * 
 * 
 * cdef void _hexlify_sha1(char *as_bin, char *as_hex): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef int i, j
 *     cdef char c
 */

static void __pyx_f_6bzrlib_21_btree_serializer_pyx__hexlify_sha1(char *__pyx_v_as_bin, char *__pyx_v_as_hex) {
  int __pyx_v_i;
  int __pyx_v_j;
  char __pyx_v_c;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_hexlify_sha1", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":400
 *     cdef char c
 * 
 *     j = 0             # <<<<<<<<<<<<<<
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]
 */
  __pyx_v_j = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":401
 * 
 *     j = 0
 *     for i from 0 <= i < 20:             # <<<<<<<<<<<<<<
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 */
  for (__pyx_v_i = 0; __pyx_v_i < 20; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":402
 *     j = 0
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]             # <<<<<<<<<<<<<<
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1
 */
    __pyx_v_c = (__pyx_v_as_bin[__pyx_v_i]);

    /* "bzrlib/_btree_serializer_pyx.pyx":403
 *     for i from 0 <= i < 20:
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]             # <<<<<<<<<<<<<<
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]
 */
    (__pyx_v_as_hex[__pyx_v_j]) = (__pyx_v_6bzrlib_21_btree_serializer_pyx__hexbuf[((__pyx_v_c >> 4) & 0xf)]);

    /* "bzrlib/_btree_serializer_pyx.pyx":404
 *         c = as_bin[i]
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1             # <<<<<<<<<<<<<<
 *         as_hex[j] = _hexbuf[(c)&0xf]
 *         j = j + 1
 */
    __pyx_v_j = (__pyx_v_j + 1);

    /* "bzrlib/_btree_serializer_pyx.pyx":405
 *         as_hex[j] = _hexbuf[(c>>4)&0xf]
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]             # <<<<<<<<<<<<<<
 *         j = j + 1
 * 
 */
    (__pyx_v_as_hex[__pyx_v_j]) = (__pyx_v_6bzrlib_21_btree_serializer_pyx__hexbuf[(__pyx_v_c & 0xf)]);

    /* "bzrlib/_btree_serializer_pyx.pyx":406
 *         j = j + 1
 *         as_hex[j] = _hexbuf[(c)&0xf]
 *         j = j + 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_j = (__pyx_v_j + 1);
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":396
 * 
 * 
 * cdef void _hexlify_sha1(char *as_bin, char *as_hex): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef int i, j
 *     cdef char c
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "bzrlib/_btree_serializer_pyx.pyx":409
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_5_py_hexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_bin); /*proto*/
static char __pyx_doc_6bzrlib_21_btree_serializer_pyx_4_py_hexlify[] = "For test infrastructure, thunk to _hexlify_sha1";
static PyMethodDef __pyx_mdef_6bzrlib_21_btree_serializer_pyx_5_py_hexlify = {"_py_hexlify", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_5_py_hexlify, METH_O, __pyx_doc_6bzrlib_21_btree_serializer_pyx_4_py_hexlify};
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_5_py_hexlify(PyObject *__pyx_self, PyObject *__pyx_v_as_bin) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_hexlify (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_4_py_hexlify(__pyx_self, ((PyObject *)__pyx_v_as_bin));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_4_py_hexlify(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_as_bin) {
  PyObject *__pyx_v_as_hex = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_hexlify", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":411
 * def _py_hexlify(as_bin):
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyString_FromStringAndSize(NULL, 40)
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_as_bin); if (unlikely(__pyx_t_2 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 411; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = ((__pyx_t_2 != 20) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((!(PyString_CheckExact(__pyx_v_as_bin) != 0)) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":412
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')             # <<<<<<<<<<<<<<
 *     as_hex = PyString_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyString_AS_STRING(as_bin), PyString_AS_STRING(as_hex))
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":411
 * def _py_hexlify(as_bin):
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):             # <<<<<<<<<<<<<<
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyString_FromStringAndSize(NULL, 40)
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":413
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyString_FromStringAndSize(NULL, 40)             # <<<<<<<<<<<<<<
 *     _hexlify_sha1(PyString_AS_STRING(as_bin), PyString_AS_STRING(as_hex))
 *     return as_hex
 */
  __pyx_t_4 = PyString_FromStringAndSize(NULL, 40); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 413; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_as_hex = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":414
 *         raise ValueError('not a 20-byte binary digest')
 *     as_hex = PyString_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyString_AS_STRING(as_bin), PyString_AS_STRING(as_hex))             # <<<<<<<<<<<<<<
 *     return as_hex
 * 
 */
  __pyx_f_6bzrlib_21_btree_serializer_pyx__hexlify_sha1(PyString_AS_STRING(__pyx_v_as_bin), PyString_AS_STRING(__pyx_v_as_hex));

  /* "bzrlib/_btree_serializer_pyx.pyx":415
 *     as_hex = PyString_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyString_AS_STRING(as_bin), PyString_AS_STRING(as_hex))
 *     return as_hex             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_as_hex);
  __pyx_r = __pyx_v_as_hex;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":409
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._py_hexlify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_hex);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":418
 * 
 * 
 * cdef int _key_to_sha1(key, char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Map a key into its sha1 content.
 * 
 */

static int __pyx_f_6bzrlib_21_btree_serializer_pyx__key_to_sha1(PyObject *__pyx_v_key, char *__pyx_v_sha1) {
  char *__pyx_v_c_val;
  PyObject *__pyx_v_p_val;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_key_to_sha1", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":428
 *     cdef PyObject *p_val
 * 
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif (PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1):
 */
  __pyx_t_2 = (StaticTuple_CheckExact(__pyx_v_key) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  if (!(likely(((__pyx_v_key) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_key, __pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 428; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = ((StaticTuple_GET_SIZE(((StaticTuple *)__pyx_v_key)) == 1) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":429
 * 
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)             # <<<<<<<<<<<<<<
 *     elif (PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1):
 *         p_val = PyTuple_GET_ITEM_ptr_object(key, 0)
 */
    if (!(likely(((__pyx_v_key) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_key, __pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 429; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_p_val = ((PyObject *)StaticTuple_GET_ITEM(((StaticTuple *)__pyx_v_key), 0));

    /* "bzrlib/_btree_serializer_pyx.pyx":428
 *     cdef PyObject *p_val
 * 
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:             # <<<<<<<<<<<<<<
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif (PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1):
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":430
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif (PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1):             # <<<<<<<<<<<<<<
 *         p_val = PyTuple_GET_ITEM_ptr_object(key, 0)
 *     else:
 */
  __pyx_t_2 = (PyTuple_CheckExact(__pyx_v_key) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = ((PyTuple_GET_SIZE(__pyx_v_key) == 1) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":431
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif (PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1):
 *         p_val = PyTuple_GET_ITEM_ptr_object(key, 0)             # <<<<<<<<<<<<<<
 *     else:
 *         # Not a tuple or a StaticTuple
 */
    __pyx_v_p_val = PyTuple_GET_ITEM(__pyx_v_key, 0);

    /* "bzrlib/_btree_serializer_pyx.pyx":430
 *     if StaticTuple_CheckExact(key) and StaticTuple_GET_SIZE(key) == 1:
 *         p_val = <PyObject *>StaticTuple_GET_ITEM(key, 0)
 *     elif (PyTuple_CheckExact(key) and PyTuple_GET_SIZE(key) == 1):             # <<<<<<<<<<<<<<
 *         p_val = PyTuple_GET_ITEM_ptr_object(key, 0)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":434
 *     else:
 *         # Not a tuple or a StaticTuple
 *         return 0             # <<<<<<<<<<<<<<
 *     if (PyString_CheckExact_ptr(p_val) and PyString_GET_SIZE_ptr(p_val) == 45):
 *         c_val = PyString_AS_STRING_ptr(p_val)
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L3:;

  /* "bzrlib/_btree_serializer_pyx.pyx":435
 *         # Not a tuple or a StaticTuple
 *         return 0
 *     if (PyString_CheckExact_ptr(p_val) and PyString_GET_SIZE_ptr(p_val) == 45):             # <<<<<<<<<<<<<<
 *         c_val = PyString_AS_STRING_ptr(p_val)
 *     else:
 */
  __pyx_t_2 = (PyString_CheckExact(__pyx_v_p_val) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_2 = ((PyString_GET_SIZE(__pyx_v_p_val) == 45) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":436
 *         return 0
 *     if (PyString_CheckExact_ptr(p_val) and PyString_GET_SIZE_ptr(p_val) == 45):
 *         c_val = PyString_AS_STRING_ptr(p_val)             # <<<<<<<<<<<<<<
 *     else:
 *         return 0
 */
    __pyx_v_c_val = PyString_AS_STRING(__pyx_v_p_val);

    /* "bzrlib/_btree_serializer_pyx.pyx":435
 *         # Not a tuple or a StaticTuple
 *         return 0
 *     if (PyString_CheckExact_ptr(p_val) and PyString_GET_SIZE_ptr(p_val) == 45):             # <<<<<<<<<<<<<<
 *         c_val = PyString_AS_STRING_ptr(p_val)
 *     else:
 */
    goto __pyx_L8;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":438
 *         c_val = PyString_AS_STRING_ptr(p_val)
 *     else:
 *         return 0             # <<<<<<<<<<<<<<
 *     if strncmp(c_val, 'sha1:', 5) != 0:
 *         return 0
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }
  __pyx_L8:;

  /* "bzrlib/_btree_serializer_pyx.pyx":439
 *     else:
 *         return 0
 *     if strncmp(c_val, 'sha1:', 5) != 0:             # <<<<<<<<<<<<<<
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 */
  __pyx_t_1 = ((strncmp(__pyx_v_c_val, __pyx_k_sha1, 5) != 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":440
 *         return 0
 *     if strncmp(c_val, 'sha1:', 5) != 0:
 *         return 0             # <<<<<<<<<<<<<<
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_btree_serializer_pyx.pyx":439
 *     else:
 *         return 0
 *     if strncmp(c_val, 'sha1:', 5) != 0:             # <<<<<<<<<<<<<<
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":441
 *     if strncmp(c_val, 'sha1:', 5) != 0:
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
 */
  __pyx_t_1 = ((!(__pyx_f_6bzrlib_21_btree_serializer_pyx__unhexlify_sha1((__pyx_v_c_val + 5), __pyx_v_sha1) != 0)) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":442
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_btree_serializer_pyx.pyx":441
 *     if strncmp(c_val, 'sha1:', 5) != 0:
 *         return 0
 *     if not _unhexlify_sha1(c_val + 5, sha1):             # <<<<<<<<<<<<<<
 *         return 0
 *     return 1
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":443
 *     if not _unhexlify_sha1(c_val + 5, sha1):
 *         return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":418
 * 
 * 
 * cdef int _key_to_sha1(key, char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     """Map a key into its sha1 content.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("bzrlib._btree_serializer_pyx._key_to_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":446
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_7_py_key_to_sha1(PyObject *__pyx_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_6bzrlib_21_btree_serializer_pyx_6_py_key_to_sha1[] = "Map a key to a simple sha1 string.\n\n    This is a testing thunk to the C function.\n    ";
static PyMethodDef __pyx_mdef_6bzrlib_21_btree_serializer_pyx_7_py_key_to_sha1 = {"_py_key_to_sha1", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_7_py_key_to_sha1, METH_O, __pyx_doc_6bzrlib_21_btree_serializer_pyx_6_py_key_to_sha1};
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_7_py_key_to_sha1(PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_key_to_sha1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_6_py_key_to_sha1(__pyx_self, ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_6_py_key_to_sha1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_as_bin_sha = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_key_to_sha1", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":451
 *     This is a testing thunk to the C function.
 *     """
 *     as_bin_sha = PyString_FromStringAndSize(NULL, 20)             # <<<<<<<<<<<<<<
 *     if _key_to_sha1(key, PyString_AS_STRING(as_bin_sha)):
 *         return as_bin_sha
 */
  __pyx_t_1 = PyString_FromStringAndSize(NULL, 20); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 451; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_as_bin_sha = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":452
 *     """
 *     as_bin_sha = PyString_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyString_AS_STRING(as_bin_sha)):             # <<<<<<<<<<<<<<
 *         return as_bin_sha
 *     return None
 */
  __pyx_t_2 = (__pyx_f_6bzrlib_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, PyString_AS_STRING(__pyx_v_as_bin_sha)) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":453
 *     as_bin_sha = PyString_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyString_AS_STRING(as_bin_sha)):
 *         return as_bin_sha             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_as_bin_sha);
    __pyx_r = __pyx_v_as_bin_sha;
    goto __pyx_L0;

    /* "bzrlib/_btree_serializer_pyx.pyx":452
 *     """
 *     as_bin_sha = PyString_FromStringAndSize(NULL, 20)
 *     if _key_to_sha1(key, PyString_AS_STRING(as_bin_sha)):             # <<<<<<<<<<<<<<
 *         return as_bin_sha
 *     return None
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":454
 *     if _key_to_sha1(key, PyString_AS_STRING(as_bin_sha)):
 *         return as_bin_sha
 *     return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_None);
  __pyx_r = Py_None;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":446
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._py_key_to_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_as_bin_sha);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":457
 * 
 * 
 * cdef StaticTuple _sha1_to_key(char *sha1):             # <<<<<<<<<<<<<<
 *     """Compute a ('sha1:abcd',) key for a given sha1."""
 *     cdef StaticTuple key
 */

static StaticTuple *__pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_key(char *__pyx_v_sha1) {
  StaticTuple *__pyx_v_key = 0;
  PyObject *__pyx_v_hexxed = 0;
  char *__pyx_v_c_buf;
  StaticTuple *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sha1_to_key", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":462
 *     cdef object hexxed
 *     cdef char *c_buf
 *     hexxed = PyString_FromStringAndSize(NULL, 45)             # <<<<<<<<<<<<<<
 *     c_buf = PyString_AS_STRING(hexxed)
 *     memcpy(c_buf, 'sha1:', 5)
 */
  __pyx_t_1 = PyString_FromStringAndSize(NULL, 45); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 462; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_hexxed = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":463
 *     cdef char *c_buf
 *     hexxed = PyString_FromStringAndSize(NULL, 45)
 *     c_buf = PyString_AS_STRING(hexxed)             # <<<<<<<<<<<<<<
 *     memcpy(c_buf, 'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)
 */
  __pyx_v_c_buf = PyString_AS_STRING(__pyx_v_hexxed);

  /* "bzrlib/_btree_serializer_pyx.pyx":464
 *     hexxed = PyString_FromStringAndSize(NULL, 45)
 *     c_buf = PyString_AS_STRING(hexxed)
 *     memcpy(c_buf, 'sha1:', 5)             # <<<<<<<<<<<<<<
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = StaticTuple_New(1)
 */
  memcpy(__pyx_v_c_buf, __pyx_k_sha1, 5);

  /* "bzrlib/_btree_serializer_pyx.pyx":465
 *     c_buf = PyString_AS_STRING(hexxed)
 *     memcpy(c_buf, 'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)             # <<<<<<<<<<<<<<
 *     key = StaticTuple_New(1)
 *     Py_INCREF(hexxed)
 */
  __pyx_f_6bzrlib_21_btree_serializer_pyx__hexlify_sha1(__pyx_v_sha1, (__pyx_v_c_buf + 5));

  /* "bzrlib/_btree_serializer_pyx.pyx":466
 *     memcpy(c_buf, 'sha1:', 5)
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = StaticTuple_New(1)             # <<<<<<<<<<<<<<
 *     Py_INCREF(hexxed)
 *     StaticTuple_SET_ITEM(key, 0, hexxed)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(1)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 466; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":467
 *     _hexlify_sha1(sha1, c_buf+5)
 *     key = StaticTuple_New(1)
 *     Py_INCREF(hexxed)             # <<<<<<<<<<<<<<
 *     StaticTuple_SET_ITEM(key, 0, hexxed)
 *     # This is a bit expensive. To parse 120 keys takes 48us, to return them all
 */
  Py_INCREF(__pyx_v_hexxed);

  /* "bzrlib/_btree_serializer_pyx.pyx":468
 *     key = StaticTuple_New(1)
 *     Py_INCREF(hexxed)
 *     StaticTuple_SET_ITEM(key, 0, hexxed)             # <<<<<<<<<<<<<<
 *     # This is a bit expensive. To parse 120 keys takes 48us, to return them all
 *     # can be done in 66.6us (so 18.6us to build them all).
 */
  StaticTuple_SET_ITEM(__pyx_v_key, 0, __pyx_v_hexxed);

  /* "bzrlib/_btree_serializer_pyx.pyx":478
 *     # We *could* hang the PyObject form off of the gc_chk_sha1_record for ones
 *     # that we have deserialized. Something to think about, at least.
 *     key = StaticTuple_Intern(key)             # <<<<<<<<<<<<<<
 *     return key
 * 
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_Intern(__pyx_v_key)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 478; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_key, ((StaticTuple *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":479
 *     # that we have deserialized. Something to think about, at least.
 *     key = StaticTuple_Intern(key)
 *     return key             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_key));
  __pyx_r = __pyx_v_key;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":457
 * 
 * 
 * cdef StaticTuple _sha1_to_key(char *sha1):             # <<<<<<<<<<<<<<
 *     """Compute a ('sha1:abcd',) key for a given sha1."""
 *     cdef StaticTuple key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._sha1_to_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_hexxed);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":482
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_9_py_sha1_to_key(PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin); /*proto*/
static char __pyx_doc_6bzrlib_21_btree_serializer_pyx_8_py_sha1_to_key[] = "Test thunk to check the sha1 mapping.";
static PyMethodDef __pyx_mdef_6bzrlib_21_btree_serializer_pyx_9_py_sha1_to_key = {"_py_sha1_to_key", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_9_py_sha1_to_key, METH_O, __pyx_doc_6bzrlib_21_btree_serializer_pyx_8_py_sha1_to_key};
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_9_py_sha1_to_key(PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_sha1_to_key (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_8_py_sha1_to_key(__pyx_self, ((PyObject *)__pyx_v_sha1_bin));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_8_py_sha1_to_key(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sha1_bin) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_sha1_to_key", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":484
 * def _py_sha1_to_key(sha1_bin):
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:             # <<<<<<<<<<<<<<
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyString_AS_STRING(sha1_bin))
 */
  __pyx_t_2 = ((!(PyString_CheckExact(__pyx_v_sha1_bin) != 0)) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((PyString_GET_SIZE(__pyx_v_sha1_bin) != 20) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":485
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')             # <<<<<<<<<<<<<<
 *     return _sha1_to_key(PyString_AS_STRING(sha1_bin))
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 485; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 485; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":484
 * def _py_sha1_to_key(sha1_bin):
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:             # <<<<<<<<<<<<<<
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyString_AS_STRING(sha1_bin))
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":486
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')
 *     return _sha1_to_key(PyString_AS_STRING(sha1_bin))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = ((PyObject *)__pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_key(PyString_AS_STRING(__pyx_v_sha1_bin))); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 486; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":482
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._py_sha1_to_key", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":489
 * 
 * 
 * cdef unsigned int _sha1_to_uint(char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned int val
 *     # Must be in MSB, because that is how the content is sorted
 */

static unsigned int __pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_uint(char *__pyx_v_sha1) {
  unsigned int __pyx_v_val;
  unsigned int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sha1_to_uint", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":495
 *            | ((<unsigned int>(sha1[1]) & 0xff) << 16)
 *            | ((<unsigned int>(sha1[2]) & 0xff) << 8)
 *            | ((<unsigned int>(sha1[3]) & 0xff) << 0))             # <<<<<<<<<<<<<<
 *     return val
 * 
 */
  __pyx_v_val = (((((((unsigned int)(__pyx_v_sha1[0])) & 0xff) << 24) | ((((unsigned int)(__pyx_v_sha1[1])) & 0xff) << 16)) | ((((unsigned int)(__pyx_v_sha1[2])) & 0xff) << 8)) | ((((unsigned int)(__pyx_v_sha1[3])) & 0xff) << 0));

  /* "bzrlib/_btree_serializer_pyx.pyx":496
 *            | ((<unsigned int>(sha1[2]) & 0xff) << 8)
 *            | ((<unsigned int>(sha1[3]) & 0xff) << 0))
 *     return val             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":489
 * 
 * 
 * cdef unsigned int _sha1_to_uint(char *sha1): # cannot_raise             # <<<<<<<<<<<<<<
 *     cdef unsigned int val
 *     # Must be in MSB, because that is how the content is sorted
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":499
 * 
 * 
 * cdef _format_record(gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *     # This is inefficient to go from a logical state back to a
 *     # string, but it makes things work a bit better internally for now.
 */

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx__format_record(__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  PyObject *__pyx_v_block_offset_str = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_format_record", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":502
 *     # This is inefficient to go from a logical state back to a
 *     # string, but it makes things work a bit better internally for now.
 *     if record.block_offset >= 0xFFFFFFFF:             # <<<<<<<<<<<<<<
 *         # %llu is what we really want, but unfortunately it was only added
 *         # in python 2.7... :(
 */
  __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_record->block_offset); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 502; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_int_4294967295, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 502; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 502; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "bzrlib/_btree_serializer_pyx.pyx":505
 *         # %llu is what we really want, but unfortunately it was only added
 *         # in python 2.7... :(
 *         block_offset_str = str(record.block_offset)             # <<<<<<<<<<<<<<
 *         value = PyString_FromFormat('%s %u %u %u',
 *                                 PyString_AS_STRING(block_offset_str),
 */
    __pyx_t_2 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_record->block_offset); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 505; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 505; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 505; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_block_offset_str = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":506
 *         # in python 2.7... :(
 *         block_offset_str = str(record.block_offset)
 *         value = PyString_FromFormat('%s %u %u %u',             # <<<<<<<<<<<<<<
 *                                 PyString_AS_STRING(block_offset_str),
 *                                 record.block_length,
 */
    __pyx_t_2 = PyString_FromFormat(__pyx_k_s_u_u_u, PyString_AS_STRING(__pyx_v_block_offset_str), __pyx_v_record->block_length, __pyx_v_record->record_start, __pyx_v_record->record_end); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 506; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_value = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":502
 *     # This is inefficient to go from a logical state back to a
 *     # string, but it makes things work a bit better internally for now.
 *     if record.block_offset >= 0xFFFFFFFF:             # <<<<<<<<<<<<<<
 *         # %llu is what we really want, but unfortunately it was only added
 *         # in python 2.7... :(
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":511
 *                                 record.record_start, record.record_end)
 *     else:
 *         value = PyString_FromFormat('%lu %u %u %u',             # <<<<<<<<<<<<<<
 *                                     <unsigned long>record.block_offset,
 *                                     record.block_length,
 */
  /*else*/ {

    /* "bzrlib/_btree_serializer_pyx.pyx":514
 *                                     <unsigned long>record.block_offset,
 *                                     record.block_length,
 *                                     record.record_start, record.record_end)             # <<<<<<<<<<<<<<
 *     return value
 * 
 */
    __pyx_t_2 = PyString_FromFormat(__pyx_k_lu_u_u_u, ((unsigned long)__pyx_v_record->block_offset), __pyx_v_record->block_length, __pyx_v_record->record_start, __pyx_v_record->record_end); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 511; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_value = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "bzrlib/_btree_serializer_pyx.pyx":515
 *                                     record.block_length,
 *                                     record.record_start, record.record_end)
 *     return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_value);
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":499
 * 
 * 
 * cdef _format_record(gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *     # This is inefficient to go from a logical state back to a
 *     # string, but it makes things work a bit better internally for now.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._format_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_block_offset_str);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":536
 *     cdef unsigned char offsets[257]
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         # :( Why doesn't Pyrex let me do a simple sizeof(GCCHKSHA1LeafNode)
 *         # like Cython? Explicitly enumerating everything here seems to leave my
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sizeof__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode___sizeof__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sizeof__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":546
 *         # return (sizeof(GCCHKSHA1LeafNode)
 *         #     + sizeof(gc_chk_sha1_record)*self.num_records)
 *         return (sizeof(PyObject) + sizeof(void*) + sizeof(int)             # <<<<<<<<<<<<<<
 *             + sizeof(gc_chk_sha1_record*) + sizeof(PyObject *)
 *             + sizeof(gc_chk_sha1_record*) + sizeof(char)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "bzrlib/_btree_serializer_pyx.pyx":550
 *             + sizeof(gc_chk_sha1_record*) + sizeof(char)
 *             + sizeof(unsigned char)*257
 *             + sizeof(gc_chk_sha1_record)*self.num_records)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((((((((((sizeof(PyObject)) + (sizeof(void *))) + (sizeof(int))) + (sizeof(__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *))) + (sizeof(PyObject *))) + (sizeof(__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *))) + (sizeof(char))) + ((sizeof(unsigned char)) * 0x101)) + ((sizeof(__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record)) * __pyx_v_self->num_records))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 550; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":536
 *     cdef unsigned char offsets[257]
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         # :( Why doesn't Pyrex let me do a simple sizeof(GCCHKSHA1LeafNode)
 *         # like Cython? Explicitly enumerating everything here seems to leave my
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.__sizeof__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":552
 *             + sizeof(gc_chk_sha1_record)*self.num_records)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
 */

/* Python wrapper */
static void __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_2__dealloc__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":553
 * 
 *     def __dealloc__(self):
 *         if self.records != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.records)
 *             self.records = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->records != NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":554
 *     def __dealloc__(self):
 *         if self.records != NULL:
 *             PyMem_Free(self.records)             # <<<<<<<<<<<<<<
 *             self.records = NULL
 * 
 */
    PyMem_Free(__pyx_v_self->records);

    /* "bzrlib/_btree_serializer_pyx.pyx":555
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
 *             self.records = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, bytes):
 */
    __pyx_v_self->records = NULL;

    /* "bzrlib/_btree_serializer_pyx.pyx":553
 * 
 *     def __dealloc__(self):
 *         if self.records != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.records)
 *             self.records = NULL
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":552
 *             + sizeof(gc_chk_sha1_record)*self.num_records)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.records != NULL:
 *             PyMem_Free(self.records)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "bzrlib/_btree_serializer_pyx.pyx":557
 *             self.records = NULL
 * 
 *     def __init__(self, bytes):             # <<<<<<<<<<<<<<
 *         self._parse_bytes(bytes)
 *         self.last_key = None
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytes = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bytes,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bytes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 557; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_bytes = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 557; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), __pyx_v_bytes);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_4__init__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":558
 * 
 *     def __init__(self, bytes):
 *         self._parse_bytes(bytes)             # <<<<<<<<<<<<<<
 *         self.last_key = None
 *         self.last_record = NULL
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_parse_bytes(__pyx_v_self, __pyx_v_bytes); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 558; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":559
 *     def __init__(self, bytes):
 *         self._parse_bytes(bytes)
 *         self.last_key = None             # <<<<<<<<<<<<<<
 *         self.last_record = NULL
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = Py_None;

  /* "bzrlib/_btree_serializer_pyx.pyx":560
 *         self._parse_bytes(bytes)
 *         self.last_key = None
 *         self.last_record = NULL             # <<<<<<<<<<<<<<
 * 
 *     property min_key:
 */
  __pyx_v_self->last_record = NULL;

  /* "bzrlib/_btree_serializer_pyx.pyx":557
 *             self.records = NULL
 * 
 *     def __init__(self, bytes):             # <<<<<<<<<<<<<<
 *         self._parse_bytes(bytes)
 *         self.last_key = None
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":563
 * 
 *     property min_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":564
 *     property min_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None
 */
  __pyx_t_1 = ((__pyx_v_self->num_records > 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":565
 *         def __get__(self):
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[0]).sha1)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 565; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "bzrlib/_btree_serializer_pyx.pyx":564
 *     property min_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":566
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     property max_key:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_None);
  __pyx_r = Py_None;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":563
 * 
 *     property min_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[0].sha1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.min_key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":569
 * 
 *     property max_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":570
 *     property max_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None
 */
  __pyx_t_1 = ((__pyx_v_self->num_records > 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":571
 *         def __get__(self):
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)             # <<<<<<<<<<<<<<
 *             return None
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[(__pyx_v_self->num_records - 1)]).sha1)); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 571; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "bzrlib/_btree_serializer_pyx.pyx":570
 *     property max_key:
 *         def __get__(self):
 *             if self.num_records > 0:             # <<<<<<<<<<<<<<
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":572
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     cdef StaticTuple _record_to_value_and_refs(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_None);
  __pyx_r = Py_None;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":569
 * 
 *     property max_key:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             if self.num_records > 0:
 *                 return _sha1_to_key(self.records[self.num_records-1].sha1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.max_key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":574
 *             return None
 * 
 *     cdef StaticTuple _record_to_value_and_refs(self,             # <<<<<<<<<<<<<<
 *                                                gc_chk_sha1_record *record):
 *         """Extract the refs and value part of this record."""
 */

static StaticTuple *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs(CYTHON_UNUSED struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  StaticTuple *__pyx_v_value_and_refs = 0;
  StaticTuple *__pyx_v_empty = 0;
  PyObject *__pyx_v_value = NULL;
  StaticTuple *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_record_to_value_and_refs", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":579
 *         cdef StaticTuple value_and_refs
 *         cdef StaticTuple empty
 *         value_and_refs = StaticTuple_New(2)             # <<<<<<<<<<<<<<
 *         value = _format_record(record)
 *         Py_INCREF(value)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(2)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 579; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value_and_refs = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":580
 *         cdef StaticTuple empty
 *         value_and_refs = StaticTuple_New(2)
 *         value = _format_record(record)             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)
 */
  __pyx_t_1 = __pyx_f_6bzrlib_21_btree_serializer_pyx__format_record(__pyx_v_record); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 580; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":581
 *         value_and_refs = StaticTuple_New(2)
 *         value = _format_record(record)
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)
 *         # Always empty refs
 */
  Py_INCREF(__pyx_v_value);

  /* "bzrlib/_btree_serializer_pyx.pyx":582
 *         value = _format_record(record)
 *         Py_INCREF(value)
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)             # <<<<<<<<<<<<<<
 *         # Always empty refs
 *         empty = StaticTuple_New(0)
 */
  StaticTuple_SET_ITEM(__pyx_v_value_and_refs, 0, __pyx_v_value);

  /* "bzrlib/_btree_serializer_pyx.pyx":584
 *         StaticTuple_SET_ITEM(value_and_refs, 0, value)
 *         # Always empty refs
 *         empty = StaticTuple_New(0)             # <<<<<<<<<<<<<<
 *         Py_INCREF(empty)
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(0)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 584; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_empty = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":585
 *         # Always empty refs
 *         empty = StaticTuple_New(0)
 *         Py_INCREF(empty)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)
 *         return value_and_refs
 */
  Py_INCREF(((PyObject *)__pyx_v_empty));

  /* "bzrlib/_btree_serializer_pyx.pyx":586
 *         empty = StaticTuple_New(0)
 *         Py_INCREF(empty)
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)             # <<<<<<<<<<<<<<
 *         return value_and_refs
 * 
 */
  StaticTuple_SET_ITEM(__pyx_v_value_and_refs, 1, ((PyObject *)__pyx_v_empty));

  /* "bzrlib/_btree_serializer_pyx.pyx":587
 *         Py_INCREF(empty)
 *         StaticTuple_SET_ITEM(value_and_refs, 1, empty)
 *         return value_and_refs             # <<<<<<<<<<<<<<
 * 
 *     cdef StaticTuple _record_to_item(self, gc_chk_sha1_record *record):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_value_and_refs));
  __pyx_r = __pyx_v_value_and_refs;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":574
 *             return None
 * 
 *     cdef StaticTuple _record_to_value_and_refs(self,             # <<<<<<<<<<<<<<
 *                                                gc_chk_sha1_record *record):
 *         """Extract the refs and value part of this record."""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._record_to_value_and_refs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_value_and_refs);
  __Pyx_XDECREF((PyObject *)__pyx_v_empty);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":589
 *         return value_and_refs
 * 
 *     cdef StaticTuple _record_to_item(self, gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *         """Turn a given record back into a fully fledged item.
 *         """
 */

static StaticTuple *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record) {
  StaticTuple *__pyx_v_item = 0;
  StaticTuple *__pyx_v_key = 0;
  StaticTuple *__pyx_v_value_and_refs = 0;
  StaticTuple *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_record_to_item", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":596
 *         cdef StaticTuple value_and_refs
 *         cdef object value
 *         key = _sha1_to_key(record.sha1)             # <<<<<<<<<<<<<<
 *         item = StaticTuple_New(2)
 *         Py_INCREF(key)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_key(__pyx_v_record->sha1)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 596; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_key = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":597
 *         cdef object value
 *         key = _sha1_to_key(record.sha1)
 *         item = StaticTuple_New(2)             # <<<<<<<<<<<<<<
 *         Py_INCREF(key)
 *         StaticTuple_SET_ITEM(item, 0, key)
 */
  __pyx_t_1 = ((PyObject *)StaticTuple_New(2)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 597; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_item = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":598
 *         key = _sha1_to_key(record.sha1)
 *         item = StaticTuple_New(2)
 *         Py_INCREF(key)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)
 */
  Py_INCREF(((PyObject *)__pyx_v_key));

  /* "bzrlib/_btree_serializer_pyx.pyx":599
 *         item = StaticTuple_New(2)
 *         Py_INCREF(key)
 *         StaticTuple_SET_ITEM(item, 0, key)             # <<<<<<<<<<<<<<
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)
 */
  StaticTuple_SET_ITEM(__pyx_v_item, 0, ((PyObject *)__pyx_v_key));

  /* "bzrlib/_btree_serializer_pyx.pyx":600
 *         Py_INCREF(key)
 *         StaticTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)             # <<<<<<<<<<<<<<
 *         Py_INCREF(value_and_refs)
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_value_and_refs(__pyx_v_self, __pyx_v_record)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 600; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_value_and_refs = ((StaticTuple *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":601
 *         StaticTuple_SET_ITEM(item, 0, key)
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)             # <<<<<<<<<<<<<<
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)
 *         return item
 */
  Py_INCREF(((PyObject *)__pyx_v_value_and_refs));

  /* "bzrlib/_btree_serializer_pyx.pyx":602
 *         value_and_refs = self._record_to_value_and_refs(record)
 *         Py_INCREF(value_and_refs)
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)             # <<<<<<<<<<<<<<
 *         return item
 * 
 */
  StaticTuple_SET_ITEM(__pyx_v_item, 1, ((PyObject *)__pyx_v_value_and_refs));

  /* "bzrlib/_btree_serializer_pyx.pyx":603
 *         Py_INCREF(value_and_refs)
 *         StaticTuple_SET_ITEM(item, 1, value_and_refs)
 *         return item             # <<<<<<<<<<<<<<
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_item));
  __pyx_r = __pyx_v_item;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":589
 *         return value_and_refs
 * 
 *     cdef StaticTuple _record_to_item(self, gc_chk_sha1_record *record):             # <<<<<<<<<<<<<<
 *         """Turn a given record back into a fully fledged item.
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._record_to_item", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_item);
  __Pyx_XDECREF((PyObject *)__pyx_v_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_value_and_refs);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":605
 *         return item
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:             # <<<<<<<<<<<<<<
 *         """Find a gc_chk_sha1_record that matches the sha1 supplied."""
 *         cdef int lo, hi, mid, the_cmp
 */

static __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1) {
  int __pyx_v_lo;
  int __pyx_v_hi;
  int __pyx_v_mid;
  int __pyx_v_the_cmp;
  int __pyx_v_offset;
  CYTHON_UNUSED long __pyx_v_local_n_cmp;
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_lookup_record", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":619
 *         # the offset array dropped us from 23us to 20us and 156 comparisions
 *         # (1.3/key)
 *         offset = self._offset_for_sha1(sha1)             # <<<<<<<<<<<<<<
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 619; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_offset = __pyx_t_1;

  /* "bzrlib/_btree_serializer_pyx.pyx":620
 *         # (1.3/key)
 *         offset = self._offset_for_sha1(sha1)
 *         lo = self.offsets[offset]             # <<<<<<<<<<<<<<
 *         hi = self.offsets[offset+1]
 *         if hi == 255:
 */
  __pyx_v_lo = (__pyx_v_self->offsets[__pyx_v_offset]);

  /* "bzrlib/_btree_serializer_pyx.pyx":621
 *         offset = self._offset_for_sha1(sha1)
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]             # <<<<<<<<<<<<<<
 *         if hi == 255:
 *             # if hi == 255 that means we potentially ran off the end of the
 */
  __pyx_v_hi = (__pyx_v_self->offsets[(__pyx_v_offset + 1)]);

  /* "bzrlib/_btree_serializer_pyx.pyx":622
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 *         if hi == 255:             # <<<<<<<<<<<<<<
 *             # if hi == 255 that means we potentially ran off the end of the
 *             # list, so push it up to num_records
 */
  __pyx_t_2 = ((__pyx_v_hi == 0xFF) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_btree_serializer_pyx.pyx":627
 *             # note that if 'lo' == 255, that is ok, because we can start
 *             # searching from that part of the list.
 *             hi = self.num_records             # <<<<<<<<<<<<<<
 *         local_n_cmp = 0
 *         while lo < hi:
 */
    __pyx_t_1 = __pyx_v_self->num_records;
    __pyx_v_hi = __pyx_t_1;

    /* "bzrlib/_btree_serializer_pyx.pyx":622
 *         lo = self.offsets[offset]
 *         hi = self.offsets[offset+1]
 *         if hi == 255:             # <<<<<<<<<<<<<<
 *             # if hi == 255 that means we potentially ran off the end of the
 *             # list, so push it up to num_records
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":628
 *             # searching from that part of the list.
 *             hi = self.num_records
 *         local_n_cmp = 0             # <<<<<<<<<<<<<<
 *         while lo < hi:
 *             mid = (lo + hi) / 2
 */
  __pyx_v_local_n_cmp = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":629
 *             hi = self.num_records
 *         local_n_cmp = 0
 *         while lo < hi:             # <<<<<<<<<<<<<<
 *             mid = (lo + hi) / 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_lo < __pyx_v_hi) != 0);
    if (!__pyx_t_2) break;

    /* "bzrlib/_btree_serializer_pyx.pyx":630
 *         local_n_cmp = 0
 *         while lo < hi:
 *             mid = (lo + hi) / 2             # <<<<<<<<<<<<<<
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:
 */
    __pyx_v_mid = __Pyx_div_long((__pyx_v_lo + __pyx_v_hi), 2);

    /* "bzrlib/_btree_serializer_pyx.pyx":631
 *         while lo < hi:
 *             mid = (lo + hi) / 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)             # <<<<<<<<<<<<<<
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 */
    __pyx_v_the_cmp = memcmp((__pyx_v_self->records[__pyx_v_mid]).sha1, __pyx_v_sha1, 20);

    /* "bzrlib/_btree_serializer_pyx.pyx":632
 *             mid = (lo + hi) / 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:             # <<<<<<<<<<<<<<
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
 */
    __pyx_t_2 = ((__pyx_v_the_cmp == 0) != 0);
    if (__pyx_t_2) {

      /* "bzrlib/_btree_serializer_pyx.pyx":633
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:
 *                 return &self.records[mid]             # <<<<<<<<<<<<<<
 *             elif the_cmp < 0:
 *                 lo = mid + 1
 */
      __pyx_r = (&(__pyx_v_self->records[__pyx_v_mid]));
      goto __pyx_L0;

      /* "bzrlib/_btree_serializer_pyx.pyx":632
 *             mid = (lo + hi) / 2
 *             the_cmp = memcmp(self.records[mid].sha1, sha1, 20)
 *             if the_cmp == 0:             # <<<<<<<<<<<<<<
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":634
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 *             elif the_cmp < 0:             # <<<<<<<<<<<<<<
 *                 lo = mid + 1
 *             else:
 */
    __pyx_t_2 = ((__pyx_v_the_cmp < 0) != 0);
    if (__pyx_t_2) {

      /* "bzrlib/_btree_serializer_pyx.pyx":635
 *                 return &self.records[mid]
 *             elif the_cmp < 0:
 *                 lo = mid + 1             # <<<<<<<<<<<<<<
 *             else:
 *                 hi = mid
 */
      __pyx_v_lo = (__pyx_v_mid + 1);

      /* "bzrlib/_btree_serializer_pyx.pyx":634
 *             if the_cmp == 0:
 *                 return &self.records[mid]
 *             elif the_cmp < 0:             # <<<<<<<<<<<<<<
 *                 lo = mid + 1
 *             else:
 */
      goto __pyx_L6;
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":637
 *                 lo = mid + 1
 *             else:
 *                 hi = mid             # <<<<<<<<<<<<<<
 *         return NULL
 * 
 */
    /*else*/ {
      __pyx_v_hi = __pyx_v_mid;
    }
    __pyx_L6:;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":638
 *             else:
 *                 hi = mid
 *         return NULL             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  __pyx_r = NULL;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":605
 *         return item
 * 
 *     cdef gc_chk_sha1_record* _lookup_record(self, char *sha1) except? NULL:             # <<<<<<<<<<<<<<
 *         """Find a gc_chk_sha1_record that matches the sha1 supplied."""
 *         cdef int lo, hi, mid, the_cmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._lookup_record", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":640
 *         return NULL
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_6__contains__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key) {
  char __pyx_v_sha1[20];
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":643
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         if _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 */
  __pyx_t_1 = (__pyx_f_6bzrlib_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, __pyx_v_sha1) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":645
 *         if _key_to_sha1(key, sha1):
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)             # <<<<<<<<<<<<<<
 *             if record != NULL:
 *                 self.last_key = key
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_lookup_record(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_2 == NULL && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 645; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_record = __pyx_t_2;

    /* "bzrlib/_btree_serializer_pyx.pyx":646
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 *             if record != NULL:             # <<<<<<<<<<<<<<
 *                 self.last_key = key
 *                 self.last_record = record
 */
    __pyx_t_1 = ((__pyx_v_record != NULL) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_btree_serializer_pyx.pyx":647
 *             record = self._lookup_record(sha1)
 *             if record != NULL:
 *                 self.last_key = key             # <<<<<<<<<<<<<<
 *                 self.last_record = record
 *                 return True
 */
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      __Pyx_GOTREF(__pyx_v_self->last_key);
      __Pyx_DECREF(__pyx_v_self->last_key);
      __pyx_v_self->last_key = __pyx_v_key;

      /* "bzrlib/_btree_serializer_pyx.pyx":648
 *             if record != NULL:
 *                 self.last_key = key
 *                 self.last_record = record             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
      __pyx_v_self->last_record = __pyx_v_record;

      /* "bzrlib/_btree_serializer_pyx.pyx":649
 *                 self.last_key = key
 *                 self.last_record = record
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "bzrlib/_btree_serializer_pyx.pyx":646
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 *             if record != NULL:             # <<<<<<<<<<<<<<
 *                 self.last_key = key
 *                 self.last_record = record
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":643
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         if _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             # If it isn't a sha1 key, then it won't be in this leaf node
 *             record = self._lookup_record(sha1)
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":650
 *                 self.last_record = record
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":640
 *         return NULL
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":652
 *         return False
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8__getitem__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_key) {
  char __pyx_v_sha1[20];
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_record;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":655
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 *         record = NULL             # <<<<<<<<<<<<<<
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 */
  __pyx_v_record = NULL;

  /* "bzrlib/_btree_serializer_pyx.pyx":656
 *         cdef gc_chk_sha1_record *record
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:             # <<<<<<<<<<<<<<
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
 */
  __pyx_t_2 = ((__pyx_v_self->last_record != NULL) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_key == __pyx_v_self->last_key);
  __pyx_t_3 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":657
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record             # <<<<<<<<<<<<<<
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 */
    __pyx_t_4 = __pyx_v_self->last_record;
    __pyx_v_record = __pyx_t_4;

    /* "bzrlib/_btree_serializer_pyx.pyx":656
 *         cdef gc_chk_sha1_record *record
 *         record = NULL
 *         if self.last_record != NULL and key is self.last_key:             # <<<<<<<<<<<<<<
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":658
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
 */
  __pyx_t_1 = (__pyx_f_6bzrlib_21_btree_serializer_pyx__key_to_sha1(__pyx_v_key, __pyx_v_sha1) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":659
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)             # <<<<<<<<<<<<<<
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_lookup_record(__pyx_v_self, __pyx_v_sha1); if (unlikely(__pyx_t_4 == NULL && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 659; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_record = __pyx_t_4;

    /* "bzrlib/_btree_serializer_pyx.pyx":658
 *         if self.last_record != NULL and key is self.last_key:
 *             record = self.last_record
 *         elif _key_to_sha1(key, sha1):             # <<<<<<<<<<<<<<
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
 */
  }
  __pyx_L3:;

  /* "bzrlib/_btree_serializer_pyx.pyx":660
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 *         if record == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)
 */
  __pyx_t_1 = ((__pyx_v_record == NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":661
 *             record = self._lookup_record(sha1)
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))             # <<<<<<<<<<<<<<
 *         return self._record_to_value_and_refs(record)
 * 
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 661; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_key);
    __Pyx_GIVEREF(__pyx_v_key);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_key);
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_key_r_is_not_present, __pyx_t_5); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 661; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 661; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_KeyError, __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 661; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 661; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":660
 *         elif _key_to_sha1(key, sha1):
 *             record = self._lookup_record(sha1)
 *         if record == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":662
 *         if record == NULL:
 *             raise KeyError('key %r is not present' % (key,))
 *         return self._record_to_value_and_refs(record)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = ((PyObject *)((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_value_and_refs(__pyx_v_self, __pyx_v_record)); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 662; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":652
 *         return False
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         cdef char sha1[20]
 *         cdef gc_chk_sha1_record *record
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":664
 *         return self._record_to_value_and_refs(record)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.num_records
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_10__len__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":665
 * 
 *     def __len__(self):
 *         return self.num_records             # <<<<<<<<<<<<<<
 * 
 *     def all_keys(self):
 */
  __pyx_r = __pyx_v_self->num_records;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":664
 *         return self._record_to_value_and_refs(record)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.num_records
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":667
 *         return self.num_records
 * 
 *     def all_keys(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_keys (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12all_keys(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_keys", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":669
 *     def all_keys(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 669; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":670
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < self.num_records:             # <<<<<<<<<<<<<<
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 *         return result
 */
  __pyx_t_2 = __pyx_v_self->num_records;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":671
 *         result = []
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_key((__pyx_v_self->records[__pyx_v_i]).sha1)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 671; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_3 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 671; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":672
 *         for i from 0 <= i < self.num_records:
 *             PyList_Append(result, _sha1_to_key(self.records[i].sha1))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def all_items(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":667
 *         return self.num_records
 * 
 *     def all_keys(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.all_keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":674
 *         return result
 * 
 *     def all_items(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_items (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_14all_items(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  StaticTuple *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_items", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":676
 *     def all_items(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 676; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":677
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < self.num_records:             # <<<<<<<<<<<<<<
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)
 */
  __pyx_t_2 = __pyx_v_self->num_records;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":678
 *         result = []
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])             # <<<<<<<<<<<<<<
 *             PyList_Append(result, item)
 *         return result
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_record_to_item(__pyx_v_self, (&(__pyx_v_self->records[__pyx_v_i])))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 678; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_item, ((StaticTuple *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":679
 *         for i from 0 <= i < self.num_records:
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_3 = PyList_Append(__pyx_v_result, ((PyObject *)__pyx_v_item)); if (unlikely(__pyx_t_3 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 679; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":680
 *             item = self._record_to_item(&self.records[i])
 *             PyList_Append(result, item)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":674
 *         return result
 * 
 *     def all_items(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.all_items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":682
 *         return result
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise             # <<<<<<<<<<<<<<
 *         """Count how many records are in this section."""
 *         cdef char *c_cur
 */

static int __pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records(CYTHON_UNUSED struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_content, char *__pyx_v_c_end) {
  char *__pyx_v_c_cur;
  int __pyx_v_num_records;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_count_records", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":687
 *         cdef int num_records
 * 
 *         c_cur = c_content             # <<<<<<<<<<<<<<
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:
 */
  __pyx_v_c_cur = __pyx_v_c_content;

  /* "bzrlib/_btree_serializer_pyx.pyx":688
 * 
 *         c_cur = c_content
 *         num_records = 0             # <<<<<<<<<<<<<<
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 */
  __pyx_v_num_records = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":689
 *         c_cur = c_content
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:             # <<<<<<<<<<<<<<
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_c_cur != NULL) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_c_cur < __pyx_v_c_end) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "bzrlib/_btree_serializer_pyx.pyx":690
 *         num_records = 0
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);             # <<<<<<<<<<<<<<
 *             if c_cur == NULL:
 *                 break
 */
    __pyx_v_c_cur = ((char *)memchr(__pyx_v_c_cur, '\n', (__pyx_v_c_end - __pyx_v_c_cur)));

    /* "bzrlib/_btree_serializer_pyx.pyx":691
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:             # <<<<<<<<<<<<<<
 *                 break
 *             c_cur = c_cur + 1
 */
    __pyx_t_1 = ((__pyx_v_c_cur == NULL) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_btree_serializer_pyx.pyx":692
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:
 *                 break             # <<<<<<<<<<<<<<
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1
 */
      goto __pyx_L4_break;

      /* "bzrlib/_btree_serializer_pyx.pyx":691
 *         while c_cur != NULL and c_cur < c_end:
 *             c_cur = <char *>memchr(c_cur, c'\n', c_end - c_cur);
 *             if c_cur == NULL:             # <<<<<<<<<<<<<<
 *                 break
 *             c_cur = c_cur + 1
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":693
 *             if c_cur == NULL:
 *                 break
 *             c_cur = c_cur + 1             # <<<<<<<<<<<<<<
 *             num_records = num_records + 1
 *         return num_records
 */
    __pyx_v_c_cur = (__pyx_v_c_cur + 1);

    /* "bzrlib/_btree_serializer_pyx.pyx":694
 *                 break
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1             # <<<<<<<<<<<<<<
 *         return num_records
 * 
 */
    __pyx_v_num_records = (__pyx_v_num_records + 1);
  }
  __pyx_L4_break:;

  /* "bzrlib/_btree_serializer_pyx.pyx":695
 *             c_cur = c_cur + 1
 *             num_records = num_records + 1
 *         return num_records             # <<<<<<<<<<<<<<
 * 
 *     cdef _parse_bytes(self, bytes):
 */
  __pyx_r = __pyx_v_num_records;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":682
 *         return result
 * 
 *     cdef int _count_records(self, char *c_content, char *c_end): # cannot_raise             # <<<<<<<<<<<<<<
 *         """Count how many records are in this section."""
 *         cdef char *c_cur
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":697
 *         return num_records
 * 
 *     cdef _parse_bytes(self, bytes):             # <<<<<<<<<<<<<<
 *         """Parse the string 'bytes' into content."""
 *         cdef char *c_bytes
 */

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_bytes) {
  char *__pyx_v_c_bytes;
  char *__pyx_v_c_cur;
  char *__pyx_v_c_end;
  Py_ssize_t __pyx_v_n_bytes;
  int __pyx_v_num_records;
  int __pyx_v_entry;
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_t_4;
  int __pyx_t_5;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_bytes", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":707
 *         cdef gc_chk_sha1_record *cur_record
 * 
 *         if not PyString_CheckExact(bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError('We only support parsing plain 8-bit strings.')
 *         # Pass 1, count how many records there will be
 */
  __pyx_t_1 = ((!(PyString_CheckExact(__pyx_v_bytes) != 0)) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":708
 * 
 *         if not PyString_CheckExact(bytes):
 *             raise TypeError('We only support parsing plain 8-bit strings.')             # <<<<<<<<<<<<<<
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyString_GET_SIZE(bytes)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":707
 *         cdef gc_chk_sha1_record *cur_record
 * 
 *         if not PyString_CheckExact(bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError('We only support parsing plain 8-bit strings.')
 *         # Pass 1, count how many records there will be
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":710
 *             raise TypeError('We only support parsing plain 8-bit strings.')
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyString_GET_SIZE(bytes)             # <<<<<<<<<<<<<<
 *         c_bytes = PyString_AS_STRING(bytes)
 *         c_end = c_bytes + n_bytes
 */
  __pyx_v_n_bytes = PyString_GET_SIZE(__pyx_v_bytes);

  /* "bzrlib/_btree_serializer_pyx.pyx":711
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyString_GET_SIZE(bytes)
 *         c_bytes = PyString_AS_STRING(bytes)             # <<<<<<<<<<<<<<
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, 'type=leaf\n', 10):
 */
  __pyx_v_c_bytes = PyString_AS_STRING(__pyx_v_bytes);

  /* "bzrlib/_btree_serializer_pyx.pyx":712
 *         n_bytes = PyString_GET_SIZE(bytes)
 *         c_bytes = PyString_AS_STRING(bytes)
 *         c_end = c_bytes + n_bytes             # <<<<<<<<<<<<<<
 *         if strncmp(c_bytes, 'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 */
  __pyx_v_c_end = (__pyx_v_c_bytes + __pyx_v_n_bytes);

  /* "bzrlib/_btree_serializer_pyx.pyx":713
 *         c_bytes = PyString_AS_STRING(bytes)
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, 'type=leaf\n', 10):             # <<<<<<<<<<<<<<
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (bytes[:10],))
 */
  __pyx_t_1 = (strncmp(__pyx_v_c_bytes, __pyx_k_type_leaf_2, 10) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":715
 *         if strncmp(c_bytes, 'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (bytes[:10],))             # <<<<<<<<<<<<<<
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)
 */
    __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_bytes, 0, 10, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_bytes_did_not_start_with_type_le, __pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":714
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, 'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"             # <<<<<<<<<<<<<<
 *                              % (bytes[:10],))
 *         c_cur = c_bytes + 10
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 714; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 714; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 714; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":713
 *         c_bytes = PyString_AS_STRING(bytes)
 *         c_end = c_bytes + n_bytes
 *         if strncmp(c_bytes, 'type=leaf\n', 10):             # <<<<<<<<<<<<<<
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (bytes[:10],))
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":716
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (bytes[:10],))
 *         c_cur = c_bytes + 10             # <<<<<<<<<<<<<<
 *         num_records = self._count_records(c_cur, c_end)
 *         # Now allocate the memory for these items, and go to town
 */
  __pyx_v_c_cur = (__pyx_v_c_bytes + 10);

  /* "bzrlib/_btree_serializer_pyx.pyx":717
 *                              % (bytes[:10],))
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)             # <<<<<<<<<<<<<<
 *         # Now allocate the memory for these items, and go to town
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *
 */
  __pyx_v_num_records = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_count_records(__pyx_v_self, __pyx_v_c_cur, __pyx_v_c_end);

  /* "bzrlib/_btree_serializer_pyx.pyx":719
 *         num_records = self._count_records(c_cur, c_end)
 *         # Now allocate the memory for these items, and go to town
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *             # <<<<<<<<<<<<<<
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records
 */
  __pyx_v_self->records = ((__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *)PyMem_Malloc((__pyx_v_num_records * ((sizeof(unsigned short)) + (sizeof(__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record))))));

  /* "bzrlib/_btree_serializer_pyx.pyx":721
 *         self.records = <gc_chk_sha1_record*>PyMem_Malloc(num_records *
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records             # <<<<<<<<<<<<<<
 *         cur_record = self.records
 *         entry = 0
 */
  __pyx_v_self->num_records = __pyx_v_num_records;

  /* "bzrlib/_btree_serializer_pyx.pyx":722
 *             (sizeof(unsigned short) + sizeof(gc_chk_sha1_record)))
 *         self.num_records = num_records
 *         cur_record = self.records             # <<<<<<<<<<<<<<
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 */
  __pyx_t_4 = __pyx_v_self->records;
  __pyx_v_cur_record = __pyx_t_4;

  /* "bzrlib/_btree_serializer_pyx.pyx":723
 *         self.num_records = num_records
 *         cur_record = self.records
 *         entry = 0             # <<<<<<<<<<<<<<
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 */
  __pyx_v_entry = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":724
 *         cur_record = self.records
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:             # <<<<<<<<<<<<<<
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1
 */
  while (1) {
    __pyx_t_5 = ((__pyx_v_c_cur != NULL) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = ((__pyx_v_c_cur < __pyx_v_c_end) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_5 = ((__pyx_v_entry < __pyx_v_num_records) != 0);
    __pyx_t_1 = __pyx_t_5;
    __pyx_L7_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "bzrlib/_btree_serializer_pyx.pyx":725
 *         entry = 0
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)             # <<<<<<<<<<<<<<
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_parse_one_entry(__pyx_v_self, __pyx_v_c_cur, __pyx_v_c_end, __pyx_v_cur_record); if (unlikely(__pyx_t_6 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 725; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_c_cur = __pyx_t_6;

    /* "bzrlib/_btree_serializer_pyx.pyx":726
 *         while c_cur != NULL and c_cur < c_end and entry < num_records:
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1             # <<<<<<<<<<<<<<
 *             entry = entry + 1
 *         if (entry != self.num_records
 */
    __pyx_v_cur_record = (__pyx_v_cur_record + 1);

    /* "bzrlib/_btree_serializer_pyx.pyx":727
 *             c_cur = self._parse_one_entry(c_cur, c_end, cur_record)
 *             cur_record = cur_record + 1
 *             entry = entry + 1             # <<<<<<<<<<<<<<
 *         if (entry != self.num_records
 *             or c_cur != c_end
 */
    __pyx_v_entry = (__pyx_v_entry + 1);
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":728
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 */
  __pyx_t_5 = ((__pyx_v_entry != __pyx_v_self->num_records) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L11_bool_binop_done;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":729
 *             entry = entry + 1
 *         if (entry != self.num_records
 *             or c_cur != c_end             # <<<<<<<<<<<<<<
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')
 */
  __pyx_t_5 = ((__pyx_v_c_cur != __pyx_v_c_end) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_1 = __pyx_t_5;
    goto __pyx_L11_bool_binop_done;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":730
 *         if (entry != self.num_records
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):             # <<<<<<<<<<<<<<
 *             raise ValueError('Something went wrong while parsing.')
 *         # Pass 3: build the offset map
 */
  __pyx_t_5 = ((__pyx_v_cur_record != (__pyx_v_self->records + __pyx_v_self->num_records)) != 0);
  __pyx_t_1 = __pyx_t_5;
  __pyx_L11_bool_binop_done:;

  /* "bzrlib/_btree_serializer_pyx.pyx":728
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 */
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":731
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')             # <<<<<<<<<<<<<<
 *         # Pass 3: build the offset map
 *         self._compute_common()
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 731; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 731; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":728
 *             cur_record = cur_record + 1
 *             entry = entry + 1
 *         if (entry != self.num_records             # <<<<<<<<<<<<<<
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":733
 *             raise ValueError('Something went wrong while parsing.')
 *         # Pass 3: build the offset map
 *         self._compute_common()             # <<<<<<<<<<<<<<
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_compute_common(__pyx_v_self); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 733; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":697
 *         return num_records
 * 
 *     cdef _parse_bytes(self, bytes):             # <<<<<<<<<<<<<<
 *         """Parse the string 'bytes' into content."""
 *         cdef char *c_bytes
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._parse_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":735
 *         self._compute_common()
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,             # <<<<<<<<<<<<<<
 *                                 gc_chk_sha1_record *cur_record) except NULL:
 *         """Read a single sha record from the bytes.
 */

static char *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry(CYTHON_UNUSED struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_c_cur, char *__pyx_v_c_end, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *__pyx_v_cur_record) {
  char *__pyx_v_c_next;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_one_entry", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":742
 *         """
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):             # <<<<<<<<<<<<<<
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
 */
  __pyx_t_1 = (strncmp(__pyx_v_c_cur, __pyx_k_sha1, 5) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":744
 *         if strncmp(c_cur, 'sha1:', 5):
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 */
    __pyx_t_2 = __pyx_f_6bzrlib_21_btree_serializer_pyx_safe_string_from_size(__pyx_v_c_cur, 10); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_line_did_not_start_with_sha1_r, __pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 744; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":743
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):
 *             raise ValueError('line did not start with sha1: %r'             # <<<<<<<<<<<<<<
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 743; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 743; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 743; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":742
 *         """
 *         cdef char *c_next
 *         if strncmp(c_cur, 'sha1:', 5):             # <<<<<<<<<<<<<<
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":745
 *             raise ValueError('line did not start with sha1: %r'
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5             # <<<<<<<<<<<<<<
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
 */
  __pyx_v_c_cur = (__pyx_v_c_cur + 5);

  /* "bzrlib/_btree_serializer_pyx.pyx":746
 *                 % (safe_string_from_size(c_cur, 10),))
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)             # <<<<<<<<<<<<<<
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 */
  __pyx_v_c_next = ((char *)memchr(__pyx_v_c_cur, '\x00', (__pyx_v_c_end - __pyx_v_c_cur)));

  /* "bzrlib/_btree_serializer_pyx.pyx":747
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):             # <<<<<<<<<<<<<<
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 */
  __pyx_t_4 = ((__pyx_v_c_next == NULL) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next - __pyx_v_c_cur) != 40) != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":748
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')             # <<<<<<<<<<<<<<
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 748; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 748; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":747
 *         c_cur = c_cur + 5
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):             # <<<<<<<<<<<<<<
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":749
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):             # <<<<<<<<<<<<<<
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 */
  __pyx_t_1 = ((!(__pyx_f_6bzrlib_21_btree_serializer_pyx__unhexlify_sha1(__pyx_v_c_cur, __pyx_v_cur_record->sha1) != 0)) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":750
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 750; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 750; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":749
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):             # <<<<<<<<<<<<<<
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":751
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":752
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':             # <<<<<<<<<<<<<<
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
 */
  __pyx_t_1 = (((__pyx_v_c_cur[0]) != '\x00') != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":753
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 753; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 753; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":752
 *             raise ValueError('We failed to unhexlify')
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':             # <<<<<<<<<<<<<<
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":754
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1             # <<<<<<<<<<<<<<
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 */
  __pyx_v_c_cur = (__pyx_v_c_cur + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":755
 *             raise ValueError('only 1 null, not 2 as expected')
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')
 */
  __pyx_v_cur_record->block_offset = strtoll(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "bzrlib/_btree_serializer_pyx.pyx":756
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != ' ') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L10_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":757
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 757; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 757; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":756
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":758
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":759
 *             raise ValueError('Failed to parse block offset')
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 */
  __pyx_v_cur_record->block_length = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "bzrlib/_btree_serializer_pyx.pyx":760
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L13_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != ' ') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L13_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":761
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 761; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 761; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":760
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":762
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":763
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 */
  __pyx_v_cur_record->record_start = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "bzrlib/_btree_serializer_pyx.pyx":764
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != ' ') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L16_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":765
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":764
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":766
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":767
 *             raise ValueError('Failed to parse block length')
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)             # <<<<<<<<<<<<<<
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')
 */
  __pyx_v_cur_record->record_end = strtoul(__pyx_v_c_cur, (&__pyx_v_c_next), 10);

  /* "bzrlib/_btree_serializer_pyx.pyx":768
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
 */
  __pyx_t_4 = ((__pyx_v_c_cur == __pyx_v_c_next) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L19_bool_binop_done;
  }
  __pyx_t_4 = (((__pyx_v_c_next[0]) != '\n') != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L19_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":769
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         return c_cur
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":768
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':             # <<<<<<<<<<<<<<
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":770
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1             # <<<<<<<<<<<<<<
 *         return c_cur
 * 
 */
  __pyx_v_c_cur = (__pyx_v_c_next + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":771
 *             raise ValueError('Failed to parse record end')
 *         c_cur = c_next + 1
 *         return c_cur             # <<<<<<<<<<<<<<
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:
 */
  __pyx_r = __pyx_v_c_cur;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":735
 *         self._compute_common()
 * 
 *     cdef char *_parse_one_entry(self, char *c_cur, char *c_end,             # <<<<<<<<<<<<<<
 *                                 gc_chk_sha1_record *cur_record) except NULL:
 *         """Read a single sha record from the bytes.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._parse_one_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":773
 *         return c_cur
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:             # <<<<<<<<<<<<<<
 *         """Find the first interesting 8-bits of this sha1."""
 *         cdef int this_offset
 */

static int __pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, char *__pyx_v_sha1) {
  int __pyx_v_this_offset;
  unsigned int __pyx_v_as_uint;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_offset_for_sha1", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":777
 *         cdef int this_offset
 *         cdef unsigned int as_uint
 *         as_uint = _sha1_to_uint(sha1)             # <<<<<<<<<<<<<<
 *         this_offset = (as_uint >> self.common_shift) & 0xFF
 *         return this_offset
 */
  __pyx_v_as_uint = __pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_uint(__pyx_v_sha1);

  /* "bzrlib/_btree_serializer_pyx.pyx":778
 *         cdef unsigned int as_uint
 *         as_uint = _sha1_to_uint(sha1)
 *         this_offset = (as_uint >> self.common_shift) & 0xFF             # <<<<<<<<<<<<<<
 *         return this_offset
 * 
 */
  __pyx_v_this_offset = ((__pyx_v_as_uint >> __pyx_v_self->common_shift) & 0xFF);

  /* "bzrlib/_btree_serializer_pyx.pyx":779
 *         as_uint = _sha1_to_uint(sha1)
 *         this_offset = (as_uint >> self.common_shift) & 0xFF
 *         return this_offset             # <<<<<<<<<<<<<<
 * 
 *     def _get_offset_for_sha1(self, sha1):
 */
  __pyx_r = __pyx_v_this_offset;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":773
 *         return c_cur
 * 
 *     cdef int _offset_for_sha1(self, char *sha1) except -1:             # <<<<<<<<<<<<<<
 *         """Find the first interesting 8-bits of this sha1."""
 *         cdef int this_offset
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":781
 *         return this_offset
 * 
 *     def _get_offset_for_sha1(self, sha1):             # <<<<<<<<<<<<<<
 *         return self._offset_for_sha1(PyString_AS_STRING(sha1))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1(PyObject *__pyx_v_self, PyObject *__pyx_v_sha1); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1(PyObject *__pyx_v_self, PyObject *__pyx_v_sha1) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_offset_for_sha1 (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_sha1));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_16_get_offset_for_sha1(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_sha1) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_offset_for_sha1", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":782
 * 
 *     def _get_offset_for_sha1(self, sha1):
 *         return self._offset_for_sha1(PyString_AS_STRING(sha1))             # <<<<<<<<<<<<<<
 * 
 *     cdef _compute_common(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, PyString_AS_STRING(__pyx_v_sha1)); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 782; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 782; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":781
 *         return this_offset
 * 
 *     def _get_offset_for_sha1(self, sha1):             # <<<<<<<<<<<<<<
 *         return self._offset_for_sha1(PyString_AS_STRING(sha1))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._get_offset_for_sha1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":784
 *         return self._offset_for_sha1(PyString_AS_STRING(sha1))
 * 
 *     cdef _compute_common(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int first
 *         cdef unsigned int this
 */

static PyObject *__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  unsigned int __pyx_v_first;
  unsigned int __pyx_v_this;
  unsigned int __pyx_v_common_mask;
  unsigned char __pyx_v_common_shift;
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_v_this_offset;
  int __pyx_v_max_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_compute_common", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":798
 *         # By XORing the records together, we can determine what bits are set in
 *         # all of them
 *         if self.num_records < 2:             # <<<<<<<<<<<<<<
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
 */
  __pyx_t_1 = ((__pyx_v_self->num_records < 2) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":801
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
 *             self.common_shift = 24             # <<<<<<<<<<<<<<
 *         else:
 *             common_mask = 0xFFFFFFFF
 */
    __pyx_v_self->common_shift = 24;

    /* "bzrlib/_btree_serializer_pyx.pyx":798
 *         # By XORing the records together, we can determine what bits are set in
 *         # all of them
 *         if self.num_records < 2:             # <<<<<<<<<<<<<<
 *             # Everything is in common if you have 0 or 1 leaves
 *             # So we'll always just shift to the first byte
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":803
 *             self.common_shift = 24
 *         else:
 *             common_mask = 0xFFFFFFFF             # <<<<<<<<<<<<<<
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:
 */
  /*else*/ {
    __pyx_v_common_mask = 0xFFFFFFFF;

    /* "bzrlib/_btree_serializer_pyx.pyx":804
 *         else:
 *             common_mask = 0xFFFFFFFF
 *             first = _sha1_to_uint(self.records[0].sha1)             # <<<<<<<<<<<<<<
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)
 */
    __pyx_v_first = __pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_uint((__pyx_v_self->records[0]).sha1);

    /* "bzrlib/_btree_serializer_pyx.pyx":805
 *             common_mask = 0xFFFFFFFF
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:             # <<<<<<<<<<<<<<
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask
 */
    __pyx_t_2 = __pyx_v_self->num_records;
    for (__pyx_v_i = 0+1; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

      /* "bzrlib/_btree_serializer_pyx.pyx":806
 *             first = _sha1_to_uint(self.records[0].sha1)
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)             # <<<<<<<<<<<<<<
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24
 */
      __pyx_v_this = __pyx_f_6bzrlib_21_btree_serializer_pyx__sha1_to_uint((__pyx_v_self->records[__pyx_v_i]).sha1);

      /* "bzrlib/_btree_serializer_pyx.pyx":807
 *             for i from 0 < i < self.num_records:
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask             # <<<<<<<<<<<<<<
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:
 */
      __pyx_v_common_mask = ((~(__pyx_v_first ^ __pyx_v_this)) & __pyx_v_common_mask);
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":808
 *                 this = _sha1_to_uint(self.records[i].sha1)
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24             # <<<<<<<<<<<<<<
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1
 */
    __pyx_v_common_shift = 24;

    /* "bzrlib/_btree_serializer_pyx.pyx":809
 *                 common_mask = (~(first ^ this)) & common_mask
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:             # <<<<<<<<<<<<<<
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_common_mask); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 809; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyNumber_And(__pyx_t_3, __pyx_int_2147483648); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 809; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 809; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_5) {
      } else {
        __pyx_t_1 = __pyx_t_5;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_5 = ((__pyx_v_common_shift > 0) != 0);
      __pyx_t_1 = __pyx_t_5;
      __pyx_L8_bool_binop_done:;
      if (!__pyx_t_1) break;

      /* "bzrlib/_btree_serializer_pyx.pyx":810
 *             common_shift = 24
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1             # <<<<<<<<<<<<<<
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift
 */
      __pyx_v_common_mask = (__pyx_v_common_mask << 1);

      /* "bzrlib/_btree_serializer_pyx.pyx":811
 *             while common_mask & 0x80000000 and common_shift > 0:
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1             # <<<<<<<<<<<<<<
 *             self.common_shift = common_shift
 *         offset = 0
 */
      __pyx_v_common_shift = (__pyx_v_common_shift - 1);
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":812
 *                 common_mask = common_mask << 1
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift             # <<<<<<<<<<<<<<
 *         offset = 0
 *         max_offset = self.num_records
 */
    __pyx_v_self->common_shift = __pyx_v_common_shift;
  }
  __pyx_L3:;

  /* "bzrlib/_btree_serializer_pyx.pyx":813
 *                 common_shift = common_shift - 1
 *             self.common_shift = common_shift
 *         offset = 0             # <<<<<<<<<<<<<<
 *         max_offset = self.num_records
 *         # We cap this loop at 254 records. All the other offsets just get
 */
  __pyx_v_offset = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":814
 *             self.common_shift = common_shift
 *         offset = 0
 *         max_offset = self.num_records             # <<<<<<<<<<<<<<
 *         # We cap this loop at 254 records. All the other offsets just get
 *         # filled with 0xff as the singleton saying 'too many'.
 */
  __pyx_t_2 = __pyx_v_self->num_records;
  __pyx_v_max_offset = __pyx_t_2;

  /* "bzrlib/_btree_serializer_pyx.pyx":819
 *         # It means that if we have >255 records we have to bisect the second
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:             # <<<<<<<<<<<<<<
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
 */
  __pyx_t_1 = ((__pyx_v_max_offset > 0xFF) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":820
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:
 *             max_offset = 255             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 */
    __pyx_v_max_offset = 0xFF;

    /* "bzrlib/_btree_serializer_pyx.pyx":819
 *         # It means that if we have >255 records we have to bisect the second
 *         # half of the list, but this is going to be very rare in practice.
 *         if max_offset > 255:             # <<<<<<<<<<<<<<
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":821
 *         if max_offset > 255:
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:             # <<<<<<<<<<<<<<
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:
 */
  __pyx_t_2 = __pyx_v_max_offset;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_2; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":822
 *             max_offset = 255
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)             # <<<<<<<<<<<<<<
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self->__pyx_vtab)->_offset_for_sha1(__pyx_v_self, (__pyx_v_self->records[__pyx_v_i]).sha1); if (unlikely(__pyx_t_6 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 822; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_this_offset = __pyx_t_6;

    /* "bzrlib/_btree_serializer_pyx.pyx":823
 *         for i from 0 <= i < max_offset:
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:             # <<<<<<<<<<<<<<
 *                 self.offsets[offset] = i
 *                 offset = offset + 1
 */
    while (1) {
      __pyx_t_1 = ((__pyx_v_offset <= __pyx_v_this_offset) != 0);
      if (!__pyx_t_1) break;

      /* "bzrlib/_btree_serializer_pyx.pyx":824
 *             this_offset = self._offset_for_sha1(self.records[i].sha1)
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i             # <<<<<<<<<<<<<<
 *                 offset = offset + 1
 *         while offset < 257:
 */
      (__pyx_v_self->offsets[__pyx_v_offset]) = __pyx_v_i;

      /* "bzrlib/_btree_serializer_pyx.pyx":825
 *             while offset <= this_offset:
 *                 self.offsets[offset] = i
 *                 offset = offset + 1             # <<<<<<<<<<<<<<
 *         while offset < 257:
 *             self.offsets[offset] = max_offset
 */
      __pyx_v_offset = (__pyx_v_offset + 1);
    }
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":826
 *                 self.offsets[offset] = i
 *                 offset = offset + 1
 *         while offset < 257:             # <<<<<<<<<<<<<<
 *             self.offsets[offset] = max_offset
 *             offset = offset + 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_offset < 0x101) != 0);
    if (!__pyx_t_1) break;

    /* "bzrlib/_btree_serializer_pyx.pyx":827
 *                 offset = offset + 1
 *         while offset < 257:
 *             self.offsets[offset] = max_offset             # <<<<<<<<<<<<<<
 *             offset = offset + 1
 * 
 */
    (__pyx_v_self->offsets[__pyx_v_offset]) = __pyx_v_max_offset;

    /* "bzrlib/_btree_serializer_pyx.pyx":828
 *         while offset < 257:
 *             self.offsets[offset] = max_offset
 *             offset = offset + 1             # <<<<<<<<<<<<<<
 * 
 *     def _get_offsets(self):
 */
    __pyx_v_offset = (__pyx_v_offset + 1);
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":784
 *         return self._offset_for_sha1(PyString_AS_STRING(sha1))
 * 
 *     cdef _compute_common(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int first
 *         cdef unsigned int this
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._compute_common", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":830
 *             offset = offset + 1
 * 
 *     def _get_offsets(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_offsets (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_18_get_offsets(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_offsets", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":832
 *     def _get_offsets(self):
 *         cdef int i
 *         result = []             # <<<<<<<<<<<<<<
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 832; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":833
 *         cdef int i
 *         result = []
 *         for i from 0 <= i < 257:             # <<<<<<<<<<<<<<
 *             PyList_Append(result, self.offsets[i])
 *         return result
 */
  for (__pyx_v_i = 0; __pyx_v_i < 0x101; __pyx_v_i++) {

    /* "bzrlib/_btree_serializer_pyx.pyx":834
 *         result = []
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_unsigned_char((__pyx_v_self->offsets[__pyx_v_i])); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_2 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 834; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":835
 *         for i from 0 <= i < 257:
 *             PyList_Append(result, self.offsets[i])
 *         return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":830
 *             offset = offset + 1
 * 
 *     def _get_offsets(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         result = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode._get_offsets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":522
 * 
 *     cdef gc_chk_sha1_record *records
 *     cdef public object last_key             # <<<<<<<<<<<<<<
 *     cdef gc_chk_sha1_record *last_record
 *     cdef public int num_records
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->last_key);
  __pyx_r = __pyx_v_self->last_key;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_2__set__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_4__del__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->last_key);
  __Pyx_DECREF(__pyx_v_self->last_key);
  __pyx_v_self->last_key = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":524
 *     cdef public object last_key
 *     cdef gc_chk_sha1_record *last_record
 *     cdef public int num_records             # <<<<<<<<<<<<<<
 *     # This is the number of bits to shift to get to the interesting byte. A
 *     # value of 24 means that the very first byte changes across all keys.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->num_records); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 524; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.num_records.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_2__set__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 524; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->num_records = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.num_records.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":530
 *     # ignore. 0 means that at least the first 3 bytes are identical, though
 *     # that is going to be very rare
 *     cdef public unsigned char common_shift             # <<<<<<<<<<<<<<
 *     # This maps an interesting byte to the first record that matches.
 *     # Equivalent to bisect.bisect_left(self.records, sha1), though only taking
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift___get__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->common_shift); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 530; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.common_shift.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_2__set__(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_char(__pyx_v_value); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 530; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->common_shift = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode.common_shift.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":838
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_11_parse_into_chk(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6bzrlib_21_btree_serializer_pyx_10_parse_into_chk[] = "Parse into a format optimized for chk records.";
static PyMethodDef __pyx_mdef_6bzrlib_21_btree_serializer_pyx_11_parse_into_chk = {"_parse_into_chk", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_11_parse_into_chk, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6bzrlib_21_btree_serializer_pyx_10_parse_into_chk};
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_11_parse_into_chk(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytes = 0;
  PyObject *__pyx_v_key_length = 0;
  PyObject *__pyx_v_ref_list_length = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_parse_into_chk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bytes,&__pyx_n_s_key_length,&__pyx_n_s_ref_list_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_bytes)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_key_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ref_list_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, 2); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_parse_into_chk") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_bytes = values[0];
    __pyx_v_key_length = values[1];
    __pyx_v_ref_list_length = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_parse_into_chk", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._parse_into_chk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_10_parse_into_chk(__pyx_self, __pyx_v_bytes, __pyx_v_key_length, __pyx_v_ref_list_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_10_parse_into_chk(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytes, PyObject *__pyx_v_key_length, PyObject *__pyx_v_ref_list_length) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_parse_into_chk", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":840
 * def _parse_into_chk(bytes, key_length, ref_list_length):
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1             # <<<<<<<<<<<<<<
 *     assert ref_list_length == 0
 *     return GCCHKSHA1LeafNode(bytes)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_key_length, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetNone(PyExc_AssertionError);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 840; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif

  /* "bzrlib/_btree_serializer_pyx.pyx":841
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 *     assert ref_list_length == 0             # <<<<<<<<<<<<<<
 *     return GCCHKSHA1LeafNode(bytes)
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_ref_list_length, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetNone(PyExc_AssertionError);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 841; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif

  /* "bzrlib/_btree_serializer_pyx.pyx":842
 *     assert key_length == 1
 *     assert ref_list_length == 0
 *     return GCCHKSHA1LeafNode(bytes)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_bytes);
  __Pyx_GIVEREF(__pyx_v_bytes);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_bytes);
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode), __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 842; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":838
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._parse_into_chk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_btree_serializer_pyx.pyx":845
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_13_flatten_node(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6bzrlib_21_btree_serializer_pyx_12_flatten_node[] = "Convert a node into the serialized form.\n\n    :param node: A tuple representing a node:\n        (index, key_tuple, value, references)\n    :param reference_lists: Does this index have reference lists?\n    :return: (string_key, flattened)\n        string_key  The serialized key for referencing this node\n        flattened   A string with the serialized form for the contents\n    ";
static PyMethodDef __pyx_mdef_6bzrlib_21_btree_serializer_pyx_13_flatten_node = {"_flatten_node", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_13_flatten_node, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6bzrlib_21_btree_serializer_pyx_12_flatten_node};
static PyObject *__pyx_pw_6bzrlib_21_btree_serializer_pyx_13_flatten_node(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_node = 0;
  PyObject *__pyx_v_reference_lists = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_flatten_node (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_node,&__pyx_n_s_reference_lists,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_node)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_reference_lists)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_flatten_node", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_flatten_node") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_node = values[0];
    __pyx_v_reference_lists = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_flatten_node", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._flatten_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_21_btree_serializer_pyx_12_flatten_node(__pyx_self, __pyx_v_node, __pyx_v_reference_lists);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_21_btree_serializer_pyx_12_flatten_node(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_node, PyObject *__pyx_v_reference_lists) {
  int __pyx_v_have_reference_lists;
  Py_ssize_t __pyx_v_flat_len;
  Py_ssize_t __pyx_v_key_len;
  Py_ssize_t __pyx_v_node_len;
  char *__pyx_v_value;
  Py_ssize_t __pyx_v_value_len;
  char *__pyx_v_out;
  Py_ssize_t __pyx_v_refs_len;
  Py_ssize_t __pyx_v_next_len;
  int __pyx_v_first_ref_list;
  int __pyx_v_first_reference;
  int __pyx_v_i;
  Py_ssize_t __pyx_v_ref_bit_len;
  PyObject *__pyx_v_string_key = NULL;
  PyObject *__pyx_v_ref_lists = NULL;
  PyObject *__pyx_v_ref_list = NULL;
  PyObject *__pyx_v_reference = NULL;
  PyObject *__pyx_v_ref_bit = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  char *__pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_flatten_node", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":869
 *     cdef Py_ssize_t ref_bit_len
 * 
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):             # <<<<<<<<<<<<<<
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))
 */
  __pyx_t_2 = ((!(PyTuple_CheckExact(__pyx_v_node) != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((!(StaticTuple_CheckExact(__pyx_v_node) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":871
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))             # <<<<<<<<<<<<<<
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 */
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_We_expected_a_tuple_or_StaticTup, ((PyObject *)Py_TYPE(__pyx_v_node))); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 871; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);

    /* "bzrlib/_btree_serializer_pyx.pyx":870
 * 
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'             # <<<<<<<<<<<<<<
 *             % type(node))
 *     node_len = len(node)
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 870; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 870; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 870; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":869
 *     cdef Py_ssize_t ref_bit_len
 * 
 *     if not PyTuple_CheckExact(node) and not StaticTuple_CheckExact(node):             # <<<<<<<<<<<<<<
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":872
 *         raise TypeError('We expected a tuple() or StaticTuple() for node not: %s'
 *             % type(node))
 *     node_len = len(node)             # <<<<<<<<<<<<<<
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 */
  __pyx_t_5 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_5 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 872; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_node_len = __pyx_t_5;

  /* "bzrlib/_btree_serializer_pyx.pyx":873
 *             % type(node))
 *     node_len = len(node)
 *     have_reference_lists = reference_lists             # <<<<<<<<<<<<<<
 *     if have_reference_lists:
 *         if node_len != 4:
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_reference_lists); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 873; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_have_reference_lists = __pyx_t_6;

  /* "bzrlib/_btree_serializer_pyx.pyx":874
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 */
  __pyx_t_1 = (__pyx_v_have_reference_lists != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":875
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 *         if node_len != 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 */
    __pyx_t_1 = ((__pyx_v_node_len != 4) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_btree_serializer_pyx.pyx":877
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))             # <<<<<<<<<<<<<<
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 */
      __pyx_t_5 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_5 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_With_ref_lists_we_expected_4_ent, __pyx_t_3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 877; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":876
 *     if have_reference_lists:
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'             # <<<<<<<<<<<<<<
 *                 % len(node))
 *     elif node_len < 3:
 */
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 876; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

      /* "bzrlib/_btree_serializer_pyx.pyx":875
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:
 *         if node_len != 4:             # <<<<<<<<<<<<<<
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":874
 *     node_len = len(node)
 *     have_reference_lists = reference_lists
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         if node_len != 4:
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 */
    goto __pyx_L6;
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":878
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 *     elif node_len < 3:             # <<<<<<<<<<<<<<
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))
 */
  __pyx_t_1 = ((__pyx_v_node_len < 3) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":880
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))             # <<<<<<<<<<<<<<
 *     # TODO: We can probably do better than string.join(), namely
 *     #       when key has only 1 item, we can just grab that string
 */
    __pyx_t_5 = PyObject_Length(__pyx_v_node); if (unlikely(__pyx_t_5 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 880; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 880; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Without_ref_lists_we_need_at_lea, __pyx_t_4); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 880; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":879
 *                 % len(node))
 *     elif node_len < 3:
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'             # <<<<<<<<<<<<<<
 *             % len(node))
 *     # TODO: We can probably do better than string.join(), namely
 */
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 879; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":878
 *             raise ValueError('With ref_lists, we expected 4 entries not: %s'
 *                 % len(node))
 *     elif node_len < 3:             # <<<<<<<<<<<<<<
 *         raise ValueError('Without ref_lists, we need at least 3 entries not: %s'
 *             % len(node))
 */
  }
  __pyx_L6:;

  /* "bzrlib/_btree_serializer_pyx.pyx":891
 *     #       We *could* do more work on our own, and grab the actual items
 *     #       lists. For now, just ask people to use a better compiler. :)
 *     string_key = '\0'.join(node[1])             # <<<<<<<<<<<<<<
 * 
 *     # TODO: instead of using string joins, precompute the final string length,
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_node, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 891; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyString_Join(__pyx_kp_s__20, __pyx_t_3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 891; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_string_key = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":905
 *     # ref := BYTES (NULL BYTES)*
 *     # value := BYTES
 *     refs_len = 0             # <<<<<<<<<<<<<<
 *     if have_reference_lists:
 *         # Figure out how many bytes it will take to store the references
 */
  __pyx_v_refs_len = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":906
 *     # value := BYTES
 *     refs_len = 0
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
 */
  __pyx_t_1 = (__pyx_v_have_reference_lists != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":908
 *     if have_reference_lists:
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]             # <<<<<<<<<<<<<<
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_node, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(__pyx_t_4 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 908; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_ref_lists = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":909
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function             # <<<<<<<<<<<<<<
 *         if next_len > 0:
 *             # If there are no nodes, we don't need to do any work
 */
    __pyx_t_5 = PyObject_Length(__pyx_v_ref_lists); if (unlikely(__pyx_t_5 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 909; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_v_next_len = __pyx_t_5;

    /* "bzrlib/_btree_serializer_pyx.pyx":910
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:             # <<<<<<<<<<<<<<
 *             # If there are no nodes, we don't need to do any work
 *             # Otherwise we will need (len - 1) '\t' characters to separate
 */
    __pyx_t_1 = ((__pyx_v_next_len > 0) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_btree_serializer_pyx.pyx":914
 *             # Otherwise we will need (len - 1) '\t' characters to separate
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 */
      __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

      /* "bzrlib/_btree_serializer_pyx.pyx":915
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *                 next_len = len(ref_list)
 *                 if next_len > 0:
 */
      if (likely(PyList_CheckExact(__pyx_v_ref_lists)) || PyTuple_CheckExact(__pyx_v_ref_lists)) {
        __pyx_t_4 = __pyx_v_ref_lists; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
        __pyx_t_7 = NULL;
      } else {
        __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_ref_lists); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      for (;;) {
        if (likely(!__pyx_t_7)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          } else {
            if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #else
            __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_3);
            #endif
          }
        } else {
          __pyx_t_3 = __pyx_t_7(__pyx_t_4);
          if (unlikely(!__pyx_t_3)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 915; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_3);
        }
        __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "bzrlib/_btree_serializer_pyx.pyx":916
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)             # <<<<<<<<<<<<<<
 *                 if next_len > 0:
 *                     # We will need (len - 1) '\r' characters to separate the
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_ref_list); if (unlikely(__pyx_t_8 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 916; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __pyx_v_next_len = __pyx_t_8;

        /* "bzrlib/_btree_serializer_pyx.pyx":917
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 *                 if next_len > 0:             # <<<<<<<<<<<<<<
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
 */
        __pyx_t_1 = ((__pyx_v_next_len > 0) != 0);
        if (__pyx_t_1) {

          /* "bzrlib/_btree_serializer_pyx.pyx":920
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)
 */
          __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

          /* "bzrlib/_btree_serializer_pyx.pyx":921
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:             # <<<<<<<<<<<<<<
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):
 */
          if (likely(PyList_CheckExact(__pyx_v_ref_list)) || PyTuple_CheckExact(__pyx_v_ref_list)) {
            __pyx_t_3 = __pyx_v_ref_list; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
            __pyx_t_9 = NULL;
          } else {
            __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_ref_list); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          for (;;) {
            if (likely(!__pyx_t_9)) {
              if (likely(PyList_CheckExact(__pyx_t_3))) {
                if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
                #if CYTHON_COMPILING_IN_CPYTHON
                __pyx_t_10 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                #else
                __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                __Pyx_GOTREF(__pyx_t_10);
                #endif
              } else {
                if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
                #if CYTHON_COMPILING_IN_CPYTHON
                __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                #else
                __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                __Pyx_GOTREF(__pyx_t_10);
                #endif
              }
            } else {
              __pyx_t_10 = __pyx_t_9(__pyx_t_3);
              if (unlikely(!__pyx_t_10)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 921; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_10);
            }
            __Pyx_XDECREF_SET(__pyx_v_reference, __pyx_t_10);
            __pyx_t_10 = 0;

            /* "bzrlib/_btree_serializer_pyx.pyx":922
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)             # <<<<<<<<<<<<<<
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(
 */
            __pyx_t_2 = ((!(PyTuple_CheckExact(__pyx_v_reference) != 0)) != 0);
            if (__pyx_t_2) {
            } else {
              __pyx_t_1 = __pyx_t_2;
              goto __pyx_L16_bool_binop_done;
            }

            /* "bzrlib/_btree_serializer_pyx.pyx":923
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):             # <<<<<<<<<<<<<<
 *                             raise TypeError(
 *                                 'We expect references to be tuples not: %s'
 */
            __pyx_t_2 = ((!(StaticTuple_CheckExact(__pyx_v_reference) != 0)) != 0);
            __pyx_t_1 = __pyx_t_2;
            __pyx_L16_bool_binop_done:;

            /* "bzrlib/_btree_serializer_pyx.pyx":922
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)             # <<<<<<<<<<<<<<
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(
 */
            if (__pyx_t_1) {

              /* "bzrlib/_btree_serializer_pyx.pyx":926
 *                             raise TypeError(
 *                                 'We expect references to be tuples not: %s'
 *                                 % type(reference))             # <<<<<<<<<<<<<<
 *                         next_len = len(reference)
 *                         if next_len > 0:
 */
              __pyx_t_10 = __Pyx_PyString_Format(__pyx_kp_s_We_expect_references_to_be_tuple, ((PyObject *)Py_TYPE(__pyx_v_reference))); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 926; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              __Pyx_GOTREF(__pyx_t_10);

              /* "bzrlib/_btree_serializer_pyx.pyx":924
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(             # <<<<<<<<<<<<<<
 *                                 'We expect references to be tuples not: %s'
 *                                 % type(reference))
 */
              __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 924; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_GIVEREF(__pyx_t_10);
              PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10);
              __pyx_t_10 = 0;
              __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_11, NULL); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 924; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_Raise(__pyx_t_10, 0, 0, 0);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              {__pyx_filename = __pyx_f[0]; __pyx_lineno = 924; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

              /* "bzrlib/_btree_serializer_pyx.pyx":922
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:
 *                         if (not PyTuple_CheckExact(reference)             # <<<<<<<<<<<<<<
 *                             and not StaticTuple_CheckExact(reference)):
 *                             raise TypeError(
 */
            }

            /* "bzrlib/_btree_serializer_pyx.pyx":927
 *                                 'We expect references to be tuples not: %s'
 *                                 % type(reference))
 *                         next_len = len(reference)             # <<<<<<<<<<<<<<
 *                         if next_len > 0:
 *                             # We will need (len - 1) '\x00' characters to
 */
            __pyx_t_12 = PyObject_Length(__pyx_v_reference); if (unlikely(__pyx_t_12 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 927; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __pyx_v_next_len = __pyx_t_12;

            /* "bzrlib/_btree_serializer_pyx.pyx":928
 *                                 % type(reference))
 *                         next_len = len(reference)
 *                         if next_len > 0:             # <<<<<<<<<<<<<<
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
 */
            __pyx_t_1 = ((__pyx_v_next_len > 0) != 0);
            if (__pyx_t_1) {

              /* "bzrlib/_btree_serializer_pyx.pyx":931
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)             # <<<<<<<<<<<<<<
 *                             for ref_bit in reference:
 *                                 if not PyString_CheckExact(ref_bit):
 */
              __pyx_v_refs_len = (__pyx_v_refs_len + (__pyx_v_next_len - 1));

              /* "bzrlib/_btree_serializer_pyx.pyx":932
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:             # <<<<<<<<<<<<<<
 *                                 if not PyString_CheckExact(ref_bit):
 *                                     raise TypeError('We expect reference bits'
 */
              if (likely(PyList_CheckExact(__pyx_v_reference)) || PyTuple_CheckExact(__pyx_v_reference)) {
                __pyx_t_10 = __pyx_v_reference; __Pyx_INCREF(__pyx_t_10); __pyx_t_12 = 0;
                __pyx_t_13 = NULL;
              } else {
                __pyx_t_12 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_reference); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_t_13 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
              }
              for (;;) {
                if (likely(!__pyx_t_13)) {
                  if (likely(PyList_CheckExact(__pyx_t_10))) {
                    if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_10)) break;
                    #if CYTHON_COMPILING_IN_CPYTHON
                    __pyx_t_11 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_12); __Pyx_INCREF(__pyx_t_11); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                    #else
                    __pyx_t_11 = PySequence_ITEM(__pyx_t_10, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                    __Pyx_GOTREF(__pyx_t_11);
                    #endif
                  } else {
                    if (__pyx_t_12 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
                    #if CYTHON_COMPILING_IN_CPYTHON
                    __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_12); __Pyx_INCREF(__pyx_t_11); __pyx_t_12++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                    #else
                    __pyx_t_11 = PySequence_ITEM(__pyx_t_10, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                    __Pyx_GOTREF(__pyx_t_11);
                    #endif
                  }
                } else {
                  __pyx_t_11 = __pyx_t_13(__pyx_t_10);
                  if (unlikely(!__pyx_t_11)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 932; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_11);
                }
                __Pyx_XDECREF_SET(__pyx_v_ref_bit, __pyx_t_11);
                __pyx_t_11 = 0;

                /* "bzrlib/_btree_serializer_pyx.pyx":933
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:
 *                                 if not PyString_CheckExact(ref_bit):             # <<<<<<<<<<<<<<
 *                                     raise TypeError('We expect reference bits'
 *                                         ' to be strings not: %s'
 */
                __pyx_t_1 = ((!(PyString_CheckExact(__pyx_v_ref_bit) != 0)) != 0);
                if (__pyx_t_1) {

                  /* "bzrlib/_btree_serializer_pyx.pyx":936
 *                                     raise TypeError('We expect reference bits'
 *                                         ' to be strings not: %s'
 *                                         % type(<object>ref_bit))             # <<<<<<<<<<<<<<
 *                                 refs_len = refs_len + PyString_GET_SIZE(ref_bit)
 * 
 */
                  __pyx_t_11 = __Pyx_PyString_Format(__pyx_kp_s_We_expect_reference_bits_to_be_s, ((PyObject *)Py_TYPE(__pyx_v_ref_bit))); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 936; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                  __Pyx_GOTREF(__pyx_t_11);

                  /* "bzrlib/_btree_serializer_pyx.pyx":934
 *                             for ref_bit in reference:
 *                                 if not PyString_CheckExact(ref_bit):
 *                                     raise TypeError('We expect reference bits'             # <<<<<<<<<<<<<<
 *                                         ' to be strings not: %s'
 *                                         % type(<object>ref_bit))
 */
                  __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 934; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                  __Pyx_GOTREF(__pyx_t_14);
                  __Pyx_GIVEREF(__pyx_t_11);
                  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_11);
                  __pyx_t_11 = 0;
                  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_14, NULL); if (unlikely(!__pyx_t_11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 934; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
                  __Pyx_GOTREF(__pyx_t_11);
                  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                  __Pyx_Raise(__pyx_t_11, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 934; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

                  /* "bzrlib/_btree_serializer_pyx.pyx":933
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:
 *                                 if not PyString_CheckExact(ref_bit):             # <<<<<<<<<<<<<<
 *                                     raise TypeError('We expect reference bits'
 *                                         ' to be strings not: %s'
 */
                }

                /* "bzrlib/_btree_serializer_pyx.pyx":937
 *                                         ' to be strings not: %s'
 *                                         % type(<object>ref_bit))
 *                                 refs_len = refs_len + PyString_GET_SIZE(ref_bit)             # <<<<<<<<<<<<<<
 * 
 *     # So we have the (key NULL refs NULL value LF)
 */
                __pyx_v_refs_len = (__pyx_v_refs_len + PyString_GET_SIZE(__pyx_v_ref_bit));

                /* "bzrlib/_btree_serializer_pyx.pyx":932
 *                             # separate the reference key
 *                             refs_len = refs_len + (next_len - 1)
 *                             for ref_bit in reference:             # <<<<<<<<<<<<<<
 *                                 if not PyString_CheckExact(ref_bit):
 *                                     raise TypeError('We expect reference bits'
 */
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

              /* "bzrlib/_btree_serializer_pyx.pyx":928
 *                                 % type(reference))
 *                         next_len = len(reference)
 *                         if next_len > 0:             # <<<<<<<<<<<<<<
 *                             # We will need (len - 1) '\x00' characters to
 *                             # separate the reference key
 */
            }

            /* "bzrlib/_btree_serializer_pyx.pyx":921
 *                     # references
 *                     refs_len = refs_len + (next_len - 1)
 *                     for reference in ref_list:             # <<<<<<<<<<<<<<
 *                         if (not PyTuple_CheckExact(reference)
 *                             and not StaticTuple_CheckExact(reference)):
 */
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "bzrlib/_btree_serializer_pyx.pyx":917
 *             for ref_list in ref_lists:
 *                 next_len = len(ref_list)
 *                 if next_len > 0:             # <<<<<<<<<<<<<<
 *                     # We will need (len - 1) '\r' characters to separate the
 *                     # references
 */
        }

        /* "bzrlib/_btree_serializer_pyx.pyx":915
 *             # the reference lists
 *             refs_len = refs_len + (next_len - 1)
 *             for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *                 next_len = len(ref_list)
 *                 if next_len > 0:
 */
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":910
 *         ref_lists = node[3]
 *         next_len = len(ref_lists) # TODO: use a Py function
 *         if next_len > 0:             # <<<<<<<<<<<<<<
 *             # If there are no nodes, we don't need to do any work
 *             # Otherwise we will need (len - 1) '\t' characters to separate
 */
    }

    /* "bzrlib/_btree_serializer_pyx.pyx":906
 *     # value := BYTES
 *     refs_len = 0
 *     if have_reference_lists:             # <<<<<<<<<<<<<<
 *         # Figure out how many bytes it will take to store the references
 *         ref_lists = node[3]
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":940
 * 
 *     # So we have the (key NULL refs NULL value LF)
 *     key_len = PyString_Size(string_key)             # <<<<<<<<<<<<<<
 *     val = node[2]
 *     if not PyString_CheckExact(val):
 */
  __pyx_v_key_len = PyString_Size(__pyx_v_string_key);

  /* "bzrlib/_btree_serializer_pyx.pyx":941
 *     # So we have the (key NULL refs NULL value LF)
 *     key_len = PyString_Size(string_key)
 *     val = node[2]             # <<<<<<<<<<<<<<
 *     if not PyString_CheckExact(val):
 *         raise TypeError('Expected a plain str for value not: %s'
 */
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_node, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(__pyx_t_4 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 941; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_val = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":942
 *     key_len = PyString_Size(string_key)
 *     val = node[2]
 *     if not PyString_CheckExact(val):             # <<<<<<<<<<<<<<
 *         raise TypeError('Expected a plain str for value not: %s'
 *                         % type(val))
 */
  __pyx_t_1 = ((!(PyString_CheckExact(__pyx_v_val) != 0)) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":944
 *     if not PyString_CheckExact(val):
 *         raise TypeError('Expected a plain str for value not: %s'
 *                         % type(val))             # <<<<<<<<<<<<<<
 *     value = PyString_AS_STRING(val)
 *     value_len = PyString_GET_SIZE(val)
 */
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Expected_a_plain_str_for_value_n, ((PyObject *)Py_TYPE(__pyx_v_val))); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 944; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);

    /* "bzrlib/_btree_serializer_pyx.pyx":943
 *     val = node[2]
 *     if not PyString_CheckExact(val):
 *         raise TypeError('Expected a plain str for value not: %s'             # <<<<<<<<<<<<<<
 *                         % type(val))
 *     value = PyString_AS_STRING(val)
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 943; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 943; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 943; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_btree_serializer_pyx.pyx":942
 *     key_len = PyString_Size(string_key)
 *     val = node[2]
 *     if not PyString_CheckExact(val):             # <<<<<<<<<<<<<<
 *         raise TypeError('Expected a plain str for value not: %s'
 *                         % type(val))
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":945
 *         raise TypeError('Expected a plain str for value not: %s'
 *                         % type(val))
 *     value = PyString_AS_STRING(val)             # <<<<<<<<<<<<<<
 *     value_len = PyString_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 */
  __pyx_v_value = PyString_AS_STRING(__pyx_v_val);

  /* "bzrlib/_btree_serializer_pyx.pyx":946
 *                         % type(val))
 *     value = PyString_AS_STRING(val)
 *     value_len = PyString_GET_SIZE(val)             # <<<<<<<<<<<<<<
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 *     line = PyString_FromStringAndSize(NULL, flat_len)
 */
  __pyx_v_value_len = PyString_GET_SIZE(__pyx_v_val);

  /* "bzrlib/_btree_serializer_pyx.pyx":947
 *     value = PyString_AS_STRING(val)
 *     value_len = PyString_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)             # <<<<<<<<<<<<<<
 *     line = PyString_FromStringAndSize(NULL, flat_len)
 *     # Get a pointer to the new buffer
 */
  __pyx_v_flat_len = (((((__pyx_v_key_len + 1) + __pyx_v_refs_len) + 1) + __pyx_v_value_len) + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":948
 *     value_len = PyString_GET_SIZE(val)
 *     flat_len = (key_len + 1 + refs_len + 1 + value_len + 1)
 *     line = PyString_FromStringAndSize(NULL, flat_len)             # <<<<<<<<<<<<<<
 *     # Get a pointer to the new buffer
 *     out = PyString_AsString(line)
 */
  __pyx_t_4 = PyString_FromStringAndSize(NULL, __pyx_v_flat_len); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 948; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_line = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":950
 *     line = PyString_FromStringAndSize(NULL, flat_len)
 *     # Get a pointer to the new buffer
 *     out = PyString_AsString(line)             # <<<<<<<<<<<<<<
 *     memcpy(out, PyString_AsString(string_key), key_len)
 *     out = out + key_len
 */
  __pyx_t_15 = PyString_AsString(__pyx_v_line); if (unlikely(__pyx_t_15 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 950; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_out = __pyx_t_15;

  /* "bzrlib/_btree_serializer_pyx.pyx":951
 *     # Get a pointer to the new buffer
 *     out = PyString_AsString(line)
 *     memcpy(out, PyString_AsString(string_key), key_len)             # <<<<<<<<<<<<<<
 *     out = out + key_len
 *     out[0] = c'\0'
 */
  __pyx_t_15 = PyString_AsString(__pyx_v_string_key); if (unlikely(__pyx_t_15 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 951; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  memcpy(__pyx_v_out, __pyx_t_15, __pyx_v_key_len);

  /* "bzrlib/_btree_serializer_pyx.pyx":952
 *     out = PyString_AsString(line)
 *     memcpy(out, PyString_AsString(string_key), key_len)
 *     out = out + key_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\0'
 *     out = out + 1
 */
  __pyx_v_out = (__pyx_v_out + __pyx_v_key_len);

  /* "bzrlib/_btree_serializer_pyx.pyx":953
 *     memcpy(out, PyString_AsString(string_key), key_len)
 *     out = out + key_len
 *     out[0] = c'\0'             # <<<<<<<<<<<<<<
 *     out = out + 1
 *     if refs_len > 0:
 */
  (__pyx_v_out[0]) = '\x00';

  /* "bzrlib/_btree_serializer_pyx.pyx":954
 *     out = out + key_len
 *     out[0] = c'\0'
 *     out = out + 1             # <<<<<<<<<<<<<<
 *     if refs_len > 0:
 *         first_ref_list = 1
 */
  __pyx_v_out = (__pyx_v_out + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":955
 *     out[0] = c'\0'
 *     out = out + 1
 *     if refs_len > 0:             # <<<<<<<<<<<<<<
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 */
  __pyx_t_1 = ((__pyx_v_refs_len > 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_btree_serializer_pyx.pyx":956
 *     out = out + 1
 *     if refs_len > 0:
 *         first_ref_list = 1             # <<<<<<<<<<<<<<
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:
 */
    __pyx_v_first_ref_list = 1;

    /* "bzrlib/_btree_serializer_pyx.pyx":957
 *     if refs_len > 0:
 *         first_ref_list = 1
 *         for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
 */
    if (unlikely(!__pyx_v_ref_lists)) { __Pyx_RaiseUnboundLocalError("ref_lists"); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;} }
    if (likely(PyList_CheckExact(__pyx_v_ref_lists)) || PyTuple_CheckExact(__pyx_v_ref_lists)) {
      __pyx_t_4 = __pyx_v_ref_lists; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_ref_lists); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 957; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_ref_list, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":958
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:             # <<<<<<<<<<<<<<
 *                 out[0] = c'\t'
 *                 out = out + 1
 */
      __pyx_t_1 = ((__pyx_v_first_ref_list == 0) != 0);
      if (__pyx_t_1) {

        /* "bzrlib/_btree_serializer_pyx.pyx":959
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'             # <<<<<<<<<<<<<<
 *                 out = out + 1
 *             first_ref_list = 0
 */
        (__pyx_v_out[0]) = '\t';

        /* "bzrlib/_btree_serializer_pyx.pyx":960
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
 *                 out = out + 1             # <<<<<<<<<<<<<<
 *             first_ref_list = 0
 *             first_reference = 1
 */
        __pyx_v_out = (__pyx_v_out + 1);

        /* "bzrlib/_btree_serializer_pyx.pyx":958
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 *             if first_ref_list == 0:             # <<<<<<<<<<<<<<
 *                 out[0] = c'\t'
 *                 out = out + 1
 */
      }

      /* "bzrlib/_btree_serializer_pyx.pyx":961
 *                 out[0] = c'\t'
 *                 out = out + 1
 *             first_ref_list = 0             # <<<<<<<<<<<<<<
 *             first_reference = 1
 *             for reference in ref_list:
 */
      __pyx_v_first_ref_list = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":962
 *                 out = out + 1
 *             first_ref_list = 0
 *             first_reference = 1             # <<<<<<<<<<<<<<
 *             for reference in ref_list:
 *                 if first_reference == 0:
 */
      __pyx_v_first_reference = 1;

      /* "bzrlib/_btree_serializer_pyx.pyx":963
 *             first_ref_list = 0
 *             first_reference = 1
 *             for reference in ref_list:             # <<<<<<<<<<<<<<
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
 */
      if (likely(PyList_CheckExact(__pyx_v_ref_list)) || PyTuple_CheckExact(__pyx_v_ref_list)) {
        __pyx_t_3 = __pyx_v_ref_list; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_ref_list); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      }
      for (;;) {
        if (likely(!__pyx_t_9)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_9(__pyx_t_3);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 963; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        __Pyx_XDECREF_SET(__pyx_v_reference, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "bzrlib/_btree_serializer_pyx.pyx":964
 *             first_reference = 1
 *             for reference in ref_list:
 *                 if first_reference == 0:             # <<<<<<<<<<<<<<
 *                     out[0] = c'\r'
 *                     out = out + 1
 */
        __pyx_t_1 = ((__pyx_v_first_reference == 0) != 0);
        if (__pyx_t_1) {

          /* "bzrlib/_btree_serializer_pyx.pyx":965
 *             for reference in ref_list:
 *                 if first_reference == 0:
 *                     out[0] = c'\r'             # <<<<<<<<<<<<<<
 *                     out = out + 1
 *                 first_reference = 0
 */
          (__pyx_v_out[0]) = '\r';

          /* "bzrlib/_btree_serializer_pyx.pyx":966
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
 *                     out = out + 1             # <<<<<<<<<<<<<<
 *                 first_reference = 0
 *                 next_len = len(reference)
 */
          __pyx_v_out = (__pyx_v_out + 1);

          /* "bzrlib/_btree_serializer_pyx.pyx":964
 *             first_reference = 1
 *             for reference in ref_list:
 *                 if first_reference == 0:             # <<<<<<<<<<<<<<
 *                     out[0] = c'\r'
 *                     out = out + 1
 */
        }

        /* "bzrlib/_btree_serializer_pyx.pyx":967
 *                     out[0] = c'\r'
 *                     out = out + 1
 *                 first_reference = 0             # <<<<<<<<<<<<<<
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 */
        __pyx_v_first_reference = 0;

        /* "bzrlib/_btree_serializer_pyx.pyx":968
 *                     out = out + 1
 *                 first_reference = 0
 *                 next_len = len(reference)             # <<<<<<<<<<<<<<
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:
 */
        __pyx_t_12 = PyObject_Length(__pyx_v_reference); if (unlikely(__pyx_t_12 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 968; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __pyx_v_next_len = __pyx_t_12;

        /* "bzrlib/_btree_serializer_pyx.pyx":969
 *                 first_reference = 0
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:             # <<<<<<<<<<<<<<
 *                     if i != 0:
 *                         out[0] = c'\x00'
 */
        __pyx_t_12 = __pyx_v_next_len;
        for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_12; __pyx_v_i++) {

          /* "bzrlib/_btree_serializer_pyx.pyx":970
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:             # <<<<<<<<<<<<<<
 *                         out[0] = c'\x00'
 *                         out = out + 1
 */
          __pyx_t_1 = ((__pyx_v_i != 0) != 0);
          if (__pyx_t_1) {

            /* "bzrlib/_btree_serializer_pyx.pyx":971
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:
 *                         out[0] = c'\x00'             # <<<<<<<<<<<<<<
 *                         out = out + 1
 *                     ref_bit = reference[i]
 */
            (__pyx_v_out[0]) = '\x00';

            /* "bzrlib/_btree_serializer_pyx.pyx":972
 *                     if i != 0:
 *                         out[0] = c'\x00'
 *                         out = out + 1             # <<<<<<<<<<<<<<
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyString_GET_SIZE(ref_bit)
 */
            __pyx_v_out = (__pyx_v_out + 1);

            /* "bzrlib/_btree_serializer_pyx.pyx":970
 *                 next_len = len(reference)
 *                 for i from 0 <= i < next_len:
 *                     if i != 0:             # <<<<<<<<<<<<<<
 *                         out[0] = c'\x00'
 *                         out = out + 1
 */
          }

          /* "bzrlib/_btree_serializer_pyx.pyx":973
 *                         out[0] = c'\x00'
 *                         out = out + 1
 *                     ref_bit = reference[i]             # <<<<<<<<<<<<<<
 *                     ref_bit_len = PyString_GET_SIZE(ref_bit)
 *                     memcpy(out, PyString_AS_STRING(ref_bit), ref_bit_len)
 */
          __pyx_t_10 = __Pyx_GetItemInt(__pyx_v_reference, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(__pyx_t_10 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 973; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_XDECREF_SET(__pyx_v_ref_bit, __pyx_t_10);
          __pyx_t_10 = 0;

          /* "bzrlib/_btree_serializer_pyx.pyx":974
 *                         out = out + 1
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyString_GET_SIZE(ref_bit)             # <<<<<<<<<<<<<<
 *                     memcpy(out, PyString_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len
 */
          __pyx_v_ref_bit_len = PyString_GET_SIZE(__pyx_v_ref_bit);

          /* "bzrlib/_btree_serializer_pyx.pyx":975
 *                     ref_bit = reference[i]
 *                     ref_bit_len = PyString_GET_SIZE(ref_bit)
 *                     memcpy(out, PyString_AS_STRING(ref_bit), ref_bit_len)             # <<<<<<<<<<<<<<
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'
 */
          memcpy(__pyx_v_out, PyString_AS_STRING(__pyx_v_ref_bit), __pyx_v_ref_bit_len);

          /* "bzrlib/_btree_serializer_pyx.pyx":976
 *                     ref_bit_len = PyString_GET_SIZE(ref_bit)
 *                     memcpy(out, PyString_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\0'
 *     out = out  + 1
 */
          __pyx_v_out = (__pyx_v_out + __pyx_v_ref_bit_len);
        }

        /* "bzrlib/_btree_serializer_pyx.pyx":963
 *             first_ref_list = 0
 *             first_reference = 1
 *             for reference in ref_list:             # <<<<<<<<<<<<<<
 *                 if first_reference == 0:
 *                     out[0] = c'\r'
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "bzrlib/_btree_serializer_pyx.pyx":957
 *     if refs_len > 0:
 *         first_ref_list = 1
 *         for ref_list in ref_lists:             # <<<<<<<<<<<<<<
 *             if first_ref_list == 0:
 *                 out[0] = c'\t'
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "bzrlib/_btree_serializer_pyx.pyx":955
 *     out[0] = c'\0'
 *     out = out + 1
 *     if refs_len > 0:             # <<<<<<<<<<<<<<
 *         first_ref_list = 1
 *         for ref_list in ref_lists:
 */
  }

  /* "bzrlib/_btree_serializer_pyx.pyx":977
 *                     memcpy(out, PyString_AS_STRING(ref_bit), ref_bit_len)
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'             # <<<<<<<<<<<<<<
 *     out = out  + 1
 *     memcpy(out, value, value_len)
 */
  (__pyx_v_out[0]) = '\x00';

  /* "bzrlib/_btree_serializer_pyx.pyx":978
 *                     out = out + ref_bit_len
 *     out[0] = c'\0'
 *     out = out  + 1             # <<<<<<<<<<<<<<
 *     memcpy(out, value, value_len)
 *     out = out + value_len
 */
  __pyx_v_out = (__pyx_v_out + 1);

  /* "bzrlib/_btree_serializer_pyx.pyx":979
 *     out[0] = c'\0'
 *     out = out  + 1
 *     memcpy(out, value, value_len)             # <<<<<<<<<<<<<<
 *     out = out + value_len
 *     out[0] = c'\n'
 */
  memcpy(__pyx_v_out, __pyx_v_value, __pyx_v_value_len);

  /* "bzrlib/_btree_serializer_pyx.pyx":980
 *     out = out  + 1
 *     memcpy(out, value, value_len)
 *     out = out + value_len             # <<<<<<<<<<<<<<
 *     out[0] = c'\n'
 *     return string_key, line
 */
  __pyx_v_out = (__pyx_v_out + __pyx_v_value_len);

  /* "bzrlib/_btree_serializer_pyx.pyx":981
 *     memcpy(out, value, value_len)
 *     out = out + value_len
 *     out[0] = c'\n'             # <<<<<<<<<<<<<<
 *     return string_key, line
 */
  (__pyx_v_out[0]) = '\n';

  /* "bzrlib/_btree_serializer_pyx.pyx":982
 *     out = out + value_len
 *     out[0] = c'\n'
 *     return string_key, line             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 982; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_string_key);
  __Pyx_GIVEREF(__pyx_v_string_key);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_string_key);
  __Pyx_INCREF(__pyx_v_line);
  __Pyx_GIVEREF(__pyx_v_line);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_line);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "bzrlib/_btree_serializer_pyx.pyx":845
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("bzrlib._btree_serializer_pyx._flatten_node", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_string_key);
  __Pyx_XDECREF(__pyx_v_ref_lists);
  __Pyx_XDECREF(__pyx_v_ref_list);
  __Pyx_XDECREF(__pyx_v_reference);
  __Pyx_XDECREF(__pyx_v_ref_bit);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser __pyx_vtable_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser;

static PyObject *__pyx_tp_new_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)o);
  p->__pyx_vtab = __pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser;
  p->bytes = Py_None; Py_INCREF(Py_None);
  p->keys = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o) {
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->bytes);
  Py_CLEAR(p->keys);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)o;
  if (p->bytes) {
    e = (*v)(p->bytes, a); if (e) return e;
  }
  if (p->keys) {
    e = (*v)(p->keys, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *p = (struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *)o;
  tmp = ((PyObject*)p->bytes);
  p->bytes = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->keys);
  p->keys = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser[] = {
  {"parse", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_3parse, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser = {
  PyVarObject_HEAD_INIT(0, 0)
  "bzrlib._btree_serializer_pyx.BTreeLeafParser", /*tp_name*/
  sizeof(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Parse the leaf nodes of a BTree index.\n\n    :ivar bytes: The PyString object containing the uncompressed text for the\n        node.\n    :ivar key_length: An integer describing how many pieces the keys have for\n        this index.\n    :ivar ref_list_length: An integer describing how many references this index\n        contains.\n    :ivar keys: A PyList of keys found in this node.\n\n    :ivar _cur_str: A pointer to the start of the next line to parse\n    :ivar _end_str: A pointer to the end of bytes\n    :ivar _start: Pointer to the location within the current line while\n        parsing.\n    :ivar _header_found: True when we have parsed the header for this node\n    ", /*tp_doc*/
  __pyx_tp_traverse_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser, /*tp_traverse*/
  __pyx_tp_clear_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode;

static PyObject *__pyx_tp_new_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o);
  p->__pyx_vtab = __pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  p->last_key = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o) {
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->last_key);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  if (p->last_key) {
    e = (*v)(p->last_key, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *p = (struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *)o;
  tmp = ((PyObject*)p->last_key);
  p->last_key = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_min_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7min_key_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_max_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7max_key_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_1__get__(o);
}

static int __pyx_setprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_3__set__(o, v);
  }
  else {
    return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_8last_key_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_1__get__(o);
}

static int __pyx_setprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11num_records_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_1__get__(o);
}

static int __pyx_setprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_12common_shift_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode[] = {
  {"__sizeof__", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_1__sizeof__, METH_NOARGS, 0},
  {"all_keys", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_13all_keys, METH_NOARGS, 0},
  {"all_items", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_15all_items, METH_NOARGS, 0},
  {"_get_offset_for_sha1", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_17_get_offset_for_sha1, METH_O, 0},
  {"_get_offsets", (PyCFunction)__pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_19_get_offsets, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode[] = {
  {(char *)"min_key", __pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_min_key, 0, 0, 0},
  {(char *)"max_key", __pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_max_key, 0, 0, 0},
  {(char *)"last_key", __pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key, __pyx_setprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_last_key, 0, 0},
  {(char *)"num_records", __pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records, __pyx_setprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_num_records, 0, 0},
  {(char *)"common_shift", __pyx_getprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift, __pyx_setprop_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_common_shift, 0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_GCCHKSHA1LeafNode = {
  __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_7__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_GCCHKSHA1LeafNode = {
  __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_11__len__, /*mp_length*/
  __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_9__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode = {
  PyVarObject_HEAD_INIT(0, 0)
  "bzrlib._btree_serializer_pyx.GCCHKSHA1LeafNode", /*tp_name*/
  sizeof(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_GCCHKSHA1LeafNode, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_GCCHKSHA1LeafNode, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Track all the entries for a given leaf node.", /*tp_doc*/
  __pyx_tp_traverse_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_traverse*/
  __pyx_tp_clear_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_btree_serializer_pyx",
    __pyx_k_Pyrex_extensions_to_btree_node_p, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_kp_s_Expected_a_plain_str_for_value_n, __pyx_k_Expected_a_plain_str_for_value_n, sizeof(__pyx_k_Expected_a_plain_str_for_value_n), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_find_the_value_area, __pyx_k_Failed_to_find_the_value_area, sizeof(__pyx_k_Failed_to_find_the_value_area), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_parse_block_length, __pyx_k_Failed_to_parse_block_length, sizeof(__pyx_k_Failed_to_parse_block_length), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_parse_block_offset, __pyx_k_Failed_to_parse_block_offset, sizeof(__pyx_k_Failed_to_parse_block_offset), 0, 0, 1, 0},
  {&__pyx_kp_s_Failed_to_parse_record_end, __pyx_k_Failed_to_parse_record_end, sizeof(__pyx_k_Failed_to_parse_record_end), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_s_Line_did_not_contain_40_hex_byte, __pyx_k_Line_did_not_contain_40_hex_byte, sizeof(__pyx_k_Line_did_not_contain_40_hex_byte), 0, 0, 1, 0},
  {&__pyx_kp_s_Node_did_not_start_with_type_lea, __pyx_k_Node_did_not_start_with_type_lea, sizeof(__pyx_k_Node_did_not_start_with_type_lea), 0, 0, 1, 0},
  {&__pyx_kp_s_Something_went_wrong_while_parsi, __pyx_k_Something_went_wrong_while_parsi, sizeof(__pyx_k_Something_went_wrong_while_parsi), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s_We_expect_reference_bits_to_be_s, __pyx_k_We_expect_reference_bits_to_be_s, sizeof(__pyx_k_We_expect_reference_bits_to_be_s), 0, 0, 1, 0},
  {&__pyx_kp_s_We_expect_references_to_be_tuple, __pyx_k_We_expect_references_to_be_tuple, sizeof(__pyx_k_We_expect_references_to_be_tuple), 0, 0, 1, 0},
  {&__pyx_kp_s_We_expected_a_tuple_or_StaticTup, __pyx_k_We_expected_a_tuple_or_StaticTup, sizeof(__pyx_k_We_expected_a_tuple_or_StaticTup), 0, 0, 1, 0},
  {&__pyx_kp_s_We_failed_to_unhexlify, __pyx_k_We_failed_to_unhexlify, sizeof(__pyx_k_We_failed_to_unhexlify), 0, 0, 1, 0},
  {&__pyx_kp_s_We_only_support_parsing_plain_8, __pyx_k_We_only_support_parsing_plain_8, sizeof(__pyx_k_We_only_support_parsing_plain_8), 0, 0, 1, 0},
  {&__pyx_kp_s_With_ref_lists_we_expected_4_ent, __pyx_k_With_ref_lists_we_expected_4_ent, sizeof(__pyx_k_With_ref_lists_we_expected_4_ent), 0, 0, 1, 0},
  {&__pyx_kp_s_Without_ref_lists_we_need_at_lea, __pyx_k_Without_ref_lists_we_need_at_lea, sizeof(__pyx_k_Without_ref_lists_we_need_at_lea), 0, 0, 1, 0},
  {&__pyx_kp_s__20, __pyx_k__20, sizeof(__pyx_k__20), 0, 0, 1, 0},
  {&__pyx_n_s_as_bin, __pyx_k_as_bin, sizeof(__pyx_k_as_bin), 0, 0, 1, 1},
  {&__pyx_n_s_as_bin_sha, __pyx_k_as_bin_sha, sizeof(__pyx_k_as_bin_sha), 0, 0, 1, 1},
  {&__pyx_n_s_as_hex, __pyx_k_as_hex, sizeof(__pyx_k_as_hex), 0, 0, 1, 1},
  {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
  {&__pyx_kp_s_bytes_did_not_start_with_type_le, __pyx_k_bytes_did_not_start_with_type_le, sizeof(__pyx_k_bytes_did_not_start_with_type_le), 0, 0, 1, 0},
  {&__pyx_n_s_bzrlib__btree_serializer_pyx, __pyx_k_bzrlib__btree_serializer_pyx, sizeof(__pyx_k_bzrlib__btree_serializer_pyx), 0, 0, 1, 1},
  {&__pyx_n_s_first_ref_list, __pyx_k_first_ref_list, sizeof(__pyx_k_first_ref_list), 0, 0, 1, 1},
  {&__pyx_n_s_first_reference, __pyx_k_first_reference, sizeof(__pyx_k_first_reference), 0, 0, 1, 1},
  {&__pyx_n_s_flat_len, __pyx_k_flat_len, sizeof(__pyx_k_flat_len), 0, 0, 1, 1},
  {&__pyx_n_s_flatten_node, __pyx_k_flatten_node, sizeof(__pyx_k_flatten_node), 0, 0, 1, 1},
  {&__pyx_n_s_have_reference_lists, __pyx_k_have_reference_lists, sizeof(__pyx_k_have_reference_lists), 0, 0, 1, 1},
  {&__pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_k_home_vila_src_bzr_integration_2, sizeof(__pyx_k_home_vila_src_bzr_integration_2), 0, 0, 1, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_kp_s_invalid_key_loop_counter_self_re, __pyx_k_invalid_key_loop_counter_self_re, sizeof(__pyx_k_invalid_key_loop_counter_self_re), 0, 0, 1, 0},
  {&__pyx_kp_s_invalid_key_wanted_segment_from, __pyx_k_invalid_key_wanted_segment_from, sizeof(__pyx_k_invalid_key_wanted_segment_from), 0, 0, 1, 0},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_key_len, __pyx_k_key_len, sizeof(__pyx_k_key_len), 0, 0, 1, 1},
  {&__pyx_n_s_key_length, __pyx_k_key_length, sizeof(__pyx_k_key_length), 0, 0, 1, 1},
  {&__pyx_kp_s_key_r_is_not_present, __pyx_k_key_r_is_not_present, sizeof(__pyx_k_key_r_is_not_present), 0, 0, 1, 0},
  {&__pyx_kp_s_last_self__start, __pyx_k_last_self__start, sizeof(__pyx_k_last_self__start), 0, 0, 1, 0},
  {&__pyx_n_s_line, __pyx_k_line, sizeof(__pyx_k_line), 0, 0, 1, 1},
  {&__pyx_kp_s_line_did_not_start_with_sha1_r, __pyx_k_line_did_not_start_with_sha1_r, sizeof(__pyx_k_line_did_not_start_with_sha1_r), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_next_len, __pyx_k_next_len, sizeof(__pyx_k_next_len), 0, 0, 1, 1},
  {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
  {&__pyx_n_s_node_len, __pyx_k_node_len, sizeof(__pyx_k_node_len), 0, 0, 1, 1},
  {&__pyx_kp_s_not_a_20_byte_binary_digest, __pyx_k_not_a_20_byte_binary_digest, sizeof(__pyx_k_not_a_20_byte_binary_digest), 0, 0, 1, 0},
  {&__pyx_kp_s_not_a_40_byte_hex_digest, __pyx_k_not_a_40_byte_hex_digest, sizeof(__pyx_k_not_a_40_byte_hex_digest), 0, 0, 1, 0},
  {&__pyx_kp_s_only_1_null_not_2_as_expected, __pyx_k_only_1_null_not_2_as_expected, sizeof(__pyx_k_only_1_null_not_2_as_expected), 0, 0, 1, 0},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_parse, __pyx_k_parse, sizeof(__pyx_k_parse), 0, 0, 1, 1},
  {&__pyx_n_s_parse_into_chk, __pyx_k_parse_into_chk, sizeof(__pyx_k_parse_into_chk), 0, 0, 1, 1},
  {&__pyx_n_s_parse_leaf_lines, __pyx_k_parse_leaf_lines, sizeof(__pyx_k_parse_leaf_lines), 0, 0, 1, 1},
  {&__pyx_n_s_parser, __pyx_k_parser, sizeof(__pyx_k_parser), 0, 0, 1, 1},
  {&__pyx_n_s_py_hexlify, __pyx_k_py_hexlify, sizeof(__pyx_k_py_hexlify), 0, 0, 1, 1},
  {&__pyx_n_s_py_key_to_sha1, __pyx_k_py_key_to_sha1, sizeof(__pyx_k_py_key_to_sha1), 0, 0, 1, 1},
  {&__pyx_n_s_py_sha1_to_key, __pyx_k_py_sha1_to_key, sizeof(__pyx_k_py_sha1_to_key), 0, 0, 1, 1},
  {&__pyx_n_s_py_unhexlify, __pyx_k_py_unhexlify, sizeof(__pyx_k_py_unhexlify), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_ref_bit, __pyx_k_ref_bit, sizeof(__pyx_k_ref_bit), 0, 0, 1, 1},
  {&__pyx_n_s_ref_bit_len, __pyx_k_ref_bit_len, sizeof(__pyx_k_ref_bit_len), 0, 0, 1, 1},
  {&__pyx_n_s_ref_list, __pyx_k_ref_list, sizeof(__pyx_k_ref_list), 0, 0, 1, 1},
  {&__pyx_n_s_ref_list_length, __pyx_k_ref_list_length, sizeof(__pyx_k_ref_list_length), 0, 0, 1, 1},
  {&__pyx_n_s_ref_lists, __pyx_k_ref_lists, sizeof(__pyx_k_ref_lists), 0, 0, 1, 1},
  {&__pyx_n_s_reference, __pyx_k_reference, sizeof(__pyx_k_reference), 0, 0, 1, 1},
  {&__pyx_n_s_reference_lists, __pyx_k_reference_lists, sizeof(__pyx_k_reference_lists), 0, 0, 1, 1},
  {&__pyx_n_s_refs_len, __pyx_k_refs_len, sizeof(__pyx_k_refs_len), 0, 0, 1, 1},
  {&__pyx_kp_s_self_bytes_is_not_a_string, __pyx_k_self_bytes_is_not_a_string, sizeof(__pyx_k_self_bytes_is_not_a_string), 0, 0, 1, 0},
  {&__pyx_n_s_sha1_bin, __pyx_k_sha1_bin, sizeof(__pyx_k_sha1_bin), 0, 0, 1, 1},
  {&__pyx_kp_s_sha1_bin_must_be_a_str_of_exactl, __pyx_k_sha1_bin_must_be_a_str_of_exactl, sizeof(__pyx_k_sha1_bin_must_be_a_str_of_exactl), 0, 0, 1, 0},
  {&__pyx_n_s_string_key, __pyx_k_string_key, sizeof(__pyx_k_string_key), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_tried_to_create_a_string_with_an, __pyx_k_tried_to_create_a_string_with_an, sizeof(__pyx_k_tried_to_create_a_string_with_an), 0, 0, 1, 0},
  {&__pyx_kp_s_unexpected_reference_data_presen, __pyx_k_unexpected_reference_data_presen, sizeof(__pyx_k_unexpected_reference_data_presen), 0, 0, 1, 0},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_value_len, __pyx_k_value_len, sizeof(__pyx_k_value_len), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 99; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 661; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "bzrlib/_btree_serializer_pyx.pyx":229
 *         if last < self._start:
 *             # Unexpected error condition - fail
 *             raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *         if 0 == self._header_found:
 *             # The first line in a leaf node is the header "type=leaf\n"
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_last_self__start); if (unlikely(!__pyx_tuple_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 229; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "bzrlib/_btree_serializer_pyx.pyx":244
 *         if temp_ptr == NULL:
 *             # Invalid line
 *             raise AssertionError("Failed to find the value area")             # <<<<<<<<<<<<<<
 *         else:
 *             # Because of how conversions were done, we ended up with *lots* of
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_find_the_value_area); if (unlikely(!__pyx_tuple__2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "bzrlib/_btree_serializer_pyx.pyx":271
 *                 loop_counter = loop_counter + 1
 *                 if last < self._start:
 *                     raise AssertionError("last < self._start")             # <<<<<<<<<<<<<<
 *                 # find the next reference list end point:
 *                 temp_ptr = <char*>memchr(self._start, c'\t', last - self._start)
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_last_self__start); if (unlikely(!__pyx_tuple__3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 271; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "bzrlib/_btree_serializer_pyx.pyx":278
 *                     if loop_counter != self.ref_list_length:
 *                         # Invalid line
 *                         raise AssertionError(             # <<<<<<<<<<<<<<
 *                             "invalid key, loop_counter != self.ref_list_length")
 *                     else:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_invalid_key_loop_counter_self_re); if (unlikely(!__pyx_tuple__4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 278; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "bzrlib/_btree_serializer_pyx.pyx":307
 *             if last != self._start:
 *                 # unexpected reference data present
 *                 raise AssertionError("unexpected reference data present")             # <<<<<<<<<<<<<<
 *             node_value = StaticTuple(value, StaticTuple())
 *         PyList_Append(self.keys, StaticTuple(key, node_value))
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_unexpected_reference_data_presen); if (unlikely(!__pyx_tuple__5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 307; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "bzrlib/_btree_serializer_pyx.pyx":315
 *         cdef Py_ssize_t byte_count
 *         if not PyString_CheckExact(self.bytes):
 *             raise AssertionError('self.bytes is not a string.')             # <<<<<<<<<<<<<<
 *         byte_count = PyString_Size(self.bytes)
 *         self._cur_str = PyString_AsString(self.bytes)
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_self_bytes_is_not_a_string); if (unlikely(!__pyx_tuple__6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 315; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "bzrlib/_btree_serializer_pyx.pyx":389
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):
 *         raise ValueError('not a 40-byte hex digest')             # <<<<<<<<<<<<<<
 *     as_bin = PyString_FromStringAndSize(NULL, 20)
 *     if _unhexlify_sha1(PyString_AS_STRING(as_hex), PyString_AS_STRING(as_bin)):
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_not_a_40_byte_hex_digest); if (unlikely(!__pyx_tuple__7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "bzrlib/_btree_serializer_pyx.pyx":412
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):
 *         raise ValueError('not a 20-byte binary digest')             # <<<<<<<<<<<<<<
 *     as_hex = PyString_FromStringAndSize(NULL, 40)
 *     _hexlify_sha1(PyString_AS_STRING(as_bin), PyString_AS_STRING(as_hex))
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_not_a_20_byte_binary_digest); if (unlikely(!__pyx_tuple__8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 412; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "bzrlib/_btree_serializer_pyx.pyx":485
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:
 *         raise ValueError('sha1_bin must be a str of exactly 20 bytes')             # <<<<<<<<<<<<<<
 *     return _sha1_to_key(PyString_AS_STRING(sha1_bin))
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_sha1_bin_must_be_a_str_of_exactl); if (unlikely(!__pyx_tuple__9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 485; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "bzrlib/_btree_serializer_pyx.pyx":708
 * 
 *         if not PyString_CheckExact(bytes):
 *             raise TypeError('We only support parsing plain 8-bit strings.')             # <<<<<<<<<<<<<<
 *         # Pass 1, count how many records there will be
 *         n_bytes = PyString_GET_SIZE(bytes)
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_We_only_support_parsing_plain_8); if (unlikely(!__pyx_tuple__10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 708; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "bzrlib/_btree_serializer_pyx.pyx":715
 *         if strncmp(c_bytes, 'type=leaf\n', 10):
 *             raise ValueError("bytes did not start with 'type=leaf\\n': %r"
 *                              % (bytes[:10],))             # <<<<<<<<<<<<<<
 *         c_cur = c_bytes + 10
 *         num_records = self._count_records(c_cur, c_end)
 */
  __pyx_slice__11 = PySlice_New(Py_None, __pyx_int_10, Py_None); if (unlikely(!__pyx_slice__11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 715; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);

  /* "bzrlib/_btree_serializer_pyx.pyx":731
 *             or c_cur != c_end
 *             or cur_record != self.records + self.num_records):
 *             raise ValueError('Something went wrong while parsing.')             # <<<<<<<<<<<<<<
 *         # Pass 3: build the offset map
 *         self._compute_common()
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_Something_went_wrong_while_parsi); if (unlikely(!__pyx_tuple__12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 731; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "bzrlib/_btree_serializer_pyx.pyx":748
 *         c_next = <char *>memchr(c_cur, c'\0', c_end - c_cur)
 *         if c_next == NULL or (c_next - c_cur != 40):
 *             raise ValueError('Line did not contain 40 hex bytes')             # <<<<<<<<<<<<<<
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Line_did_not_contain_40_hex_byte); if (unlikely(!__pyx_tuple__13)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 748; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "bzrlib/_btree_serializer_pyx.pyx":750
 *             raise ValueError('Line did not contain 40 hex bytes')
 *         if not _unhexlify_sha1(c_cur, cur_record.sha1):
 *             raise ValueError('We failed to unhexlify')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_We_failed_to_unhexlify); if (unlikely(!__pyx_tuple__14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 750; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "bzrlib/_btree_serializer_pyx.pyx":753
 *         c_cur = c_next + 1
 *         if c_cur[0] != c'\0':
 *             raise ValueError('only 1 null, not 2 as expected')             # <<<<<<<<<<<<<<
 *         c_cur = c_cur + 1
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_only_1_null_not_2_as_expected); if (unlikely(!__pyx_tuple__15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 753; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "bzrlib/_btree_serializer_pyx.pyx":757
 *         cur_record.block_offset = strtoll(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block offset')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_parse_block_offset); if (unlikely(!__pyx_tuple__16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 757; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "bzrlib/_btree_serializer_pyx.pyx":761
 *         cur_record.block_length = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_parse_block_length); if (unlikely(!__pyx_tuple__17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 761; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "bzrlib/_btree_serializer_pyx.pyx":765
 *         cur_record.record_start = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c' ':
 *             raise ValueError('Failed to parse block length')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_parse_block_length); if (unlikely(!__pyx_tuple__18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 765; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "bzrlib/_btree_serializer_pyx.pyx":769
 *         cur_record.record_end = strtoul(c_cur, &c_next, 10)
 *         if c_cur == c_next or c_next[0] != c'\n':
 *             raise ValueError('Failed to parse record end')             # <<<<<<<<<<<<<<
 *         c_cur = c_next + 1
 *         return c_cur
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_Failed_to_parse_record_end); if (unlikely(!__pyx_tuple__19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 769; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "bzrlib/_btree_serializer_pyx.pyx":325
 * 
 * 
 * def _parse_leaf_lines(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(bytes, key_length, ref_list_length)
 *     return parser.parse()
 */
  __pyx_tuple__21 = PyTuple_Pack(4, __pyx_n_s_bytes, __pyx_n_s_key_length, __pyx_n_s_ref_list_length, __pyx_n_s_parser); if (unlikely(!__pyx_tuple__21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(3, 0, 4, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_parse_leaf_lines, 325, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_btree_serializer_pyx.pyx":386
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):
 */
  __pyx_tuple__23 = PyTuple_Pack(2, __pyx_n_s_as_hex, __pyx_n_s_as_bin); if (unlikely(!__pyx_tuple__23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_py_unhexlify, 386, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_btree_serializer_pyx.pyx":409
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_n_s_as_bin, __pyx_n_s_as_hex); if (unlikely(!__pyx_tuple__25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_py_hexlify, 409, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_btree_serializer_pyx.pyx":446
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_n_s_key, __pyx_n_s_as_bin_sha); if (unlikely(!__pyx_tuple__27)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_py_key_to_sha1, 446, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_btree_serializer_pyx.pyx":482
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_n_s_sha1_bin); if (unlikely(!__pyx_tuple__29)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_py_sha1_to_key, 482, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_btree_serializer_pyx.pyx":838
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */
  __pyx_tuple__31 = PyTuple_Pack(3, __pyx_n_s_bytes, __pyx_n_s_key_length, __pyx_n_s_ref_list_length); if (unlikely(!__pyx_tuple__31)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_parse_into_chk, 838, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_btree_serializer_pyx.pyx":845
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */
  __pyx_tuple__33 = PyTuple_Pack(22, __pyx_n_s_node, __pyx_n_s_reference_lists, __pyx_n_s_have_reference_lists, __pyx_n_s_flat_len, __pyx_n_s_key_len, __pyx_n_s_node_len, __pyx_n_s_value, __pyx_n_s_value_len, __pyx_n_s_out, __pyx_n_s_refs_len, __pyx_n_s_next_len, __pyx_n_s_first_ref_list, __pyx_n_s_first_reference, __pyx_n_s_i, __pyx_n_s_ref_bit_len, __pyx_n_s_string_key, __pyx_n_s_ref_lists, __pyx_n_s_ref_list, __pyx_n_s_reference, __pyx_n_s_ref_bit, __pyx_n_s_val, __pyx_n_s_line); if (unlikely(!__pyx_tuple__33)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(2, 0, 22, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_flatten_node, 845, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_2147483648 = PyInt_FromString((char *)"2147483648", 0, 0); if (unlikely(!__pyx_int_2147483648)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_int_4294967295 = PyInt_FromString((char *)"4294967295", 0, 0); if (unlikely(!__pyx_int_4294967295)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_btree_serializer_pyx(void); /*proto*/
PyMODINIT_FUNC init_btree_serializer_pyx(void)
#else
PyMODINIT_FUNC PyInit__btree_serializer_pyx(void); /*proto*/
PyMODINIT_FUNC PyInit__btree_serializer_pyx(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__btree_serializer_pyx(void)", 0);
  if (__Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_btree_serializer_pyx", __pyx_methods, __pyx_k_Pyrex_extensions_to_btree_node_p, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  if (__pyx_module_is_main_bzrlib___btree_serializer_pyx) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "bzrlib._btree_serializer_pyx")) {
      if (unlikely(PyDict_SetItemString(modules, "bzrlib._btree_serializer_pyx", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser = &__pyx_vtable_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser.extract_key = (PyObject *(*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *, char *))__pyx_f_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_extract_key;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser.process_line = (int (*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser *))__pyx_f_6bzrlib_21_btree_serializer_pyx_15BTreeLeafParser_process_line;
  if (PyType_Ready(&__pyx_type_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 123; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_type_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser.tp_dict, __pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 123; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyObject_SetAttrString(__pyx_m, "BTreeLeafParser", (PyObject *)&__pyx_type_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 123; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser = &__pyx_type_6bzrlib_21_btree_serializer_pyx_BTreeLeafParser;
  __pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode = &__pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._record_to_value_and_refs = (StaticTuple *(*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_value_and_refs;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._record_to_item = (StaticTuple *(*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__record_to_item;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._lookup_record = (__pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *(*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__lookup_record;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._count_records = (int (*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__count_records;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._parse_bytes = (PyObject *(*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, PyObject *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_bytes;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._parse_one_entry = (char *(*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *, char *, __pyx_t_6bzrlib_21_btree_serializer_pyx_gc_chk_sha1_record *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__parse_one_entry;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._offset_for_sha1 = (int (*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *, char *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__offset_for_sha1;
  __pyx_vtable_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode._compute_common = (PyObject *(*)(struct __pyx_obj_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode *))__pyx_f_6bzrlib_21_btree_serializer_pyx_17GCCHKSHA1LeafNode__compute_common;
  if (PyType_Ready(&__pyx_type_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 518; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_type_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode.tp_dict, __pyx_vtabptr_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 518; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyObject_SetAttrString(__pyx_m, "GCCHKSHA1LeafNode", (PyObject *)&__pyx_type_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 518; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode = &__pyx_type_6bzrlib_21_btree_serializer_pyx_GCCHKSHA1LeafNode;
  /*--- Type import code ---*/
  __pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple = __Pyx_ImportType("bzrlib._static_tuple_c", "StaticTuple", sizeof(StaticTuple), 0); if (unlikely(!__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 26; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif

  /* "bzrlib/_btree_serializer_pyx.pyx":77
 * # have to worry about exception checking.
 * ## extern cdef class StaticTuple
 * import sys             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":120
 * 
 * # This sets up the StaticTuple C_API functionality
 * import_static_tuple_c()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = import_static_tuple_c(); if (unlikely(__pyx_t_2 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_btree_serializer_pyx.pyx":325
 * 
 * 
 * def _parse_leaf_lines(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     parser = BTreeLeafParser(bytes, key_length, ref_list_length)
 *     return parser.parse()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_21_btree_serializer_pyx_1_parse_leaf_lines, NULL, __pyx_n_s_bzrlib__btree_serializer_pyx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_leaf_lines, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 325; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":347
 * cdef int _unhexbuf[256]
 * cdef char *_hexbuf
 * _hexbuf = '0123456789abcdef'             # <<<<<<<<<<<<<<
 * 
 * cdef _populate_unhexbuf():
 */
  __pyx_v_6bzrlib_21_btree_serializer_pyx__hexbuf = __pyx_k_0123456789abcdef;

  /* "bzrlib/_btree_serializer_pyx.pyx":359
 *     for i from 10 <= i < 16: # ABCDEF => 10, 11, 12, 13, 14, 15, 16
 *         _unhexbuf[(i - 10 + c'A')] = i
 * _populate_unhexbuf()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_6bzrlib_21_btree_serializer_pyx__populate_unhexbuf(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 359; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":386
 * 
 * 
 * def _py_unhexlify(as_hex):             # <<<<<<<<<<<<<<
 *     """For the test infrastructure, just thunks to _unhexlify_sha1"""
 *     if len(as_hex) != 40 or not PyString_CheckExact(as_hex):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_21_btree_serializer_pyx_3_py_unhexlify, NULL, __pyx_n_s_bzrlib__btree_serializer_pyx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_unhexlify, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 386; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":409
 * 
 * 
 * def _py_hexlify(as_bin):             # <<<<<<<<<<<<<<
 *     """For test infrastructure, thunk to _hexlify_sha1"""
 *     if len(as_bin) != 20 or not PyString_CheckExact(as_bin):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_21_btree_serializer_pyx_5_py_hexlify, NULL, __pyx_n_s_bzrlib__btree_serializer_pyx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_hexlify, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 409; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":446
 * 
 * 
 * def _py_key_to_sha1(key):             # <<<<<<<<<<<<<<
 *     """Map a key to a simple sha1 string.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_21_btree_serializer_pyx_7_py_key_to_sha1, NULL, __pyx_n_s_bzrlib__btree_serializer_pyx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_key_to_sha1, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 446; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":482
 * 
 * 
 * def _py_sha1_to_key(sha1_bin):             # <<<<<<<<<<<<<<
 *     """Test thunk to check the sha1 mapping."""
 *     if not PyString_CheckExact(sha1_bin) or PyString_GET_SIZE(sha1_bin) != 20:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_21_btree_serializer_pyx_9_py_sha1_to_key, NULL, __pyx_n_s_bzrlib__btree_serializer_pyx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_py_sha1_to_key, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 482; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":838
 * 
 * 
 * def _parse_into_chk(bytes, key_length, ref_list_length):             # <<<<<<<<<<<<<<
 *     """Parse into a format optimized for chk records."""
 *     assert key_length == 1
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_21_btree_serializer_pyx_11_parse_into_chk, NULL, __pyx_n_s_bzrlib__btree_serializer_pyx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_into_chk, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 838; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":845
 * 
 * 
 * def _flatten_node(node, reference_lists):             # <<<<<<<<<<<<<<
 *     """Convert a node into the serialized form.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_21_btree_serializer_pyx_13_flatten_node, NULL, __pyx_n_s_bzrlib__btree_serializer_pyx); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_flatten_node, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 845; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_btree_serializer_pyx.pyx":1
 * # Copyright (C) 2008, 2009, 2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init bzrlib._btree_serializer_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init bzrlib._btree_serializer_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject* args = PyTuple_Pack(1, arg);
    return (likely(args)) ? __Pyx_PyObject_Call(func, args, NULL) : NULL;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#endif
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (PyErr_ExceptionMatches(PyExc_OverflowError))
                    PyErr_Clear();
                else
                    goto bad;
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_COMPILING_IN_CPYTHON
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                #if PyLong_SHIFT < 30 && PyLong_SHIFT != 15
                default: return PyLong_Type.tp_richcompare(op1, op2, Py_EQ);
                #else
                default: Py_RETURN_FALSE;
                #endif
            }
        }
            if (a == b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            if ((double)a == (double)b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    return PyObject_RichCompare(op1, op2, Py_EQ);
}
#endif

#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (PyErr_ExceptionMatches(PyExc_OverflowError))
                        PyErr_Clear();
                    else
                        return NULL;
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, -(sdigit) digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value) {
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(unsigned char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned char),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *x) {
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 2 * PyLong_SHIFT) {
                            return (unsigned char) (((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 3 * PyLong_SHIFT) {
                            return (unsigned char) (((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 4 * PyLong_SHIFT) {
                            return (unsigned char) (((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned char, sdigit, -(sdigit) digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) ((((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, long, PyLong_AsLong(x))
            } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned char val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned char) -1;
        }
    } else {
        unsigned char val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (unsigned char) -1;
        val = __Pyx_PyInt_As_unsigned_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned char");
    return (unsigned char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned char");
    return (unsigned char) -1;
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, -(sdigit) digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
