/* Generated by Cython 0.23.3 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_23_3"
#include <stddef.h>
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
#define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
#define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) (void_result, Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__bzrlib___simple_set_pyx
#define __PYX_HAVE_API__bzrlib___simple_set_pyx
#include "python-compat.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "bzrlib/_simple_set_pyx.pyx",
};

/*--- Type declarations ---*/
struct SimpleSetObject;
struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator;

/* "bzrlib/_simple_set_pyx.pxd":31
 * 
 * 
 * cdef public api class SimpleSet [object SimpleSetObject, type SimpleSet_Type]:             # <<<<<<<<<<<<<<
 *     """A class similar to PySet, but with simpler implementation.
 * 
 */
struct SimpleSetObject {
  PyObject_HEAD
  struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *__pyx_vtab;
  Py_ssize_t _used;
  Py_ssize_t _fill;
  Py_ssize_t _mask;
  PyObject **_table;
};

__PYX_EXTERN_C DL_EXPORT(PyTypeObject) SimpleSet_Type;

/* "bzrlib/_simple_set_pyx.pyx":361
 * 
 * 
 * cdef class _SimpleSet_iterator:             # <<<<<<<<<<<<<<
 *     """Iterator over the SimpleSet structure."""
 * 
 */
struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator {
  PyObject_HEAD
  Py_ssize_t pos;
  struct SimpleSetObject *set;
  Py_ssize_t _used;
  Py_ssize_t len;
};


__PYX_EXTERN_C DL_EXPORT(PyTypeObject) SimpleSet_Type;

/* "bzrlib/_simple_set_pyx.pyx":88
 * 
 * 
 * cdef public api class SimpleSet [object SimpleSetObject, type SimpleSet_Type]:             # <<<<<<<<<<<<<<
 *     """This class can be used to track canonical forms for objects.
 * 
 */

struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet {
  PyObject *(*_get)(struct SimpleSetObject *, PyObject *);
  PyObject *(*_add)(struct SimpleSetObject *, PyObject *);
  int (*_discard)(struct SimpleSetObject *, PyObject *);
  int (*_insert_clean)(struct SimpleSetObject *, PyObject *);
  Py_ssize_t (*_resize)(struct SimpleSetObject *, Py_ssize_t);
};
static struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *__pyx_vtabptr_6bzrlib_15_simple_set_pyx_SimpleSet;

/* --- Runtime support code (head) --- */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name);

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb);

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

static int __Pyx_check_binary_version(void);

static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__get(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto*/
static int __pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__insert_clean(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto*/
static Py_ssize_t __pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__resize(struct SimpleSetObject *__pyx_v_self, Py_ssize_t __pyx_v_min_used); /* proto*/
static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__add(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto*/
static int __pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__discard(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto*/

/* Module declarations from 'bzrlib._simple_set_pyx' */
static PyTypeObject *__pyx_ptype_6bzrlib_15_simple_set_pyx_SimpleSet = 0;
static PyTypeObject *__pyx_ptype_6bzrlib_15_simple_set_pyx__SimpleSet_iterator = 0;
static PyObject *__pyx_v_6bzrlib_15_simple_set_pyx__dummy_obj = 0;
static PyObject *__pyx_v_6bzrlib_15_simple_set_pyx__dummy;
static PyObject *__pyx_v_6bzrlib_15_simple_set_pyx__NotImplemented = 0;
static struct SimpleSetObject *__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_New(void); /*proto*/
static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Add(PyObject *, PyObject *); /*proto*/
static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Contains(PyObject *, PyObject *); /*proto*/
static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Discard(PyObject *, PyObject *); /*proto*/
static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Get(struct SimpleSetObject *, PyObject *); /*proto*/
static Py_ssize_t __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Size(PyObject *); /*proto*/
static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Next(PyObject *, Py_ssize_t *, PyObject **); /*proto*/
static int __pyx_f_6bzrlib_15_simple_set_pyx__is_equal(PyObject *, long, PyObject *); /*proto*/
static struct SimpleSetObject *__pyx_f_6bzrlib_15_simple_set_pyx__check_self(PyObject *); /*proto*/
static PyObject **__pyx_f_6bzrlib_15_simple_set_pyx__lookup(struct SimpleSetObject *, PyObject *); /*proto*/
static PyObject **__pyx_f_6bzrlib_15_simple_set_pyx__SimpleSet_Lookup(PyObject *, PyObject *); /*proto*/
static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_traverse(struct SimpleSetObject *, visitproc, void *); /*proto*/
#define __Pyx_MODULE_NAME "bzrlib._simple_set_pyx"
int __pyx_module_is_main_bzrlib___simple_set_pyx = 0;

/* Implementation of 'bzrlib._simple_set_pyx' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_AssertionError;
static char __pyx_k_obj[] = "obj";
static char __pyx_k_main[] = "__main__";
static char __pyx_k_null[] = "<null>";
static char __pyx_k_test[] = "__test__";
static char __pyx_k_dummy[] = "<dummy>";
static char __pyx_k_object[] = "object";
static char __pyx_k_KeyError[] = "KeyError";
static char __pyx_k_TypeError[] = "TypeError";
static char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static char __pyx_k_MemoryError[] = "MemoryError";
static char __pyx_k_RuntimeError[] = "RuntimeError";
static char __pyx_k_StopIteration[] = "StopIteration";
static char __pyx_k_AssertionError[] = "AssertionError";
static char __pyx_k_NotImplemented[] = "NotImplemented";
static char __pyx_k_ran_out_of_slots[] = "ran out of slots.";
static char __pyx_k_Key_s_is_not_present[] = "Key %s is not present";
static char __pyx_k_self_must_not_be_None[] = "self must not be None";
static char __pyx_k_should_never_get_here[] = "should never get here";
static char __pyx_k_Definition_of_a_class_that_is_si[] = "Definition of a class that is similar to Set with some small changes.";
static char __pyx_k_Set_size_changed_during_iteratio[] = "Set size changed during iteration";
static char __pyx_k_Types_added_to_SimpleSet_must_im[] = "Types added to SimpleSet must implement both tp_richcompare and tp_hash";
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_s_Key_s_is_not_present;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_kp_s_Set_size_changed_during_iteratio;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Types_added_to_SimpleSet_must_im;
static PyObject *__pyx_kp_s_dummy;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_kp_s_null;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_kp_s_ran_out_of_slots;
static PyObject *__pyx_kp_s_self_must_not_be_None;
static PyObject *__pyx_kp_s_should_never_get_here;
static PyObject *__pyx_n_s_test;
static int __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet___init__(struct SimpleSetObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_2__sizeof__(struct SimpleSetObject *__pyx_v_self); /* proto */
static void __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4__dealloc__(struct SimpleSetObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4used___get__(struct SimpleSetObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4fill___get__(struct SimpleSetObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4mask___get__(struct SimpleSetObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_6_memory_size(struct SimpleSetObject *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_8__len__(struct SimpleSetObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_10_test_lookup(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_16_py_resize(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_min_used); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_18add(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_20discard(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_22__iter__(struct SimpleSetObject *__pyx_v_self); /* proto */
static int __pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator___init__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self, PyObject *__pyx_v_obj); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_2__iter__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_4__next__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_6__length_hint__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self); /* proto */
static PyObject *__pyx_tp_new_6bzrlib_15_simple_set_pyx_SimpleSet(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6bzrlib_15_simple_set_pyx__SimpleSet_iterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;

/* "bzrlib/_simple_set_pyx.pyx":62
 * 
 * 
 * cdef int _is_equal(PyObject *this, long this_hash, PyObject *other) except -1:             # <<<<<<<<<<<<<<
 *     cdef long other_hash
 * 
 */

static int __pyx_f_6bzrlib_15_simple_set_pyx__is_equal(PyObject *__pyx_v_this, long __pyx_v_this_hash, PyObject *__pyx_v_other) {
  long __pyx_v_other_hash;
  PyObject *__pyx_v_res = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_equal", 0);

  /* "bzrlib/_simple_set_pyx.pyx":65
 *     cdef long other_hash
 * 
 *     if this == other:             # <<<<<<<<<<<<<<
 *         return 1
 *     other_hash = PyObject_Hash(other)
 */
  __pyx_t_1 = ((__pyx_v_this == __pyx_v_other) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":66
 * 
 *     if this == other:
 *         return 1             # <<<<<<<<<<<<<<
 *     other_hash = PyObject_Hash(other)
 *     if other_hash != this_hash:
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":65
 *     cdef long other_hash
 * 
 *     if this == other:             # <<<<<<<<<<<<<<
 *         return 1
 *     other_hash = PyObject_Hash(other)
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":67
 *     if this == other:
 *         return 1
 *     other_hash = PyObject_Hash(other)             # <<<<<<<<<<<<<<
 *     if other_hash != this_hash:
 *         return 0
 */
  __pyx_t_2 = PyObject_Hash(__pyx_v_other); if (unlikely(__pyx_t_2 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 67; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_other_hash = __pyx_t_2;

  /* "bzrlib/_simple_set_pyx.pyx":68
 *         return 1
 *     other_hash = PyObject_Hash(other)
 *     if other_hash != this_hash:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_other_hash != __pyx_v_this_hash) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":69
 *     other_hash = PyObject_Hash(other)
 *     if other_hash != this_hash:
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     # This implements a subset of the PyObject_RichCompareBool functionality.
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":68
 *         return 1
 *     other_hash = PyObject_Hash(other)
 *     if other_hash != this_hash:             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":78
 *     #      equal. (It doesn't try to cast them both to some intermediate form
 *     #      that would compare equal.)
 *     res = Py_TYPE(this).tp_richcompare(this, other, Py_EQ)             # <<<<<<<<<<<<<<
 *     if res is _NotImplemented:
 *         res = Py_TYPE(other).tp_richcompare(other, this, Py_EQ)
 */
  __pyx_t_3 = Py_TYPE(__pyx_v_this)->tp_richcompare(__pyx_v_this, __pyx_v_other, Py_EQ); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 78; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_res = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":79
 *     #      that would compare equal.)
 *     res = Py_TYPE(this).tp_richcompare(this, other, Py_EQ)
 *     if res is _NotImplemented:             # <<<<<<<<<<<<<<
 *         res = Py_TYPE(other).tp_richcompare(other, this, Py_EQ)
 *         if res is _NotImplemented:
 */
  __pyx_t_1 = (__pyx_v_res == __pyx_v_6bzrlib_15_simple_set_pyx__NotImplemented);
  __pyx_t_4 = (__pyx_t_1 != 0);
  if (__pyx_t_4) {

    /* "bzrlib/_simple_set_pyx.pyx":80
 *     res = Py_TYPE(this).tp_richcompare(this, other, Py_EQ)
 *     if res is _NotImplemented:
 *         res = Py_TYPE(other).tp_richcompare(other, this, Py_EQ)             # <<<<<<<<<<<<<<
 *         if res is _NotImplemented:
 *             return 0
 */
    __pyx_t_3 = Py_TYPE(__pyx_v_other)->tp_richcompare(__pyx_v_other, __pyx_v_this, Py_EQ); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 80; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_res, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "bzrlib/_simple_set_pyx.pyx":81
 *     if res is _NotImplemented:
 *         res = Py_TYPE(other).tp_richcompare(other, this, Py_EQ)
 *         if res is _NotImplemented:             # <<<<<<<<<<<<<<
 *             return 0
 *     if res:
 */
    __pyx_t_4 = (__pyx_v_res == __pyx_v_6bzrlib_15_simple_set_pyx__NotImplemented);
    __pyx_t_1 = (__pyx_t_4 != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_simple_set_pyx.pyx":82
 *         res = Py_TYPE(other).tp_richcompare(other, this, Py_EQ)
 *         if res is _NotImplemented:
 *             return 0             # <<<<<<<<<<<<<<
 *     if res:
 *         return 1
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "bzrlib/_simple_set_pyx.pyx":81
 *     if res is _NotImplemented:
 *         res = Py_TYPE(other).tp_richcompare(other, this, Py_EQ)
 *         if res is _NotImplemented:             # <<<<<<<<<<<<<<
 *             return 0
 *     if res:
 */
    }

    /* "bzrlib/_simple_set_pyx.pyx":79
 *     #      that would compare equal.)
 *     res = Py_TYPE(this).tp_richcompare(this, other, Py_EQ)
 *     if res is _NotImplemented:             # <<<<<<<<<<<<<<
 *         res = Py_TYPE(other).tp_richcompare(other, this, Py_EQ)
 *         if res is _NotImplemented:
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":83
 *         if res is _NotImplemented:
 *             return 0
 *     if res:             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_res); if (unlikely(__pyx_t_1 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 83; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":84
 *             return 0
 *     if res:
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":83
 *         if res is _NotImplemented:
 *             return 0
 *     if res:             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":85
 *     if res:
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":62
 * 
 * 
 * cdef int _is_equal(PyObject *this, long this_hash, PyObject *other) except -1:             # <<<<<<<<<<<<<<
 *     cdef long other_hash
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._is_equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":105
 *     DEF DEFAULT_SIZE=1024
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t size, n_bytes
 * 
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet___init__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet___init__(struct SimpleSetObject *__pyx_v_self) {
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_v_n_bytes;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":108
 *         cdef Py_ssize_t size, n_bytes
 * 
 *         size = DEFAULT_SIZE             # <<<<<<<<<<<<<<
 *         self._mask = size - 1
 *         self._used = 0
 */
  __pyx_v_size = 0x400;

  /* "bzrlib/_simple_set_pyx.pyx":109
 * 
 *         size = DEFAULT_SIZE
 *         self._mask = size - 1             # <<<<<<<<<<<<<<
 *         self._used = 0
 *         self._fill = 0
 */
  __pyx_v_self->_mask = (__pyx_v_size - 1);

  /* "bzrlib/_simple_set_pyx.pyx":110
 *         size = DEFAULT_SIZE
 *         self._mask = size - 1
 *         self._used = 0             # <<<<<<<<<<<<<<
 *         self._fill = 0
 *         n_bytes = sizeof(PyObject*) * size;
 */
  __pyx_v_self->_used = 0;

  /* "bzrlib/_simple_set_pyx.pyx":111
 *         self._mask = size - 1
 *         self._used = 0
 *         self._fill = 0             # <<<<<<<<<<<<<<
 *         n_bytes = sizeof(PyObject*) * size;
 *         self._table = <PyObject **>PyMem_Malloc(n_bytes)
 */
  __pyx_v_self->_fill = 0;

  /* "bzrlib/_simple_set_pyx.pyx":112
 *         self._used = 0
 *         self._fill = 0
 *         n_bytes = sizeof(PyObject*) * size;             # <<<<<<<<<<<<<<
 *         self._table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if self._table == NULL:
 */
  __pyx_v_n_bytes = ((sizeof(PyObject *)) * __pyx_v_size);

  /* "bzrlib/_simple_set_pyx.pyx":113
 *         self._fill = 0
 *         n_bytes = sizeof(PyObject*) * size;
 *         self._table = <PyObject **>PyMem_Malloc(n_bytes)             # <<<<<<<<<<<<<<
 *         if self._table == NULL:
 *             raise MemoryError()
 */
  __pyx_v_self->_table = ((PyObject **)PyMem_Malloc(__pyx_v_n_bytes));

  /* "bzrlib/_simple_set_pyx.pyx":114
 *         n_bytes = sizeof(PyObject*) * size;
 *         self._table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if self._table == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         memset(self._table, 0, n_bytes)
 */
  __pyx_t_1 = ((__pyx_v_self->_table == NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":115
 *         self._table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if self._table == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         memset(self._table, 0, n_bytes)
 * 
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":114
 *         n_bytes = sizeof(PyObject*) * size;
 *         self._table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if self._table == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         memset(self._table, 0, n_bytes)
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":116
 *         if self._table == NULL:
 *             raise MemoryError()
 *         memset(self._table, 0, n_bytes)             # <<<<<<<<<<<<<<
 * 
 *     def __sizeof__(self):
 */
  memset(__pyx_v_self->_table, 0, __pyx_v_n_bytes);

  /* "bzrlib/_simple_set_pyx.pyx":105
 *     DEF DEFAULT_SIZE=1024
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t size, n_bytes
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":118
 *         memset(self._table, 0, n_bytes)
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         # Note: Pyrex doesn't allow sizeof(class) so we re-implement it here.
 *         # Bits are:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_3__sizeof__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_3__sizeof__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sizeof__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_2__sizeof__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_2__sizeof__(struct SimpleSetObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sizeof__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":128
 *         # better than no estimate at all
 *         # return sizeof(SimpleSet) + (self._mask + 1) * (sizeof(PyObject*))
 *         return (sizeof(PyObject) + sizeof(void*)             # <<<<<<<<<<<<<<
 *                 + 3*sizeof(Py_ssize_t) + sizeof(PyObject**)
 *                 + (self._mask + 1) * sizeof(PyObject*))
 */
  __Pyx_XDECREF(__pyx_r);

  /* "bzrlib/_simple_set_pyx.pyx":130
 *         return (sizeof(PyObject) + sizeof(void*)
 *                 + 3*sizeof(Py_ssize_t) + sizeof(PyObject**)
 *                 + (self._mask + 1) * sizeof(PyObject*))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((((((sizeof(PyObject)) + (sizeof(void *))) + (3 * (sizeof(Py_ssize_t)))) + (sizeof(PyObject **))) + ((__pyx_v_self->_mask + 1) * (sizeof(PyObject *))))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 130; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":118
 *         memset(self._table, 0, n_bytes)
 * 
 *     def __sizeof__(self):             # <<<<<<<<<<<<<<
 *         # Note: Pyrex doesn't allow sizeof(class) so we re-implement it here.
 *         # Bits are:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.__sizeof__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":132
 *                 + (self._mask + 1) * sizeof(PyObject*))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._table != NULL:
 *             PyMem_Free(self._table)
 */

/* Python wrapper */
static void __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_5__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4__dealloc__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4__dealloc__(struct SimpleSetObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":133
 * 
 *     def __dealloc__(self):
 *         if self._table != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self._table)
 *             self._table = NULL
 */
  __pyx_t_1 = ((__pyx_v_self->_table != NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":134
 *     def __dealloc__(self):
 *         if self._table != NULL:
 *             PyMem_Free(self._table)             # <<<<<<<<<<<<<<
 *             self._table = NULL
 * 
 */
    PyMem_Free(__pyx_v_self->_table);

    /* "bzrlib/_simple_set_pyx.pyx":135
 *         if self._table != NULL:
 *             PyMem_Free(self._table)
 *             self._table = NULL             # <<<<<<<<<<<<<<
 * 
 *     property used:
 */
    __pyx_v_self->_table = NULL;

    /* "bzrlib/_simple_set_pyx.pyx":133
 * 
 *     def __dealloc__(self):
 *         if self._table != NULL:             # <<<<<<<<<<<<<<
 *             PyMem_Free(self._table)
 *             self._table = NULL
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":132
 *                 + (self._mask + 1) * sizeof(PyObject*))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._table != NULL:
 *             PyMem_Free(self._table)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "bzrlib/_simple_set_pyx.pyx":138
 * 
 *     property used:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._used
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4used_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4used_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4used___get__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4used___get__(struct SimpleSetObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":139
 *     property used:
 *         def __get__(self):
 *             return self._used             # <<<<<<<<<<<<<<
 * 
 *     property fill:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->_used); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 139; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":138
 * 
 *     property used:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._used
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.used.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":142
 * 
 *     property fill:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._fill
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4fill_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4fill_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4fill___get__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4fill___get__(struct SimpleSetObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":143
 *     property fill:
 *         def __get__(self):
 *             return self._fill             # <<<<<<<<<<<<<<
 * 
 *     property mask:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->_fill); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 143; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":142
 * 
 *     property fill:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._fill
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.fill.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":146
 * 
 *     property mask:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._mask
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4mask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4mask_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4mask___get__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_4mask___get__(struct SimpleSetObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":147
 *     property mask:
 *         def __get__(self):
 *             return self._mask             # <<<<<<<<<<<<<<
 * 
 *     def _memory_size(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->_mask); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 147; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":146
 * 
 *     property mask:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._mask
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.mask.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":149
 *             return self._mask
 * 
 *     def _memory_size(self):             # <<<<<<<<<<<<<<
 *         """Return the number of bytes of memory consumed by this class."""
 *         return sizeof(self) + (sizeof(PyObject*)*(self._mask + 1))
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_7_memory_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_6_memory_size[] = "Return the number of bytes of memory consumed by this class.";
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_7_memory_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_memory_size (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_6_memory_size(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_6_memory_size(struct SimpleSetObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_memory_size", 0);

  /* "bzrlib/_simple_set_pyx.pyx":151
 *     def _memory_size(self):
 *         """Return the number of bytes of memory consumed by this class."""
 *         return sizeof(self) + (sizeof(PyObject*)*(self._mask + 1))             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(((sizeof(__pyx_v_self)) + ((sizeof(PyObject *)) * (__pyx_v_self->_mask + 1)))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 151; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":149
 *             return self._mask
 * 
 *     def _memory_size(self):             # <<<<<<<<<<<<<<
 *         """Return the number of bytes of memory consumed by this class."""
 *         return sizeof(self) + (sizeof(PyObject*)*(self._mask + 1))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._memory_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":153
 *         return sizeof(self) + (sizeof(PyObject*)*(self._mask + 1))
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._used
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_9__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_9__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_8__len__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_8__len__(struct SimpleSetObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":154
 * 
 *     def __len__(self):
 *         return self._used             # <<<<<<<<<<<<<<
 * 
 *     def _test_lookup(self, key):
 */
  __pyx_r = __pyx_v_self->_used;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":153
 *         return sizeof(self) + (sizeof(PyObject*)*(self._mask + 1))
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._used
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":156
 *         return self._used
 * 
 *     def _test_lookup(self, key):             # <<<<<<<<<<<<<<
 *         cdef PyObject **slot
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_11_test_lookup(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_11_test_lookup(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_test_lookup (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_10_test_lookup(((struct SimpleSetObject *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_10_test_lookup(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject **__pyx_v_slot;
  PyObject *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject **__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_test_lookup", 0);

  /* "bzrlib/_simple_set_pyx.pyx":159
 *         cdef PyObject **slot
 * 
 *         slot = _lookup(self, key)             # <<<<<<<<<<<<<<
 *         if slot[0] == NULL:
 *             res = '<null>'
 */
  __pyx_t_1 = __pyx_f_6bzrlib_15_simple_set_pyx__lookup(__pyx_v_self, __pyx_v_key); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 159; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_slot = __pyx_t_1;

  /* "bzrlib/_simple_set_pyx.pyx":160
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL:             # <<<<<<<<<<<<<<
 *             res = '<null>'
 *         elif slot[0] == _dummy:
 */
  __pyx_t_2 = (((__pyx_v_slot[0]) == NULL) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":161
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL:
 *             res = '<null>'             # <<<<<<<<<<<<<<
 *         elif slot[0] == _dummy:
 *             res = '<dummy>'
 */
    __Pyx_INCREF(__pyx_kp_s_null);
    __pyx_v_res = __pyx_kp_s_null;

    /* "bzrlib/_simple_set_pyx.pyx":160
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL:             # <<<<<<<<<<<<<<
 *             res = '<null>'
 *         elif slot[0] == _dummy:
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_simple_set_pyx.pyx":162
 *         if slot[0] == NULL:
 *             res = '<null>'
 *         elif slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             res = '<dummy>'
 *         else:
 */
  __pyx_t_2 = (((__pyx_v_slot[0]) == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":163
 *             res = '<null>'
 *         elif slot[0] == _dummy:
 *             res = '<dummy>'             # <<<<<<<<<<<<<<
 *         else:
 *             res = <object>slot[0]
 */
    __Pyx_INCREF(__pyx_kp_s_dummy);
    __pyx_v_res = __pyx_kp_s_dummy;

    /* "bzrlib/_simple_set_pyx.pyx":162
 *         if slot[0] == NULL:
 *             res = '<null>'
 *         elif slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             res = '<dummy>'
 *         else:
 */
    goto __pyx_L3;
  }

  /* "bzrlib/_simple_set_pyx.pyx":165
 *             res = '<dummy>'
 *         else:
 *             res = <object>slot[0]             # <<<<<<<<<<<<<<
 *         return <int>(slot - self._table), res
 * 
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)(__pyx_v_slot[0]));
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_res = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "bzrlib/_simple_set_pyx.pyx":166
 *         else:
 *             res = <object>slot[0]
 *         return <int>(slot - self._table), res             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int(((int)(__pyx_v_slot - __pyx_v_self->_table))); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_res);
  __Pyx_GIVEREF(__pyx_v_res);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_res);
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":156
 *         return self._used
 * 
 *     def _test_lookup(self, key):             # <<<<<<<<<<<<<<
 *         cdef PyObject **slot
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._test_lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":168
 *         return <int>(slot - self._table), res
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         """Is key present in this SimpleSet."""
 *         cdef PyObject **slot
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_13__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__[] = "Is key present in this SimpleSet.";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__;
#endif
static int __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_13__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__(((struct SimpleSetObject *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject **__pyx_v_slot;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject **__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":172
 *         cdef PyObject **slot
 * 
 *         slot = _lookup(self, key)             # <<<<<<<<<<<<<<
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return False
 */
  __pyx_t_1 = __pyx_f_6bzrlib_15_simple_set_pyx__lookup(__pyx_v_self, __pyx_v_key); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 172; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_slot = __pyx_t_1;

  /* "bzrlib/_simple_set_pyx.pyx":173
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  __pyx_t_3 = (((__pyx_v_slot[0]) == NULL) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((__pyx_v_slot[0]) == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":174
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":173
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             return False
 *         return True
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":175
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cdef PyObject *_get(self, object key) except? NULL:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":168
 *         return <int>(slot - self._table), res
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         """Is key present in this SimpleSet."""
 *         cdef PyObject **slot
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":177
 *         return True
 * 
 *     cdef PyObject *_get(self, object key) except? NULL:             # <<<<<<<<<<<<<<
 *         """Return the object (or nothing) define at the given location."""
 *         cdef PyObject **slot
 */

static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__get(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject **__pyx_v_slot;
  PyObject *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject **__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get", 0);

  /* "bzrlib/_simple_set_pyx.pyx":181
 *         cdef PyObject **slot
 * 
 *         slot = _lookup(self, key)             # <<<<<<<<<<<<<<
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return NULL
 */
  __pyx_t_1 = __pyx_f_6bzrlib_15_simple_set_pyx__lookup(__pyx_v_self, __pyx_v_key); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 181; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_slot = __pyx_t_1;

  /* "bzrlib/_simple_set_pyx.pyx":182
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             return NULL
 *         return slot[0]
 */
  __pyx_t_3 = (((__pyx_v_slot[0]) == NULL) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((__pyx_v_slot[0]) == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":183
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return NULL             # <<<<<<<<<<<<<<
 *         return slot[0]
 * 
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":182
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             return NULL
 *         return slot[0]
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":184
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return NULL
 *         return slot[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __pyx_r = (__pyx_v_slot[0]);
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":177
 *         return True
 * 
 *     cdef PyObject *_get(self, object key) except? NULL:             # <<<<<<<<<<<<<<
 *         """Return the object (or nothing) define at the given location."""
 *         cdef PyObject **slot
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":186
 *         return slot[0]
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         """Return a stored item that is equivalent to key."""
 *         cdef PyObject *py_val
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_15__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__[] = "Return a stored item that is equivalent to key.";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__;
#endif
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_15__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__(((struct SimpleSetObject *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_v_py_val;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":190
 *         cdef PyObject *py_val
 * 
 *         py_val = self._get(key)             # <<<<<<<<<<<<<<
 *         if py_val == NULL:
 *             raise KeyError("Key %s is not present" % key)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)__pyx_v_self->__pyx_vtab)->_get(__pyx_v_self, __pyx_v_key); if (unlikely(__pyx_t_1 == NULL && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 190; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_py_val = __pyx_t_1;

  /* "bzrlib/_simple_set_pyx.pyx":191
 * 
 *         py_val = self._get(key)
 *         if py_val == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError("Key %s is not present" % key)
 *         val = <object>(py_val)
 */
  __pyx_t_2 = ((__pyx_v_py_val == NULL) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":192
 *         py_val = self._get(key)
 *         if py_val == NULL:
 *             raise KeyError("Key %s is not present" % key)             # <<<<<<<<<<<<<<
 *         val = <object>(py_val)
 *         return val
 */
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Key_s_is_not_present, __pyx_v_key); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_KeyError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":191
 * 
 *         py_val = self._get(key)
 *         if py_val == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError("Key %s is not present" % key)
 *         val = <object>(py_val)
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":193
 *         if py_val == NULL:
 *             raise KeyError("Key %s is not present" % key)
 *         val = <object>(py_val)             # <<<<<<<<<<<<<<
 *         return val
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_py_val);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_val = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":194
 *             raise KeyError("Key %s is not present" % key)
 *         val = <object>(py_val)
 *         return val             # <<<<<<<<<<<<<<
 * 
 *     cdef int _insert_clean(self, PyObject *key) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_val);
  __pyx_r = __pyx_v_val;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":186
 *         return slot[0]
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         """Return a stored item that is equivalent to key."""
 *         cdef PyObject *py_val
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":196
 *         return val
 * 
 *     cdef int _insert_clean(self, PyObject *key) except -1:             # <<<<<<<<<<<<<<
 *         """Insert a key into self.table.
 * 
 */

static int __pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__insert_clean(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  size_t __pyx_v_i;
  size_t __pyx_v_n_lookup;
  long __pyx_v_the_hash;
  PyObject **__pyx_v_table;
  PyObject **__pyx_v_slot;
  Py_ssize_t __pyx_v_mask;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject **__pyx_t_2;
  long __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_insert_clean", 0);

  /* "bzrlib/_simple_set_pyx.pyx":208
 *         cdef Py_ssize_t mask
 * 
 *         mask = self._mask             # <<<<<<<<<<<<<<
 *         table = self._table
 * 
 */
  __pyx_t_1 = __pyx_v_self->_mask;
  __pyx_v_mask = __pyx_t_1;

  /* "bzrlib/_simple_set_pyx.pyx":209
 * 
 *         mask = self._mask
 *         table = self._table             # <<<<<<<<<<<<<<
 * 
 *         the_hash = PyObject_Hash(key)
 */
  __pyx_t_2 = __pyx_v_self->_table;
  __pyx_v_table = __pyx_t_2;

  /* "bzrlib/_simple_set_pyx.pyx":211
 *         table = self._table
 * 
 *         the_hash = PyObject_Hash(key)             # <<<<<<<<<<<<<<
 *         i = the_hash
 *         for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 */
  __pyx_t_3 = PyObject_Hash(__pyx_v_key); if (unlikely(__pyx_t_3 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_the_hash = __pyx_t_3;

  /* "bzrlib/_simple_set_pyx.pyx":212
 * 
 *         the_hash = PyObject_Hash(key)
 *         i = the_hash             # <<<<<<<<<<<<<<
 *         for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 *             slot = &table[i & mask]
 */
  __pyx_v_i = __pyx_v_the_hash;

  /* "bzrlib/_simple_set_pyx.pyx":213
 *         the_hash = PyObject_Hash(key)
 *         i = the_hash
 *         for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever             # <<<<<<<<<<<<<<
 *             slot = &table[i & mask]
 *             if slot[0] == NULL:
 */
  __pyx_t_4 = ((size_t)__pyx_v_mask);
  for (__pyx_v_n_lookup = 0; __pyx_v_n_lookup <= __pyx_t_4; __pyx_v_n_lookup++) {

    /* "bzrlib/_simple_set_pyx.pyx":214
 *         i = the_hash
 *         for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 *             slot = &table[i & mask]             # <<<<<<<<<<<<<<
 *             if slot[0] == NULL:
 *                 slot[0] = key
 */
    __pyx_v_slot = (&(__pyx_v_table[(__pyx_v_i & __pyx_v_mask)]));

    /* "bzrlib/_simple_set_pyx.pyx":215
 *         for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 *             slot = &table[i & mask]
 *             if slot[0] == NULL:             # <<<<<<<<<<<<<<
 *                 slot[0] = key
 *                 self._fill = self._fill + 1
 */
    __pyx_t_5 = (((__pyx_v_slot[0]) == NULL) != 0);
    if (__pyx_t_5) {

      /* "bzrlib/_simple_set_pyx.pyx":216
 *             slot = &table[i & mask]
 *             if slot[0] == NULL:
 *                 slot[0] = key             # <<<<<<<<<<<<<<
 *                 self._fill = self._fill + 1
 *                 self._used = self._used + 1
 */
      (__pyx_v_slot[0]) = __pyx_v_key;

      /* "bzrlib/_simple_set_pyx.pyx":217
 *             if slot[0] == NULL:
 *                 slot[0] = key
 *                 self._fill = self._fill + 1             # <<<<<<<<<<<<<<
 *                 self._used = self._used + 1
 *                 return 1
 */
      __pyx_v_self->_fill = (__pyx_v_self->_fill + 1);

      /* "bzrlib/_simple_set_pyx.pyx":218
 *                 slot[0] = key
 *                 self._fill = self._fill + 1
 *                 self._used = self._used + 1             # <<<<<<<<<<<<<<
 *                 return 1
 *             i = i + 1 + n_lookup
 */
      __pyx_v_self->_used = (__pyx_v_self->_used + 1);

      /* "bzrlib/_simple_set_pyx.pyx":219
 *                 self._fill = self._fill + 1
 *                 self._used = self._used + 1
 *                 return 1             # <<<<<<<<<<<<<<
 *             i = i + 1 + n_lookup
 *         raise RuntimeError('ran out of slots.')
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "bzrlib/_simple_set_pyx.pyx":215
 *         for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 *             slot = &table[i & mask]
 *             if slot[0] == NULL:             # <<<<<<<<<<<<<<
 *                 slot[0] = key
 *                 self._fill = self._fill + 1
 */
    }

    /* "bzrlib/_simple_set_pyx.pyx":220
 *                 self._used = self._used + 1
 *                 return 1
 *             i = i + 1 + n_lookup             # <<<<<<<<<<<<<<
 *         raise RuntimeError('ran out of slots.')
 * 
 */
    __pyx_v_i = ((__pyx_v_i + 1) + __pyx_v_n_lookup);
  }

  /* "bzrlib/_simple_set_pyx.pyx":221
 *                 return 1
 *             i = i + 1 + n_lookup
 *         raise RuntimeError('ran out of slots.')             # <<<<<<<<<<<<<<
 * 
 *     def _py_resize(self, min_used):
 */
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 221; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_Raise(__pyx_t_6, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 221; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_simple_set_pyx.pyx":196
 *         return val
 * 
 *     cdef int _insert_clean(self, PyObject *key) except -1:             # <<<<<<<<<<<<<<
 *         """Insert a key into self.table.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._insert_clean", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":223
 *         raise RuntimeError('ran out of slots.')
 * 
 *     def _py_resize(self, min_used):             # <<<<<<<<<<<<<<
 *         """Do not use this directly, it is only exposed for testing."""
 *         return self._resize(min_used)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_17_py_resize(PyObject *__pyx_v_self, PyObject *__pyx_v_min_used); /*proto*/
static char __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_16_py_resize[] = "Do not use this directly, it is only exposed for testing.";
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_17_py_resize(PyObject *__pyx_v_self, PyObject *__pyx_v_min_used) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_py_resize (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_16_py_resize(((struct SimpleSetObject *)__pyx_v_self), ((PyObject *)__pyx_v_min_used));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_16_py_resize(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_min_used) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_py_resize", 0);

  /* "bzrlib/_simple_set_pyx.pyx":225
 *     def _py_resize(self, min_used):
 *         """Do not use this directly, it is only exposed for testing."""
 *         return self._resize(min_used)             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t _resize(self, Py_ssize_t min_used) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_min_used); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)__pyx_v_self->__pyx_vtab)->_resize(__pyx_v_self, __pyx_t_1); if (unlikely(__pyx_t_2 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":223
 *         raise RuntimeError('ran out of slots.')
 * 
 *     def _py_resize(self, min_used):             # <<<<<<<<<<<<<<
 *         """Do not use this directly, it is only exposed for testing."""
 *         return self._resize(min_used)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._py_resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":227
 *         return self._resize(min_used)
 * 
 *     cdef Py_ssize_t _resize(self, Py_ssize_t min_used) except -1:             # <<<<<<<<<<<<<<
 *         """Resize the internal table.
 * 
 */

static Py_ssize_t __pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__resize(struct SimpleSetObject *__pyx_v_self, Py_ssize_t __pyx_v_min_used) {
  Py_ssize_t __pyx_v_new_size;
  Py_ssize_t __pyx_v_n_bytes;
  Py_ssize_t __pyx_v_remaining;
  PyObject **__pyx_v_new_table;
  PyObject **__pyx_v_old_table;
  PyObject **__pyx_v_slot;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject **__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_resize", 0);

  /* "bzrlib/_simple_set_pyx.pyx":239
 *         cdef PyObject **new_table, **old_table, **slot
 * 
 *         new_size = DEFAULT_SIZE             # <<<<<<<<<<<<<<
 *         while new_size <= min_used and new_size > 0:
 *             new_size = new_size << 1
 */
  __pyx_v_new_size = 0x400;

  /* "bzrlib/_simple_set_pyx.pyx":240
 * 
 *         new_size = DEFAULT_SIZE
 *         while new_size <= min_used and new_size > 0:             # <<<<<<<<<<<<<<
 *             new_size = new_size << 1
 *         # We rolled over our signed size field
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_new_size <= __pyx_v_min_used) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_new_size > 0) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "bzrlib/_simple_set_pyx.pyx":241
 *         new_size = DEFAULT_SIZE
 *         while new_size <= min_used and new_size > 0:
 *             new_size = new_size << 1             # <<<<<<<<<<<<<<
 *         # We rolled over our signed size field
 *         if new_size <= 0:
 */
    __pyx_v_new_size = (__pyx_v_new_size << 1);
  }

  /* "bzrlib/_simple_set_pyx.pyx":243
 *             new_size = new_size << 1
 *         # We rolled over our signed size field
 *         if new_size <= 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         # Even if min_used == self._mask + 1, and we aren't changing the actual
 */
  __pyx_t_1 = ((__pyx_v_new_size <= 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":244
 *         # We rolled over our signed size field
 *         if new_size <= 0:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         # Even if min_used == self._mask + 1, and we aren't changing the actual
 *         # size, we will still run the algorithm so that dummy entries are
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 244; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":243
 *             new_size = new_size << 1
 *         # We rolled over our signed size field
 *         if new_size <= 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         # Even if min_used == self._mask + 1, and we aren't changing the actual
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":252
 *         #     raise RuntimeError('cannot shrink SimpleSet to something'
 *         #                        ' smaller than the number of used slots.')
 *         n_bytes = sizeof(PyObject*) * new_size;             # <<<<<<<<<<<<<<
 *         new_table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if new_table == NULL:
 */
  __pyx_v_n_bytes = ((sizeof(PyObject *)) * __pyx_v_new_size);

  /* "bzrlib/_simple_set_pyx.pyx":253
 *         #                        ' smaller than the number of used slots.')
 *         n_bytes = sizeof(PyObject*) * new_size;
 *         new_table = <PyObject **>PyMem_Malloc(n_bytes)             # <<<<<<<<<<<<<<
 *         if new_table == NULL:
 *             raise MemoryError()
 */
  __pyx_v_new_table = ((PyObject **)PyMem_Malloc(__pyx_v_n_bytes));

  /* "bzrlib/_simple_set_pyx.pyx":254
 *         n_bytes = sizeof(PyObject*) * new_size;
 *         new_table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if new_table == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((__pyx_v_new_table == NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":255
 *         new_table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if new_table == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         old_table = self._table
 */
    PyErr_NoMemory(); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 255; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":254
 *         n_bytes = sizeof(PyObject*) * new_size;
 *         new_table = <PyObject **>PyMem_Malloc(n_bytes)
 *         if new_table == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":257
 *             raise MemoryError()
 * 
 *         old_table = self._table             # <<<<<<<<<<<<<<
 *         self._table = new_table
 *         memset(self._table, 0, n_bytes)
 */
  __pyx_t_3 = __pyx_v_self->_table;
  __pyx_v_old_table = __pyx_t_3;

  /* "bzrlib/_simple_set_pyx.pyx":258
 * 
 *         old_table = self._table
 *         self._table = new_table             # <<<<<<<<<<<<<<
 *         memset(self._table, 0, n_bytes)
 *         self._mask = new_size - 1
 */
  __pyx_v_self->_table = __pyx_v_new_table;

  /* "bzrlib/_simple_set_pyx.pyx":259
 *         old_table = self._table
 *         self._table = new_table
 *         memset(self._table, 0, n_bytes)             # <<<<<<<<<<<<<<
 *         self._mask = new_size - 1
 *         self._used = 0
 */
  memset(__pyx_v_self->_table, 0, __pyx_v_n_bytes);

  /* "bzrlib/_simple_set_pyx.pyx":260
 *         self._table = new_table
 *         memset(self._table, 0, n_bytes)
 *         self._mask = new_size - 1             # <<<<<<<<<<<<<<
 *         self._used = 0
 *         remaining = self._fill
 */
  __pyx_v_self->_mask = (__pyx_v_new_size - 1);

  /* "bzrlib/_simple_set_pyx.pyx":261
 *         memset(self._table, 0, n_bytes)
 *         self._mask = new_size - 1
 *         self._used = 0             # <<<<<<<<<<<<<<
 *         remaining = self._fill
 *         self._fill = 0
 */
  __pyx_v_self->_used = 0;

  /* "bzrlib/_simple_set_pyx.pyx":262
 *         self._mask = new_size - 1
 *         self._used = 0
 *         remaining = self._fill             # <<<<<<<<<<<<<<
 *         self._fill = 0
 * 
 */
  __pyx_t_4 = __pyx_v_self->_fill;
  __pyx_v_remaining = __pyx_t_4;

  /* "bzrlib/_simple_set_pyx.pyx":263
 *         self._used = 0
 *         remaining = self._fill
 *         self._fill = 0             # <<<<<<<<<<<<<<
 * 
 *         # Moving everything to the other table is refcount neutral, so we don't
 */
  __pyx_v_self->_fill = 0;

  /* "bzrlib/_simple_set_pyx.pyx":267
 *         # Moving everything to the other table is refcount neutral, so we don't
 *         # worry about it.
 *         slot = old_table             # <<<<<<<<<<<<<<
 *         while remaining > 0:
 *             if slot[0] == NULL: # unused slot
 */
  __pyx_v_slot = __pyx_v_old_table;

  /* "bzrlib/_simple_set_pyx.pyx":268
 *         # worry about it.
 *         slot = old_table
 *         while remaining > 0:             # <<<<<<<<<<<<<<
 *             if slot[0] == NULL: # unused slot
 *                 pass
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_remaining > 0) != 0);
    if (!__pyx_t_1) break;

    /* "bzrlib/_simple_set_pyx.pyx":269
 *         slot = old_table
 *         while remaining > 0:
 *             if slot[0] == NULL: # unused slot             # <<<<<<<<<<<<<<
 *                 pass
 *             elif slot[0] == _dummy: # dummy slot
 */
    __pyx_t_1 = (((__pyx_v_slot[0]) == NULL) != 0);
    if (__pyx_t_1) {
      goto __pyx_L11;
    }

    /* "bzrlib/_simple_set_pyx.pyx":271
 *             if slot[0] == NULL: # unused slot
 *                 pass
 *             elif slot[0] == _dummy: # dummy slot             # <<<<<<<<<<<<<<
 *                 remaining = remaining - 1
 *             else: # active slot
 */
    __pyx_t_1 = (((__pyx_v_slot[0]) == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_simple_set_pyx.pyx":272
 *                 pass
 *             elif slot[0] == _dummy: # dummy slot
 *                 remaining = remaining - 1             # <<<<<<<<<<<<<<
 *             else: # active slot
 *                 remaining = remaining - 1
 */
      __pyx_v_remaining = (__pyx_v_remaining - 1);

      /* "bzrlib/_simple_set_pyx.pyx":271
 *             if slot[0] == NULL: # unused slot
 *                 pass
 *             elif slot[0] == _dummy: # dummy slot             # <<<<<<<<<<<<<<
 *                 remaining = remaining - 1
 *             else: # active slot
 */
      goto __pyx_L11;
    }

    /* "bzrlib/_simple_set_pyx.pyx":274
 *                 remaining = remaining - 1
 *             else: # active slot
 *                 remaining = remaining - 1             # <<<<<<<<<<<<<<
 *                 self._insert_clean(slot[0])
 *             slot = slot + 1
 */
    /*else*/ {
      __pyx_v_remaining = (__pyx_v_remaining - 1);

      /* "bzrlib/_simple_set_pyx.pyx":275
 *             else: # active slot
 *                 remaining = remaining - 1
 *                 self._insert_clean(slot[0])             # <<<<<<<<<<<<<<
 *             slot = slot + 1
 *         PyMem_Free(old_table)
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)__pyx_v_self->__pyx_vtab)->_insert_clean(__pyx_v_self, (__pyx_v_slot[0])); if (unlikely(__pyx_t_5 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 275; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
    __pyx_L11:;

    /* "bzrlib/_simple_set_pyx.pyx":276
 *                 remaining = remaining - 1
 *                 self._insert_clean(slot[0])
 *             slot = slot + 1             # <<<<<<<<<<<<<<
 *         PyMem_Free(old_table)
 *         return new_size
 */
    __pyx_v_slot = (__pyx_v_slot + 1);
  }

  /* "bzrlib/_simple_set_pyx.pyx":277
 *                 self._insert_clean(slot[0])
 *             slot = slot + 1
 *         PyMem_Free(old_table)             # <<<<<<<<<<<<<<
 *         return new_size
 * 
 */
  PyMem_Free(__pyx_v_old_table);

  /* "bzrlib/_simple_set_pyx.pyx":278
 *             slot = slot + 1
 *         PyMem_Free(old_table)
 *         return new_size             # <<<<<<<<<<<<<<
 * 
 *     def add(self, key):
 */
  __pyx_r = __pyx_v_new_size;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":227
 *         return self._resize(min_used)
 * 
 *     cdef Py_ssize_t _resize(self, Py_ssize_t min_used) except -1:             # <<<<<<<<<<<<<<
 *         """Resize the internal table.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._resize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":280
 *         return new_size
 * 
 *     def add(self, key):             # <<<<<<<<<<<<<<
 *         """Similar to set.add(), start tracking this key.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_19add(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_18add[] = "Similar to set.add(), start tracking this key.\n        \n        There is one small difference, which is that we return the object that\n        is stored at the given location. (which is closer to the\n        dict.setdefault() functionality.)\n        ";
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_19add(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_18add(((struct SimpleSetObject *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_18add(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 0);

  /* "bzrlib/_simple_set_pyx.pyx":287
 *         dict.setdefault() functionality.)
 *         """
 *         return self._add(key)             # <<<<<<<<<<<<<<
 * 
 *     cdef object _add(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)__pyx_v_self->__pyx_vtab)->_add(__pyx_v_self, __pyx_v_key); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 287; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":280
 *         return new_size
 * 
 *     def add(self, key):             # <<<<<<<<<<<<<<
 *         """Similar to set.add(), start tracking this key.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":289
 *         return self._add(key)
 * 
 *     cdef object _add(self, key):             # <<<<<<<<<<<<<<
 *         cdef PyObject **slot, *py_key
 *         cdef int added
 */

static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__add(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject **__pyx_v_slot;
  PyObject *__pyx_v_py_key;
  int __pyx_v_added;
  PyObject *__pyx_v_retval = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject **__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_add", 0);

  /* "bzrlib/_simple_set_pyx.pyx":293
 *         cdef int added
 * 
 *         py_key = <PyObject *>key             # <<<<<<<<<<<<<<
 *         if (Py_TYPE(py_key).tp_richcompare == NULL
 *             or Py_TYPE(py_key).tp_hash == NULL):
 */
  __pyx_v_py_key = ((PyObject *)__pyx_v_key);

  /* "bzrlib/_simple_set_pyx.pyx":294
 * 
 *         py_key = <PyObject *>key
 *         if (Py_TYPE(py_key).tp_richcompare == NULL             # <<<<<<<<<<<<<<
 *             or Py_TYPE(py_key).tp_hash == NULL):
 *             raise TypeError('Types added to SimpleSet must implement'
 */
  __pyx_t_2 = ((Py_TYPE(__pyx_v_py_key)->tp_richcompare == NULL) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "bzrlib/_simple_set_pyx.pyx":295
 *         py_key = <PyObject *>key
 *         if (Py_TYPE(py_key).tp_richcompare == NULL
 *             or Py_TYPE(py_key).tp_hash == NULL):             # <<<<<<<<<<<<<<
 *             raise TypeError('Types added to SimpleSet must implement'
 *                             ' both tp_richcompare and tp_hash')
 */
  __pyx_t_2 = ((Py_TYPE(__pyx_v_py_key)->tp_hash == NULL) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "bzrlib/_simple_set_pyx.pyx":294
 * 
 *         py_key = <PyObject *>key
 *         if (Py_TYPE(py_key).tp_richcompare == NULL             # <<<<<<<<<<<<<<
 *             or Py_TYPE(py_key).tp_hash == NULL):
 *             raise TypeError('Types added to SimpleSet must implement'
 */
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":296
 *         if (Py_TYPE(py_key).tp_richcompare == NULL
 *             or Py_TYPE(py_key).tp_hash == NULL):
 *             raise TypeError('Types added to SimpleSet must implement'             # <<<<<<<<<<<<<<
 *                             ' both tp_richcompare and tp_hash')
 *         added = 0
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":294
 * 
 *         py_key = <PyObject *>key
 *         if (Py_TYPE(py_key).tp_richcompare == NULL             # <<<<<<<<<<<<<<
 *             or Py_TYPE(py_key).tp_hash == NULL):
 *             raise TypeError('Types added to SimpleSet must implement'
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":298
 *             raise TypeError('Types added to SimpleSet must implement'
 *                             ' both tp_richcompare and tp_hash')
 *         added = 0             # <<<<<<<<<<<<<<
 *         # We need at least one empty slot
 *         assert self._used < self._mask
 */
  __pyx_v_added = 0;

  /* "bzrlib/_simple_set_pyx.pyx":300
 *         added = 0
 *         # We need at least one empty slot
 *         assert self._used < self._mask             # <<<<<<<<<<<<<<
 *         slot = _lookup(self, key)
 *         if (slot[0] == NULL):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_self->_used < __pyx_v_self->_mask) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 300; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif

  /* "bzrlib/_simple_set_pyx.pyx":301
 *         # We need at least one empty slot
 *         assert self._used < self._mask
 *         slot = _lookup(self, key)             # <<<<<<<<<<<<<<
 *         if (slot[0] == NULL):
 *             Py_INCREF(py_key)
 */
  __pyx_t_4 = __pyx_f_6bzrlib_15_simple_set_pyx__lookup(__pyx_v_self, __pyx_v_key); if (unlikely(__pyx_t_4 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 301; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_slot = __pyx_t_4;

  /* "bzrlib/_simple_set_pyx.pyx":302
 *         assert self._used < self._mask
 *         slot = _lookup(self, key)
 *         if (slot[0] == NULL):             # <<<<<<<<<<<<<<
 *             Py_INCREF(py_key)
 *             self._fill = self._fill + 1
 */
  __pyx_t_1 = (((__pyx_v_slot[0]) == NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":303
 *         slot = _lookup(self, key)
 *         if (slot[0] == NULL):
 *             Py_INCREF(py_key)             # <<<<<<<<<<<<<<
 *             self._fill = self._fill + 1
 *             self._used = self._used + 1
 */
    Py_INCREF(__pyx_v_py_key);

    /* "bzrlib/_simple_set_pyx.pyx":304
 *         if (slot[0] == NULL):
 *             Py_INCREF(py_key)
 *             self._fill = self._fill + 1             # <<<<<<<<<<<<<<
 *             self._used = self._used + 1
 *             slot[0] = py_key
 */
    __pyx_v_self->_fill = (__pyx_v_self->_fill + 1);

    /* "bzrlib/_simple_set_pyx.pyx":305
 *             Py_INCREF(py_key)
 *             self._fill = self._fill + 1
 *             self._used = self._used + 1             # <<<<<<<<<<<<<<
 *             slot[0] = py_key
 *             added = 1
 */
    __pyx_v_self->_used = (__pyx_v_self->_used + 1);

    /* "bzrlib/_simple_set_pyx.pyx":306
 *             self._fill = self._fill + 1
 *             self._used = self._used + 1
 *             slot[0] = py_key             # <<<<<<<<<<<<<<
 *             added = 1
 *         elif (slot[0] == _dummy):
 */
    (__pyx_v_slot[0]) = __pyx_v_py_key;

    /* "bzrlib/_simple_set_pyx.pyx":307
 *             self._used = self._used + 1
 *             slot[0] = py_key
 *             added = 1             # <<<<<<<<<<<<<<
 *         elif (slot[0] == _dummy):
 *             Py_INCREF(py_key)
 */
    __pyx_v_added = 1;

    /* "bzrlib/_simple_set_pyx.pyx":302
 *         assert self._used < self._mask
 *         slot = _lookup(self, key)
 *         if (slot[0] == NULL):             # <<<<<<<<<<<<<<
 *             Py_INCREF(py_key)
 *             self._fill = self._fill + 1
 */
    goto __pyx_L6;
  }

  /* "bzrlib/_simple_set_pyx.pyx":308
 *             slot[0] = py_key
 *             added = 1
 *         elif (slot[0] == _dummy):             # <<<<<<<<<<<<<<
 *             Py_INCREF(py_key)
 *             self._used = self._used + 1
 */
  __pyx_t_1 = (((__pyx_v_slot[0]) == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":309
 *             added = 1
 *         elif (slot[0] == _dummy):
 *             Py_INCREF(py_key)             # <<<<<<<<<<<<<<
 *             self._used = self._used + 1
 *             slot[0] = py_key
 */
    Py_INCREF(__pyx_v_py_key);

    /* "bzrlib/_simple_set_pyx.pyx":310
 *         elif (slot[0] == _dummy):
 *             Py_INCREF(py_key)
 *             self._used = self._used + 1             # <<<<<<<<<<<<<<
 *             slot[0] = py_key
 *             added = 1
 */
    __pyx_v_self->_used = (__pyx_v_self->_used + 1);

    /* "bzrlib/_simple_set_pyx.pyx":311
 *             Py_INCREF(py_key)
 *             self._used = self._used + 1
 *             slot[0] = py_key             # <<<<<<<<<<<<<<
 *             added = 1
 *         # No else: clause. If _lookup returns a pointer to
 */
    (__pyx_v_slot[0]) = __pyx_v_py_key;

    /* "bzrlib/_simple_set_pyx.pyx":312
 *             self._used = self._used + 1
 *             slot[0] = py_key
 *             added = 1             # <<<<<<<<<<<<<<
 *         # No else: clause. If _lookup returns a pointer to
 *         # a live object, then we already have a value at this location.
 */
    __pyx_v_added = 1;

    /* "bzrlib/_simple_set_pyx.pyx":308
 *             slot[0] = py_key
 *             added = 1
 *         elif (slot[0] == _dummy):             # <<<<<<<<<<<<<<
 *             Py_INCREF(py_key)
 *             self._used = self._used + 1
 */
  }
  __pyx_L6:;

  /* "bzrlib/_simple_set_pyx.pyx":315
 *         # No else: clause. If _lookup returns a pointer to
 *         # a live object, then we already have a value at this location.
 *         retval = <object>(slot[0])             # <<<<<<<<<<<<<<
 *         # PySet and PyDict use a 2-3rds full algorithm, we'll follow suit
 *         if added and (self._fill * 3) >= ((self._mask + 1) * 2):
 */
  __pyx_t_3 = ((PyObject *)(__pyx_v_slot[0]));
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_retval = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":317
 *         retval = <object>(slot[0])
 *         # PySet and PyDict use a 2-3rds full algorithm, we'll follow suit
 *         if added and (self._fill * 3) >= ((self._mask + 1) * 2):             # <<<<<<<<<<<<<<
 *             # However, we always work for a load factor of 2:1
 *             self._resize(self._used * 2)
 */
  __pyx_t_2 = (__pyx_v_added != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_self->_fill * 3) >= ((__pyx_v_self->_mask + 1) * 2)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":319
 *         if added and (self._fill * 3) >= ((self._mask + 1) * 2):
 *             # However, we always work for a load factor of 2:1
 *             self._resize(self._used * 2)             # <<<<<<<<<<<<<<
 *         # Even if we resized and ended up moving retval into a different slot,
 *         # it is still the value that is held at the slot equivalent to 'key',
 */
    __pyx_t_5 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)__pyx_v_self->__pyx_vtab)->_resize(__pyx_v_self, (__pyx_v_self->_used * 2)); if (unlikely(__pyx_t_5 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 319; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":317
 *         retval = <object>(slot[0])
 *         # PySet and PyDict use a 2-3rds full algorithm, we'll follow suit
 *         if added and (self._fill * 3) >= ((self._mask + 1) * 2):             # <<<<<<<<<<<<<<
 *             # However, we always work for a load factor of 2:1
 *             self._resize(self._used * 2)
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":323
 *         # it is still the value that is held at the slot equivalent to 'key',
 *         # so we can still return it
 *         return retval             # <<<<<<<<<<<<<<
 * 
 *     def discard(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retval);
  __pyx_r = __pyx_v_retval;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":289
 *         return self._add(key)
 * 
 *     cdef object _add(self, key):             # <<<<<<<<<<<<<<
 *         cdef PyObject **slot, *py_key
 *         cdef int added
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retval);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":325
 *         return retval
 * 
 *     def discard(self, key):             # <<<<<<<<<<<<<<
 *         """Remove key from the set, whether it exists or not.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_21discard(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_20discard[] = "Remove key from the set, whether it exists or not.\n\n        :return: False if the item did not exist, True if it did\n        ";
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_21discard(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("discard (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_20discard(((struct SimpleSetObject *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_20discard(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("discard", 0);

  /* "bzrlib/_simple_set_pyx.pyx":330
 *         :return: False if the item did not exist, True if it did
 *         """
 *         if self._discard(key):             # <<<<<<<<<<<<<<
 *             return True
 *         return False
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)__pyx_v_self->__pyx_vtab)->_discard(__pyx_v_self, __pyx_v_key); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 330; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":331
 *         """
 *         if self._discard(key):
 *             return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":330
 *         :return: False if the item did not exist, True if it did
 *         """
 *         if self._discard(key):             # <<<<<<<<<<<<<<
 *             return True
 *         return False
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":332
 *         if self._discard(key):
 *             return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef int _discard(self, key) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_False);
  __pyx_r = Py_False;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":325
 *         return retval
 * 
 *     def discard(self, key):             # <<<<<<<<<<<<<<
 *         """Remove key from the set, whether it exists or not.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.discard", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":334
 *         return False
 * 
 *     cdef int _discard(self, key) except -1:             # <<<<<<<<<<<<<<
 *         cdef PyObject **slot, *py_key
 * 
 */

static int __pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__discard(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject **__pyx_v_slot;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject **__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_discard", 0);

  /* "bzrlib/_simple_set_pyx.pyx":337
 *         cdef PyObject **slot, *py_key
 * 
 *         slot = _lookup(self, key)             # <<<<<<<<<<<<<<
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return 0
 */
  __pyx_t_1 = __pyx_f_6bzrlib_15_simple_set_pyx__lookup(__pyx_v_self, __pyx_v_key); if (unlikely(__pyx_t_1 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 337; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_slot = __pyx_t_1;

  /* "bzrlib/_simple_set_pyx.pyx":338
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             return 0
 *         self._used = self._used - 1
 */
  __pyx_t_3 = (((__pyx_v_slot[0]) == NULL) != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (((__pyx_v_slot[0]) == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":339
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return 0             # <<<<<<<<<<<<<<
 *         self._used = self._used - 1
 *         Py_DECREF(slot[0])
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":338
 * 
 *         slot = _lookup(self, key)
 *         if slot[0] == NULL or slot[0] == _dummy:             # <<<<<<<<<<<<<<
 *             return 0
 *         self._used = self._used - 1
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":340
 *         if slot[0] == NULL or slot[0] == _dummy:
 *             return 0
 *         self._used = self._used - 1             # <<<<<<<<<<<<<<
 *         Py_DECREF(slot[0])
 *         slot[0] = _dummy
 */
  __pyx_v_self->_used = (__pyx_v_self->_used - 1);

  /* "bzrlib/_simple_set_pyx.pyx":341
 *             return 0
 *         self._used = self._used - 1
 *         Py_DECREF(slot[0])             # <<<<<<<<<<<<<<
 *         slot[0] = _dummy
 *         # PySet uses the heuristic: If more than 1/5 are dummies, then resize
 */
  Py_DECREF((__pyx_v_slot[0]));

  /* "bzrlib/_simple_set_pyx.pyx":342
 *         self._used = self._used - 1
 *         Py_DECREF(slot[0])
 *         slot[0] = _dummy             # <<<<<<<<<<<<<<
 *         # PySet uses the heuristic: If more than 1/5 are dummies, then resize
 *         #                           them away
 */
  (__pyx_v_slot[0]) = __pyx_v_6bzrlib_15_simple_set_pyx__dummy;

  /* "bzrlib/_simple_set_pyx.pyx":353
 *         # For now, we'll just use their algorithm, but we may want to revisit
 *         # it
 *         if ((self._fill - self._used) * 5 > self._mask):             # <<<<<<<<<<<<<<
 *             self._resize(self._used * 2)
 *         return 1
 */
  __pyx_t_2 = ((((__pyx_v_self->_fill - __pyx_v_self->_used) * 5) > __pyx_v_self->_mask) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":354
 *         # it
 *         if ((self._fill - self._used) * 5 > self._mask):
 *             self._resize(self._used * 2)             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)__pyx_v_self->__pyx_vtab)->_resize(__pyx_v_self, (__pyx_v_self->_used * 2)); if (unlikely(__pyx_t_4 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 354; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":353
 *         # For now, we'll just use their algorithm, but we may want to revisit
 *         # it
 *         if ((self._fill - self._used) * 5 > self._mask):             # <<<<<<<<<<<<<<
 *             self._resize(self._used * 2)
 *         return 1
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":355
 *         if ((self._fill - self._used) * 5 > self._mask):
 *             self._resize(self._used * 2)
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":334
 *         return False
 * 
 *     cdef int _discard(self, key) except -1:             # <<<<<<<<<<<<<<
 *         cdef PyObject **slot, *py_key
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet._discard", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":357
 *         return 1
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return _SimpleSet_iterator(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_23__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_23__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_22__iter__(((struct SimpleSetObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_9SimpleSet_22__iter__(struct SimpleSetObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":358
 * 
 *     def __iter__(self):
 *         return _SimpleSet_iterator(self)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 358; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_15_simple_set_pyx__SimpleSet_iterator), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 358; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":357
 *         return 1
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return _SimpleSet_iterator(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":369
 *     cdef Py_ssize_t len # number of entries left
 * 
 *     def __init__(self, obj):             # <<<<<<<<<<<<<<
 *         self.set = obj
 *         self.pos = 0
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 369; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_obj = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 369; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._SimpleSet_iterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator___init__(((struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)__pyx_v_self), __pyx_v_obj);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator___init__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self, PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":370
 * 
 *     def __init__(self, obj):
 *         self.set = obj             # <<<<<<<<<<<<<<
 *         self.pos = 0
 *         self._used = self.set._used
 */
  if (!(likely(((__pyx_v_obj) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_obj, __pyx_ptype_6bzrlib_15_simple_set_pyx_SimpleSet))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 370; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_1 = __pyx_v_obj;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->set);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->set));
  __pyx_v_self->set = ((struct SimpleSetObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":371
 *     def __init__(self, obj):
 *         self.set = obj
 *         self.pos = 0             # <<<<<<<<<<<<<<
 *         self._used = self.set._used
 *         self.len = self.set._used
 */
  __pyx_v_self->pos = 0;

  /* "bzrlib/_simple_set_pyx.pyx":372
 *         self.set = obj
 *         self.pos = 0
 *         self._used = self.set._used             # <<<<<<<<<<<<<<
 *         self.len = self.set._used
 * 
 */
  __pyx_t_2 = __pyx_v_self->set->_used;
  __pyx_v_self->_used = __pyx_t_2;

  /* "bzrlib/_simple_set_pyx.pyx":373
 *         self.pos = 0
 *         self._used = self.set._used
 *         self.len = self.set._used             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_2 = __pyx_v_self->set->_used;
  __pyx_v_self->len = __pyx_t_2;

  /* "bzrlib/_simple_set_pyx.pyx":369
 *     cdef Py_ssize_t len # number of entries left
 * 
 *     def __init__(self, obj):             # <<<<<<<<<<<<<<
 *         self.set = obj
 *         self.pos = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._SimpleSet_iterator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":375
 *         self.len = self.set._used
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_2__iter__(((struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_2__iter__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":376
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":375
 *         self.len = self.set._used
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":378
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t mask, i
 *         cdef PyObject *key
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_5__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_5__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_4__next__(((struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_4__next__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self) {
  PyObject *__pyx_v_key;
  PyObject *__pyx_v_the_key = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":382
 *         cdef PyObject *key
 * 
 *         if self.set is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 *         if self.set._used != self._used:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->set) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":383
 * 
 *         if self.set is None:
 *             raise StopIteration             # <<<<<<<<<<<<<<
 *         if self.set._used != self._used:
 *             # Force this exception to continue to be raised
 */
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 383; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":382
 *         cdef PyObject *key
 * 
 *         if self.set is None:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 *         if self.set._used != self._used:
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":384
 *         if self.set is None:
 *             raise StopIteration
 *         if self.set._used != self._used:             # <<<<<<<<<<<<<<
 *             # Force this exception to continue to be raised
 *             self._used = -1
 */
  __pyx_t_2 = ((__pyx_v_self->set->_used != __pyx_v_self->_used) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":386
 *         if self.set._used != self._used:
 *             # Force this exception to continue to be raised
 *             self._used = -1             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Set size changed during iteration")
 *         if not SimpleSet_Next(self.set, &self.pos, &key):
 */
    __pyx_v_self->_used = -1L;

    /* "bzrlib/_simple_set_pyx.pyx":387
 *             # Force this exception to continue to be raised
 *             self._used = -1
 *             raise RuntimeError("Set size changed during iteration")             # <<<<<<<<<<<<<<
 *         if not SimpleSet_Next(self.set, &self.pos, &key):
 *             self.set = None
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 387; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 387; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":384
 *         if self.set is None:
 *             raise StopIteration
 *         if self.set._used != self._used:             # <<<<<<<<<<<<<<
 *             # Force this exception to continue to be raised
 *             self._used = -1
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":388
 *             self._used = -1
 *             raise RuntimeError("Set size changed during iteration")
 *         if not SimpleSet_Next(self.set, &self.pos, &key):             # <<<<<<<<<<<<<<
 *             self.set = None
 *             raise StopIteration
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->set);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Next(__pyx_t_3, (&__pyx_v_self->pos), (&__pyx_v_key)); if (unlikely(__pyx_t_4 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!(__pyx_t_4 != 0)) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":389
 *             raise RuntimeError("Set size changed during iteration")
 *         if not SimpleSet_Next(self.set, &self.pos, &key):
 *             self.set = None             # <<<<<<<<<<<<<<
 *             raise StopIteration
 *         # we found something
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->set);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->set));
    __pyx_v_self->set = ((struct SimpleSetObject *)Py_None);

    /* "bzrlib/_simple_set_pyx.pyx":390
 *         if not SimpleSet_Next(self.set, &self.pos, &key):
 *             self.set = None
 *             raise StopIteration             # <<<<<<<<<<<<<<
 *         # we found something
 *         the_key = <object>key # INCREF
 */
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 390; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":388
 *             self._used = -1
 *             raise RuntimeError("Set size changed during iteration")
 *         if not SimpleSet_Next(self.set, &self.pos, &key):             # <<<<<<<<<<<<<<
 *             self.set = None
 *             raise StopIteration
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":392
 *             raise StopIteration
 *         # we found something
 *         the_key = <object>key # INCREF             # <<<<<<<<<<<<<<
 *         self.len = self.len - 1
 *         return the_key
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_key);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_the_key = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":393
 *         # we found something
 *         the_key = <object>key # INCREF
 *         self.len = self.len - 1             # <<<<<<<<<<<<<<
 *         return the_key
 * 
 */
  __pyx_v_self->len = (__pyx_v_self->len - 1);

  /* "bzrlib/_simple_set_pyx.pyx":394
 *         the_key = <object>key # INCREF
 *         self.len = self.len - 1
 *         return the_key             # <<<<<<<<<<<<<<
 * 
 *     def __length_hint__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_the_key);
  __pyx_r = __pyx_v_the_key;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":378
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t mask, i
 *         cdef PyObject *key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._SimpleSet_iterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_the_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":396
 *         return the_key
 * 
 *     def __length_hint__(self):             # <<<<<<<<<<<<<<
 *         if self.set is not None and self._used == self.set._used:
 *             return self.len
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_7__length_hint__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_7__length_hint__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__length_hint__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_6__length_hint__(((struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_6__length_hint__(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__length_hint__", 0);

  /* "bzrlib/_simple_set_pyx.pyx":397
 * 
 *     def __length_hint__(self):
 *         if self.set is not None and self._used == self.set._used:             # <<<<<<<<<<<<<<
 *             return self.len
 *         return 0
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_self->set) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_self->_used == __pyx_v_self->set->_used) != 0);
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_simple_set_pyx.pyx":398
 *     def __length_hint__(self):
 *         if self.set is not None and self._used == self.set._used:
 *             return self.len             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_self->len); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 398; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":397
 * 
 *     def __length_hint__(self):
 *         if self.set is not None and self._used == self.set._used:             # <<<<<<<<<<<<<<
 *             return self.len
 *         return 0
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":399
 *         if self.set is not None and self._used == self.set._used:
 *             return self.len
 *         return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":396
 *         return the_key
 * 
 *     def __length_hint__(self):             # <<<<<<<<<<<<<<
 *         if self.set is not None and self._used == self.set._used:
 *             return self.len
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._SimpleSet_iterator.__length_hint__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":403
 * 
 * 
 * cdef api SimpleSet SimpleSet_New():             # <<<<<<<<<<<<<<
 *     """Create a new SimpleSet object."""
 *     return SimpleSet()
 */

static struct SimpleSetObject *__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_New(void) {
  struct SimpleSetObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_New", 0);

  /* "bzrlib/_simple_set_pyx.pyx":405
 * cdef api SimpleSet SimpleSet_New():
 *     """Create a new SimpleSet object."""
 *     return SimpleSet()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_15_simple_set_pyx_SimpleSet), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 405; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct SimpleSetObject *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":403
 * 
 * 
 * cdef api SimpleSet SimpleSet_New():             # <<<<<<<<<<<<<<
 *     """Create a new SimpleSet object."""
 *     return SimpleSet()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_New", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":408
 * 
 * 
 * cdef SimpleSet _check_self(object self):             # <<<<<<<<<<<<<<
 *     """Check that the parameter is not None.
 * 
 */

static struct SimpleSetObject *__pyx_f_6bzrlib_15_simple_set_pyx__check_self(PyObject *__pyx_v_self) {
  struct SimpleSetObject *__pyx_v_true_self = 0;
  struct SimpleSetObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_check_self", 0);

  /* "bzrlib/_simple_set_pyx.pyx":417
 *     """
 *     cdef SimpleSet true_self
 *     if self is None:             # <<<<<<<<<<<<<<
 *         raise TypeError('self must not be None')
 *     true_self = self
 */
  __pyx_t_1 = (__pyx_v_self == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":418
 *     cdef SimpleSet true_self
 *     if self is None:
 *         raise TypeError('self must not be None')             # <<<<<<<<<<<<<<
 *     true_self = self
 *     return true_self
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_simple_set_pyx.pyx":417
 *     """
 *     cdef SimpleSet true_self
 *     if self is None:             # <<<<<<<<<<<<<<
 *         raise TypeError('self must not be None')
 *     true_self = self
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":419
 *     if self is None:
 *         raise TypeError('self must not be None')
 *     true_self = self             # <<<<<<<<<<<<<<
 *     return true_self
 * 
 */
  if (!(likely(((__pyx_v_self) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self, __pyx_ptype_6bzrlib_15_simple_set_pyx_SimpleSet))))) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 419; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_t_3 = __pyx_v_self;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_true_self = ((struct SimpleSetObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":420
 *         raise TypeError('self must not be None')
 *     true_self = self
 *     return true_self             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_true_self));
  __pyx_r = __pyx_v_true_self;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":408
 * 
 * 
 * cdef SimpleSet _check_self(object self):             # <<<<<<<<<<<<<<
 *     """Check that the parameter is not None.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._check_self", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_true_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":423
 * 
 * 
 * cdef PyObject **_lookup(SimpleSet self, object key) except NULL:             # <<<<<<<<<<<<<<
 *     """Find the slot where 'key' would fit.
 * 
 */

static PyObject **__pyx_f_6bzrlib_15_simple_set_pyx__lookup(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  size_t __pyx_v_i;
  size_t __pyx_v_n_lookup;
  Py_ssize_t __pyx_v_mask;
  long __pyx_v_key_hash;
  PyObject **__pyx_v_table;
  PyObject **__pyx_v_slot;
  PyObject *__pyx_v_cur;
  PyObject **__pyx_v_free_slot;
  PyObject *__pyx_v_py_key;
  PyObject **__pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject **__pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_lookup", 0);

  /* "bzrlib/_simple_set_pyx.pyx":464
 *     cdef PyObject **table, **slot, *cur, **free_slot, *py_key
 * 
 *     py_key = <PyObject *>key             # <<<<<<<<<<<<<<
 *     # Note: avoid using hash(obj) because of a bug w/ pyrex 0.9.8.5 and 64-bit
 *     #       (it treats hash() as returning an 'int' rather than a 'long')
 */
  __pyx_v_py_key = ((PyObject *)__pyx_v_key);

  /* "bzrlib/_simple_set_pyx.pyx":467
 *     # Note: avoid using hash(obj) because of a bug w/ pyrex 0.9.8.5 and 64-bit
 *     #       (it treats hash() as returning an 'int' rather than a 'long')
 *     key_hash = PyObject_Hash(py_key)             # <<<<<<<<<<<<<<
 *     i = <size_t>key_hash
 *     mask = self._mask
 */
  __pyx_t_1 = PyObject_Hash(__pyx_v_py_key); if (unlikely(__pyx_t_1 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 467; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_key_hash = __pyx_t_1;

  /* "bzrlib/_simple_set_pyx.pyx":468
 *     #       (it treats hash() as returning an 'int' rather than a 'long')
 *     key_hash = PyObject_Hash(py_key)
 *     i = <size_t>key_hash             # <<<<<<<<<<<<<<
 *     mask = self._mask
 *     table = self._table
 */
  __pyx_v_i = ((size_t)__pyx_v_key_hash);

  /* "bzrlib/_simple_set_pyx.pyx":469
 *     key_hash = PyObject_Hash(py_key)
 *     i = <size_t>key_hash
 *     mask = self._mask             # <<<<<<<<<<<<<<
 *     table = self._table
 *     free_slot = NULL
 */
  __pyx_t_2 = __pyx_v_self->_mask;
  __pyx_v_mask = __pyx_t_2;

  /* "bzrlib/_simple_set_pyx.pyx":470
 *     i = <size_t>key_hash
 *     mask = self._mask
 *     table = self._table             # <<<<<<<<<<<<<<
 *     free_slot = NULL
 *     for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 */
  __pyx_t_3 = __pyx_v_self->_table;
  __pyx_v_table = __pyx_t_3;

  /* "bzrlib/_simple_set_pyx.pyx":471
 *     mask = self._mask
 *     table = self._table
 *     free_slot = NULL             # <<<<<<<<<<<<<<
 *     for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 *         slot = &table[i & mask]
 */
  __pyx_v_free_slot = NULL;

  /* "bzrlib/_simple_set_pyx.pyx":472
 *     table = self._table
 *     free_slot = NULL
 *     for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever             # <<<<<<<<<<<<<<
 *         slot = &table[i & mask]
 *         cur = slot[0]
 */
  __pyx_t_4 = ((size_t)__pyx_v_mask);
  for (__pyx_v_n_lookup = 0; __pyx_v_n_lookup <= __pyx_t_4; __pyx_v_n_lookup++) {

    /* "bzrlib/_simple_set_pyx.pyx":473
 *     free_slot = NULL
 *     for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 *         slot = &table[i & mask]             # <<<<<<<<<<<<<<
 *         cur = slot[0]
 *         if cur == NULL:
 */
    __pyx_v_slot = (&(__pyx_v_table[(__pyx_v_i & __pyx_v_mask)]));

    /* "bzrlib/_simple_set_pyx.pyx":474
 *     for n_lookup from 0 <= n_lookup <= <size_t>mask: # Don't loop forever
 *         slot = &table[i & mask]
 *         cur = slot[0]             # <<<<<<<<<<<<<<
 *         if cur == NULL:
 *             # Found a blank spot
 */
    __pyx_v_cur = (__pyx_v_slot[0]);

    /* "bzrlib/_simple_set_pyx.pyx":475
 *         slot = &table[i & mask]
 *         cur = slot[0]
 *         if cur == NULL:             # <<<<<<<<<<<<<<
 *             # Found a blank spot
 *             if free_slot != NULL:
 */
    __pyx_t_5 = ((__pyx_v_cur == NULL) != 0);
    if (__pyx_t_5) {

      /* "bzrlib/_simple_set_pyx.pyx":477
 *         if cur == NULL:
 *             # Found a blank spot
 *             if free_slot != NULL:             # <<<<<<<<<<<<<<
 *                 # Did we find an earlier _dummy entry?
 *                 return free_slot
 */
      __pyx_t_5 = ((__pyx_v_free_slot != NULL) != 0);
      if (__pyx_t_5) {

        /* "bzrlib/_simple_set_pyx.pyx":479
 *             if free_slot != NULL:
 *                 # Did we find an earlier _dummy entry?
 *                 return free_slot             # <<<<<<<<<<<<<<
 *             else:
 *                 return slot
 */
        __pyx_r = __pyx_v_free_slot;
        goto __pyx_L0;

        /* "bzrlib/_simple_set_pyx.pyx":477
 *         if cur == NULL:
 *             # Found a blank spot
 *             if free_slot != NULL:             # <<<<<<<<<<<<<<
 *                 # Did we find an earlier _dummy entry?
 *                 return free_slot
 */
      }

      /* "bzrlib/_simple_set_pyx.pyx":481
 *                 return free_slot
 *             else:
 *                 return slot             # <<<<<<<<<<<<<<
 *         if cur == py_key:
 *             # Found an exact pointer to the key
 */
      /*else*/ {
        __pyx_r = __pyx_v_slot;
        goto __pyx_L0;
      }

      /* "bzrlib/_simple_set_pyx.pyx":475
 *         slot = &table[i & mask]
 *         cur = slot[0]
 *         if cur == NULL:             # <<<<<<<<<<<<<<
 *             # Found a blank spot
 *             if free_slot != NULL:
 */
    }

    /* "bzrlib/_simple_set_pyx.pyx":482
 *             else:
 *                 return slot
 *         if cur == py_key:             # <<<<<<<<<<<<<<
 *             # Found an exact pointer to the key
 *             return slot
 */
    __pyx_t_5 = ((__pyx_v_cur == __pyx_v_py_key) != 0);
    if (__pyx_t_5) {

      /* "bzrlib/_simple_set_pyx.pyx":484
 *         if cur == py_key:
 *             # Found an exact pointer to the key
 *             return slot             # <<<<<<<<<<<<<<
 *         if cur == _dummy:
 *             if free_slot == NULL:
 */
      __pyx_r = __pyx_v_slot;
      goto __pyx_L0;

      /* "bzrlib/_simple_set_pyx.pyx":482
 *             else:
 *                 return slot
 *         if cur == py_key:             # <<<<<<<<<<<<<<
 *             # Found an exact pointer to the key
 *             return slot
 */
    }

    /* "bzrlib/_simple_set_pyx.pyx":485
 *             # Found an exact pointer to the key
 *             return slot
 *         if cur == _dummy:             # <<<<<<<<<<<<<<
 *             if free_slot == NULL:
 *                 free_slot = slot
 */
    __pyx_t_5 = ((__pyx_v_cur == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
    if (__pyx_t_5) {

      /* "bzrlib/_simple_set_pyx.pyx":486
 *             return slot
 *         if cur == _dummy:
 *             if free_slot == NULL:             # <<<<<<<<<<<<<<
 *                 free_slot = slot
 *         elif _is_equal(py_key, key_hash, cur):
 */
      __pyx_t_5 = ((__pyx_v_free_slot == NULL) != 0);
      if (__pyx_t_5) {

        /* "bzrlib/_simple_set_pyx.pyx":487
 *         if cur == _dummy:
 *             if free_slot == NULL:
 *                 free_slot = slot             # <<<<<<<<<<<<<<
 *         elif _is_equal(py_key, key_hash, cur):
 *             # Both py_key and cur belong in this slot, return it
 */
        __pyx_v_free_slot = __pyx_v_slot;

        /* "bzrlib/_simple_set_pyx.pyx":486
 *             return slot
 *         if cur == _dummy:
 *             if free_slot == NULL:             # <<<<<<<<<<<<<<
 *                 free_slot = slot
 *         elif _is_equal(py_key, key_hash, cur):
 */
      }

      /* "bzrlib/_simple_set_pyx.pyx":485
 *             # Found an exact pointer to the key
 *             return slot
 *         if cur == _dummy:             # <<<<<<<<<<<<<<
 *             if free_slot == NULL:
 *                 free_slot = slot
 */
      goto __pyx_L8;
    }

    /* "bzrlib/_simple_set_pyx.pyx":488
 *             if free_slot == NULL:
 *                 free_slot = slot
 *         elif _is_equal(py_key, key_hash, cur):             # <<<<<<<<<<<<<<
 *             # Both py_key and cur belong in this slot, return it
 *             return slot
 */
    __pyx_t_6 = __pyx_f_6bzrlib_15_simple_set_pyx__is_equal(__pyx_v_py_key, __pyx_v_key_hash, __pyx_v_cur); if (unlikely(__pyx_t_6 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 488; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (__pyx_t_5) {

      /* "bzrlib/_simple_set_pyx.pyx":490
 *         elif _is_equal(py_key, key_hash, cur):
 *             # Both py_key and cur belong in this slot, return it
 *             return slot             # <<<<<<<<<<<<<<
 *         i = i + 1 + n_lookup
 *     raise AssertionError('should never get here')
 */
      __pyx_r = __pyx_v_slot;
      goto __pyx_L0;

      /* "bzrlib/_simple_set_pyx.pyx":488
 *             if free_slot == NULL:
 *                 free_slot = slot
 *         elif _is_equal(py_key, key_hash, cur):             # <<<<<<<<<<<<<<
 *             # Both py_key and cur belong in this slot, return it
 *             return slot
 */
    }
    __pyx_L8:;

    /* "bzrlib/_simple_set_pyx.pyx":491
 *             # Both py_key and cur belong in this slot, return it
 *             return slot
 *         i = i + 1 + n_lookup             # <<<<<<<<<<<<<<
 *     raise AssertionError('should never get here')
 * 
 */
    __pyx_v_i = ((__pyx_v_i + 1) + __pyx_v_n_lookup);
  }

  /* "bzrlib/_simple_set_pyx.pyx":492
 *             return slot
 *         i = i + 1 + n_lookup
 *     raise AssertionError('should never get here')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_7)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 492; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_Raise(__pyx_t_7, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 492; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_simple_set_pyx.pyx":423
 * 
 * 
 * cdef PyObject **_lookup(SimpleSet self, object key) except NULL:             # <<<<<<<<<<<<<<
 *     """Find the slot where 'key' would fit.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":495
 * 
 * 
 * cdef api PyObject **_SimpleSet_Lookup(object self, object key) except NULL:             # <<<<<<<<<<<<<<
 *     """Find the slot where 'key' would fit.
 * 
 */

static PyObject **__pyx_f_6bzrlib_15_simple_set_pyx__SimpleSet_Lookup(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject **__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject **__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_SimpleSet_Lookup", 0);

  /* "bzrlib/_simple_set_pyx.pyx":507
 *         should never be NULL, but may reference a NULL (PyObject*)
 *     """
 *     return _lookup(_check_self(self), key)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_15_simple_set_pyx__check_self(__pyx_v_self)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 507; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_6bzrlib_15_simple_set_pyx__lookup(((struct SimpleSetObject *)__pyx_t_1), __pyx_v_key); if (unlikely(__pyx_t_2 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 507; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":495
 * 
 * 
 * cdef api PyObject **_SimpleSet_Lookup(object self, object key) except NULL:             # <<<<<<<<<<<<<<
 *     """Find the slot where 'key' would fit.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx._SimpleSet_Lookup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":510
 * 
 * 
 * cdef api object SimpleSet_Add(object self, object key):             # <<<<<<<<<<<<<<
 *     """Add a key to the SimpleSet (set).
 * 
 */

static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Add(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_Add", 0);

  /* "bzrlib/_simple_set_pyx.pyx":520
 *         (consider dict.setdefault(key, key))
 *     """
 *     return _check_self(self)._add(key)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_15_simple_set_pyx__check_self(__pyx_v_self)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 520; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)((struct SimpleSetObject *)__pyx_t_1)->__pyx_vtab)->_add(((struct SimpleSetObject *)__pyx_t_1), __pyx_v_key); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 520; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":510
 * 
 * 
 * cdef api object SimpleSet_Add(object self, object key):             # <<<<<<<<<<<<<<
 *     """Add a key to the SimpleSet (set).
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_Add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":523
 * 
 * 
 * cdef api int SimpleSet_Contains(object self, object key) except -1:             # <<<<<<<<<<<<<<
 *     """Is key present in self?"""
 *     return (key in _check_self(self))
 */

static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Contains(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_Contains", 0);

  /* "bzrlib/_simple_set_pyx.pyx":525
 * cdef api int SimpleSet_Contains(object self, object key) except -1:
 *     """Is key present in self?"""
 *     return (key in _check_self(self))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_15_simple_set_pyx__check_self(__pyx_v_self)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 525; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 525; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":523
 * 
 * 
 * cdef api int SimpleSet_Contains(object self, object key) except -1:             # <<<<<<<<<<<<<<
 *     """Is key present in self?"""
 *     return (key in _check_self(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_Contains", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":528
 * 
 * 
 * cdef api int SimpleSet_Discard(object self, object key) except -1:             # <<<<<<<<<<<<<<
 *     """Remove the object referenced at location 'key'.
 * 
 */

static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Discard(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_Discard", 0);

  /* "bzrlib/_simple_set_pyx.pyx":536
 *         error.
 *     """
 *     return _check_self(self)._discard(key)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_15_simple_set_pyx__check_self(__pyx_v_self)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 536; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)((struct SimpleSetObject *)__pyx_t_1)->__pyx_vtab)->_discard(((struct SimpleSetObject *)__pyx_t_1), __pyx_v_key); if (unlikely(__pyx_t_2 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 536; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":528
 * 
 * 
 * cdef api int SimpleSet_Discard(object self, object key) except -1:             # <<<<<<<<<<<<<<
 *     """Remove the object referenced at location 'key'.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_Discard", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":539
 * 
 * 
 * cdef api PyObject *SimpleSet_Get(SimpleSet self, object key) except? NULL:             # <<<<<<<<<<<<<<
 *     """Get a pointer to the object present at location 'key'.
 * 
 */

static PyObject *__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Get(struct SimpleSetObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_Get", 0);

  /* "bzrlib/_simple_set_pyx.pyx":549
 *     :return: The object present at that location
 *     """
 *     return _check_self(self)._get(key)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_15_simple_set_pyx__check_self(((PyObject *)__pyx_v_self))); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 549; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet *)((struct SimpleSetObject *)__pyx_t_1)->__pyx_vtab)->_get(((struct SimpleSetObject *)__pyx_t_1), __pyx_v_key); if (unlikely(__pyx_t_2 == NULL && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 549; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":539
 * 
 * 
 * cdef api PyObject *SimpleSet_Get(SimpleSet self, object key) except? NULL:             # <<<<<<<<<<<<<<
 *     """Get a pointer to the object present at location 'key'.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_Get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":552
 * 
 * 
 * cdef api Py_ssize_t SimpleSet_Size(object self) except -1:             # <<<<<<<<<<<<<<
 *     """Get the number of active entries in 'self'"""
 *     return _check_self(self)._used
 */

static Py_ssize_t __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Size(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_Size", 0);

  /* "bzrlib/_simple_set_pyx.pyx":554
 * cdef api Py_ssize_t SimpleSet_Size(object self) except -1:
 *     """Get the number of active entries in 'self'"""
 *     return _check_self(self)._used             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_15_simple_set_pyx__check_self(__pyx_v_self)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 554; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct SimpleSetObject *)__pyx_t_1)->_used;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":552
 * 
 * 
 * cdef api Py_ssize_t SimpleSet_Size(object self) except -1:             # <<<<<<<<<<<<<<
 *     """Get the number of active entries in 'self'"""
 *     return _check_self(self)._used
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_Size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":557
 * 
 * 
 * cdef api int SimpleSet_Next(object self, Py_ssize_t *pos,             # <<<<<<<<<<<<<<
 *                             PyObject **key) except -1:
 *     """Walk over items in a SimpleSet.
 */

static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Next(PyObject *__pyx_v_self, Py_ssize_t *__pyx_v_pos, PyObject **__pyx_v_key) {
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_mask;
  struct SimpleSetObject *__pyx_v_true_self = 0;
  PyObject **__pyx_v_table;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject **__pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_Next", 0);

  /* "bzrlib/_simple_set_pyx.pyx":569
 *     cdef SimpleSet true_self
 *     cdef PyObject **table
 *     true_self = _check_self(self)             # <<<<<<<<<<<<<<
 *     i = pos[0]
 *     if (i < 0):
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_6bzrlib_15_simple_set_pyx__check_self(__pyx_v_self)); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 569; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_true_self = ((struct SimpleSetObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":570
 *     cdef PyObject **table
 *     true_self = _check_self(self)
 *     i = pos[0]             # <<<<<<<<<<<<<<
 *     if (i < 0):
 *         return 0
 */
  __pyx_v_i = (__pyx_v_pos[0]);

  /* "bzrlib/_simple_set_pyx.pyx":571
 *     true_self = _check_self(self)
 *     i = pos[0]
 *     if (i < 0):             # <<<<<<<<<<<<<<
 *         return 0
 *     mask = true_self._mask
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":572
 *     i = pos[0]
 *     if (i < 0):
 *         return 0             # <<<<<<<<<<<<<<
 *     mask = true_self._mask
 *     table= true_self._table
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":571
 *     true_self = _check_self(self)
 *     i = pos[0]
 *     if (i < 0):             # <<<<<<<<<<<<<<
 *         return 0
 *     mask = true_self._mask
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":573
 *     if (i < 0):
 *         return 0
 *     mask = true_self._mask             # <<<<<<<<<<<<<<
 *     table= true_self._table
 *     while (i <= mask and (table[i] == NULL or table[i] == _dummy)):
 */
  __pyx_t_3 = __pyx_v_true_self->_mask;
  __pyx_v_mask = __pyx_t_3;

  /* "bzrlib/_simple_set_pyx.pyx":574
 *         return 0
 *     mask = true_self._mask
 *     table= true_self._table             # <<<<<<<<<<<<<<
 *     while (i <= mask and (table[i] == NULL or table[i] == _dummy)):
 *         i = i + 1
 */
  __pyx_t_4 = __pyx_v_true_self->_table;
  __pyx_v_table = __pyx_t_4;

  /* "bzrlib/_simple_set_pyx.pyx":575
 *     mask = true_self._mask
 *     table= true_self._table
 *     while (i <= mask and (table[i] == NULL or table[i] == _dummy)):             # <<<<<<<<<<<<<<
 *         i = i + 1
 *     pos[0] = i + 1
 */
  while (1) {
    __pyx_t_5 = ((__pyx_v_i <= __pyx_v_mask) != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_5 = (((__pyx_v_table[__pyx_v_i]) == NULL) != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_5 = (((__pyx_v_table[__pyx_v_i]) == __pyx_v_6bzrlib_15_simple_set_pyx__dummy) != 0);
    __pyx_t_2 = __pyx_t_5;
    __pyx_L6_bool_binop_done:;
    if (!__pyx_t_2) break;

    /* "bzrlib/_simple_set_pyx.pyx":576
 *     table= true_self._table
 *     while (i <= mask and (table[i] == NULL or table[i] == _dummy)):
 *         i = i + 1             # <<<<<<<<<<<<<<
 *     pos[0] = i + 1
 *     if (i > mask):
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "bzrlib/_simple_set_pyx.pyx":577
 *     while (i <= mask and (table[i] == NULL or table[i] == _dummy)):
 *         i = i + 1
 *     pos[0] = i + 1             # <<<<<<<<<<<<<<
 *     if (i > mask):
 *         return 0 # All done
 */
  (__pyx_v_pos[0]) = (__pyx_v_i + 1);

  /* "bzrlib/_simple_set_pyx.pyx":578
 *         i = i + 1
 *     pos[0] = i + 1
 *     if (i > mask):             # <<<<<<<<<<<<<<
 *         return 0 # All done
 *     if (key != NULL):
 */
  __pyx_t_2 = ((__pyx_v_i > __pyx_v_mask) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":579
 *     pos[0] = i + 1
 *     if (i > mask):
 *         return 0 # All done             # <<<<<<<<<<<<<<
 *     if (key != NULL):
 *         key[0] = table[i]
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "bzrlib/_simple_set_pyx.pyx":578
 *         i = i + 1
 *     pos[0] = i + 1
 *     if (i > mask):             # <<<<<<<<<<<<<<
 *         return 0 # All done
 *     if (key != NULL):
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":580
 *     if (i > mask):
 *         return 0 # All done
 *     if (key != NULL):             # <<<<<<<<<<<<<<
 *         key[0] = table[i]
 *     return 1
 */
  __pyx_t_2 = ((__pyx_v_key != NULL) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_simple_set_pyx.pyx":581
 *         return 0 # All done
 *     if (key != NULL):
 *         key[0] = table[i]             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    (__pyx_v_key[0]) = (__pyx_v_table[__pyx_v_i]);

    /* "bzrlib/_simple_set_pyx.pyx":580
 *     if (i > mask):
 *         return 0 # All done
 *     if (key != NULL):             # <<<<<<<<<<<<<<
 *         key[0] = table[i]
 *     return 1
 */
  }

  /* "bzrlib/_simple_set_pyx.pyx":582
 *     if (key != NULL):
 *         key[0] = table[i]
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":557
 * 
 * 
 * cdef api int SimpleSet_Next(object self, Py_ssize_t *pos,             # <<<<<<<<<<<<<<
 *                             PyObject **key) except -1:
 *     """Walk over items in a SimpleSet.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_Next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_true_self);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_simple_set_pyx.pyx":585
 * 
 * 
 * cdef int SimpleSet_traverse(SimpleSet self, visitproc visit,             # <<<<<<<<<<<<<<
 *                             void *arg) except -1:
 *     """This is an implementation of 'tp_traverse' that hits the whole table.
 */

static int __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_traverse(struct SimpleSetObject *__pyx_v_self, visitproc __pyx_v_visit, void *__pyx_v_arg) {
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_next_key;
  int __pyx_v_ret;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SimpleSet_traverse", 0);

  /* "bzrlib/_simple_set_pyx.pyx":597
 *     cdef int ret
 * 
 *     pos = 0             # <<<<<<<<<<<<<<
 *     while SimpleSet_Next(self, &pos, &next_key):
 *         ret = visit(next_key, arg)
 */
  __pyx_v_pos = 0;

  /* "bzrlib/_simple_set_pyx.pyx":598
 * 
 *     pos = 0
 *     while SimpleSet_Next(self, &pos, &next_key):             # <<<<<<<<<<<<<<
 *         ret = visit(next_key, arg)
 *         if ret:
 */
  while (1) {
    __pyx_t_1 = __pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Next(((PyObject *)__pyx_v_self), (&__pyx_v_pos), (&__pyx_v_next_key)); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 598; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (!__pyx_t_2) break;

    /* "bzrlib/_simple_set_pyx.pyx":599
 *     pos = 0
 *     while SimpleSet_Next(self, &pos, &next_key):
 *         ret = visit(next_key, arg)             # <<<<<<<<<<<<<<
 *         if ret:
 *             return ret
 */
    __pyx_v_ret = __pyx_v_visit(__pyx_v_next_key, __pyx_v_arg);

    /* "bzrlib/_simple_set_pyx.pyx":600
 *     while SimpleSet_Next(self, &pos, &next_key):
 *         ret = visit(next_key, arg)
 *         if ret:             # <<<<<<<<<<<<<<
 *             return ret
 *     return 0
 */
    __pyx_t_2 = (__pyx_v_ret != 0);
    if (__pyx_t_2) {

      /* "bzrlib/_simple_set_pyx.pyx":601
 *         ret = visit(next_key, arg)
 *         if ret:
 *             return ret             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
      __pyx_r = __pyx_v_ret;
      goto __pyx_L0;

      /* "bzrlib/_simple_set_pyx.pyx":600
 *     while SimpleSet_Next(self, &pos, &next_key):
 *         ret = visit(next_key, arg)
 *         if ret:             # <<<<<<<<<<<<<<
 *             return ret
 *     return 0
 */
    }
  }

  /* "bzrlib/_simple_set_pyx.pyx":602
 *         if ret:
 *             return ret
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * # It is a little bit ugly to do this, but it works, and means that Meliae can
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "bzrlib/_simple_set_pyx.pyx":585
 * 
 * 
 * cdef int SimpleSet_traverse(SimpleSet self, visitproc visit,             # <<<<<<<<<<<<<<
 *                             void *arg) except -1:
 *     """This is an implementation of 'tp_traverse' that hits the whole table.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._simple_set_pyx.SimpleSet_traverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6bzrlib_15_simple_set_pyx_SimpleSet __pyx_vtable_6bzrlib_15_simple_set_pyx_SimpleSet;

static PyObject *__pyx_tp_new_6bzrlib_15_simple_set_pyx_SimpleSet(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct SimpleSetObject *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct SimpleSetObject *)o);
  p->__pyx_vtab = __pyx_vtabptr_6bzrlib_15_simple_set_pyx_SimpleSet;
  return o;
}

static void __pyx_tp_dealloc_6bzrlib_15_simple_set_pyx_SimpleSet(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_5__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_6bzrlib_15_simple_set_pyx_SimpleSet(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_6bzrlib_15_simple_set_pyx_9SimpleSet_used(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4used_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_15_simple_set_pyx_9SimpleSet_fill(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4fill_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_15_simple_set_pyx_9SimpleSet_mask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_4mask_1__get__(o);
}

static PyMethodDef __pyx_methods_6bzrlib_15_simple_set_pyx_SimpleSet[] = {
  {"__sizeof__", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_3__sizeof__, METH_NOARGS, 0},
  {"_memory_size", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_7_memory_size, METH_NOARGS, __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_6_memory_size},
  {"_test_lookup", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_11_test_lookup, METH_O, 0},
  {"_py_resize", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_17_py_resize, METH_O, __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_16_py_resize},
  {"add", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_19add, METH_O, __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_18add},
  {"discard", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_21discard, METH_O, __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_20discard},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6bzrlib_15_simple_set_pyx_SimpleSet[] = {
  {(char *)"used", __pyx_getprop_6bzrlib_15_simple_set_pyx_9SimpleSet_used, 0, 0, 0},
  {(char *)"fill", __pyx_getprop_6bzrlib_15_simple_set_pyx_9SimpleSet_fill, 0, 0, 0},
  {(char *)"mask", __pyx_getprop_6bzrlib_15_simple_set_pyx_9SimpleSet_mask, 0, 0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_SimpleSet = {
  __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_9__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_6bzrlib_15_simple_set_pyx_SimpleSet, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_13__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_SimpleSet = {
  __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_9__len__, /*mp_length*/
  __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_15__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

DL_EXPORT(PyTypeObject) SimpleSet_Type = {
  PyVarObject_HEAD_INIT(0, 0)
  "bzrlib._simple_set_pyx.SimpleSet", /*tp_name*/
  sizeof(struct SimpleSetObject), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6bzrlib_15_simple_set_pyx_SimpleSet, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_SimpleSet, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_SimpleSet, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "This class can be used to track canonical forms for objects.\n\n    It is similar in function to the interned dictionary that is used by\n    strings. However:\n\n      1) It assumes that hash(obj) is cheap, so does not need to inline a copy\n         of it\n      2) It only stores one reference to the object, rather than 2 (key vs\n         key:value)\n\n    As such, it uses 1/3rd the amount of memory to store a pointer to the\n    interned object.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_23__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6bzrlib_15_simple_set_pyx_SimpleSet, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6bzrlib_15_simple_set_pyx_SimpleSet, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6bzrlib_15_simple_set_pyx_9SimpleSet_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6bzrlib_15_simple_set_pyx_SimpleSet, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_6bzrlib_15_simple_set_pyx__SimpleSet_iterator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)o);
  p->set = ((struct SimpleSetObject *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6bzrlib_15_simple_set_pyx__SimpleSet_iterator(PyObject *o) {
  struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *p = (struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->set);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6bzrlib_15_simple_set_pyx__SimpleSet_iterator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *p = (struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)o;
  if (p->set) {
    e = (*v)(((PyObject*)p->set), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6bzrlib_15_simple_set_pyx__SimpleSet_iterator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *p = (struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator *)o;
  tmp = ((PyObject*)p->set);
  p->set = ((struct SimpleSetObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_6bzrlib_15_simple_set_pyx__SimpleSet_iterator[] = {
  {"__next__", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_5__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__length_hint__", (PyCFunction)__pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_7__length_hint__, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6bzrlib_15_simple_set_pyx__SimpleSet_iterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "bzrlib._simple_set_pyx._SimpleSet_iterator", /*tp_name*/
  sizeof(struct __pyx_obj_6bzrlib_15_simple_set_pyx__SimpleSet_iterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6bzrlib_15_simple_set_pyx__SimpleSet_iterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Iterator over the SimpleSet structure.", /*tp_doc*/
  __pyx_tp_traverse_6bzrlib_15_simple_set_pyx__SimpleSet_iterator, /*tp_traverse*/
  __pyx_tp_clear_6bzrlib_15_simple_set_pyx__SimpleSet_iterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_3__iter__, /*tp_iter*/
  __pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_5__next__, /*tp_iternext*/
  __pyx_methods_6bzrlib_15_simple_set_pyx__SimpleSet_iterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6bzrlib_15_simple_set_pyx_19_SimpleSet_iterator_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6bzrlib_15_simple_set_pyx__SimpleSet_iterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_simple_set_pyx",
    __pyx_k_Definition_of_a_class_that_is_si, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_s_Key_s_is_not_present, __pyx_k_Key_s_is_not_present, sizeof(__pyx_k_Key_s_is_not_present), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Set_size_changed_during_iteratio, __pyx_k_Set_size_changed_during_iteratio, sizeof(__pyx_k_Set_size_changed_during_iteratio), 0, 0, 1, 0},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Types_added_to_SimpleSet_must_im, __pyx_k_Types_added_to_SimpleSet_must_im, sizeof(__pyx_k_Types_added_to_SimpleSet_must_im), 0, 0, 1, 0},
  {&__pyx_kp_s_dummy, __pyx_k_dummy, sizeof(__pyx_k_dummy), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_kp_s_null, __pyx_k_null, sizeof(__pyx_k_null), 0, 0, 1, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_kp_s_ran_out_of_slots, __pyx_k_ran_out_of_slots, sizeof(__pyx_k_ran_out_of_slots), 0, 0, 1, 0},
  {&__pyx_kp_s_self_must_not_be_None, __pyx_k_self_must_not_be_None, sizeof(__pyx_k_self_must_not_be_None), 0, 0, 1, 0},
  {&__pyx_kp_s_should_never_get_here, __pyx_k_should_never_get_here, sizeof(__pyx_k_should_never_get_here), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 54; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 59; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 115; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 192; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 221; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 383; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 492; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "bzrlib/_simple_set_pyx.pyx":221
 *                 return 1
 *             i = i + 1 + n_lookup
 *         raise RuntimeError('ran out of slots.')             # <<<<<<<<<<<<<<
 * 
 *     def _py_resize(self, min_used):
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_ran_out_of_slots); if (unlikely(!__pyx_tuple_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 221; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "bzrlib/_simple_set_pyx.pyx":296
 *         if (Py_TYPE(py_key).tp_richcompare == NULL
 *             or Py_TYPE(py_key).tp_hash == NULL):
 *             raise TypeError('Types added to SimpleSet must implement'             # <<<<<<<<<<<<<<
 *                             ' both tp_richcompare and tp_hash')
 *         added = 0
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Types_added_to_SimpleSet_must_im); if (unlikely(!__pyx_tuple__2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 296; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "bzrlib/_simple_set_pyx.pyx":387
 *             # Force this exception to continue to be raised
 *             self._used = -1
 *             raise RuntimeError("Set size changed during iteration")             # <<<<<<<<<<<<<<
 *         if not SimpleSet_Next(self.set, &self.pos, &key):
 *             self.set = None
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_Set_size_changed_during_iteratio); if (unlikely(!__pyx_tuple__3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 387; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "bzrlib/_simple_set_pyx.pyx":418
 *     cdef SimpleSet true_self
 *     if self is None:
 *         raise TypeError('self must not be None')             # <<<<<<<<<<<<<<
 *     true_self = self
 *     return true_self
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_self_must_not_be_None); if (unlikely(!__pyx_tuple__4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 418; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "bzrlib/_simple_set_pyx.pyx":492
 *             return slot
 *         i = i + 1 + n_lookup
 *     raise AssertionError('should never get here')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_should_never_get_here); if (unlikely(!__pyx_tuple__5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 492; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_simple_set_pyx(void); /*proto*/
PyMODINIT_FUNC init_simple_set_pyx(void)
#else
PyMODINIT_FUNC PyInit__simple_set_pyx(void); /*proto*/
PyMODINIT_FUNC PyInit__simple_set_pyx(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__simple_set_pyx(void)", 0);
  if (__Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_simple_set_pyx", __pyx_methods, __pyx_k_Definition_of_a_class_that_is_si, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  if (__pyx_module_is_main_bzrlib___simple_set_pyx) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "bzrlib._simple_set_pyx")) {
      if (unlikely(PyDict_SetItemString(modules, "bzrlib._simple_set_pyx", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  __pyx_v_6bzrlib_15_simple_set_pyx__dummy_obj = Py_None; Py_INCREF(Py_None);
  __pyx_v_6bzrlib_15_simple_set_pyx__NotImplemented = Py_None; Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("SimpleSet_New", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_New, "struct SimpleSetObject *(void)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_ExportFunction("SimpleSet_Add", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Add, "PyObject *(PyObject *, PyObject *)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_ExportFunction("SimpleSet_Contains", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Contains, "int (PyObject *, PyObject *)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_ExportFunction("SimpleSet_Discard", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Discard, "int (PyObject *, PyObject *)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_ExportFunction("SimpleSet_Get", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Get, "PyObject *(struct SimpleSetObject *, PyObject *)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_ExportFunction("SimpleSet_Size", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Size, "Py_ssize_t (PyObject *)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_ExportFunction("SimpleSet_Next", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_Next, "int (PyObject *, Py_ssize_t *, PyObject **)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (__Pyx_ExportFunction("_SimpleSet_Lookup", (void (*)(void))__pyx_f_6bzrlib_15_simple_set_pyx__SimpleSet_Lookup, "PyObject **(PyObject *, PyObject *)") < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Type init code ---*/
  __pyx_vtabptr_6bzrlib_15_simple_set_pyx_SimpleSet = &__pyx_vtable_6bzrlib_15_simple_set_pyx_SimpleSet;
  __pyx_vtable_6bzrlib_15_simple_set_pyx_SimpleSet._get = (PyObject *(*)(struct SimpleSetObject *, PyObject *))__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__get;
  __pyx_vtable_6bzrlib_15_simple_set_pyx_SimpleSet._add = (PyObject *(*)(struct SimpleSetObject *, PyObject *))__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__add;
  __pyx_vtable_6bzrlib_15_simple_set_pyx_SimpleSet._discard = (int (*)(struct SimpleSetObject *, PyObject *))__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__discard;
  __pyx_vtable_6bzrlib_15_simple_set_pyx_SimpleSet._insert_clean = (int (*)(struct SimpleSetObject *, PyObject *))__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__insert_clean;
  __pyx_vtable_6bzrlib_15_simple_set_pyx_SimpleSet._resize = (Py_ssize_t (*)(struct SimpleSetObject *, Py_ssize_t))__pyx_f_6bzrlib_15_simple_set_pyx_9SimpleSet__resize;
  if (PyType_Ready(&SimpleSet_Type) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  SimpleSet_Type.tp_print = 0;
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&SimpleSet_Type, "__contains__"); if (unlikely(!wrapper)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__.doc = __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_12__contains__;
    }
  }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&SimpleSet_Type, "__getitem__"); if (unlikely(!wrapper)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__.doc = __pyx_doc_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6bzrlib_15_simple_set_pyx_9SimpleSet_14__getitem__;
    }
  }
  #endif
  if (__Pyx_SetVtable(SimpleSet_Type.tp_dict, __pyx_vtabptr_6bzrlib_15_simple_set_pyx_SimpleSet) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyObject_SetAttrString(__pyx_m, "SimpleSet", (PyObject *)&SimpleSet_Type) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 88; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_6bzrlib_15_simple_set_pyx_SimpleSet = &SimpleSet_Type;
  if (PyType_Ready(&__pyx_type_6bzrlib_15_simple_set_pyx__SimpleSet_iterator) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 361; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_type_6bzrlib_15_simple_set_pyx__SimpleSet_iterator.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "_SimpleSet_iterator", (PyObject *)&__pyx_type_6bzrlib_15_simple_set_pyx__SimpleSet_iterator) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 361; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_6bzrlib_15_simple_set_pyx__SimpleSet_iterator = &__pyx_type_6bzrlib_15_simple_set_pyx__SimpleSet_iterator;
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif

  /* "bzrlib/_simple_set_pyx.pyx":54
 * cdef object _dummy_obj
 * cdef PyObject *_dummy
 * _dummy_obj = object()             # <<<<<<<<<<<<<<
 * _dummy = <PyObject *>_dummy_obj
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_object); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 54; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(__pyx_v_6bzrlib_15_simple_set_pyx__dummy_obj);
  __Pyx_DECREF_SET(__pyx_v_6bzrlib_15_simple_set_pyx__dummy_obj, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_simple_set_pyx.pyx":55
 * cdef PyObject *_dummy
 * _dummy_obj = object()
 * _dummy = <PyObject *>_dummy_obj             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_6bzrlib_15_simple_set_pyx__dummy = ((PyObject *)__pyx_v_6bzrlib_15_simple_set_pyx__dummy_obj);

  /* "bzrlib/_simple_set_pyx.pyx":59
 * 
 * cdef object _NotImplemented
 * _NotImplemented = NotImplemented             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __Pyx_XGOTREF(__pyx_v_6bzrlib_15_simple_set_pyx__NotImplemented);
  __Pyx_DECREF_SET(__pyx_v_6bzrlib_15_simple_set_pyx__NotImplemented, __pyx_builtin_NotImplemented);
  __Pyx_GIVEREF(__pyx_builtin_NotImplemented);

  /* "bzrlib/_simple_set_pyx.pyx":606
 * # It is a little bit ugly to do this, but it works, and means that Meliae can
 * # dump the total memory consumed by all child objects.
 * (<PyTypeObject *>SimpleSet).tp_traverse = <traverseproc>SimpleSet_traverse             # <<<<<<<<<<<<<<
 */
  ((PyTypeObject *)__pyx_ptype_6bzrlib_15_simple_set_pyx_SimpleSet)->tp_traverse = ((traverseproc)__pyx_f_6bzrlib_15_simple_set_pyx_SimpleSet_traverse);

  /* "bzrlib/_simple_set_pyx.pyx":1
 * # Copyright (C) 2009, 2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init bzrlib._simple_set_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init bzrlib._simple_set_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
#endif

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, -(sdigit) digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, -(sdigit) digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
