/* Generated by Cython 0.23.3 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_23_3"
#include <stddef.h>
#ifndef offsetof
#define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
#define CYTHON_COMPILING_IN_PYPY 1
#define CYTHON_COMPILING_IN_CPYTHON 0
#else
#define CYTHON_COMPILING_IN_PYPY 0
#define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
#define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
#define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) (void_result, Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__bzrlib___bencode_pyx
#define __PYX_HAVE_API__bzrlib___bencode_pyx
#include "stddef.h"
#include "stdlib.h"
#include "string.h"
#include "python-compat.h"
#include "_bencode_pyx.h"
#include "_static_tuple_c.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "bzrlib/_bencode_pyx.pyx",
  "bzrlib/_static_tuple_c.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder;
struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder;

/* "bzrlib/_bencode_pyx.pyx":240
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     INITSIZE = 1024     # initial size for encoder buffer
 *     INT_BUF_SIZE = 32
 */
enum  {
  __pyx_e_6bzrlib_12_bencode_pyx_INITSIZE = 0x400,
  __pyx_e_6bzrlib_12_bencode_pyx_INT_BUF_SIZE = 32
};

/* "bzrlib/_bencode_pyx.pyx":54
 *     int snprintf(char* buffer, size_t nsize, char* fmt, ...)
 * 
 * cdef class Decoder             # <<<<<<<<<<<<<<
 * cdef class Encoder
 * 
 */
struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder {
  PyObject_HEAD
  struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *__pyx_vtab;
  char *tail;
  int size;
  int _yield_tuples;
  PyObject *text;
};


/* "bzrlib/_bencode_pyx.pyx":55
 * 
 * cdef class Decoder
 * cdef class Encoder             # <<<<<<<<<<<<<<
 * 
 * cdef extern from "_bencode_pyx.h":
 */
struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder {
  PyObject_HEAD
  struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *__pyx_vtab;
  char *tail;
  int size;
  char *buffer;
  int maxsize;
};



/* "bzrlib/_bencode_pyx.pyx":69
 * 
 * 
 * cdef class Decoder:             # <<<<<<<<<<<<<<
 *     """Bencode decoder"""
 * 
 */

struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder {
  PyObject *(*_decode_object)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *);
  int (*_read_digits)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *, char);
  PyObject *(*_decode_int)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *);
  PyObject *(*_decode_string)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *);
  PyObject *(*_decode_list)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *);
  PyObject *(*_decode_dict)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *);
};
static struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *__pyx_vtabptr_6bzrlib_12_bencode_pyx_Decoder;


/* "bzrlib/_bencode_pyx.pyx":245
 * 
 * 
 * cdef class Encoder:             # <<<<<<<<<<<<<<
 *     """Bencode encoder"""
 * 
 */

struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder {
  int (*_ensure_buffer)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, int);
  int (*_encode_int)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, int);
  int (*_encode_long)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *);
  int (*_append_string)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *);
  int (*_encode_string)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *);
  int (*_encode_list)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *);
  int (*_encode_dict)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *);
};
static struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *__pyx_vtabptr_6bzrlib_12_bencode_pyx_Encoder;

/* --- Runtime support code (head) --- */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name);

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb);

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);

static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);

static CYTHON_INLINE void __Pyx_ExceptionSave(PyObject **type, PyObject **value, PyObject **tb);
static void __Pyx_ExceptionReset(PyObject *type, PyObject *value, PyObject *tb);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

static int __Pyx_SetVtable(PyObject *dict, void *vtable);

static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name);

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_object(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Decoder__read_digits(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self, char __pyx_v_stop_char); /* proto*/
static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_int(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_string(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_list(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_dict(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__ensure_buffer(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, int __pyx_v_required); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_int(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, int __pyx_v_x); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_long(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__append_string(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_s); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_string(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_list(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x); /* proto*/
static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_dict(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x); /* proto*/

/* Module declarations from 'bzrlib._static_tuple_c' */

/* Module declarations from 'bzrlib._static_tuple_c' */
static PyTypeObject *__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple = 0;

/* Module declarations from 'bzrlib._bencode_pyx' */
static PyTypeObject *__pyx_ptype_6bzrlib_12_bencode_pyx_Decoder = 0;
static PyTypeObject *__pyx_ptype_6bzrlib_12_bencode_pyx_Encoder = 0;
#define __Pyx_MODULE_NAME "bzrlib._bencode_pyx"
int __pyx_module_is_main_bzrlib___bencode_pyx = 0;

/* Implementation of 'bzrlib._bencode_pyx' */
static PyObject *__pyx_builtin_object;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_MemoryError;
static char __pyx_k_d[] = "%d:";
static char __pyx_k_e[] = "e";
static char __pyx_k_i[] = "i";
static char __pyx_k_s[] = "s";
static char __pyx_k_x[] = "x";
static char __pyx_k__7[] = "";
static char __pyx_k_doc[] = "__doc__";
static char __pyx_k_i_de[] = "i%de";
static char __pyx_k_init[] = "__init__";
static char __pyx_k_join[] = "join";
static char __pyx_k_keys[] = "keys";
static char __pyx_k_main[] = "__main__";
static char __pyx_k_self[] = "self";
static char __pyx_k_sort[] = "sort";
static char __pyx_k_test[] = "__test__";
static char __pyx_k_slots[] = "__slots__";
static char __pyx_k_decode[] = "decode";
static char __pyx_k_encode[] = "encode";
static char __pyx_k_module[] = "__module__";
static char __pyx_k_object[] = "object";
static char __pyx_k_bdecode[] = "bdecode";
static char __pyx_k_bencode[] = "bencode";
static char __pyx_k_encoder[] = "encoder";
static char __pyx_k_maxsize[] = "maxsize";
static char __pyx_k_prepare[] = "__prepare__";
static char __pyx_k_process[] = "process";
static char __pyx_k_bencoded[] = "bencoded";
static char __pyx_k_qualname[] = "__qualname__";
static char __pyx_k_Bencached[] = "Bencached";
static char __pyx_k_TypeError[] = "TypeError";
static char __pyx_k_metaclass[] = "__metaclass__";
static char __pyx_k_ValueError[] = "ValueError";
static char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static char __pyx_k_MemoryError[] = "MemoryError";
static char __pyx_k_RuntimeError[] = "RuntimeError";
static char __pyx_k_yield_tuples[] = "yield_tuples";
static char __pyx_k_decode_object[] = "_decode_object";
static char __pyx_k_junk_in_stream[] = "junk in stream";
static char __pyx_k_malformed_dict[] = "malformed dict";
static char __pyx_k_malformed_list[] = "malformed list";
static char __pyx_k_String_required[] = "String required";
static char __pyx_k_Bencached___init[] = "Bencached.__init__";
static char __pyx_k_bdecode_as_tuple[] = "bdecode_as_tuple";
static char __pyx_k_stream_underflow[] = "stream underflow";
static char __pyx_k_too_deeply_nested[] = "too deeply nested";
static char __pyx_k_unsupported_type_r[] = "unsupported type %r";
static char __pyx_k_bzrlib__bencode_pyx[] = "bzrlib._bencode_pyx";
static char __pyx_k_dict_keys_disordered[] = "dict keys disordered";
static char __pyx_k_int_d_too_big_to_encode[] = "int %d too big to encode";
static char __pyx_k_string_size_below_zero_d[] = "string size below zero: %d";
static char __pyx_k_string_s_too_big_to_encode[] = "string %s too big to encode";
static char __pyx_k_key_was_not_a_simple_string[] = "key was not a simple string.";
static char __pyx_k_Stop_character_c_not_found_c[] = "Stop character %c not found: %c";
static char __pyx_k_key_in_dict_should_be_string[] = "key in dict should be string";
static char __pyx_k_string_len_not_terminated_by[] = "string len not terminated by \":\"";
static char __pyx_k_leading_zeros_are_not_allowed[] = "leading zeros are not allowed";
static char __pyx_k_home_vila_src_bzr_integration_2[] = "/home/vila/src/bzr/integration/2.7.0-dev/bzrlib/_bencode_pyx.pyx";
static char __pyx_k_Cannot_realloc_buffer_for_encode[] = "Cannot realloc buffer for encoder";
static char __pyx_k_Not_enough_memory_to_allocate_bu[] = "Not enough memory to allocate buffer for encoder";
static char __pyx_k_Pyrex_implementation_for_bencode[] = "Pyrex implementation for bencode coder/decoder";
static char __pyx_k_unknown_object_type_identifier_r[] = "unknown object type identifier %r";
static PyObject *__pyx_n_s_Bencached;
static PyObject *__pyx_n_s_Bencached___init;
static PyObject *__pyx_kp_s_Cannot_realloc_buffer_for_encode;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_Not_enough_memory_to_allocate_bu;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_kp_s_Stop_character_c_not_found_c;
static PyObject *__pyx_kp_s_String_required;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__7;
static PyObject *__pyx_n_s_bdecode;
static PyObject *__pyx_n_s_bdecode_as_tuple;
static PyObject *__pyx_n_s_bencode;
static PyObject *__pyx_n_s_bencoded;
static PyObject *__pyx_n_s_bzrlib__bencode_pyx;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_kp_s_dict_keys_disordered;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_encoder;
static PyObject *__pyx_kp_s_home_vila_src_bzr_integration_2;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_kp_s_int_d_too_big_to_encode;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_kp_s_junk_in_stream;
static PyObject *__pyx_kp_s_key_in_dict_should_be_string;
static PyObject *__pyx_kp_s_key_was_not_a_simple_string;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_kp_s_leading_zeros_are_not_allowed;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_kp_s_malformed_dict;
static PyObject *__pyx_kp_s_malformed_list;
static PyObject *__pyx_n_s_maxsize;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_process;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_slots;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_kp_s_stream_underflow;
static PyObject *__pyx_kp_s_string_len_not_terminated_by;
static PyObject *__pyx_kp_s_string_s_too_big_to_encode;
static PyObject *__pyx_kp_s_string_size_below_zero_d;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_too_deeply_nested;
static PyObject *__pyx_kp_s_unknown_object_type_identifier_r;
static PyObject *__pyx_kp_s_unsupported_type_r;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_yield_tuples;
static int __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder___init__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self, PyObject *__pyx_v_s, PyObject *__pyx_v_yield_tuples); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_2decode(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4decode_object(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4tail___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4size___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_13_yield_tuples___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_bdecode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_2bdecode_as_tuple(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_9Bencached___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_s); /* proto */
static int __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder___init__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, int __pyx_v_maxsize); /* proto */
static void __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_2__dealloc__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4__str__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_6process(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4tail___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4size___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_6buffer___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_7maxsize___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_4bencode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_tp_new_6bzrlib_12_bencode_pyx_Decoder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6bzrlib_12_bencode_pyx_Encoder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static int __pyx_k__13;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;

/* "bzrlib/_bencode_pyx.pyx":77
 *     cdef object text
 * 
 *     def __init__(self, s, yield_tuples=0):             # <<<<<<<<<<<<<<
 *         """Initialize decoder engine.
 *         @param  s:  Python string.
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6bzrlib_12_bencode_pyx_7Decoder___init__[] = "Initialize decoder engine.\n        @param  s:  Python string.\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Decoder___init__;
#endif
static int __pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_v_yield_tuples = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_s,&__pyx_n_s_yield_tuples,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_s)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yield_tuples);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_s = values[0];
    __pyx_v_yield_tuples = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 77; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder___init__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self), __pyx_v_s, __pyx_v_yield_tuples);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder___init__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self, PyObject *__pyx_v_s, PyObject *__pyx_v_yield_tuples) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  char *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "bzrlib/_bencode_pyx.pyx":81
 *         @param  s:  Python string.
 *         """
 *         if not PyString_CheckExact(s):             # <<<<<<<<<<<<<<
 *             raise TypeError("String required")
 * 
 */
  __pyx_t_1 = ((!(PyString_CheckExact(__pyx_v_s) != 0)) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":82
 *         """
 *         if not PyString_CheckExact(s):
 *             raise TypeError("String required")             # <<<<<<<<<<<<<<
 * 
 *         self.text = s
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 82; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 82; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":81
 *         @param  s:  Python string.
 *         """
 *         if not PyString_CheckExact(s):             # <<<<<<<<<<<<<<
 *             raise TypeError("String required")
 * 
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":84
 *             raise TypeError("String required")
 * 
 *         self.text = s             # <<<<<<<<<<<<<<
 *         self.tail = PyString_AS_STRING(s)
 *         self.size = PyString_GET_SIZE(s)
 */
  __Pyx_INCREF(__pyx_v_s);
  __Pyx_GIVEREF(__pyx_v_s);
  __Pyx_GOTREF(__pyx_v_self->text);
  __Pyx_DECREF(__pyx_v_self->text);
  __pyx_v_self->text = __pyx_v_s;

  /* "bzrlib/_bencode_pyx.pyx":85
 * 
 *         self.text = s
 *         self.tail = PyString_AS_STRING(s)             # <<<<<<<<<<<<<<
 *         self.size = PyString_GET_SIZE(s)
 *         self._yield_tuples = int(yield_tuples)
 */
  __pyx_t_3 = PyString_AS_STRING(__pyx_v_s); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 85; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->tail = __pyx_t_3;

  /* "bzrlib/_bencode_pyx.pyx":86
 *         self.text = s
 *         self.tail = PyString_AS_STRING(s)
 *         self.size = PyString_GET_SIZE(s)             # <<<<<<<<<<<<<<
 *         self._yield_tuples = int(yield_tuples)
 * 
 */
  __pyx_t_4 = PyString_GET_SIZE(__pyx_v_s); if (unlikely(__pyx_t_4 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 86; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_self->size = __pyx_t_4;

  /* "bzrlib/_bencode_pyx.pyx":87
 *         self.tail = PyString_AS_STRING(s)
 *         self.size = PyString_GET_SIZE(s)
 *         self._yield_tuples = int(yield_tuples)             # <<<<<<<<<<<<<<
 * 
 *     def decode(self):
 */
  __pyx_t_2 = PyNumber_Int(__pyx_v_yield_tuples); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 87; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 87; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->_yield_tuples = __pyx_t_5;

  /* "bzrlib/_bencode_pyx.pyx":77
 *     cdef object text
 * 
 *     def __init__(self, s, yield_tuples=0):             # <<<<<<<<<<<<<<
 *         """Initialize decoder engine.
 *         @param  s:  Python string.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":89
 *         self._yield_tuples = int(yield_tuples)
 * 
 *     def decode(self):             # <<<<<<<<<<<<<<
 *         result = self._decode_object()
 *         if self.size != 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_3decode(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_3decode(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decode (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_2decode(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_2decode(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode", 0);

  /* "bzrlib/_bencode_pyx.pyx":90
 * 
 *     def decode(self):
 *         result = self._decode_object()             # <<<<<<<<<<<<<<
 *         if self.size != 0:
 *             raise ValueError('junk in stream')
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 90; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "bzrlib/_bencode_pyx.pyx":91
 *     def decode(self):
 *         result = self._decode_object()
 *         if self.size != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('junk in stream')
 *         return result
 */
  __pyx_t_2 = ((__pyx_v_self->size != 0) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_bencode_pyx.pyx":92
 *         result = self._decode_object()
 *         if self.size != 0:
 *             raise ValueError('junk in stream')             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":91
 *     def decode(self):
 *         result = self._decode_object()
 *         if self.size != 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('junk in stream')
 *         return result
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":93
 *         if self.size != 0:
 *             raise ValueError('junk in stream')
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def decode_object(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":89
 *         self._yield_tuples = int(yield_tuples)
 * 
 *     def decode(self):             # <<<<<<<<<<<<<<
 *         result = self._decode_object()
 *         if self.size != 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":95
 *         return result
 * 
 *     def decode_object(self):             # <<<<<<<<<<<<<<
 *         return self._decode_object()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_5decode_object(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_5decode_object(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decode_object (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4decode_object(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4decode_object(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("decode_object", 0);

  /* "bzrlib/_bencode_pyx.pyx":96
 * 
 *     def decode_object(self):
 *         return self._decode_object()             # <<<<<<<<<<<<<<
 * 
 *     cdef object _decode_object(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 96; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":95
 *         return result
 * 
 *     def decode_object(self):             # <<<<<<<<<<<<<<
 *         return self._decode_object()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder.decode_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":98
 *         return self._decode_object()
 * 
 *     cdef object _decode_object(self):             # <<<<<<<<<<<<<<
 *         cdef char ch
 * 
 */

static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_object(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  char __pyx_v_ch;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_decode_object", 0);

  /* "bzrlib/_bencode_pyx.pyx":101
 *         cdef char ch
 * 
 *         if 0 == self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError('stream underflow')
 * 
 */
  __pyx_t_1 = ((0 == __pyx_v_self->size) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":102
 * 
 *         if 0 == self.size:
 *             raise ValueError('stream underflow')             # <<<<<<<<<<<<<<
 * 
 *         if Py_EnterRecursiveCall("_decode_object"):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":101
 *         cdef char ch
 * 
 *         if 0 == self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError('stream underflow')
 * 
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":104
 *             raise ValueError('stream underflow')
 * 
 *         if Py_EnterRecursiveCall("_decode_object"):             # <<<<<<<<<<<<<<
 *             raise RuntimeError("too deeply nested")
 *         try:
 */
  __pyx_t_1 = (Py_EnterRecursiveCall(__pyx_k_decode_object) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":105
 * 
 *         if Py_EnterRecursiveCall("_decode_object"):
 *             raise RuntimeError("too deeply nested")             # <<<<<<<<<<<<<<
 *         try:
 *             ch = self.tail[0]
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":104
 *             raise ValueError('stream underflow')
 * 
 *         if Py_EnterRecursiveCall("_decode_object"):             # <<<<<<<<<<<<<<
 *             raise RuntimeError("too deeply nested")
 *         try:
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":106
 *         if Py_EnterRecursiveCall("_decode_object"):
 *             raise RuntimeError("too deeply nested")
 *         try:             # <<<<<<<<<<<<<<
 *             ch = self.tail[0]
 *             if c'0' <= ch <= c'9':
 */
  /*try:*/ {

    /* "bzrlib/_bencode_pyx.pyx":107
 *             raise RuntimeError("too deeply nested")
 *         try:
 *             ch = self.tail[0]             # <<<<<<<<<<<<<<
 *             if c'0' <= ch <= c'9':
 *                 return self._decode_string()
 */
    __pyx_v_ch = (__pyx_v_self->tail[0]);

    /* "bzrlib/_bencode_pyx.pyx":108
 *         try:
 *             ch = self.tail[0]
 *             if c'0' <= ch <= c'9':             # <<<<<<<<<<<<<<
 *                 return self._decode_string()
 *             elif ch == c'l':
 */
    __pyx_t_1 = ('0' <= __pyx_v_ch);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_ch <= '9');
    }
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "bzrlib/_bencode_pyx.pyx":109
 *             ch = self.tail[0]
 *             if c'0' <= ch <= c'9':
 *                 return self._decode_string()             # <<<<<<<<<<<<<<
 *             elif ch == c'l':
 *                 D_UPDATE_TAIL(self, 1)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_string(__pyx_v_self); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 109; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L5_return;

      /* "bzrlib/_bencode_pyx.pyx":108
 *         try:
 *             ch = self.tail[0]
 *             if c'0' <= ch <= c'9':             # <<<<<<<<<<<<<<
 *                 return self._decode_string()
 *             elif ch == c'l':
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":110
 *             if c'0' <= ch <= c'9':
 *                 return self._decode_string()
 *             elif ch == c'l':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_list()
 */
    __pyx_t_3 = ((__pyx_v_ch == 'l') != 0);
    if (__pyx_t_3) {

      /* "bzrlib/_bencode_pyx.pyx":111
 *                 return self._decode_string()
 *             elif ch == c'l':
 *                 D_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 *                 return self._decode_list()
 *             elif ch == c'i':
 */
      D_UPDATE_TAIL(__pyx_v_self, 1);

      /* "bzrlib/_bencode_pyx.pyx":112
 *             elif ch == c'l':
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_list()             # <<<<<<<<<<<<<<
 *             elif ch == c'i':
 *                 D_UPDATE_TAIL(self, 1)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_list(__pyx_v_self); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 112; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L5_return;

      /* "bzrlib/_bencode_pyx.pyx":110
 *             if c'0' <= ch <= c'9':
 *                 return self._decode_string()
 *             elif ch == c'l':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_list()
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":113
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_list()
 *             elif ch == c'i':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_int()
 */
    __pyx_t_3 = ((__pyx_v_ch == 'i') != 0);
    if (__pyx_t_3) {

      /* "bzrlib/_bencode_pyx.pyx":114
 *                 return self._decode_list()
 *             elif ch == c'i':
 *                 D_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 *                 return self._decode_int()
 *             elif ch == c'd':
 */
      D_UPDATE_TAIL(__pyx_v_self, 1);

      /* "bzrlib/_bencode_pyx.pyx":115
 *             elif ch == c'i':
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_int()             # <<<<<<<<<<<<<<
 *             elif ch == c'd':
 *                 D_UPDATE_TAIL(self, 1)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_int(__pyx_v_self); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 115; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L5_return;

      /* "bzrlib/_bencode_pyx.pyx":113
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_list()
 *             elif ch == c'i':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_int()
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":116
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_int()
 *             elif ch == c'd':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_dict()
 */
    __pyx_t_3 = ((__pyx_v_ch == 'd') != 0);
    if (__pyx_t_3) {

      /* "bzrlib/_bencode_pyx.pyx":117
 *                 return self._decode_int()
 *             elif ch == c'd':
 *                 D_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 *                 return self._decode_dict()
 *             else:
 */
      D_UPDATE_TAIL(__pyx_v_self, 1);

      /* "bzrlib/_bencode_pyx.pyx":118
 *             elif ch == c'd':
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_dict()             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError('unknown object type identifier %r' % ch)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_dict(__pyx_v_self); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 118; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L5_return;

      /* "bzrlib/_bencode_pyx.pyx":116
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_int()
 *             elif ch == c'd':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return self._decode_dict()
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":120
 *                 return self._decode_dict()
 *             else:
 *                 raise ValueError('unknown object type identifier %r' % ch)             # <<<<<<<<<<<<<<
 *         finally:
 *             Py_LeaveRecursiveCall()
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyInt_From_char(__pyx_v_ch); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_unknown_object_type_identifier_r, __pyx_t_2); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 120; __pyx_clineno = __LINE__; goto __pyx_L6_error;}
    }
  }

  /* "bzrlib/_bencode_pyx.pyx":122
 *                 raise ValueError('unknown object type identifier %r' % ch)
 *         finally:
 *             Py_LeaveRecursiveCall()             # <<<<<<<<<<<<<<
 * 
 *     cdef int _read_digits(self, char stop_char) except -1:
 */
  /*finally:*/ {
    /*exception exit:*/{
      __pyx_L6_error:;
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        Py_LeaveRecursiveCall();
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
    }
    __pyx_L5_return: {
      __pyx_t_13 = __pyx_r;
      __pyx_r = 0;
      Py_LeaveRecursiveCall();
      __pyx_r = __pyx_t_13;
      __pyx_t_13 = 0;
      goto __pyx_L0;
    }
  }

  /* "bzrlib/_bencode_pyx.pyx":98
 *         return self._decode_object()
 * 
 *     cdef object _decode_object(self):             # <<<<<<<<<<<<<<
 *         cdef char ch
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder._decode_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":124
 *             Py_LeaveRecursiveCall()
 * 
 *     cdef int _read_digits(self, char stop_char) except -1:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         i = 0
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Decoder__read_digits(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self, char __pyx_v_stop_char) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_read_digits", 0);

  /* "bzrlib/_bencode_pyx.pyx":126
 *     cdef int _read_digits(self, char stop_char) except -1:
 *         cdef int i
 *         i = 0             # <<<<<<<<<<<<<<
 *         while ((self.tail[i] >= c'0' and self.tail[i] <= c'9') or
 *                self.tail[i] == c'-') and i < self.size:
 */
  __pyx_v_i = 0;

  /* "bzrlib/_bencode_pyx.pyx":127
 *         cdef int i
 *         i = 0
 *         while ((self.tail[i] >= c'0' and self.tail[i] <= c'9') or             # <<<<<<<<<<<<<<
 *                self.tail[i] == c'-') and i < self.size:
 *             i = i + 1
 */
  while (1) {

    /* "bzrlib/_bencode_pyx.pyx":128
 *         i = 0
 *         while ((self.tail[i] >= c'0' and self.tail[i] <= c'9') or
 *                self.tail[i] == c'-') and i < self.size:             # <<<<<<<<<<<<<<
 *             i = i + 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_self->tail[__pyx_v_i]) >= '0') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L7_next_or;
    } else {
    }

    /* "bzrlib/_bencode_pyx.pyx":127
 *         cdef int i
 *         i = 0
 *         while ((self.tail[i] >= c'0' and self.tail[i] <= c'9') or             # <<<<<<<<<<<<<<
 *                self.tail[i] == c'-') and i < self.size:
 *             i = i + 1
 */
    __pyx_t_2 = (((__pyx_v_self->tail[__pyx_v_i]) <= '9') != 0);
    if (!__pyx_t_2) {
    } else {
      goto __pyx_L6_next_and;
    }
    __pyx_L7_next_or:;

    /* "bzrlib/_bencode_pyx.pyx":128
 *         i = 0
 *         while ((self.tail[i] >= c'0' and self.tail[i] <= c'9') or
 *                self.tail[i] == c'-') and i < self.size:             # <<<<<<<<<<<<<<
 *             i = i + 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_self->tail[__pyx_v_i]) == '-') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_L6_next_and:;
    __pyx_t_2 = ((__pyx_v_i < __pyx_v_self->size) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L5_bool_binop_done:;
    if (!__pyx_t_1) break;

    /* "bzrlib/_bencode_pyx.pyx":129
 *         while ((self.tail[i] >= c'0' and self.tail[i] <= c'9') or
 *                self.tail[i] == c'-') and i < self.size:
 *             i = i + 1             # <<<<<<<<<<<<<<
 * 
 *         if self.tail[i] != stop_char:
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "bzrlib/_bencode_pyx.pyx":131
 *             i = i + 1
 * 
 *         if self.tail[i] != stop_char:             # <<<<<<<<<<<<<<
 *             raise ValueError("Stop character %c not found: %c" %
 *                 (stop_char, self.tail[i]))
 */
  __pyx_t_1 = (((__pyx_v_self->tail[__pyx_v_i]) != __pyx_v_stop_char) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":133
 *         if self.tail[i] != stop_char:
 *             raise ValueError("Stop character %c not found: %c" %
 *                 (stop_char, self.tail[i]))             # <<<<<<<<<<<<<<
 *         if (self.tail[0] == c'0' or
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):
 */
    __pyx_t_3 = __Pyx_PyInt_From_char(__pyx_v_stop_char); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 133; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_char((__pyx_v_self->tail[__pyx_v_i])); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 133; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 133; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;

    /* "bzrlib/_bencode_pyx.pyx":132
 * 
 *         if self.tail[i] != stop_char:
 *             raise ValueError("Stop character %c not found: %c" %             # <<<<<<<<<<<<<<
 *                 (stop_char, self.tail[i]))
 *         if (self.tail[0] == c'0' or
 */
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Stop_character_c_not_found_c, __pyx_t_5); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 132; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 132; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 132; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 132; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":131
 *             i = i + 1
 * 
 *         if self.tail[i] != stop_char:             # <<<<<<<<<<<<<<
 *             raise ValueError("Stop character %c not found: %c" %
 *                 (stop_char, self.tail[i]))
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":134
 *             raise ValueError("Stop character %c not found: %c" %
 *                 (stop_char, self.tail[i]))
 *         if (self.tail[0] == c'0' or             # <<<<<<<<<<<<<<
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):
 *             if i == 1:
 */
  __pyx_t_2 = (((__pyx_v_self->tail[0]) == '0') != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L11_bool_binop_done;
  }

  /* "bzrlib/_bencode_pyx.pyx":135
 *                 (stop_char, self.tail[i]))
 *         if (self.tail[0] == c'0' or
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):             # <<<<<<<<<<<<<<
 *             if i == 1:
 *                 return i
 */
  __pyx_t_2 = (((__pyx_v_self->tail[0]) == '-') != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_self->tail[1]) == '0') != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L11_bool_binop_done:;

  /* "bzrlib/_bencode_pyx.pyx":134
 *             raise ValueError("Stop character %c not found: %c" %
 *                 (stop_char, self.tail[i]))
 *         if (self.tail[0] == c'0' or             # <<<<<<<<<<<<<<
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):
 *             if i == 1:
 */
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":136
 *         if (self.tail[0] == c'0' or
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):
 *             if i == 1:             # <<<<<<<<<<<<<<
 *                 return i
 *             else:
 */
    __pyx_t_1 = ((__pyx_v_i == 1) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_bencode_pyx.pyx":137
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):
 *             if i == 1:
 *                 return i             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError # leading zeroes are not allowed
 */
      __pyx_r = __pyx_v_i;
      goto __pyx_L0;

      /* "bzrlib/_bencode_pyx.pyx":136
 *         if (self.tail[0] == c'0' or
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):
 *             if i == 1:             # <<<<<<<<<<<<<<
 *                 return i
 *             else:
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":139
 *                 return i
 *             else:
 *                 raise ValueError # leading zeroes are not allowed             # <<<<<<<<<<<<<<
 *         return i
 * 
 */
    /*else*/ {
      __Pyx_Raise(__pyx_builtin_ValueError, 0, 0, 0);
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 139; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }

    /* "bzrlib/_bencode_pyx.pyx":134
 *             raise ValueError("Stop character %c not found: %c" %
 *                 (stop_char, self.tail[i]))
 *         if (self.tail[0] == c'0' or             # <<<<<<<<<<<<<<
 *                 (self.tail[0] == c'-' and self.tail[1] == c'0')):
 *             if i == 1:
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":140
 *             else:
 *                 raise ValueError # leading zeroes are not allowed
 *         return i             # <<<<<<<<<<<<<<
 * 
 *     cdef object _decode_int(self):
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":124
 *             Py_LeaveRecursiveCall()
 * 
 *     cdef int _read_digits(self, char stop_char) except -1:             # <<<<<<<<<<<<<<
 *         cdef int i
 *         i = 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder._read_digits", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":142
 *         return i
 * 
 *     cdef object _decode_int(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         i = self._read_digits(c'e')
 */

static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_int(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  char const *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_decode_int", 0);

  /* "bzrlib/_bencode_pyx.pyx":144
 *     cdef object _decode_int(self):
 *         cdef int i
 *         i = self._read_digits(c'e')             # <<<<<<<<<<<<<<
 *         self.tail[i] = 0
 *         try:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_read_digits(__pyx_v_self, 'e'); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 144; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_i = __pyx_t_1;

  /* "bzrlib/_bencode_pyx.pyx":145
 *         cdef int i
 *         i = self._read_digits(c'e')
 *         self.tail[i] = 0             # <<<<<<<<<<<<<<
 *         try:
 *             ret = PyInt_FromString(self.tail, NULL, 10)
 */
  (__pyx_v_self->tail[__pyx_v_i]) = 0;

  /* "bzrlib/_bencode_pyx.pyx":146
 *         i = self._read_digits(c'e')
 *         self.tail[i] = 0
 *         try:             # <<<<<<<<<<<<<<
 *             ret = PyInt_FromString(self.tail, NULL, 10)
 *         finally:
 */
  /*try:*/ {

    /* "bzrlib/_bencode_pyx.pyx":147
 *         self.tail[i] = 0
 *         try:
 *             ret = PyInt_FromString(self.tail, NULL, 10)             # <<<<<<<<<<<<<<
 *         finally:
 *             self.tail[i] = c'e'
 */
    __pyx_t_2 = PyInt_FromString(__pyx_v_self->tail, NULL, 10); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 147; __pyx_clineno = __LINE__; goto __pyx_L4_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_ret = __pyx_t_2;
    __pyx_t_2 = 0;
  }

  /* "bzrlib/_bencode_pyx.pyx":149
 *             ret = PyInt_FromString(self.tail, NULL, 10)
 *         finally:
 *             self.tail[i] = c'e'             # <<<<<<<<<<<<<<
 *         D_UPDATE_TAIL(self, i+1)
 *         return ret
 */
  /*finally:*/ {
    /*normal exit:*/{
      (__pyx_v_self->tail[__pyx_v_i]) = 'e';
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __pyx_L4_error:;
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __pyx_t_1 = __pyx_lineno; __pyx_t_3 = __pyx_clineno; __pyx_t_4 = __pyx_filename;
      {
        (__pyx_v_self->tail[__pyx_v_i]) = 'e';
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      }
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0;
      __pyx_lineno = __pyx_t_1; __pyx_clineno = __pyx_t_3; __pyx_filename = __pyx_t_4;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "bzrlib/_bencode_pyx.pyx":150
 *         finally:
 *             self.tail[i] = c'e'
 *         D_UPDATE_TAIL(self, i+1)             # <<<<<<<<<<<<<<
 *         return ret
 * 
 */
  D_UPDATE_TAIL(__pyx_v_self, (__pyx_v_i + 1));

  /* "bzrlib/_bencode_pyx.pyx":151
 *             self.tail[i] = c'e'
 *         D_UPDATE_TAIL(self, i+1)
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     cdef object _decode_string(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":142
 *         return i
 * 
 *     cdef object _decode_int(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         i = self._read_digits(c'e')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder._decode_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":153
 *         return ret
 * 
 *     cdef object _decode_string(self):             # <<<<<<<<<<<<<<
 *         cdef int n
 *         cdef char *next_tail
 */

static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_string(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  int __pyx_v_n;
  char *__pyx_v_next_tail;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_decode_string", 0);

  /* "bzrlib/_bencode_pyx.pyx":160
 *         # or they wouldn't have called _decode_string
 *         # strtol will stop at trailing whitespace, etc
 *         n = strtol(self.tail, &next_tail, 10)             # <<<<<<<<<<<<<<
 *         if next_tail == NULL or next_tail[0] != c':':
 *             raise ValueError('string len not terminated by ":"')
 */
  __pyx_v_n = strtol(__pyx_v_self->tail, (&__pyx_v_next_tail), 10);

  /* "bzrlib/_bencode_pyx.pyx":161
 *         # strtol will stop at trailing whitespace, etc
 *         n = strtol(self.tail, &next_tail, 10)
 *         if next_tail == NULL or next_tail[0] != c':':             # <<<<<<<<<<<<<<
 *             raise ValueError('string len not terminated by ":"')
 *         # strtol allows leading zeros, so validate that we don't have that
 */
  __pyx_t_2 = ((__pyx_v_next_tail == NULL) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_next_tail[0]) != ':') != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":162
 *         n = strtol(self.tail, &next_tail, 10)
 *         if next_tail == NULL or next_tail[0] != c':':
 *             raise ValueError('string len not terminated by ":"')             # <<<<<<<<<<<<<<
 *         # strtol allows leading zeros, so validate that we don't have that
 *         if (self.tail[0] == c'0'
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 162; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 162; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":161
 *         # strtol will stop at trailing whitespace, etc
 *         n = strtol(self.tail, &next_tail, 10)
 *         if next_tail == NULL or next_tail[0] != c':':             # <<<<<<<<<<<<<<
 *             raise ValueError('string len not terminated by ":"')
 *         # strtol allows leading zeros, so validate that we don't have that
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":164
 *             raise ValueError('string len not terminated by ":"')
 *         # strtol allows leading zeros, so validate that we don't have that
 *         if (self.tail[0] == c'0'             # <<<<<<<<<<<<<<
 *             and (n != 0 or (next_tail - self.tail != 1))):
 *             raise ValueError('leading zeros are not allowed')
 */
  __pyx_t_2 = (((__pyx_v_self->tail[0]) == '0') != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "bzrlib/_bencode_pyx.pyx":165
 *         # strtol allows leading zeros, so validate that we don't have that
 *         if (self.tail[0] == c'0'
 *             and (n != 0 or (next_tail - self.tail != 1))):             # <<<<<<<<<<<<<<
 *             raise ValueError('leading zeros are not allowed')
 *         D_UPDATE_TAIL(self, next_tail - self.tail + 1)
 */
  __pyx_t_2 = ((__pyx_v_n != 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_next_tail - __pyx_v_self->tail) != 1) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "bzrlib/_bencode_pyx.pyx":164
 *             raise ValueError('string len not terminated by ":"')
 *         # strtol allows leading zeros, so validate that we don't have that
 *         if (self.tail[0] == c'0'             # <<<<<<<<<<<<<<
 *             and (n != 0 or (next_tail - self.tail != 1))):
 *             raise ValueError('leading zeros are not allowed')
 */
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":166
 *         if (self.tail[0] == c'0'
 *             and (n != 0 or (next_tail - self.tail != 1))):
 *             raise ValueError('leading zeros are not allowed')             # <<<<<<<<<<<<<<
 *         D_UPDATE_TAIL(self, next_tail - self.tail + 1)
 *         if n == 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":164
 *             raise ValueError('string len not terminated by ":"')
 *         # strtol allows leading zeros, so validate that we don't have that
 *         if (self.tail[0] == c'0'             # <<<<<<<<<<<<<<
 *             and (n != 0 or (next_tail - self.tail != 1))):
 *             raise ValueError('leading zeros are not allowed')
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":167
 *             and (n != 0 or (next_tail - self.tail != 1))):
 *             raise ValueError('leading zeros are not allowed')
 *         D_UPDATE_TAIL(self, next_tail - self.tail + 1)             # <<<<<<<<<<<<<<
 *         if n == 0:
 *             return ''
 */
  D_UPDATE_TAIL(__pyx_v_self, ((__pyx_v_next_tail - __pyx_v_self->tail) + 1));

  /* "bzrlib/_bencode_pyx.pyx":168
 *             raise ValueError('leading zeros are not allowed')
 *         D_UPDATE_TAIL(self, next_tail - self.tail + 1)
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return ''
 *         if n > self.size:
 */
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":169
 *         D_UPDATE_TAIL(self, next_tail - self.tail + 1)
 *         if n == 0:
 *             return ''             # <<<<<<<<<<<<<<
 *         if n > self.size:
 *             raise ValueError('stream underflow')
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_s__7);
    __pyx_r = __pyx_kp_s__7;
    goto __pyx_L0;

    /* "bzrlib/_bencode_pyx.pyx":168
 *             raise ValueError('leading zeros are not allowed')
 *         D_UPDATE_TAIL(self, next_tail - self.tail + 1)
 *         if n == 0:             # <<<<<<<<<<<<<<
 *             return ''
 *         if n > self.size:
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":170
 *         if n == 0:
 *             return ''
 *         if n > self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError('stream underflow')
 *         if n < 0:
 */
  __pyx_t_1 = ((__pyx_v_n > __pyx_v_self->size) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":171
 *             return ''
 *         if n > self.size:
 *             raise ValueError('stream underflow')             # <<<<<<<<<<<<<<
 *         if n < 0:
 *             raise ValueError('string size below zero: %d' % n)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":170
 *         if n == 0:
 *             return ''
 *         if n > self.size:             # <<<<<<<<<<<<<<
 *             raise ValueError('stream underflow')
 *         if n < 0:
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":172
 *         if n > self.size:
 *             raise ValueError('stream underflow')
 *         if n < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('string size below zero: %d' % n)
 * 
 */
  __pyx_t_1 = ((__pyx_v_n < 0) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":173
 *             raise ValueError('stream underflow')
 *         if n < 0:
 *             raise ValueError('string size below zero: %d' % n)             # <<<<<<<<<<<<<<
 * 
 *         result = PyString_FromStringAndSize(self.tail, n)
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_n); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_string_size_below_zero_d, __pyx_t_3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 173; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":172
 *         if n > self.size:
 *             raise ValueError('stream underflow')
 *         if n < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError('string size below zero: %d' % n)
 * 
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":175
 *             raise ValueError('string size below zero: %d' % n)
 * 
 *         result = PyString_FromStringAndSize(self.tail, n)             # <<<<<<<<<<<<<<
 *         D_UPDATE_TAIL(self, n)
 *         return result
 */
  __pyx_t_4 = PyString_FromStringAndSize(__pyx_v_self->tail, __pyx_v_n); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 175; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "bzrlib/_bencode_pyx.pyx":176
 * 
 *         result = PyString_FromStringAndSize(self.tail, n)
 *         D_UPDATE_TAIL(self, n)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  D_UPDATE_TAIL(__pyx_v_self, __pyx_v_n);

  /* "bzrlib/_bencode_pyx.pyx":177
 *         result = PyString_FromStringAndSize(self.tail, n)
 *         D_UPDATE_TAIL(self, n)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cdef object _decode_list(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":153
 *         return ret
 * 
 *     cdef object _decode_string(self):             # <<<<<<<<<<<<<<
 *         cdef int n
 *         cdef char *next_tail
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder._decode_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":179
 *         return result
 * 
 *     cdef object _decode_list(self):             # <<<<<<<<<<<<<<
 *         result = []
 * 
 */

static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_list(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_decode_list", 0);

  /* "bzrlib/_bencode_pyx.pyx":180
 * 
 *     cdef object _decode_list(self):
 *         result = []             # <<<<<<<<<<<<<<
 * 
 *         while self.size > 0:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 180; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_bencode_pyx.pyx":182
 *         result = []
 * 
 *         while self.size > 0:             # <<<<<<<<<<<<<<
 *             if self.tail[0] == c'e':
 *                 D_UPDATE_TAIL(self, 1)
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_self->size > 0) != 0);
    if (!__pyx_t_2) break;

    /* "bzrlib/_bencode_pyx.pyx":183
 * 
 *         while self.size > 0:
 *             if self.tail[0] == c'e':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 if self._yield_tuples:
 */
    __pyx_t_2 = (((__pyx_v_self->tail[0]) == 'e') != 0);
    if (__pyx_t_2) {

      /* "bzrlib/_bencode_pyx.pyx":184
 *         while self.size > 0:
 *             if self.tail[0] == c'e':
 *                 D_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 *                 if self._yield_tuples:
 *                     return tuple(result)
 */
      D_UPDATE_TAIL(__pyx_v_self, 1);

      /* "bzrlib/_bencode_pyx.pyx":185
 *             if self.tail[0] == c'e':
 *                 D_UPDATE_TAIL(self, 1)
 *                 if self._yield_tuples:             # <<<<<<<<<<<<<<
 *                     return tuple(result)
 *                 else:
 */
      __pyx_t_2 = (__pyx_v_self->_yield_tuples != 0);
      if (__pyx_t_2) {

        /* "bzrlib/_bencode_pyx.pyx":186
 *                 D_UPDATE_TAIL(self, 1)
 *                 if self._yield_tuples:
 *                     return tuple(result)             # <<<<<<<<<<<<<<
 *                 else:
 *                     return result
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 186; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L0;

        /* "bzrlib/_bencode_pyx.pyx":185
 *             if self.tail[0] == c'e':
 *                 D_UPDATE_TAIL(self, 1)
 *                 if self._yield_tuples:             # <<<<<<<<<<<<<<
 *                     return tuple(result)
 *                 else:
 */
      }

      /* "bzrlib/_bencode_pyx.pyx":188
 *                     return tuple(result)
 *                 else:
 *                     return result             # <<<<<<<<<<<<<<
 *             else:
 *                 # As a quick shortcut, check to see if the next object is a
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_result);
        __pyx_r = __pyx_v_result;
        goto __pyx_L0;
      }

      /* "bzrlib/_bencode_pyx.pyx":183
 * 
 *         while self.size > 0:
 *             if self.tail[0] == c'e':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 if self._yield_tuples:
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":193
 *                 # string, since we know that won't be creating recursion
 *                 # if self.tail[0] >= c'0' and self.tail[0] <= c'9':
 *                 PyList_Append(result, self._decode_object())             # <<<<<<<<<<<<<<
 * 
 *         raise ValueError('malformed list')
 */
    /*else*/ {
      __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 193; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_3 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 193; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }

  /* "bzrlib/_bencode_pyx.pyx":195
 *                 PyList_Append(result, self._decode_object())
 * 
 *         raise ValueError('malformed list')             # <<<<<<<<<<<<<<
 * 
 *     cdef object _decode_dict(self):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":179
 *         return result
 * 
 *     cdef object _decode_list(self):             # <<<<<<<<<<<<<<
 *         result = []
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder._decode_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":197
 *         raise ValueError('malformed list')
 * 
 *     cdef object _decode_dict(self):             # <<<<<<<<<<<<<<
 *         cdef char ch
 * 
 */

static PyObject *__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_dict(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  char __pyx_v_ch;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_lastkey = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_decode_dict", 0);

  /* "bzrlib/_bencode_pyx.pyx":200
 *         cdef char ch
 * 
 *         result = {}             # <<<<<<<<<<<<<<
 *         lastkey = None
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 200; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_bencode_pyx.pyx":201
 * 
 *         result = {}
 *         lastkey = None             # <<<<<<<<<<<<<<
 * 
 *         while self.size > 0:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_lastkey = Py_None;

  /* "bzrlib/_bencode_pyx.pyx":203
 *         lastkey = None
 * 
 *         while self.size > 0:             # <<<<<<<<<<<<<<
 *             ch = self.tail[0]
 *             if ch == c'e':
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_self->size > 0) != 0);
    if (!__pyx_t_2) break;

    /* "bzrlib/_bencode_pyx.pyx":204
 * 
 *         while self.size > 0:
 *             ch = self.tail[0]             # <<<<<<<<<<<<<<
 *             if ch == c'e':
 *                 D_UPDATE_TAIL(self, 1)
 */
    __pyx_v_ch = (__pyx_v_self->tail[0]);

    /* "bzrlib/_bencode_pyx.pyx":205
 *         while self.size > 0:
 *             ch = self.tail[0]
 *             if ch == c'e':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return result
 */
    __pyx_t_2 = ((__pyx_v_ch == 'e') != 0);
    if (__pyx_t_2) {

      /* "bzrlib/_bencode_pyx.pyx":206
 *             ch = self.tail[0]
 *             if ch == c'e':
 *                 D_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 *                 return result
 *             else:
 */
      D_UPDATE_TAIL(__pyx_v_self, 1);

      /* "bzrlib/_bencode_pyx.pyx":207
 *             if ch == c'e':
 *                 D_UPDATE_TAIL(self, 1)
 *                 return result             # <<<<<<<<<<<<<<
 *             else:
 *                 # keys should be strings only
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L0;

      /* "bzrlib/_bencode_pyx.pyx":205
 *         while self.size > 0:
 *             ch = self.tail[0]
 *             if ch == c'e':             # <<<<<<<<<<<<<<
 *                 D_UPDATE_TAIL(self, 1)
 *                 return result
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":210
 *             else:
 *                 # keys should be strings only
 *                 if self.tail[0] < c'0' or self.tail[0] > c'9':             # <<<<<<<<<<<<<<
 *                     raise ValueError('key was not a simple string.')
 *                 key = self._decode_string()
 */
    /*else*/ {
      __pyx_t_3 = (((__pyx_v_self->tail[0]) < '0') != 0);
      if (!__pyx_t_3) {
      } else {
        __pyx_t_2 = __pyx_t_3;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_3 = (((__pyx_v_self->tail[0]) > '9') != 0);
      __pyx_t_2 = __pyx_t_3;
      __pyx_L7_bool_binop_done:;
      if (__pyx_t_2) {

        /* "bzrlib/_bencode_pyx.pyx":211
 *                 # keys should be strings only
 *                 if self.tail[0] < c'0' or self.tail[0] > c'9':
 *                     raise ValueError('key was not a simple string.')             # <<<<<<<<<<<<<<
 *                 key = self._decode_string()
 *                 if lastkey >= key:
 */
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

        /* "bzrlib/_bencode_pyx.pyx":210
 *             else:
 *                 # keys should be strings only
 *                 if self.tail[0] < c'0' or self.tail[0] > c'9':             # <<<<<<<<<<<<<<
 *                     raise ValueError('key was not a simple string.')
 *                 key = self._decode_string()
 */
      }

      /* "bzrlib/_bencode_pyx.pyx":212
 *                 if self.tail[0] < c'0' or self.tail[0] > c'9':
 *                     raise ValueError('key was not a simple string.')
 *                 key = self._decode_string()             # <<<<<<<<<<<<<<
 *                 if lastkey >= key:
 *                     raise ValueError('dict keys disordered')
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_string(__pyx_v_self); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 212; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "bzrlib/_bencode_pyx.pyx":213
 *                     raise ValueError('key was not a simple string.')
 *                 key = self._decode_string()
 *                 if lastkey >= key:             # <<<<<<<<<<<<<<
 *                     raise ValueError('dict keys disordered')
 *                 else:
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_lastkey, __pyx_v_key, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 213; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "bzrlib/_bencode_pyx.pyx":214
 *                 key = self._decode_string()
 *                 if lastkey >= key:
 *                     raise ValueError('dict keys disordered')             # <<<<<<<<<<<<<<
 *                 else:
 *                     lastkey = key
 */
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_Raise(__pyx_t_1, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        {__pyx_filename = __pyx_f[0]; __pyx_lineno = 214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

        /* "bzrlib/_bencode_pyx.pyx":213
 *                     raise ValueError('key was not a simple string.')
 *                 key = self._decode_string()
 *                 if lastkey >= key:             # <<<<<<<<<<<<<<
 *                     raise ValueError('dict keys disordered')
 *                 else:
 */
      }

      /* "bzrlib/_bencode_pyx.pyx":216
 *                     raise ValueError('dict keys disordered')
 *                 else:
 *                     lastkey = key             # <<<<<<<<<<<<<<
 *                 value = self._decode_object()
 *                 result[key] = value
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_v_key);
        __Pyx_DECREF_SET(__pyx_v_lastkey, __pyx_v_key);
      }

      /* "bzrlib/_bencode_pyx.pyx":217
 *                 else:
 *                     lastkey = key
 *                 value = self._decode_object()             # <<<<<<<<<<<<<<
 *                 result[key] = value
 * 
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self->__pyx_vtab)->_decode_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 217; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "bzrlib/_bencode_pyx.pyx":218
 *                     lastkey = key
 *                 value = self._decode_object()
 *                 result[key] = value             # <<<<<<<<<<<<<<
 * 
 *         raise ValueError('malformed dict')
 */
      if (unlikely(PyDict_SetItem(__pyx_v_result, __pyx_v_key, __pyx_v_value) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 218; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }

  /* "bzrlib/_bencode_pyx.pyx":220
 *                 result[key] = value
 * 
 *         raise ValueError('malformed dict')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 220; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  {__pyx_filename = __pyx_f[0]; __pyx_lineno = 220; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":197
 *         raise ValueError('malformed list')
 * 
 *     cdef object _decode_dict(self):             # <<<<<<<<<<<<<<
 *         cdef char ch
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder._decode_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_lastkey);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":72
 *     """Bencode decoder"""
 * 
 *     cdef readonly char *tail             # <<<<<<<<<<<<<<
 *     cdef readonly int size
 *     cdef readonly int _yield_tuples
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_4tail_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_4tail_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4tail___get__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4tail___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->tail); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 72; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder.tail.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":73
 * 
 *     cdef readonly char *tail
 *     cdef readonly int size             # <<<<<<<<<<<<<<
 *     cdef readonly int _yield_tuples
 *     cdef object text
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4size___get__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_4size___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 73; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":74
 *     cdef readonly char *tail
 *     cdef readonly int size
 *     cdef readonly int _yield_tuples             # <<<<<<<<<<<<<<
 *     cdef object text
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_13_yield_tuples_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_13_yield_tuples_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_13_yield_tuples___get__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Decoder_13_yield_tuples___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_yield_tuples); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 74; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Decoder._yield_tuples.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":223
 * 
 * 
 * def bdecode(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object"""
 *     return Decoder(s).decode()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_1bdecode(PyObject *__pyx_self, PyObject *__pyx_v_s); /*proto*/
static char __pyx_doc_6bzrlib_12_bencode_pyx_bdecode[] = "Decode string x to Python object";
static PyMethodDef __pyx_mdef_6bzrlib_12_bencode_pyx_1bdecode = {"bdecode", (PyCFunction)__pyx_pw_6bzrlib_12_bencode_pyx_1bdecode, METH_O, __pyx_doc_6bzrlib_12_bencode_pyx_bdecode};
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_1bdecode(PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bdecode (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_bdecode(__pyx_self, ((PyObject *)__pyx_v_s));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_bdecode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bdecode", 0);

  /* "bzrlib/_bencode_pyx.pyx":225
 * def bdecode(object s):
 *     """Decode string x to Python object"""
 *     return Decoder(s).decode()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_s);
  __Pyx_GIVEREF(__pyx_v_s);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_s);
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_12_bencode_pyx_Decoder), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 225; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":223
 * 
 * 
 * def bdecode(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object"""
 *     return Decoder(s).decode()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.bdecode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":228
 * 
 * 
 * def bdecode_as_tuple(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object, using tuples rather than lists."""
 *     return Decoder(s, True).decode()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_3bdecode_as_tuple(PyObject *__pyx_self, PyObject *__pyx_v_s); /*proto*/
static char __pyx_doc_6bzrlib_12_bencode_pyx_2bdecode_as_tuple[] = "Decode string x to Python object, using tuples rather than lists.";
static PyMethodDef __pyx_mdef_6bzrlib_12_bencode_pyx_3bdecode_as_tuple = {"bdecode_as_tuple", (PyCFunction)__pyx_pw_6bzrlib_12_bencode_pyx_3bdecode_as_tuple, METH_O, __pyx_doc_6bzrlib_12_bencode_pyx_2bdecode_as_tuple};
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_3bdecode_as_tuple(PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bdecode_as_tuple (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_2bdecode_as_tuple(__pyx_self, ((PyObject *)__pyx_v_s));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_2bdecode_as_tuple(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bdecode_as_tuple", 0);

  /* "bzrlib/_bencode_pyx.pyx":230
 * def bdecode_as_tuple(object s):
 *     """Decode string x to Python object, using tuples rather than lists."""
 *     return Decoder(s, True).decode()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_s);
  __Pyx_GIVEREF(__pyx_v_s);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_s);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_2, 1, Py_True);
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_12_bencode_pyx_Decoder), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 230; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":228
 * 
 * 
 * def bdecode_as_tuple(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object, using tuples rather than lists."""
 *     return Decoder(s, True).decode()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.bdecode_as_tuple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":236
 *     __slots__ = ['bencoded']
 * 
 *     def __init__(self, s):             # <<<<<<<<<<<<<<
 *         self.bencoded = s
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_9Bencached_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6bzrlib_12_bencode_pyx_9Bencached_1__init__ = {"__init__", (PyCFunction)__pyx_pw_6bzrlib_12_bencode_pyx_9Bencached_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_9Bencached_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_s = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_s,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_s = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Bencached.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_9Bencached___init__(__pyx_self, __pyx_v_self, __pyx_v_s);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_9Bencached___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "bzrlib/_bencode_pyx.pyx":237
 * 
 *     def __init__(self, s):
 *         self.bencoded = s             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_bencoded, __pyx_v_s) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 237; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":236
 *     __slots__ = ['bencoded']
 * 
 *     def __init__(self, s):             # <<<<<<<<<<<<<<
 *         self.bencoded = s
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Bencached.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":253
 *     cdef readonly int maxsize
 * 
 *     def __init__(self, int maxsize=INITSIZE):             # <<<<<<<<<<<<<<
 *         """Initialize encoder engine
 *         @param  maxsize:    initial size of internal char buffer
 */

/* Python wrapper */
static int __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6bzrlib_12_bencode_pyx_7Encoder___init__[] = "Initialize encoder engine\n        @param  maxsize:    initial size of internal char buffer\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Encoder___init__;
#endif
static int __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_maxsize;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_maxsize,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_maxsize);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_maxsize = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_maxsize == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
    } else {
      __pyx_v_maxsize = __pyx_k__13;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); {__pyx_filename = __pyx_f[0]; __pyx_lineno = 253; __pyx_clineno = __LINE__; goto __pyx_L3_error;}
  __pyx_L3_error:;
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder___init__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self), __pyx_v_maxsize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder___init__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, int __pyx_v_maxsize) {
  char *__pyx_v_p;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "bzrlib/_bencode_pyx.pyx":259
 *         cdef char *p
 * 
 *         self.maxsize = 0             # <<<<<<<<<<<<<<
 *         self.size = 0
 *         self.tail = NULL
 */
  __pyx_v_self->maxsize = 0;

  /* "bzrlib/_bencode_pyx.pyx":260
 * 
 *         self.maxsize = 0
 *         self.size = 0             # <<<<<<<<<<<<<<
 *         self.tail = NULL
 * 
 */
  __pyx_v_self->size = 0;

  /* "bzrlib/_bencode_pyx.pyx":261
 *         self.maxsize = 0
 *         self.size = 0
 *         self.tail = NULL             # <<<<<<<<<<<<<<
 * 
 *         p = <char*>malloc(maxsize)
 */
  __pyx_v_self->tail = NULL;

  /* "bzrlib/_bencode_pyx.pyx":263
 *         self.tail = NULL
 * 
 *         p = <char*>malloc(maxsize)             # <<<<<<<<<<<<<<
 *         if p == NULL:
 *             raise MemoryError('Not enough memory to allocate buffer '
 */
  __pyx_v_p = ((char *)malloc(__pyx_v_maxsize));

  /* "bzrlib/_bencode_pyx.pyx":264
 * 
 *         p = <char*>malloc(maxsize)
 *         if p == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError('Not enough memory to allocate buffer '
 *                               'for encoder')
 */
  __pyx_t_1 = ((__pyx_v_p == NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":265
 *         p = <char*>malloc(maxsize)
 *         if p == NULL:
 *             raise MemoryError('Not enough memory to allocate buffer '             # <<<<<<<<<<<<<<
 *                               'for encoder')
 *         self.buffer = p
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":264
 * 
 *         p = <char*>malloc(maxsize)
 *         if p == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError('Not enough memory to allocate buffer '
 *                               'for encoder')
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":267
 *             raise MemoryError('Not enough memory to allocate buffer '
 *                               'for encoder')
 *         self.buffer = p             # <<<<<<<<<<<<<<
 *         self.maxsize = maxsize
 *         self.tail = p
 */
  __pyx_v_self->buffer = __pyx_v_p;

  /* "bzrlib/_bencode_pyx.pyx":268
 *                               'for encoder')
 *         self.buffer = p
 *         self.maxsize = maxsize             # <<<<<<<<<<<<<<
 *         self.tail = p
 * 
 */
  __pyx_v_self->maxsize = __pyx_v_maxsize;

  /* "bzrlib/_bencode_pyx.pyx":269
 *         self.buffer = p
 *         self.maxsize = maxsize
 *         self.tail = p             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->tail = __pyx_v_p;

  /* "bzrlib/_bencode_pyx.pyx":253
 *     cdef readonly int maxsize
 * 
 *     def __init__(self, int maxsize=INITSIZE):             # <<<<<<<<<<<<<<
 *         """Initialize encoder engine
 *         @param  maxsize:    initial size of internal char buffer
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":271
 *         self.tail = p
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.buffer)
 *         self.buffer = NULL
 */

/* Python wrapper */
static void __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_2__dealloc__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_2__dealloc__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "bzrlib/_bencode_pyx.pyx":272
 * 
 *     def __dealloc__(self):
 *         free(self.buffer)             # <<<<<<<<<<<<<<
 *         self.buffer = NULL
 *         self.maxsize = 0
 */
  free(__pyx_v_self->buffer);

  /* "bzrlib/_bencode_pyx.pyx":273
 *     def __dealloc__(self):
 *         free(self.buffer)
 *         self.buffer = NULL             # <<<<<<<<<<<<<<
 *         self.maxsize = 0
 * 
 */
  __pyx_v_self->buffer = NULL;

  /* "bzrlib/_bencode_pyx.pyx":274
 *         free(self.buffer)
 *         self.buffer = NULL
 *         self.maxsize = 0             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_v_self->maxsize = 0;

  /* "bzrlib/_bencode_pyx.pyx":271
 *         self.tail = p
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free(self.buffer)
 *         self.buffer = NULL
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "bzrlib/_bencode_pyx.pyx":276
 *         self.maxsize = 0
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.buffer != NULL and self.size != 0:
 *             return PyString_FromStringAndSize(self.buffer, self.size)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_5__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_5__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4__str__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4__str__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "bzrlib/_bencode_pyx.pyx":277
 * 
 *     def __str__(self):
 *         if self.buffer != NULL and self.size != 0:             # <<<<<<<<<<<<<<
 *             return PyString_FromStringAndSize(self.buffer, self.size)
 *         else:
 */
  __pyx_t_2 = ((__pyx_v_self->buffer != NULL) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->size != 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":278
 *     def __str__(self):
 *         if self.buffer != NULL and self.size != 0:
 *             return PyString_FromStringAndSize(self.buffer, self.size)             # <<<<<<<<<<<<<<
 *         else:
 *             return ''
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyString_FromStringAndSize(__pyx_v_self->buffer, __pyx_v_self->size); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 278; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "bzrlib/_bencode_pyx.pyx":277
 * 
 *     def __str__(self):
 *         if self.buffer != NULL and self.size != 0:             # <<<<<<<<<<<<<<
 *             return PyString_FromStringAndSize(self.buffer, self.size)
 *         else:
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":280
 *             return PyString_FromStringAndSize(self.buffer, self.size)
 *         else:
 *             return ''             # <<<<<<<<<<<<<<
 * 
 *     cdef int _ensure_buffer(self, int required) except 0:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_s__7);
    __pyx_r = __pyx_kp_s__7;
    goto __pyx_L0;
  }

  /* "bzrlib/_bencode_pyx.pyx":276
 *         self.maxsize = 0
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self.buffer != NULL and self.size != 0:
 *             return PyString_FromStringAndSize(self.buffer, self.size)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":282
 *             return ''
 * 
 *     cdef int _ensure_buffer(self, int required) except 0:             # <<<<<<<<<<<<<<
 *         """Ensure that tail of CharTail buffer has enough size.
 *         If buffer is not big enough then function try to
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__ensure_buffer(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, int __pyx_v_required) {
  char *__pyx_v_new_buffer;
  int __pyx_v_new_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ensure_buffer", 0);

  /* "bzrlib/_bencode_pyx.pyx":290
 *         cdef int   new_size
 * 
 *         if self.size + required < self.maxsize:             # <<<<<<<<<<<<<<
 *             return 1
 * 
 */
  __pyx_t_1 = (((__pyx_v_self->size + __pyx_v_required) < __pyx_v_self->maxsize) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":291
 * 
 *         if self.size + required < self.maxsize:
 *             return 1             # <<<<<<<<<<<<<<
 * 
 *         new_size = self.maxsize
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "bzrlib/_bencode_pyx.pyx":290
 *         cdef int   new_size
 * 
 *         if self.size + required < self.maxsize:             # <<<<<<<<<<<<<<
 *             return 1
 * 
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":293
 *             return 1
 * 
 *         new_size = self.maxsize             # <<<<<<<<<<<<<<
 *         while new_size < self.size + required:
 *             new_size = new_size * 2
 */
  __pyx_t_2 = __pyx_v_self->maxsize;
  __pyx_v_new_size = __pyx_t_2;

  /* "bzrlib/_bencode_pyx.pyx":294
 * 
 *         new_size = self.maxsize
 *         while new_size < self.size + required:             # <<<<<<<<<<<<<<
 *             new_size = new_size * 2
 *         new_buffer = <char*>realloc(self.buffer, <size_t>new_size)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_new_size < (__pyx_v_self->size + __pyx_v_required)) != 0);
    if (!__pyx_t_1) break;

    /* "bzrlib/_bencode_pyx.pyx":295
 *         new_size = self.maxsize
 *         while new_size < self.size + required:
 *             new_size = new_size * 2             # <<<<<<<<<<<<<<
 *         new_buffer = <char*>realloc(self.buffer, <size_t>new_size)
 *         if new_buffer == NULL:
 */
    __pyx_v_new_size = (__pyx_v_new_size * 2);
  }

  /* "bzrlib/_bencode_pyx.pyx":296
 *         while new_size < self.size + required:
 *             new_size = new_size * 2
 *         new_buffer = <char*>realloc(self.buffer, <size_t>new_size)             # <<<<<<<<<<<<<<
 *         if new_buffer == NULL:
 *             raise MemoryError('Cannot realloc buffer for encoder')
 */
  __pyx_v_new_buffer = ((char *)realloc(__pyx_v_self->buffer, ((size_t)__pyx_v_new_size)));

  /* "bzrlib/_bencode_pyx.pyx":297
 *             new_size = new_size * 2
 *         new_buffer = <char*>realloc(self.buffer, <size_t>new_size)
 *         if new_buffer == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError('Cannot realloc buffer for encoder')
 * 
 */
  __pyx_t_1 = ((__pyx_v_new_buffer == NULL) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":298
 *         new_buffer = <char*>realloc(self.buffer, <size_t>new_size)
 *         if new_buffer == NULL:
 *             raise MemoryError('Cannot realloc buffer for encoder')             # <<<<<<<<<<<<<<
 * 
 *         self.buffer = new_buffer
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":297
 *             new_size = new_size * 2
 *         new_buffer = <char*>realloc(self.buffer, <size_t>new_size)
 *         if new_buffer == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError('Cannot realloc buffer for encoder')
 * 
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":300
 *             raise MemoryError('Cannot realloc buffer for encoder')
 * 
 *         self.buffer = new_buffer             # <<<<<<<<<<<<<<
 *         self.maxsize = new_size
 *         self.tail = &new_buffer[self.size]
 */
  __pyx_v_self->buffer = __pyx_v_new_buffer;

  /* "bzrlib/_bencode_pyx.pyx":301
 * 
 *         self.buffer = new_buffer
 *         self.maxsize = new_size             # <<<<<<<<<<<<<<
 *         self.tail = &new_buffer[self.size]
 *         return 1
 */
  __pyx_v_self->maxsize = __pyx_v_new_size;

  /* "bzrlib/_bencode_pyx.pyx":302
 *         self.buffer = new_buffer
 *         self.maxsize = new_size
 *         self.tail = &new_buffer[self.size]             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  __pyx_v_self->tail = (&(__pyx_v_new_buffer[__pyx_v_self->size]));

  /* "bzrlib/_bencode_pyx.pyx":303
 *         self.maxsize = new_size
 *         self.tail = &new_buffer[self.size]
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int _encode_int(self, int x) except 0:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":282
 *             return ''
 * 
 *     cdef int _ensure_buffer(self, int required) except 0:             # <<<<<<<<<<<<<<
 *         """Ensure that tail of CharTail buffer has enough size.
 *         If buffer is not big enough then function try to
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder._ensure_buffer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":305
 *         return 1
 * 
 *     cdef int _encode_int(self, int x) except 0:             # <<<<<<<<<<<<<<
 *         """Encode int to bencode string iNNNe
 *         @param  x:  value to encode
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_int(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, int __pyx_v_x) {
  int __pyx_v_n;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_encode_int", 0);

  /* "bzrlib/_bencode_pyx.pyx":310
 *         """
 *         cdef int n
 *         self._ensure_buffer(INT_BUF_SIZE)             # <<<<<<<<<<<<<<
 *         n = snprintf(self.tail, INT_BUF_SIZE, "i%de", x)
 *         if n < 0:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_ensure_buffer(__pyx_v_self, __pyx_e_6bzrlib_12_bencode_pyx_INT_BUF_SIZE); if (unlikely(__pyx_t_1 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 310; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":311
 *         cdef int n
 *         self._ensure_buffer(INT_BUF_SIZE)
 *         n = snprintf(self.tail, INT_BUF_SIZE, "i%de", x)             # <<<<<<<<<<<<<<
 *         if n < 0:
 *             raise MemoryError('int %d too big to encode' % x)
 */
  __pyx_v_n = snprintf(__pyx_v_self->tail, __pyx_e_6bzrlib_12_bencode_pyx_INT_BUF_SIZE, __pyx_k_i_de, __pyx_v_x);

  /* "bzrlib/_bencode_pyx.pyx":312
 *         self._ensure_buffer(INT_BUF_SIZE)
 *         n = snprintf(self.tail, INT_BUF_SIZE, "i%de", x)
 *         if n < 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError('int %d too big to encode' % x)
 *         E_UPDATE_TAIL(self, n)
 */
  __pyx_t_2 = ((__pyx_v_n < 0) != 0);
  if (__pyx_t_2) {

    /* "bzrlib/_bencode_pyx.pyx":313
 *         n = snprintf(self.tail, INT_BUF_SIZE, "i%de", x)
 *         if n < 0:
 *             raise MemoryError('int %d too big to encode' % x)             # <<<<<<<<<<<<<<
 *         E_UPDATE_TAIL(self, n)
 *         return 1
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_x); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 313; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_int_d_too_big_to_encode, __pyx_t_3); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 313; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 313; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 313; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 313; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":312
 *         self._ensure_buffer(INT_BUF_SIZE)
 *         n = snprintf(self.tail, INT_BUF_SIZE, "i%de", x)
 *         if n < 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError('int %d too big to encode' % x)
 *         E_UPDATE_TAIL(self, n)
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":314
 *         if n < 0:
 *             raise MemoryError('int %d too big to encode' % x)
 *         E_UPDATE_TAIL(self, n)             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  E_UPDATE_TAIL(__pyx_v_self, __pyx_v_n);

  /* "bzrlib/_bencode_pyx.pyx":315
 *             raise MemoryError('int %d too big to encode' % x)
 *         E_UPDATE_TAIL(self, n)
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int _encode_long(self, x) except 0:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":305
 *         return 1
 * 
 *     cdef int _encode_int(self, int x) except 0:             # <<<<<<<<<<<<<<
 *         """Encode int to bencode string iNNNe
 *         @param  x:  value to encode
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder._encode_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":317
 *         return 1
 * 
 *     cdef int _encode_long(self, x) except 0:             # <<<<<<<<<<<<<<
 *         return self._append_string(''.join(('i', str(x), 'e')))
 * 
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_long(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_encode_long", 0);

  /* "bzrlib/_bencode_pyx.pyx":318
 * 
 *     cdef int _encode_long(self, x) except 0:
 *         return self._append_string(''.join(('i', str(x), 'e')))             # <<<<<<<<<<<<<<
 * 
 *     cdef int _append_string(self, s) except 0:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 318; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_x);
  __Pyx_GIVEREF(__pyx_v_x);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_x);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 318; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 318; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_i);
  __Pyx_GIVEREF(__pyx_n_s_i);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_i);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_e);
  __Pyx_GIVEREF(__pyx_n_s_e);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_e);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__7, __pyx_t_1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 318; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_append_string(__pyx_v_self, __pyx_t_2); if (unlikely(__pyx_t_3 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 318; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":317
 *         return 1
 * 
 *     cdef int _encode_long(self, x) except 0:             # <<<<<<<<<<<<<<
 *         return self._append_string(''.join(('i', str(x), 'e')))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder._encode_long", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":320
 *         return self._append_string(''.join(('i', str(x), 'e')))
 * 
 *     cdef int _append_string(self, s) except 0:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t n
 *         n = PyString_GET_SIZE(s)
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__append_string(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_s) {
  Py_ssize_t __pyx_v_n;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  char *__pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_append_string", 0);

  /* "bzrlib/_bencode_pyx.pyx":322
 *     cdef int _append_string(self, s) except 0:
 *         cdef Py_ssize_t n
 *         n = PyString_GET_SIZE(s)             # <<<<<<<<<<<<<<
 *         self._ensure_buffer(n)
 *         memcpy(self.tail, PyString_AS_STRING(s), n)
 */
  __pyx_t_1 = PyString_GET_SIZE(__pyx_v_s); if (unlikely(__pyx_t_1 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 322; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_n = __pyx_t_1;

  /* "bzrlib/_bencode_pyx.pyx":323
 *         cdef Py_ssize_t n
 *         n = PyString_GET_SIZE(s)
 *         self._ensure_buffer(n)             # <<<<<<<<<<<<<<
 *         memcpy(self.tail, PyString_AS_STRING(s), n)
 *         E_UPDATE_TAIL(self, n)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_ensure_buffer(__pyx_v_self, __pyx_v_n); if (unlikely(__pyx_t_2 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 323; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":324
 *         n = PyString_GET_SIZE(s)
 *         self._ensure_buffer(n)
 *         memcpy(self.tail, PyString_AS_STRING(s), n)             # <<<<<<<<<<<<<<
 *         E_UPDATE_TAIL(self, n)
 *         return 1
 */
  __pyx_t_3 = PyString_AS_STRING(__pyx_v_s); if (unlikely(__pyx_t_3 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 324; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  memcpy(__pyx_v_self->tail, __pyx_t_3, __pyx_v_n);

  /* "bzrlib/_bencode_pyx.pyx":325
 *         self._ensure_buffer(n)
 *         memcpy(self.tail, PyString_AS_STRING(s), n)
 *         E_UPDATE_TAIL(self, n)             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  E_UPDATE_TAIL(__pyx_v_self, __pyx_v_n);

  /* "bzrlib/_bencode_pyx.pyx":326
 *         memcpy(self.tail, PyString_AS_STRING(s), n)
 *         E_UPDATE_TAIL(self, n)
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int _encode_string(self, x) except 0:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":320
 *         return self._append_string(''.join(('i', str(x), 'e')))
 * 
 *     cdef int _append_string(self, s) except 0:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t n
 *         n = PyString_GET_SIZE(s)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder._append_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":328
 *         return 1
 * 
 *     cdef int _encode_string(self, x) except 0:             # <<<<<<<<<<<<<<
 *         cdef int n
 *         cdef Py_ssize_t x_len
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_string(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x) {
  int __pyx_v_n;
  Py_ssize_t __pyx_v_x_len;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_encode_string", 0);

  /* "bzrlib/_bencode_pyx.pyx":331
 *         cdef int n
 *         cdef Py_ssize_t x_len
 *         x_len = PyString_GET_SIZE(x)             # <<<<<<<<<<<<<<
 *         self._ensure_buffer(x_len + INT_BUF_SIZE)
 *         n = snprintf(self.tail, INT_BUF_SIZE, '%d:', x_len)
 */
  __pyx_t_1 = PyString_GET_SIZE(__pyx_v_x); if (unlikely(__pyx_t_1 == -1L)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 331; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_v_x_len = __pyx_t_1;

  /* "bzrlib/_bencode_pyx.pyx":332
 *         cdef Py_ssize_t x_len
 *         x_len = PyString_GET_SIZE(x)
 *         self._ensure_buffer(x_len + INT_BUF_SIZE)             # <<<<<<<<<<<<<<
 *         n = snprintf(self.tail, INT_BUF_SIZE, '%d:', x_len)
 *         if n < 0:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_ensure_buffer(__pyx_v_self, (__pyx_v_x_len + __pyx_e_6bzrlib_12_bencode_pyx_INT_BUF_SIZE)); if (unlikely(__pyx_t_2 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 332; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":333
 *         x_len = PyString_GET_SIZE(x)
 *         self._ensure_buffer(x_len + INT_BUF_SIZE)
 *         n = snprintf(self.tail, INT_BUF_SIZE, '%d:', x_len)             # <<<<<<<<<<<<<<
 *         if n < 0:
 *             raise MemoryError('string %s too big to encode' % x)
 */
  __pyx_v_n = snprintf(__pyx_v_self->tail, __pyx_e_6bzrlib_12_bencode_pyx_INT_BUF_SIZE, __pyx_k_d, __pyx_v_x_len);

  /* "bzrlib/_bencode_pyx.pyx":334
 *         self._ensure_buffer(x_len + INT_BUF_SIZE)
 *         n = snprintf(self.tail, INT_BUF_SIZE, '%d:', x_len)
 *         if n < 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError('string %s too big to encode' % x)
 *         memcpy(<void *>(self.tail+n), PyString_AS_STRING(x), x_len)
 */
  __pyx_t_3 = ((__pyx_v_n < 0) != 0);
  if (__pyx_t_3) {

    /* "bzrlib/_bencode_pyx.pyx":335
 *         n = snprintf(self.tail, INT_BUF_SIZE, '%d:', x_len)
 *         if n < 0:
 *             raise MemoryError('string %s too big to encode' % x)             # <<<<<<<<<<<<<<
 *         memcpy(<void *>(self.tail+n), PyString_AS_STRING(x), x_len)
 *         E_UPDATE_TAIL(self, n + x_len)
 */
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_string_s_too_big_to_encode, __pyx_v_x); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 335; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":334
 *         self._ensure_buffer(x_len + INT_BUF_SIZE)
 *         n = snprintf(self.tail, INT_BUF_SIZE, '%d:', x_len)
 *         if n < 0:             # <<<<<<<<<<<<<<
 *             raise MemoryError('string %s too big to encode' % x)
 *         memcpy(<void *>(self.tail+n), PyString_AS_STRING(x), x_len)
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":336
 *         if n < 0:
 *             raise MemoryError('string %s too big to encode' % x)
 *         memcpy(<void *>(self.tail+n), PyString_AS_STRING(x), x_len)             # <<<<<<<<<<<<<<
 *         E_UPDATE_TAIL(self, n + x_len)
 *         return 1
 */
  __pyx_t_6 = PyString_AS_STRING(__pyx_v_x); if (unlikely(__pyx_t_6 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 336; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  memcpy(((void *)(__pyx_v_self->tail + __pyx_v_n)), __pyx_t_6, __pyx_v_x_len);

  /* "bzrlib/_bencode_pyx.pyx":337
 *             raise MemoryError('string %s too big to encode' % x)
 *         memcpy(<void *>(self.tail+n), PyString_AS_STRING(x), x_len)
 *         E_UPDATE_TAIL(self, n + x_len)             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  E_UPDATE_TAIL(__pyx_v_self, (__pyx_v_n + __pyx_v_x_len));

  /* "bzrlib/_bencode_pyx.pyx":338
 *         memcpy(<void *>(self.tail+n), PyString_AS_STRING(x), x_len)
 *         E_UPDATE_TAIL(self, n + x_len)
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int _encode_list(self, x) except 0:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":328
 *         return 1
 * 
 *     cdef int _encode_string(self, x) except 0:             # <<<<<<<<<<<<<<
 *         cdef int n
 *         cdef Py_ssize_t x_len
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder._encode_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":340
 *         return 1
 * 
 *     cdef int _encode_list(self, x) except 0:             # <<<<<<<<<<<<<<
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'l'
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_list(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_i = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_encode_list", 0);

  /* "bzrlib/_bencode_pyx.pyx":341
 * 
 *     cdef int _encode_list(self, x) except 0:
 *         self._ensure_buffer(1)             # <<<<<<<<<<<<<<
 *         self.tail[0] = c'l'
 *         E_UPDATE_TAIL(self, 1)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_ensure_buffer(__pyx_v_self, 1); if (unlikely(__pyx_t_1 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 341; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":342
 *     cdef int _encode_list(self, x) except 0:
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'l'             # <<<<<<<<<<<<<<
 *         E_UPDATE_TAIL(self, 1)
 * 
 */
  (__pyx_v_self->tail[0]) = 'l';

  /* "bzrlib/_bencode_pyx.pyx":343
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'l'
 *         E_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 * 
 *         for i in x:
 */
  E_UPDATE_TAIL(__pyx_v_self, 1);

  /* "bzrlib/_bencode_pyx.pyx":345
 *         E_UPDATE_TAIL(self, 1)
 * 
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self.process(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_2 = __pyx_v_x; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 345; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 345; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 345; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 345; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 345; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 345; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 345; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "bzrlib/_bencode_pyx.pyx":346
 * 
 *         for i in x:
 *             self.process(i)             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_buffer(1)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_process); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 346; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_i); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 346; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 346; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_v_i);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 346; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "bzrlib/_bencode_pyx.pyx":345
 *         E_UPDATE_TAIL(self, 1)
 * 
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self.process(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":348
 *             self.process(i)
 * 
 *         self._ensure_buffer(1)             # <<<<<<<<<<<<<<
 *         self.tail[0] = c'e'
 *         E_UPDATE_TAIL(self, 1)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_ensure_buffer(__pyx_v_self, 1); if (unlikely(__pyx_t_1 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 348; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":349
 * 
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'e'             # <<<<<<<<<<<<<<
 *         E_UPDATE_TAIL(self, 1)
 *         return 1
 */
  (__pyx_v_self->tail[0]) = 'e';

  /* "bzrlib/_bencode_pyx.pyx":350
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'e'
 *         E_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  E_UPDATE_TAIL(__pyx_v_self, 1);

  /* "bzrlib/_bencode_pyx.pyx":351
 *         self.tail[0] = c'e'
 *         E_UPDATE_TAIL(self, 1)
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int _encode_dict(self, x) except 0:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":340
 *         return 1
 * 
 *     cdef int _encode_list(self, x) except 0:             # <<<<<<<<<<<<<<
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'l'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder._encode_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":353
 *         return 1
 * 
 *     cdef int _encode_dict(self, x) except 0:             # <<<<<<<<<<<<<<
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'd'
 */

static int __pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_dict(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_keys = NULL;
  PyObject *__pyx_v_k = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_encode_dict", 0);

  /* "bzrlib/_bencode_pyx.pyx":354
 * 
 *     cdef int _encode_dict(self, x) except 0:
 *         self._ensure_buffer(1)             # <<<<<<<<<<<<<<
 *         self.tail[0] = c'd'
 *         E_UPDATE_TAIL(self, 1)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_ensure_buffer(__pyx_v_self, 1); if (unlikely(__pyx_t_1 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 354; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":355
 *     cdef int _encode_dict(self, x) except 0:
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'd'             # <<<<<<<<<<<<<<
 *         E_UPDATE_TAIL(self, 1)
 * 
 */
  (__pyx_v_self->tail[0]) = 'd';

  /* "bzrlib/_bencode_pyx.pyx":356
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'd'
 *         E_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 * 
 *         keys = x.keys()
 */
  E_UPDATE_TAIL(__pyx_v_self, 1);

  /* "bzrlib/_bencode_pyx.pyx":358
 *         E_UPDATE_TAIL(self, 1)
 * 
 *         keys = x.keys()             # <<<<<<<<<<<<<<
 *         keys.sort()
 *         for k in keys:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_keys); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 358; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 358; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 358; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_keys = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":359
 * 
 *         keys = x.keys()
 *         keys.sort()             # <<<<<<<<<<<<<<
 *         for k in keys:
 *             if not PyString_CheckExact(k):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_keys, __pyx_n_s_sort); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 359; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_4) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 359; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 359; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":360
 *         keys = x.keys()
 *         keys.sort()
 *         for k in keys:             # <<<<<<<<<<<<<<
 *             if not PyString_CheckExact(k):
 *                 raise TypeError('key in dict should be string')
 */
  if (likely(PyList_CheckExact(__pyx_v_keys)) || PyTuple_CheckExact(__pyx_v_keys)) {
    __pyx_t_2 = __pyx_v_keys; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_keys); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else {__pyx_filename = __pyx_f[0]; __pyx_lineno = 360; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "bzrlib/_bencode_pyx.pyx":361
 *         keys.sort()
 *         for k in keys:
 *             if not PyString_CheckExact(k):             # <<<<<<<<<<<<<<
 *                 raise TypeError('key in dict should be string')
 *             self._encode_string(k)
 */
    __pyx_t_7 = ((!(PyString_CheckExact(__pyx_v_k) != 0)) != 0);
    if (__pyx_t_7) {

      /* "bzrlib/_bencode_pyx.pyx":362
 *         for k in keys:
 *             if not PyString_CheckExact(k):
 *                 raise TypeError('key in dict should be string')             # <<<<<<<<<<<<<<
 *             self._encode_string(k)
 *             self.process(x[k])
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

      /* "bzrlib/_bencode_pyx.pyx":361
 *         keys.sort()
 *         for k in keys:
 *             if not PyString_CheckExact(k):             # <<<<<<<<<<<<<<
 *                 raise TypeError('key in dict should be string')
 *             self._encode_string(k)
 */
    }

    /* "bzrlib/_bencode_pyx.pyx":363
 *             if not PyString_CheckExact(k):
 *                 raise TypeError('key in dict should be string')
 *             self._encode_string(k)             # <<<<<<<<<<<<<<
 *             self.process(x[k])
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_encode_string(__pyx_v_self, __pyx_v_k); if (unlikely(__pyx_t_1 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 363; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":364
 *                 raise TypeError('key in dict should be string')
 *             self._encode_string(k)
 *             self.process(x[k])             # <<<<<<<<<<<<<<
 * 
 *         self._ensure_buffer(1)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_process); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 364; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PyObject_GetItem(__pyx_v_x, __pyx_v_k); if (unlikely(__pyx_t_8 == NULL)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 364; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_9) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_8); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 364; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 364; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 364; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "bzrlib/_bencode_pyx.pyx":360
 *         keys = x.keys()
 *         keys.sort()
 *         for k in keys:             # <<<<<<<<<<<<<<
 *             if not PyString_CheckExact(k):
 *                 raise TypeError('key in dict should be string')
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":366
 *             self.process(x[k])
 * 
 *         self._ensure_buffer(1)             # <<<<<<<<<<<<<<
 *         self.tail[0] = c'e'
 *         E_UPDATE_TAIL(self, 1)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_ensure_buffer(__pyx_v_self, 1); if (unlikely(__pyx_t_1 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 366; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":367
 * 
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'e'             # <<<<<<<<<<<<<<
 *         E_UPDATE_TAIL(self, 1)
 *         return 1
 */
  (__pyx_v_self->tail[0]) = 'e';

  /* "bzrlib/_bencode_pyx.pyx":368
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'e'
 *         E_UPDATE_TAIL(self, 1)             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */
  E_UPDATE_TAIL(__pyx_v_self, 1);

  /* "bzrlib/_bencode_pyx.pyx":369
 *         self.tail[0] = c'e'
 *         E_UPDATE_TAIL(self, 1)
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     def process(self, object x):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":353
 *         return 1
 * 
 *     cdef int _encode_dict(self, x) except 0:             # <<<<<<<<<<<<<<
 *         self._ensure_buffer(1)
 *         self.tail[0] = c'd'
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder._encode_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_keys);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":371
 *         return 1
 * 
 *     def process(self, object x):             # <<<<<<<<<<<<<<
 *         if Py_EnterRecursiveCall("encode"):
 *             raise RuntimeError("too deeply nested")
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_7process(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_7process(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_6process(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_6process(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("process", 0);

  /* "bzrlib/_bencode_pyx.pyx":372
 * 
 *     def process(self, object x):
 *         if Py_EnterRecursiveCall("encode"):             # <<<<<<<<<<<<<<
 *             raise RuntimeError("too deeply nested")
 *         try:
 */
  __pyx_t_1 = (Py_EnterRecursiveCall(__pyx_k_encode) != 0);
  if (__pyx_t_1) {

    /* "bzrlib/_bencode_pyx.pyx":373
 *     def process(self, object x):
 *         if Py_EnterRecursiveCall("encode"):
 *             raise RuntimeError("too deeply nested")             # <<<<<<<<<<<<<<
 *         try:
 *             if PyString_CheckExact(x):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 373; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    {__pyx_filename = __pyx_f[0]; __pyx_lineno = 373; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

    /* "bzrlib/_bencode_pyx.pyx":372
 * 
 *     def process(self, object x):
 *         if Py_EnterRecursiveCall("encode"):             # <<<<<<<<<<<<<<
 *             raise RuntimeError("too deeply nested")
 *         try:
 */
  }

  /* "bzrlib/_bencode_pyx.pyx":374
 *         if Py_EnterRecursiveCall("encode"):
 *             raise RuntimeError("too deeply nested")
 *         try:             # <<<<<<<<<<<<<<
 *             if PyString_CheckExact(x):
 *                 self._encode_string(x)
 */
  /*try:*/ {

    /* "bzrlib/_bencode_pyx.pyx":375
 *             raise RuntimeError("too deeply nested")
 *         try:
 *             if PyString_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_string(x)
 *             elif PyInt_CheckExact(x):
 */
    __pyx_t_1 = (PyString_CheckExact(__pyx_v_x) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_bencode_pyx.pyx":376
 *         try:
 *             if PyString_CheckExact(x):
 *                 self._encode_string(x)             # <<<<<<<<<<<<<<
 *             elif PyInt_CheckExact(x):
 *                 self._encode_int(x)
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_encode_string(__pyx_v_self, __pyx_v_x); if (unlikely(__pyx_t_3 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 376; __pyx_clineno = __LINE__; goto __pyx_L5_error;}

      /* "bzrlib/_bencode_pyx.pyx":375
 *             raise RuntimeError("too deeply nested")
 *         try:
 *             if PyString_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_string(x)
 *             elif PyInt_CheckExact(x):
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_bencode_pyx.pyx":377
 *             if PyString_CheckExact(x):
 *                 self._encode_string(x)
 *             elif PyInt_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_int(x)
 *             elif PyLong_CheckExact(x):
 */
    __pyx_t_1 = (PyInt_CheckExact(__pyx_v_x) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_bencode_pyx.pyx":378
 *                 self._encode_string(x)
 *             elif PyInt_CheckExact(x):
 *                 self._encode_int(x)             # <<<<<<<<<<<<<<
 *             elif PyLong_CheckExact(x):
 *                 self._encode_long(x)
 */
      __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_x); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 378; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_encode_int(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 378; __pyx_clineno = __LINE__; goto __pyx_L5_error;}

      /* "bzrlib/_bencode_pyx.pyx":377
 *             if PyString_CheckExact(x):
 *                 self._encode_string(x)
 *             elif PyInt_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_int(x)
 *             elif PyLong_CheckExact(x):
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_bencode_pyx.pyx":379
 *             elif PyInt_CheckExact(x):
 *                 self._encode_int(x)
 *             elif PyLong_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_long(x)
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)
 */
    __pyx_t_1 = (PyLong_CheckExact(__pyx_v_x) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_bencode_pyx.pyx":380
 *                 self._encode_int(x)
 *             elif PyLong_CheckExact(x):
 *                 self._encode_long(x)             # <<<<<<<<<<<<<<
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)
 *                   or StaticTuple_CheckExact(x)):
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_encode_long(__pyx_v_self, __pyx_v_x); if (unlikely(__pyx_t_4 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 380; __pyx_clineno = __LINE__; goto __pyx_L5_error;}

      /* "bzrlib/_bencode_pyx.pyx":379
 *             elif PyInt_CheckExact(x):
 *                 self._encode_int(x)
 *             elif PyLong_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_long(x)
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_bencode_pyx.pyx":381
 *             elif PyLong_CheckExact(x):
 *                 self._encode_long(x)
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)             # <<<<<<<<<<<<<<
 *                   or StaticTuple_CheckExact(x)):
 *                 self._encode_list(x)
 */
    __pyx_t_5 = (PyList_CheckExact(__pyx_v_x) != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }

    /* "bzrlib/_bencode_pyx.pyx":382
 *                 self._encode_long(x)
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)
 *                   or StaticTuple_CheckExact(x)):             # <<<<<<<<<<<<<<
 *                 self._encode_list(x)
 *             elif PyDict_CheckExact(x):
 */
    __pyx_t_5 = (PyTuple_CheckExact(__pyx_v_x) != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_1 = __pyx_t_5;
      goto __pyx_L8_bool_binop_done;
    }
    __pyx_t_5 = (StaticTuple_CheckExact(__pyx_v_x) != 0);
    __pyx_t_1 = __pyx_t_5;
    __pyx_L8_bool_binop_done:;

    /* "bzrlib/_bencode_pyx.pyx":381
 *             elif PyLong_CheckExact(x):
 *                 self._encode_long(x)
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)             # <<<<<<<<<<<<<<
 *                   or StaticTuple_CheckExact(x)):
 *                 self._encode_list(x)
 */
    if (__pyx_t_1) {

      /* "bzrlib/_bencode_pyx.pyx":383
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)
 *                   or StaticTuple_CheckExact(x)):
 *                 self._encode_list(x)             # <<<<<<<<<<<<<<
 *             elif PyDict_CheckExact(x):
 *                 self._encode_dict(x)
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_encode_list(__pyx_v_self, __pyx_v_x); if (unlikely(__pyx_t_4 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 383; __pyx_clineno = __LINE__; goto __pyx_L5_error;}

      /* "bzrlib/_bencode_pyx.pyx":381
 *             elif PyLong_CheckExact(x):
 *                 self._encode_long(x)
 *             elif (PyList_CheckExact(x) or PyTuple_CheckExact(x)             # <<<<<<<<<<<<<<
 *                   or StaticTuple_CheckExact(x)):
 *                 self._encode_list(x)
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_bencode_pyx.pyx":384
 *                   or StaticTuple_CheckExact(x)):
 *                 self._encode_list(x)
 *             elif PyDict_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_dict(x)
 *             elif PyBool_Check(x):
 */
    __pyx_t_1 = (PyDict_CheckExact(__pyx_v_x) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_bencode_pyx.pyx":385
 *                 self._encode_list(x)
 *             elif PyDict_CheckExact(x):
 *                 self._encode_dict(x)             # <<<<<<<<<<<<<<
 *             elif PyBool_Check(x):
 *                 self._encode_int(int(x))
 */
      __pyx_t_4 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_encode_dict(__pyx_v_self, __pyx_v_x); if (unlikely(__pyx_t_4 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 385; __pyx_clineno = __LINE__; goto __pyx_L5_error;}

      /* "bzrlib/_bencode_pyx.pyx":384
 *                   or StaticTuple_CheckExact(x)):
 *                 self._encode_list(x)
 *             elif PyDict_CheckExact(x):             # <<<<<<<<<<<<<<
 *                 self._encode_dict(x)
 *             elif PyBool_Check(x):
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_bencode_pyx.pyx":386
 *             elif PyDict_CheckExact(x):
 *                 self._encode_dict(x)
 *             elif PyBool_Check(x):             # <<<<<<<<<<<<<<
 *                 self._encode_int(int(x))
 *             elif isinstance(x, Bencached):
 */
    __pyx_t_1 = (PyBool_Check(__pyx_v_x) != 0);
    if (__pyx_t_1) {

      /* "bzrlib/_bencode_pyx.pyx":387
 *                 self._encode_dict(x)
 *             elif PyBool_Check(x):
 *                 self._encode_int(int(x))             # <<<<<<<<<<<<<<
 *             elif isinstance(x, Bencached):
 *                 self._append_string(x.bencoded)
 */
      __pyx_t_2 = PyNumber_Int(__pyx_v_x); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 387; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 387; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_3 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_encode_int(__pyx_v_self, __pyx_t_4); if (unlikely(__pyx_t_3 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 387; __pyx_clineno = __LINE__; goto __pyx_L5_error;}

      /* "bzrlib/_bencode_pyx.pyx":386
 *             elif PyDict_CheckExact(x):
 *                 self._encode_dict(x)
 *             elif PyBool_Check(x):             # <<<<<<<<<<<<<<
 *                 self._encode_int(int(x))
 *             elif isinstance(x, Bencached):
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_bencode_pyx.pyx":388
 *             elif PyBool_Check(x):
 *                 self._encode_int(int(x))
 *             elif isinstance(x, Bencached):             # <<<<<<<<<<<<<<
 *                 self._append_string(x.bencoded)
 *             else:
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_Bencached); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_IsInstance(__pyx_v_x, __pyx_t_2); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 388; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_t_1 != 0);
    if (__pyx_t_5) {

      /* "bzrlib/_bencode_pyx.pyx":389
 *                 self._encode_int(int(x))
 *             elif isinstance(x, Bencached):
 *                 self._append_string(x.bencoded)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError('unsupported type %r' % x)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_bencoded); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = ((struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self->__pyx_vtab)->_append_string(__pyx_v_self, __pyx_t_2); if (unlikely(__pyx_t_3 == 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 389; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "bzrlib/_bencode_pyx.pyx":388
 *             elif PyBool_Check(x):
 *                 self._encode_int(int(x))
 *             elif isinstance(x, Bencached):             # <<<<<<<<<<<<<<
 *                 self._append_string(x.bencoded)
 *             else:
 */
      goto __pyx_L7;
    }

    /* "bzrlib/_bencode_pyx.pyx":391
 *                 self._append_string(x.bencoded)
 *             else:
 *                 raise TypeError('unsupported type %r' % x)             # <<<<<<<<<<<<<<
 *         finally:
 *             Py_LeaveRecursiveCall()
 */
    /*else*/ {
      __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_unsupported_type_r, __pyx_v_x); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      {__pyx_filename = __pyx_f[0]; __pyx_lineno = 391; __pyx_clineno = __LINE__; goto __pyx_L5_error;}
    }
    __pyx_L7:;
  }

  /* "bzrlib/_bencode_pyx.pyx":393
 *                 raise TypeError('unsupported type %r' % x)
 *         finally:
 *             Py_LeaveRecursiveCall()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      Py_LeaveRecursiveCall();
      goto __pyx_L6;
    }
    /*exception exit:*/{
      __pyx_L5_error:;
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_3 = __pyx_lineno; __pyx_t_4 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        Py_LeaveRecursiveCall();
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_3; __pyx_clineno = __pyx_t_4; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "bzrlib/_bencode_pyx.pyx":371
 *         return 1
 * 
 *     def process(self, object x):             # <<<<<<<<<<<<<<
 *         if Py_EnterRecursiveCall("encode"):
 *             raise RuntimeError("too deeply nested")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.process", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":248
 *     """Bencode encoder"""
 * 
 *     cdef readonly char *tail             # <<<<<<<<<<<<<<
 *     cdef readonly int size
 *     cdef readonly char *buffer
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_4tail_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_4tail_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4tail___get__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4tail___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->tail); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 248; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.tail.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":249
 * 
 *     cdef readonly char *tail
 *     cdef readonly int size             # <<<<<<<<<<<<<<
 *     cdef readonly char *buffer
 *     cdef readonly int maxsize
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4size___get__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_4size___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 249; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":250
 *     cdef readonly char *tail
 *     cdef readonly int size
 *     cdef readonly char *buffer             # <<<<<<<<<<<<<<
 *     cdef readonly int maxsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_6buffer_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_6buffer_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_6buffer___get__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_6buffer___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->buffer); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 250; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.buffer.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":251
 *     cdef readonly int size
 *     cdef readonly char *buffer
 *     cdef readonly int maxsize             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, int maxsize=INITSIZE):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_7maxsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_7maxsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_7maxsize___get__(((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_7Encoder_7maxsize___get__(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->maxsize); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 251; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.Encoder.maxsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "bzrlib/_bencode_pyx.pyx":396
 * 
 * 
 * def bencode(x):             # <<<<<<<<<<<<<<
 *     """Encode Python object x to string"""
 *     encoder = Encoder()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_5bencode(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_6bzrlib_12_bencode_pyx_4bencode[] = "Encode Python object x to string";
static PyMethodDef __pyx_mdef_6bzrlib_12_bencode_pyx_5bencode = {"bencode", (PyCFunction)__pyx_pw_6bzrlib_12_bencode_pyx_5bencode, METH_O, __pyx_doc_6bzrlib_12_bencode_pyx_4bencode};
static PyObject *__pyx_pw_6bzrlib_12_bencode_pyx_5bencode(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bencode (wrapper)", 0);
  __pyx_r = __pyx_pf_6bzrlib_12_bencode_pyx_4bencode(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6bzrlib_12_bencode_pyx_4bencode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *__pyx_v_encoder = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bencode", 0);

  /* "bzrlib/_bencode_pyx.pyx":398
 * def bencode(x):
 *     """Encode Python object x to string"""
 *     encoder = Encoder()             # <<<<<<<<<<<<<<
 *     encoder.process(x)
 *     return str(encoder)
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6bzrlib_12_bencode_pyx_Encoder), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 398; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_encoder = ((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "bzrlib/_bencode_pyx.pyx":399
 *     """Encode Python object x to string"""
 *     encoder = Encoder()
 *     encoder.process(x)             # <<<<<<<<<<<<<<
 *     return str(encoder)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_encoder), __pyx_n_s_process); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_x); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(__pyx_v_x);
    __Pyx_GIVEREF(__pyx_v_x);
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_x);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 399; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "bzrlib/_bencode_pyx.pyx":400
 *     encoder = Encoder()
 *     encoder.process(x)
 *     return str(encoder)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_encoder));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_encoder));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_encoder));
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 400; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "bzrlib/_bencode_pyx.pyx":396
 * 
 * 
 * def bencode(x):             # <<<<<<<<<<<<<<
 *     """Encode Python object x to string"""
 *     encoder = Encoder()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("bzrlib._bencode_pyx.bencode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_encoder);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Decoder __pyx_vtable_6bzrlib_12_bencode_pyx_Decoder;

static PyObject *__pyx_tp_new_6bzrlib_12_bencode_pyx_Decoder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)o);
  p->__pyx_vtab = __pyx_vtabptr_6bzrlib_12_bencode_pyx_Decoder;
  p->text = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6bzrlib_12_bencode_pyx_Decoder(PyObject *o) {
  struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *p = (struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->text);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6bzrlib_12_bencode_pyx_Decoder(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *p = (struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)o;
  if (p->text) {
    e = (*v)(p->text, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6bzrlib_12_bencode_pyx_Decoder(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *p = (struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *)o;
  tmp = ((PyObject*)p->text);
  p->text = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6bzrlib_12_bencode_pyx_7Decoder_tail(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_4tail_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_12_bencode_pyx_7Decoder_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_4size_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_12_bencode_pyx_7Decoder__yield_tuples(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_13_yield_tuples_1__get__(o);
}

static PyMethodDef __pyx_methods_6bzrlib_12_bencode_pyx_Decoder[] = {
  {"decode", (PyCFunction)__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_3decode, METH_NOARGS, 0},
  {"decode_object", (PyCFunction)__pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_5decode_object, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6bzrlib_12_bencode_pyx_Decoder[] = {
  {(char *)"tail", __pyx_getprop_6bzrlib_12_bencode_pyx_7Decoder_tail, 0, 0, 0},
  {(char *)"size", __pyx_getprop_6bzrlib_12_bencode_pyx_7Decoder_size, 0, 0, 0},
  {(char *)"_yield_tuples", __pyx_getprop_6bzrlib_12_bencode_pyx_7Decoder__yield_tuples, 0, 0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6bzrlib_12_bencode_pyx_Decoder = {
  PyVarObject_HEAD_INIT(0, 0)
  "bzrlib._bencode_pyx.Decoder", /*tp_name*/
  sizeof(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6bzrlib_12_bencode_pyx_Decoder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Bencode decoder", /*tp_doc*/
  __pyx_tp_traverse_6bzrlib_12_bencode_pyx_Decoder, /*tp_traverse*/
  __pyx_tp_clear_6bzrlib_12_bencode_pyx_Decoder, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6bzrlib_12_bencode_pyx_Decoder, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6bzrlib_12_bencode_pyx_Decoder, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6bzrlib_12_bencode_pyx_7Decoder_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6bzrlib_12_bencode_pyx_Decoder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6bzrlib_12_bencode_pyx_Encoder __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder;

static PyObject *__pyx_tp_new_6bzrlib_12_bencode_pyx_Encoder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *)o);
  p->__pyx_vtab = __pyx_vtabptr_6bzrlib_12_bencode_pyx_Encoder;
  return o;
}

static void __pyx_tp_dealloc_6bzrlib_12_bencode_pyx_Encoder(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_tail(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_4tail_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_4size_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_buffer(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_6buffer_1__get__(o);
}

static PyObject *__pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_maxsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_7maxsize_1__get__(o);
}

static PyMethodDef __pyx_methods_6bzrlib_12_bencode_pyx_Encoder[] = {
  {"process", (PyCFunction)__pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_7process, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6bzrlib_12_bencode_pyx_Encoder[] = {
  {(char *)"tail", __pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_tail, 0, 0, 0},
  {(char *)"size", __pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_size, 0, 0, 0},
  {(char *)"buffer", __pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_buffer, 0, 0, 0},
  {(char *)"maxsize", __pyx_getprop_6bzrlib_12_bencode_pyx_7Encoder_maxsize, 0, 0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6bzrlib_12_bencode_pyx_Encoder = {
  PyVarObject_HEAD_INIT(0, 0)
  "bzrlib._bencode_pyx.Encoder", /*tp_name*/
  sizeof(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6bzrlib_12_bencode_pyx_Encoder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_5__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Bencode encoder", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6bzrlib_12_bencode_pyx_Encoder, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6bzrlib_12_bencode_pyx_Encoder, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6bzrlib_12_bencode_pyx_7Encoder_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6bzrlib_12_bencode_pyx_Encoder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_bencode_pyx",
    __pyx_k_Pyrex_implementation_for_bencode, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_Bencached, __pyx_k_Bencached, sizeof(__pyx_k_Bencached), 0, 0, 1, 1},
  {&__pyx_n_s_Bencached___init, __pyx_k_Bencached___init, sizeof(__pyx_k_Bencached___init), 0, 0, 1, 1},
  {&__pyx_kp_s_Cannot_realloc_buffer_for_encode, __pyx_k_Cannot_realloc_buffer_for_encode, sizeof(__pyx_k_Cannot_realloc_buffer_for_encode), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_Not_enough_memory_to_allocate_bu, __pyx_k_Not_enough_memory_to_allocate_bu, sizeof(__pyx_k_Not_enough_memory_to_allocate_bu), 0, 0, 1, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Stop_character_c_not_found_c, __pyx_k_Stop_character_c_not_found_c, sizeof(__pyx_k_Stop_character_c_not_found_c), 0, 0, 1, 0},
  {&__pyx_kp_s_String_required, __pyx_k_String_required, sizeof(__pyx_k_String_required), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 0, 1, 0},
  {&__pyx_n_s_bdecode, __pyx_k_bdecode, sizeof(__pyx_k_bdecode), 0, 0, 1, 1},
  {&__pyx_n_s_bdecode_as_tuple, __pyx_k_bdecode_as_tuple, sizeof(__pyx_k_bdecode_as_tuple), 0, 0, 1, 1},
  {&__pyx_n_s_bencode, __pyx_k_bencode, sizeof(__pyx_k_bencode), 0, 0, 1, 1},
  {&__pyx_n_s_bencoded, __pyx_k_bencoded, sizeof(__pyx_k_bencoded), 0, 0, 1, 1},
  {&__pyx_n_s_bzrlib__bencode_pyx, __pyx_k_bzrlib__bencode_pyx, sizeof(__pyx_k_bzrlib__bencode_pyx), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_kp_s_dict_keys_disordered, __pyx_k_dict_keys_disordered, sizeof(__pyx_k_dict_keys_disordered), 0, 0, 1, 0},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_encoder, __pyx_k_encoder, sizeof(__pyx_k_encoder), 0, 0, 1, 1},
  {&__pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_k_home_vila_src_bzr_integration_2, sizeof(__pyx_k_home_vila_src_bzr_integration_2), 0, 0, 1, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_kp_s_int_d_too_big_to_encode, __pyx_k_int_d_too_big_to_encode, sizeof(__pyx_k_int_d_too_big_to_encode), 0, 0, 1, 0},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_kp_s_junk_in_stream, __pyx_k_junk_in_stream, sizeof(__pyx_k_junk_in_stream), 0, 0, 1, 0},
  {&__pyx_kp_s_key_in_dict_should_be_string, __pyx_k_key_in_dict_should_be_string, sizeof(__pyx_k_key_in_dict_should_be_string), 0, 0, 1, 0},
  {&__pyx_kp_s_key_was_not_a_simple_string, __pyx_k_key_was_not_a_simple_string, sizeof(__pyx_k_key_was_not_a_simple_string), 0, 0, 1, 0},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_kp_s_leading_zeros_are_not_allowed, __pyx_k_leading_zeros_are_not_allowed, sizeof(__pyx_k_leading_zeros_are_not_allowed), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_kp_s_malformed_dict, __pyx_k_malformed_dict, sizeof(__pyx_k_malformed_dict), 0, 0, 1, 0},
  {&__pyx_kp_s_malformed_list, __pyx_k_malformed_list, sizeof(__pyx_k_malformed_list), 0, 0, 1, 0},
  {&__pyx_n_s_maxsize, __pyx_k_maxsize, sizeof(__pyx_k_maxsize), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_process, __pyx_k_process, sizeof(__pyx_k_process), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_slots, __pyx_k_slots, sizeof(__pyx_k_slots), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_kp_s_stream_underflow, __pyx_k_stream_underflow, sizeof(__pyx_k_stream_underflow), 0, 0, 1, 0},
  {&__pyx_kp_s_string_len_not_terminated_by, __pyx_k_string_len_not_terminated_by, sizeof(__pyx_k_string_len_not_terminated_by), 0, 0, 1, 0},
  {&__pyx_kp_s_string_s_too_big_to_encode, __pyx_k_string_s_too_big_to_encode, sizeof(__pyx_k_string_s_too_big_to_encode), 0, 0, 1, 0},
  {&__pyx_kp_s_string_size_below_zero_d, __pyx_k_string_size_below_zero_d, sizeof(__pyx_k_string_size_below_zero_d), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_too_deeply_nested, __pyx_k_too_deeply_nested, sizeof(__pyx_k_too_deeply_nested), 0, 0, 1, 0},
  {&__pyx_kp_s_unknown_object_type_identifier_r, __pyx_k_unknown_object_type_identifier_r, sizeof(__pyx_k_unknown_object_type_identifier_r), 0, 0, 1, 0},
  {&__pyx_kp_s_unsupported_type_r, __pyx_k_unsupported_type_r, sizeof(__pyx_k_unsupported_type_r), 0, 0, 1, 0},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_yield_tuples, __pyx_k_yield_tuples, sizeof(__pyx_k_yield_tuples), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_object = __Pyx_GetBuiltinName(__pyx_n_s_object); if (!__pyx_builtin_object) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 82; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "bzrlib/_bencode_pyx.pyx":82
 *         """
 *         if not PyString_CheckExact(s):
 *             raise TypeError("String required")             # <<<<<<<<<<<<<<
 * 
 *         self.text = s
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_String_required); if (unlikely(!__pyx_tuple_)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 82; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "bzrlib/_bencode_pyx.pyx":92
 *         result = self._decode_object()
 *         if self.size != 0:
 *             raise ValueError('junk in stream')             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_junk_in_stream); if (unlikely(!__pyx_tuple__2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 92; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "bzrlib/_bencode_pyx.pyx":102
 * 
 *         if 0 == self.size:
 *             raise ValueError('stream underflow')             # <<<<<<<<<<<<<<
 * 
 *         if Py_EnterRecursiveCall("_decode_object"):
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_stream_underflow); if (unlikely(!__pyx_tuple__3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 102; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "bzrlib/_bencode_pyx.pyx":105
 * 
 *         if Py_EnterRecursiveCall("_decode_object"):
 *             raise RuntimeError("too deeply nested")             # <<<<<<<<<<<<<<
 *         try:
 *             ch = self.tail[0]
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_too_deeply_nested); if (unlikely(!__pyx_tuple__4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 105; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "bzrlib/_bencode_pyx.pyx":162
 *         n = strtol(self.tail, &next_tail, 10)
 *         if next_tail == NULL or next_tail[0] != c':':
 *             raise ValueError('string len not terminated by ":"')             # <<<<<<<<<<<<<<
 *         # strtol allows leading zeros, so validate that we don't have that
 *         if (self.tail[0] == c'0'
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_string_len_not_terminated_by); if (unlikely(!__pyx_tuple__5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 162; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "bzrlib/_bencode_pyx.pyx":166
 *         if (self.tail[0] == c'0'
 *             and (n != 0 or (next_tail - self.tail != 1))):
 *             raise ValueError('leading zeros are not allowed')             # <<<<<<<<<<<<<<
 *         D_UPDATE_TAIL(self, next_tail - self.tail + 1)
 *         if n == 0:
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_leading_zeros_are_not_allowed); if (unlikely(!__pyx_tuple__6)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 166; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "bzrlib/_bencode_pyx.pyx":171
 *             return ''
 *         if n > self.size:
 *             raise ValueError('stream underflow')             # <<<<<<<<<<<<<<
 *         if n < 0:
 *             raise ValueError('string size below zero: %d' % n)
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_stream_underflow); if (unlikely(!__pyx_tuple__8)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 171; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "bzrlib/_bencode_pyx.pyx":195
 *                 PyList_Append(result, self._decode_object())
 * 
 *         raise ValueError('malformed list')             # <<<<<<<<<<<<<<
 * 
 *     cdef object _decode_dict(self):
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_malformed_list); if (unlikely(!__pyx_tuple__9)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 195; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "bzrlib/_bencode_pyx.pyx":211
 *                 # keys should be strings only
 *                 if self.tail[0] < c'0' or self.tail[0] > c'9':
 *                     raise ValueError('key was not a simple string.')             # <<<<<<<<<<<<<<
 *                 key = self._decode_string()
 *                 if lastkey >= key:
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_key_was_not_a_simple_string); if (unlikely(!__pyx_tuple__10)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 211; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "bzrlib/_bencode_pyx.pyx":214
 *                 key = self._decode_string()
 *                 if lastkey >= key:
 *                     raise ValueError('dict keys disordered')             # <<<<<<<<<<<<<<
 *                 else:
 *                     lastkey = key
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_dict_keys_disordered); if (unlikely(!__pyx_tuple__11)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 214; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "bzrlib/_bencode_pyx.pyx":220
 *                 result[key] = value
 * 
 *         raise ValueError('malformed dict')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_malformed_dict); if (unlikely(!__pyx_tuple__12)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 220; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "bzrlib/_bencode_pyx.pyx":265
 *         p = <char*>malloc(maxsize)
 *         if p == NULL:
 *             raise MemoryError('Not enough memory to allocate buffer '             # <<<<<<<<<<<<<<
 *                               'for encoder')
 *         self.buffer = p
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_Not_enough_memory_to_allocate_bu); if (unlikely(!__pyx_tuple__14)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 265; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "bzrlib/_bencode_pyx.pyx":298
 *         new_buffer = <char*>realloc(self.buffer, <size_t>new_size)
 *         if new_buffer == NULL:
 *             raise MemoryError('Cannot realloc buffer for encoder')             # <<<<<<<<<<<<<<
 * 
 *         self.buffer = new_buffer
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_Cannot_realloc_buffer_for_encode); if (unlikely(!__pyx_tuple__15)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 298; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "bzrlib/_bencode_pyx.pyx":362
 *         for k in keys:
 *             if not PyString_CheckExact(k):
 *                 raise TypeError('key in dict should be string')             # <<<<<<<<<<<<<<
 *             self._encode_string(k)
 *             self.process(x[k])
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_key_in_dict_should_be_string); if (unlikely(!__pyx_tuple__16)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 362; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "bzrlib/_bencode_pyx.pyx":373
 *     def process(self, object x):
 *         if Py_EnterRecursiveCall("encode"):
 *             raise RuntimeError("too deeply nested")             # <<<<<<<<<<<<<<
 *         try:
 *             if PyString_CheckExact(x):
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_too_deeply_nested); if (unlikely(!__pyx_tuple__17)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 373; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "bzrlib/_bencode_pyx.pyx":223
 * 
 * 
 * def bdecode(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object"""
 *     return Decoder(s).decode()
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_n_s_s); if (unlikely(!__pyx_tuple__18)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_bdecode, 223, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":228
 * 
 * 
 * def bdecode_as_tuple(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object, using tuples rather than lists."""
 *     return Decoder(s, True).decode()
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_n_s_s); if (unlikely(!__pyx_tuple__20)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 228; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_bdecode_as_tuple, 228, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 228; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":236
 *     __slots__ = ['bencoded']
 * 
 *     def __init__(self, s):             # <<<<<<<<<<<<<<
 *         self.bencoded = s
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_s); if (unlikely(!__pyx_tuple__22)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_init, 236, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":396
 * 
 * 
 * def bencode(x):             # <<<<<<<<<<<<<<
 *     """Encode Python object x to string"""
 *     encoder = Encoder()
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_n_s_x, __pyx_n_s_encoder); if (unlikely(!__pyx_tuple__24)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 396; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_vila_src_bzr_integration_2, __pyx_n_s_bencode, 396, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 396; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_bencode_pyx(void); /*proto*/
PyMODINIT_FUNC init_bencode_pyx(void)
#else
PyMODINIT_FUNC PyInit__bencode_pyx(void); /*proto*/
PyMODINIT_FUNC PyInit__bencode_pyx(void)
#endif
{
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__bencode_pyx(void)", 0);
  if (__Pyx_check_binary_version() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_bencode_pyx", __pyx_methods, __pyx_k_Pyrex_implementation_for_bencode, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;};
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif
  if (__pyx_module_is_main_bzrlib___bencode_pyx) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (!PyDict_GetItemString(modules, "bzrlib._bencode_pyx")) {
      if (unlikely(PyDict_SetItemString(modules, "bzrlib._bencode_pyx", __pyx_m) < 0)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_6bzrlib_12_bencode_pyx_Decoder = &__pyx_vtable_6bzrlib_12_bencode_pyx_Decoder;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Decoder._decode_object = (PyObject *(*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *))__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_object;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Decoder._read_digits = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *, char))__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__read_digits;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Decoder._decode_int = (PyObject *(*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *))__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_int;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Decoder._decode_string = (PyObject *(*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *))__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_string;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Decoder._decode_list = (PyObject *(*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *))__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_list;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Decoder._decode_dict = (PyObject *(*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Decoder *))__pyx_f_6bzrlib_12_bencode_pyx_7Decoder__decode_dict;
  if (PyType_Ready(&__pyx_type_6bzrlib_12_bencode_pyx_Decoder) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 69; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_type_6bzrlib_12_bencode_pyx_Decoder.tp_print = 0;
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_6bzrlib_12_bencode_pyx_Decoder, "__init__"); if (unlikely(!wrapper)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 69; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Decoder___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Decoder___init__.doc = __pyx_doc_6bzrlib_12_bencode_pyx_7Decoder___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Decoder___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_6bzrlib_12_bencode_pyx_Decoder.tp_dict, __pyx_vtabptr_6bzrlib_12_bencode_pyx_Decoder) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 69; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyObject_SetAttrString(__pyx_m, "Decoder", (PyObject *)&__pyx_type_6bzrlib_12_bencode_pyx_Decoder) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 69; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_6bzrlib_12_bencode_pyx_Decoder = &__pyx_type_6bzrlib_12_bencode_pyx_Decoder;
  __pyx_vtabptr_6bzrlib_12_bencode_pyx_Encoder = &__pyx_vtable_6bzrlib_12_bencode_pyx_Encoder;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder._ensure_buffer = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, int))__pyx_f_6bzrlib_12_bencode_pyx_7Encoder__ensure_buffer;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder._encode_int = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, int))__pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_int;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder._encode_long = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *))__pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_long;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder._append_string = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *))__pyx_f_6bzrlib_12_bencode_pyx_7Encoder__append_string;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder._encode_string = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *))__pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_string;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder._encode_list = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *))__pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_list;
  __pyx_vtable_6bzrlib_12_bencode_pyx_Encoder._encode_dict = (int (*)(struct __pyx_obj_6bzrlib_12_bencode_pyx_Encoder *, PyObject *))__pyx_f_6bzrlib_12_bencode_pyx_7Encoder__encode_dict;
  if (PyType_Ready(&__pyx_type_6bzrlib_12_bencode_pyx_Encoder) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_type_6bzrlib_12_bencode_pyx_Encoder.tp_print = 0;
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_6bzrlib_12_bencode_pyx_Encoder, "__init__"); if (unlikely(!wrapper)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Encoder___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Encoder___init__.doc = __pyx_doc_6bzrlib_12_bencode_pyx_7Encoder___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_6bzrlib_12_bencode_pyx_7Encoder___init__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_6bzrlib_12_bencode_pyx_Encoder.tp_dict, __pyx_vtabptr_6bzrlib_12_bencode_pyx_Encoder) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  if (PyObject_SetAttrString(__pyx_m, "Encoder", (PyObject *)&__pyx_type_6bzrlib_12_bencode_pyx_Encoder) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 245; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __pyx_ptype_6bzrlib_12_bencode_pyx_Encoder = &__pyx_type_6bzrlib_12_bencode_pyx_Encoder;
  /*--- Type import code ---*/
  __pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple = __Pyx_ImportType("bzrlib._static_tuple_c", "StaticTuple", sizeof(StaticTuple), 0); if (unlikely(!__pyx_ptype_6bzrlib_15_static_tuple_c_StaticTuple)) {__pyx_filename = __pyx_f[1]; __pyx_lineno = 26; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  #endif

  /* "bzrlib/_bencode_pyx.pyx":66
 *     import_static_tuple_c
 * 
 * import_static_tuple_c()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = import_static_tuple_c(); if (unlikely(__pyx_t_1 == -1)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 66; __pyx_clineno = __LINE__; goto __pyx_L1_error;}

  /* "bzrlib/_bencode_pyx.pyx":223
 * 
 * 
 * def bdecode(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object"""
 *     return Decoder(s).decode()
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_12_bencode_pyx_1bdecode, NULL, __pyx_n_s_bzrlib__bencode_pyx); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bdecode, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 223; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":228
 * 
 * 
 * def bdecode_as_tuple(object s):             # <<<<<<<<<<<<<<
 *     """Decode string x to Python object, using tuples rather than lists."""
 *     return Decoder(s, True).decode()
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_12_bencode_pyx_3bdecode_as_tuple, NULL, __pyx_n_s_bzrlib__bencode_pyx); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 228; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bdecode_as_tuple, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 228; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":233
 * 
 * 
 * class Bencached(object):             # <<<<<<<<<<<<<<
 *     __slots__ = ['bencoded']
 * 
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_builtin_object);
  __Pyx_GIVEREF(__pyx_builtin_object);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_builtin_object);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_3)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_2, __pyx_n_s_Bencached, __pyx_n_s_Bencached, (PyObject *) NULL, __pyx_n_s_bzrlib__bencode_pyx, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_4);

  /* "bzrlib/_bencode_pyx.pyx":234
 * 
 * class Bencached(object):
 *     __slots__ = ['bencoded']             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, s):
 */
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_s_bencoded);
  __Pyx_GIVEREF(__pyx_n_s_bencoded);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_s_bencoded);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_slots, __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 234; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "bzrlib/_bencode_pyx.pyx":236
 *     __slots__ = ['bencoded']
 * 
 *     def __init__(self, s):             # <<<<<<<<<<<<<<
 *         self.bencoded = s
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6bzrlib_12_bencode_pyx_9Bencached_1__init__, 0, __pyx_n_s_Bencached___init, NULL, __pyx_n_s_bzrlib__bencode_pyx, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_init, __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 236; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "bzrlib/_bencode_pyx.pyx":233
 * 
 * 
 * class Bencached(object):             # <<<<<<<<<<<<<<
 *     __slots__ = ['bencoded']
 * 
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_Bencached, __pyx_t_2, __pyx_t_4, NULL, 0, 1); if (unlikely(!__pyx_t_5)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Bencached, __pyx_t_5) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 233; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":253
 *     cdef readonly int maxsize
 * 
 *     def __init__(self, int maxsize=INITSIZE):             # <<<<<<<<<<<<<<
 *         """Initialize encoder engine
 *         @param  maxsize:    initial size of internal char buffer
 */
  __pyx_k__13 = __pyx_e_6bzrlib_12_bencode_pyx_INITSIZE;

  /* "bzrlib/_bencode_pyx.pyx":396
 * 
 * 
 * def bencode(x):             # <<<<<<<<<<<<<<
 *     """Encode Python object x to string"""
 *     encoder = Encoder()
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6bzrlib_12_bencode_pyx_5bencode, NULL, __pyx_n_s_bzrlib__bencode_pyx); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 396; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bencode, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 396; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "bzrlib/_bencode_pyx.pyx":1
 * # Copyright (C) 2007, 2009, 2010 Canonical Ltd             # <<<<<<<<<<<<<<
 * #
 * # This program is free software; you can redistribute it and/or modify
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) {__pyx_filename = __pyx_f[0]; __pyx_lineno = 1; __pyx_clineno = __LINE__; goto __pyx_L1_error;}
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init bzrlib._bencode_pyx", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init bzrlib._bencode_pyx");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

static CYTHON_INLINE void __Pyx_ErrRestore(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_Restore(type, value, tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetch(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(type, value, tb);
#endif
}

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
#else
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
#endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}

static CYTHON_INLINE void __Pyx_ExceptionSave(PyObject **type, PyObject **value, PyObject **tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
#else
    PyErr_GetExcInfo(type, value, tb);
#endif
}
static void __Pyx_ExceptionReset(PyObject *type, PyObject *value, PyObject *tb) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyThreadState *tstate = PyThreadState_GET();
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(type, value, tb);
#endif
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject* args = PyTuple_Pack(1, arg);
    return (likely(args)) ? __Pyx_PyObject_Call(func, args, NULL) : NULL;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_COMPILING_IN_CPYTHON
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyMem_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    PyObject *self = f->m_self;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_Call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyMem_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
#endif

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, -(sdigit) digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value) {
    const char neg_one = (char) -1, const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(char),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, -(sdigit) digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_Int(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Int(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility",
            module_name, class_name);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling",
            module_name, class_name);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Int(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
